package com.chelvc.framework.base.config;

import java.nio.charset.StandardCharsets;
import java.security.GeneralSecurityException;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;
import javax.net.ssl.SSLContext;

import com.fasterxml.jackson.databind.ObjectMapper;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.HttpRequestRetryHandler;
import org.apache.http.config.RegistryBuilder;
import org.apache.http.conn.ConnectionKeepAliveStrategy;
import org.apache.http.conn.socket.ConnectionSocketFactory;
import org.apache.http.conn.socket.PlainConnectionSocketFactory;
import org.apache.http.conn.ssl.NoopHostnameVerifier;
import org.apache.http.conn.ssl.SSLConnectionSocketFactory;
import org.apache.http.impl.client.DefaultConnectionKeepAliveStrategy;
import org.apache.http.impl.client.DefaultHttpRequestRetryHandler;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import org.apache.http.protocol.HttpContext;
import org.apache.http.ssl.SSLContextBuilder;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.http.MediaType;
import org.springframework.http.client.ClientHttpRequestFactory;
import org.springframework.http.client.HttpComponentsClientHttpRequestFactory;
import org.springframework.http.converter.HttpMessageConverter;
import org.springframework.http.converter.StringHttpMessageConverter;
import org.springframework.http.converter.json.MappingJackson2HttpMessageConverter;
import org.springframework.web.client.RestTemplate;

/**
 * Http客户端配置
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Configuration
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class HttpClientConfigurer {
    private final ObjectMapper objectMapper;
    private final HttpClientProperties httpClientProperties;

    /**
     * 修改默认的消息转换配置
     *
     * @param restTemplate RestTemplate实例
     */
    private void modifyDefaultMessageConverters(@NonNull RestTemplate restTemplate) {
        List<HttpMessageConverter<?>> messageConverters = restTemplate.getMessageConverters().stream()
                .map(converter -> {
                    if (converter instanceof StringHttpMessageConverter) {
                        return new StringHttpMessageConverter(StandardCharsets.UTF_8);
                    } else if (converter instanceof MappingJackson2HttpMessageConverter) {
                        MappingJackson2HttpMessageConverter mappingJackson2HttpMessageConverter =
                                new MappingJackson2HttpMessageConverter(this.objectMapper);
                        mappingJackson2HttpMessageConverter.setSupportedMediaTypes(
                                Arrays.asList(
                                        MediaType.TEXT_PLAIN,
                                        MediaType.APPLICATION_JSON,
                                        new MediaType("application", "*+json")
                                )
                        );
                        return mappingJackson2HttpMessageConverter;
                    }
                    return converter;
                }).collect(Collectors.toList());
        restTemplate.setMessageConverters(messageConverters);
    }

    @Bean
    public RestTemplate restTemplate(ClientHttpRequestFactory factory) {
        RestTemplate template = new RestTemplate(factory);
        this.modifyDefaultMessageConverters(template);
        return template;
    }

    @Bean
    public ClientHttpRequestFactory clientHttpRequestFactory() {
        HttpComponentsClientHttpRequestFactory factory = new HttpComponentsClientHttpRequestFactory(this.httpClient());
        // 连接超时
        factory.setConnectTimeout(this.httpClientProperties.getConnectTimeout());
        // 数据读取超时时间，即SocketTimeout
        factory.setReadTimeout(this.httpClientProperties.getSocketTimeout());
        // 从连接池获取请求连接的超时时间，不宜过长，必须设置，比如连接不够用时，时间过长将是灾难性的
        factory.setConnectionRequestTimeout(this.httpClientProperties.getConnectionRequestTimeout());
        return factory;
    }

    @Bean
    public HttpClient httpClient() {
        HttpClientBuilder builder = HttpClientBuilder.create();
        RegistryBuilder<ConnectionSocketFactory> registryBuilder = RegistryBuilder.create();
        registryBuilder.register("http", PlainConnectionSocketFactory.getSocketFactory());

        // 设置信任ssl访问
        try {
            SSLContext ssl = new SSLContextBuilder().loadTrustMaterial(null, (chain, type) -> true).build();
            builder.setSSLContext(ssl);
            SSLConnectionSocketFactory factory = new SSLConnectionSocketFactory(ssl, NoopHostnameVerifier.INSTANCE);
            registryBuilder.register("https", factory);
        } catch (GeneralSecurityException e) {
            log.warn("SSL setting failed: {}", e.getMessage());
        }

        // 使用Httpclient连接池的方式配置(推荐)，同时支持netty，okHttp以及其他http框架
        PoolingHttpClientConnectionManager manager = new PoolingHttpClientConnectionManager(registryBuilder.build());
        // 最大连接数
        manager.setMaxTotal(this.httpClientProperties.getMaxTotal());
        // 同路由并发数
        manager.setDefaultMaxPerRoute(this.httpClientProperties.getMaxPerRoute());
        // 配置连接池
        builder.setConnectionManager(manager);
        // 重试次数
        builder.setRetryHandler(this.httpRequestRetryHandler());
        // 设置长连接保持策略
        builder.setKeepAliveStrategy(this.connectionKeepAliveStrategy());

        // 设置后台线程剔除失效连接
        builder.evictExpiredConnections();
        builder.evictIdleConnections(this.httpClientProperties.getMaxIdleTime(), TimeUnit.MILLISECONDS);

        // 禁用Cookie回话保持功能
        builder.disableCookieManagement();
        return builder.build();
    }

    @Bean
    public HttpRequestRetryHandler httpRequestRetryHandler() {
        return new DefaultHttpRequestRetryHandler(this.httpClientProperties.getRetryCount(), true);
    }

    @Bean
    public ConnectionKeepAliveStrategy connectionKeepAliveStrategy() {
        return new DefaultConnectionKeepAliveStrategy() {
            @Override
            public long getKeepAliveDuration(HttpResponse response, HttpContext context) {
                long duration = super.getKeepAliveDuration(response, context);
                return duration < 0 ? httpClientProperties.getKeepAliveTime() : duration;
            }
        };
    }
}
