package com.chelvc.framework.base.context;

import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import java.util.Properties;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import com.beust.jcommander.internal.Sets;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.fasterxml.jackson.core.type.TypeReference;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.NoSuchBeanDefinitionException;
import org.springframework.beans.factory.config.YamlPropertiesFactoryBean;
import org.springframework.boot.autoconfigure.SpringBootApplication;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.context.annotation.ComponentScan;
import org.springframework.core.annotation.AnnotationUtils;
import org.springframework.core.env.Environment;
import org.springframework.core.io.Resource;
import org.springframework.core.io.support.PathMatchingResourcePatternResolver;
import org.springframework.core.io.support.ResourcePatternResolver;
import org.springframework.stereotype.Component;

/**
 * 应用上下文工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Component
public class ApplicationContextHolder implements ApplicationContextAware {
    /**
     * 应用名称属性
     */
    public static final String APPLICATION_NAME_PROPERTY = "spring.application.name";

    /**
     * application.yml文件搜索路径
     */
    public static final String APPLICATION_YML_PATH =
            ResourcePatternResolver.CLASSPATH_ALL_URL_PREFIX + "/application.yml";

    /**
     * 应用上下文对象
     */
    private static ApplicationContext APPLICATION_CONTEXT;

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        APPLICATION_CONTEXT = applicationContext;
    }

    /**
     * 获取应用上下文对象
     *
     * @return 应用上下文对象
     */
    public static ApplicationContext getApplicationContext() {
        return Objects.requireNonNull(APPLICATION_CONTEXT, "Application context has not been initialized");
    }

    /**
     * 获取应用环境变量信息
     *
     * @return 环境变量信息
     */
    public static Environment getEnvironment() {
        return getApplicationContext().getEnvironment();
    }

    /**
     * 获取环境变量值
     *
     * @param key 变量名称
     * @return 变量值
     */
    public static String getProperty(String key) {
        return getEnvironment().getProperty(key);
    }

    /**
     * 获取安全的环境变量值
     *
     * @param key 变量名称
     * @return 变量值
     */
    public static String getSafeProperty(String key) {
        try {
            return getProperty(key);
        } catch (Exception e) {
            log.warn("Application property get failed: {}, {}", key, e.getMessage());
        }
        return null;
    }

    /**
     * 获取环境变量值
     *
     * @param key          变量名称
     * @param defaultValue 默认值
     * @return 变量值
     */
    public static String getProperty(String key, String defaultValue) {
        return getEnvironment().getProperty(key, defaultValue);
    }

    /**
     * 获取安全的环境变量值
     *
     * @param key          变量名称
     * @param defaultValue 默认值
     * @return 变量值
     */
    public static String getSafeProperty(String key, String defaultValue) {
        try {
            return getProperty(key, defaultValue);
        } catch (Exception e) {
            log.warn("Application property get failed: {}, {}", key, e.getMessage());
        }
        return defaultValue;
    }

    /**
     * 获取环境变量值
     *
     * @param key        变量名称
     * @param targetType 变量值类型
     * @param <T>        变量值类型
     * @return 变量值
     */
    public static <T> T getProperty(String key, Class<T> targetType) {
        return getEnvironment().getProperty(key, targetType);
    }

    /**
     * 获取安全的环境变量值
     *
     * @param key        变量名称
     * @param targetType 变量值类型
     * @param <T>        变量值类型
     * @return 变量值
     */
    public static <T> T getSafeProperty(String key, Class<T> targetType) {
        try {
            return getProperty(key, targetType);
        } catch (Exception e) {
            log.warn("Application property get failed: {}, {}", key, e.getMessage());
        }
        return null;
    }

    /**
     * 获取环境变量值
     *
     * @param key        变量名称
     * @param targetType 变量值类型
     * @param <T>        变量值类型
     * @return 变量值
     */
    public static <T> T getProperty(String key, TypeReference<T> targetType) {
        return JacksonContextHolder.deserialize(getProperty(key), targetType);
    }

    /**
     * 获取安全的环境变量值
     *
     * @param key        变量名称
     * @param targetType 变量值类型
     * @param <T>        变量值类型
     * @return 变量值
     */
    public static <T> T getSafeProperty(String key, TypeReference<T> targetType) {
        try {
            return getProperty(key, targetType);
        } catch (Exception e) {
            log.warn("Application property get failed: {}, {}", key, e.getMessage());
        }
        return null;
    }

    /**
     * 获取环境变量值
     *
     * @param key          变量名称
     * @param targetType   变量值类型
     * @param defaultValue 默认值
     * @param <T>          变量值类型
     * @return 环境变量值
     */
    public static <T> T getProperty(String key, Class<T> targetType, T defaultValue) {
        return getEnvironment().getProperty(key, targetType, defaultValue);
    }

    /**
     * 获取安全的环境变量值
     *
     * @param key          变量名称
     * @param targetType   变量值类型
     * @param defaultValue 默认值
     * @param <T>          变量值类型
     * @return 环境变量值
     */
    public static <T> T getSafeProperty(String key, Class<T> targetType, T defaultValue) {
        try {
            return getProperty(key, targetType, defaultValue);
        } catch (Exception e) {
            log.warn("Application property get failed: {}, {}", key, e.getMessage());
        }
        return defaultValue;
    }

    /**
     * 获取环境变量值
     *
     * @param key          变量名称
     * @param targetType   变量值类型
     * @param defaultValue 默认值
     * @param <T>          变量值类型
     * @return 环境变量值
     */
    public static <T> T getProperty(String key, TypeReference<T> targetType, T defaultValue) {
        return ObjectUtils.ifNull(getProperty(key, targetType), defaultValue);
    }

    /**
     * 获取安全的环境变量值
     *
     * @param key          变量名称
     * @param targetType   变量值类型
     * @param defaultValue 默认值
     * @param <T>          变量值类型
     * @return 环境变量值
     */
    public static <T> T getSafeProperty(String key, TypeReference<T> targetType, T defaultValue) {
        try {
            return getProperty(key, targetType, defaultValue);
        } catch (Exception e) {
            log.warn("Application property get failed: {}, {}", key, e.getMessage());
        }
        return defaultValue;
    }

    /**
     * 获取必须的环境变量
     *
     * @param key 变量名称
     * @return 变量值
     * @throws IllegalStateException 状态异常
     */
    public static String getRequiredProperty(String key) throws IllegalStateException {
        return getEnvironment().getRequiredProperty(key);
    }

    /**
     * 获取必须的环境变量
     *
     * @param key        变量名称
     * @param targetType 变量值类型
     * @param <T>        变量值类型
     * @return 变量值
     * @throws IllegalStateException 状态异常
     */
    public static <T> T getRequiredProperty(String key, Class<T> targetType) throws IllegalStateException {
        return getEnvironment().getRequiredProperty(key, targetType);
    }

    /**
     * 获取必须的环境变量
     *
     * @param key        变量名称
     * @param targetType 变量值类型
     * @param <T>        变量值类型
     * @return 变量值
     * @throws IllegalStateException 状态异常
     */
    public static <T> T getRequiredProperty(String key, TypeReference<T> targetType) throws IllegalStateException {
        return JacksonContextHolder.deserialize(getRequiredProperty(key), targetType);
    }

    /**
     * 获取对象实例
     *
     * @param type 对象类型
     * @param <T>  对象类型泛型
     * @return 对象实例
     */
    public static <T> T getBean(Class<T> type) {
        return getBean(type, true);
    }

    /**
     * 获取可以为空的对象实例
     *
     * @param type     对象类型
     * @param required 对象实例是否必须存在
     * @param <T>      对象类型泛型
     * @return 对象实例
     */
    public static <T> T getBean(Class<T> type, boolean required) {
        try {
            return getApplicationContext().getBean(type);
        } catch (NoSuchBeanDefinitionException e) {
            if (required) {
                throw e;
            }
            return null;
        }
    }

    /**
     * 获取对象实例
     *
     * @param name 对象名称
     * @param <T>  对象类型泛型
     * @return 对象实例
     */
    public static <T> T getBean(String name) {
        return getBean(name, true);
    }

    /**
     * 获取可以为空的对象实例
     *
     * @param name     对象名称
     * @param required 对象实例是否必须存在
     * @param <T>      对象类型泛型
     * @return 对象实例
     */
    @SuppressWarnings("unchecked")
    public static <T> T getBean(String name, boolean required) {
        try {
            return (T) getApplicationContext().getBean(name);
        } catch (NoSuchBeanDefinitionException e) {
            if (required) {
                throw e;
            }
            return null;
        }
    }

    /**
     * 获取所有application.yml文件资源
     *
     * @return 资源对象数组
     */
    public static Resource[] getApplicationResources() {
        ResourcePatternResolver resourcePatternResolver = new PathMatchingResourcePatternResolver();
        try {
            return resourcePatternResolver.getResources(APPLICATION_YML_PATH);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 获取应用名称
     *
     * @return 应用名称
     */
    public static String getApplicationName() {
        return getApplicationName(getApplicationContext());
    }

    /**
     * 获取应用名称
     *
     * @param applicationContext 应用上下文实例
     * @return 应用名称
     */
    public static String getApplicationName(@NonNull ApplicationContext applicationContext) {
        return applicationContext.getEnvironment().getProperty(APPLICATION_NAME_PROPERTY);
    }

    /**
     * 根据application.yml文件资源获取应用名称
     *
     * @param resources application.yml文件资源对象实例
     * @return 应用名称
     */
    public static String getApplicationName(Resource... resources) {
        YamlPropertiesFactoryBean yaml = new YamlPropertiesFactoryBean();
        yaml.setResources(resources);
        Properties properties = Objects.requireNonNull(yaml.getObject());
        return properties.getProperty(APPLICATION_NAME_PROPERTY);
    }

    /**
     * 查找应用中的类对象
     *
     * @param condition 类对象条件判断
     * @return 类对象列表
     */
    public static List<Class<?>> lookupApplicationClasses(@NonNull Predicate<Class<?>> condition) {
        return lookupApplicationClasses(getApplicationContext(), condition);
    }

    /**
     * 查找应用中的类对象
     *
     * @param applicationContext 应用上下文对象
     * @param condition          类对象条件判断
     * @return 类对象列表
     */
    public static List<Class<?>> lookupApplicationClasses(@NonNull ApplicationContext applicationContext,
                                                          @NonNull Predicate<Class<?>> condition) {
        Set<String> packages = Sets.newHashSet();
        applicationContext.getBeansWithAnnotation(SpringBootApplication.class).forEach((key, value) -> {
            Class<?> clazz = value.getClass();

            // 加载@ComponentScan包扫描路径配置
            ComponentScan componentScan = AnnotationUtils.findAnnotation(clazz, ComponentScan.class);
            if (componentScan != null) {
                packages.addAll(Arrays.asList(componentScan.basePackages()));
            }

            // 加载@SpringBootApplication包扫描路径配置
            SpringBootApplication springBootApplication =
                    AnnotationUtils.findAnnotation(clazz, SpringBootApplication.class);
            if (springBootApplication != null) {
                packages.addAll(Arrays.asList(springBootApplication.scanBasePackages()));
            }
        });

        // 根据包扫描路径查找类对象
        if (packages.isEmpty()) {
            return ObjectUtils.lookupClasses(StringUtils.EMPTY, condition);
        }
        return packages.stream().flatMap(pack -> ObjectUtils.lookupClasses(pack, condition).stream())
                .distinct().collect(Collectors.toList());
    }
}
