package com.chelvc.framework.base.context;

import java.io.IOException;
import java.util.Objects;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.PropertyAccessor;
import com.fasterxml.jackson.core.JsonParser;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.stereotype.Component;

/**
 * Jackson上下文工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Component
public class JacksonContextHolder implements ApplicationContextAware {
    /**
     * 默认对象序列化处理器实例
     */
    private static final ObjectMapper DEFAULT_SERIALIZER =
            new ObjectMapper().setVisibility(PropertyAccessor.ALL, JsonAutoDetect.Visibility.ANY)
                    .configure(JsonParser.Feature.ALLOW_SINGLE_QUOTES, true)
                    .configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
    /**
     * 全局对象序列化处理器实例
     */
    private static ObjectMapper GLOBAL_SERIALIZER;

    /**
     * 获取全局序列化处理器实例
     *
     * @return 对象序列化处理器实例
     */
    public static ObjectMapper getGlobalSerializer() {
        return Objects.requireNonNull(GLOBAL_SERIALIZER, "Global serializer has not been initialized");
    }

    /**
     * 获取默认序列化处理器实例
     *
     * @return 对象序列化处理器实例
     */
    public static ObjectMapper getDefaultSerializer() {
        return DEFAULT_SERIALIZER;
    }

    /**
     * 对象序列化
     *
     * @param object 目标对象实例
     * @return JSON字符串
     */
    public static String serialize(Object object) {
        return serialize(getDefaultSerializer(), object);
    }

    /**
     * 对象序列化
     *
     * @param mapper 对象序列化处理器实例
     * @param object 目标对象实例
     * @return JSON字符串
     */
    public static String serialize(@NonNull ObjectMapper mapper, Object object) {
        if (object == null || object instanceof String) {
            return (String) object;
        }
        try {
            return mapper.writeValueAsString(object);
        } catch (JsonProcessingException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 对象反序列化
     *
     * @param json JSON字符串
     * @return 对象实例
     */
    public static Object deserialize(String json) {
        return deserialize(getDefaultSerializer(), json, Object.class);
    }

    /**
     * 对象反序列化
     *
     * @param mapper 对象序列化处理器实例
     * @param json   JSON字符串
     * @return 对象实例
     */
    public static Object deserialize(@NonNull ObjectMapper mapper, String json) {
        return deserialize(mapper, json, Object.class);
    }

    /**
     * 对象反序列化
     *
     * @param bytes JSON字节数组
     * @return 对象实例
     */
    public static Object deserialize(byte[] bytes) {
        return deserialize(getDefaultSerializer(), bytes, Object.class);
    }

    /**
     * 对象反序列化
     *
     * @param mapper 对象序列化处理器实例
     * @param bytes  JSON字节数组
     * @return 对象实例
     */
    public static Object deserialize(@NonNull ObjectMapper mapper, byte[] bytes) {
        return deserialize(mapper, bytes, Object.class);
    }

    /**
     * 对象反序列化
     *
     * @param json JSON字符串
     * @param type 类型对象
     * @param <T>  对象类型
     * @return 对象实例
     */
    public static <T> T deserialize(String json, Class<T> type) {
        return deserialize(getDefaultSerializer(), json, type);
    }

    /**
     * 对象反序列化
     *
     * @param mapper 对象序列化处理器实例
     * @param json   JSON字符串
     * @param type   类型对象
     * @param <T>    对象类型
     * @return 对象实例
     */
    @SuppressWarnings("unchecked")
    public static <T> T deserialize(@NonNull ObjectMapper mapper, String json, Class<T> type) {
        if (json == null || type == String.class) {
            return (T) json;
        }
        try {
            return mapper.readValue(json, type);
        } catch (JsonProcessingException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 对象反序列化
     *
     * @param json JSON字符串
     * @param type 类型引用对象
     * @param <T>  对象类型
     * @return 对象实例
     */
    public static <T> T deserialize(String json, TypeReference<T> type) {
        return deserialize(getDefaultSerializer(), json, type);
    }

    /**
     * 对象反序列化
     *
     * @param mapper 对象序列化处理器实例
     * @param json   JSON字符串
     * @param type   类型引用对象
     * @param <T>    对象类型
     * @return 对象实例
     */
    @SuppressWarnings("unchecked")
    public static <T> T deserialize(@NonNull ObjectMapper mapper, String json, TypeReference<T> type) {
        if (json == null || (type != null && type.getType() == String.class)) {
            return (T) json;
        }
        try {
            return mapper.readValue(json, type);
        } catch (JsonProcessingException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 对象反序列化
     *
     * @param bytes JSON字节数组
     * @param type  类型对象
     * @param <T>   对象类型
     * @return 对象实例
     */
    public static <T> T deserialize(byte[] bytes, Class<T> type) {
        return deserialize(getDefaultSerializer(), bytes, type);
    }

    /**
     * 对象反序列化
     *
     * @param mapper 对象序列化处理器实例
     * @param bytes  JSON字节数组
     * @param type   类型对象
     * @param <T>    对象类型
     * @return 对象实例
     */
    public static <T> T deserialize(@NonNull ObjectMapper mapper, byte[] bytes, Class<T> type) {
        if (bytes == null || type == byte[].class) {
            return (T) bytes;
        }
        try {
            return mapper.readValue(bytes, type);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 对象反序列化
     *
     * @param bytes JSON字节数组
     * @param type  类型引用对象
     * @param <T>   对象类型
     * @return 对象实例
     */
    public static <T> T deserialize(byte[] bytes, TypeReference<T> type) {
        return deserialize(getDefaultSerializer(), bytes, type);
    }

    /**
     * 对象反序列化
     *
     * @param mapper 对象序列化处理器实例
     * @param bytes  JSON字节数组
     * @param type   类型引用对象
     * @param <T>    对象类型
     * @return 对象实例
     */
    public static <T> T deserialize(@NonNull ObjectMapper mapper, byte[] bytes, TypeReference<T> type) {
        if (bytes == null || (type != null && type.getType() == byte[].class)) {
            return (T) bytes;
        }
        try {
            return mapper.readValue(bytes, type);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        GLOBAL_SERIALIZER = applicationContext.getBean(ObjectMapper.class);
    }
}
