package com.chelvc.framework.base.context;

import java.util.ArrayDeque;
import java.util.Collection;
import java.util.Deque;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.function.Function;
import javax.servlet.ServletRequestEvent;
import javax.servlet.ServletRequestListener;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.chelvc.framework.base.model.Platform;
import com.chelvc.framework.base.model.Session;
import com.chelvc.framework.base.model.Terminal;
import com.chelvc.framework.base.model.Version;
import com.chelvc.framework.base.util.HttpUtils;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import lombok.NonNull;
import org.apache.commons.lang3.ArrayUtils;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;
import org.springframework.web.context.request.RequestContextHolder;
import org.springframework.web.context.request.ServletRequestAttributes;

/**
 * 会话上下文工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
@Component
public class SessionContextHolder implements ServletRequestListener {
    /**
     * 身份标识请求头
     */
    public static final String HEADER_ID = "Id";

    /**
     * 电话号码请求头
     */
    public static final String HEADER_MOBILE = "Mobile";

    /**
     * 设备标识请求头
     */
    public static final String HEADER_DEVICE = "Device";

    /**
     * 租户标识请求头
     */
    public static final String HEADER_TENANT = "Tenant";

    /**
     * 是否首次请求头
     */
    public static final String HEADER_INITIAL = "Initial";

    /**
     * 渠道来源请求头
     */
    public static final String HEADER_CHANNEL = "Channel";

    /**
     * 应用平台请求头
     */
    public static final String HEADER_PLATFORM = "Platform";

    /**
     * 终端编号请求头
     */
    public static final String HEADER_TERMINAL = "Terminal";

    /**
     * 终端版本号请求头
     */
    public static final String HEADER_VERSION = "Version";

    /**
     * 业务类型请求头
     */
    public static final String HEADER_BUSINESS = "Business";

    /**
     * 终端序号请求头
     */
    public static final String HEADER_SEQUENCE = "Sequence";

    /**
     * 签名信息请求头
     */
    public static final String HEADER_SIGNATURE = "Signature";

    /**
     * 空会话对象实例
     */
    private static final Session EMPTY_SESSION = new Session();

    /**
     * 对象缓存映射上下文
     */
    private static final ThreadLocal<Map<Object, Object>> CACHE_CONTEXT = ThreadLocal.withInitial(HashMap::new);

    /**
     * 用户会话上下文
     */
    private static final ThreadLocal<Deque<Session>> SESSION_CONTEXT = ThreadLocal.withInitial(ArrayDeque::new);

    /**
     * 获取会话信息
     *
     * @return 会话信息
     */
    public static Session getSession() {
        return getSession(true);
    }

    /**
     * 设置会话信息
     *
     * @param session 会话信息
     */
    public static void setSession(Session session) {
        SESSION_CONTEXT.get().push(ObjectUtils.ifNull(session, EMPTY_SESSION));
    }

    /**
     * 获取会话信息
     *
     * @param required 会话是否是必须
     * @return 会话信息
     */
    public static Session getSession(boolean required) {
        Session session = SESSION_CONTEXT.get().peek();
        if (session == EMPTY_SESSION) {
            session = null;
        }
        return required ? Objects.requireNonNull(session, "Session has not been initialized") : session;
    }

    /**
     * 初始化会话信息
     *
     * @param session 会话信息
     */
    public static void initializeSession(@NonNull Session session) {
        Deque<Session> deque = SESSION_CONTEXT.get();
        deque.poll();
        deque.push(session);
    }

    /**
     * 初始化会话主体
     *
     * @param id       主体标识
     * @param business 业务类型
     * @param mobile   电话号码
     * @return 会话信息
     */
    public static Session initializeSessionPrincipal(@NonNull Long id, @NonNull String business, String mobile) {
        return initializeSessionPrincipal(id, business, mobile, true);
    }

    /**
     * 初始化会话主体
     *
     * @param id       主体标识
     * @param business 业务类型
     * @param mobile   电话号码
     * @param cover    是否覆盖已认证主体信息
     * @return 会话信息
     */
    public static Session initializeSessionPrincipal(@NonNull Long id, @NonNull String business, String mobile,
                                                     boolean cover) {
        Session session = null;
        Deque<Session> deque = null;
        if (cover || ObjectUtils.ifNull(session = (deque = SESSION_CONTEXT.get()).peek(), Session::getId) == null) {
            session = Session.builder().id(id).business(business).mobile(mobile)
                    .host(ObjectUtils.ifNull(session, Session::getHost))
                    .device(ObjectUtils.ifNull(session, Session::getDevice))
                    .tenant(ObjectUtils.ifNull(session, Session::getTenant))
                    .initial(ObjectUtils.ifNull(session, Session::getInitial))
                    .channel(ObjectUtils.ifNull(session, Session::getChannel))
                    .platform(ObjectUtils.ifNull(session, Session::getPlatform))
                    .terminal(ObjectUtils.ifNull(session, Session::getTerminal))
                    .version(ObjectUtils.ifNull(session, Session::getVersion))
                    .sequence(ObjectUtils.ifNull(session, Session::getSequence))
                    .signature(ObjectUtils.ifNull(session, Session::getSignature)).build();
            if (deque == null) {
                deque = SESSION_CONTEXT.get();
            }
            deque.poll();
            deque.push(session);
        }
        return session;
    }

    /**
     * 获取当前用户ID
     *
     * @return 用户ID
     */
    public static Long getId() {
        return getId(true);
    }

    /**
     * 获取当前用户ID
     *
     * @param requireSession 会话是否是必须
     * @return 用户ID
     */
    public static Long getId(boolean requireSession) {
        return ObjectUtils.ifNull(getSession(requireSession), Session::getId);
    }

    /**
     * 获取当前会话请求地址
     *
     * @return 请求地址
     */
    public static String getHost() {
        return getHost(true);
    }

    /**
     * 获取当前会话请求地址
     *
     * @param requireSession 会话是否是必须
     * @return 请求地址
     */
    public static String getHost(boolean requireSession) {
        return ObjectUtils.ifNull(getSession(requireSession), Session::getHost);
    }

    /**
     * 获取当前会话电话号码
     *
     * @return 电话号码
     */
    public static String getMobile() {
        return getMobile(true);
    }

    /**
     * 获取当前会话电话号码
     *
     * @param requireSession 会话是否是必须
     * @return 电话号码
     */
    public static String getMobile(boolean requireSession) {
        return ObjectUtils.ifNull(getSession(requireSession), Session::getMobile);
    }

    /**
     * 获取当前会话设备标识
     *
     * @return 设备标识
     */
    public static String getDevice() {
        return getDevice(true);
    }

    /**
     * 获取当前会话设备标识
     *
     * @param requireSession 会话是否是必须
     * @return 设备标识
     */
    public static String getDevice(boolean requireSession) {
        return ObjectUtils.ifNull(getSession(requireSession), Session::getDevice);
    }

    /**
     * 获取当前会话租户编号
     *
     * @return 租户编号
     */
    public static Integer getTenant() {
        return getTenant(true);
    }

    /**
     * 获取当前会话租户编号
     *
     * @param requireSession 会话是否是必须
     * @return 租户编号
     */
    public static Integer getTenant(boolean requireSession) {
        return ObjectUtils.ifNull(getSession(requireSession), Session::getTenant);
    }

    /**
     * 判断当前会话是否是首次请求
     *
     * @return true/false
     */
    public static boolean isInitial() {
        return isInitial(true);
    }

    /**
     * 判断当前会话是否是首次请求
     *
     * @param requireSession 会话是否是必须
     * @return true/false
     */
    public static boolean isInitial(boolean requireSession) {
        return Boolean.TRUE.equals(ObjectUtils.ifNull(getSession(requireSession), Session::getInitial));
    }

    /**
     * 获取当前会话渠道来源
     *
     * @return 渠道来源
     */
    public static String getChannel() {
        return getChannel(true);
    }

    /**
     * 获取当前会话渠道来源
     *
     * @param requireSession 会话是否是必须
     * @return 渠道来源
     */
    public static String getChannel(boolean requireSession) {
        return ObjectUtils.ifNull(getSession(requireSession), Session::getChannel);
    }

    /**
     * 获取当前会话平台信息
     *
     * @return 平台信息
     */
    public static Platform getPlatform() {
        return getPlatform(true);
    }

    /**
     * 获取当前会话平台信息
     *
     * @param requireSession 会话是否是必须
     * @return 平台信息
     */
    public static Platform getPlatform(boolean requireSession) {
        return ObjectUtils.ifNull(getSession(requireSession), Session::getPlatform);
    }

    /**
     * 获取当前会话终端信息
     *
     * @return 终端信息
     */
    public static Terminal getTerminal() {
        return getTerminal(true);
    }

    /**
     * 获取当前会话终端信息
     *
     * @param requireSession 会话是否是必须
     * @return 终端信息
     */
    public static Terminal getTerminal(boolean requireSession) {
        return ObjectUtils.ifNull(getSession(requireSession), Session::getTerminal);
    }

    /**
     * 获取当前会话版本信息
     *
     * @return 版本信息
     */
    public static String getVersion() {
        return getVersion(true);
    }

    /**
     * 获取当前会话版本信息
     *
     * @param requireSession 会话是否是必须
     * @return 版本信息
     */
    public static String getVersion(boolean requireSession) {
        return ObjectUtils.ifNull(getSession(requireSession), Session::getVersion);
    }

    /**
     * 获取当前业务类型
     *
     * @return 业务类型
     */
    public static String getBusiness() {
        return getBusiness(true);
    }

    /**
     * 获取当前业务类型
     *
     * @param requireSession 会话是否是必须
     * @return 业务类型
     */
    public static String getBusiness(boolean requireSession) {
        return ObjectUtils.ifNull(getSession(requireSession), Session::getBusiness);
    }

    /**
     * 获取当前会话签名信息
     *
     * @return 签名信息
     */
    public static String getSignature() {
        return getSignature(true);
    }

    /**
     * 获取当前会话签名信息
     *
     * @param requireSession 会话是否是必须
     * @return 签名信息
     */
    public static String getSignature(boolean requireSession) {
        return ObjectUtils.ifNull(getSession(requireSession), Session::getSignature);
    }

    /**
     * 获取当前终端序列号
     *
     * @return 序列号
     */
    public static Integer getSequence() {
        return getSequence(true);
    }

    /**
     * 获取当前终端序列号
     *
     * @param requireSession 会话是否是必须
     * @return 序列号
     */
    public static Integer getSequence(boolean requireSession) {
        return ObjectUtils.ifNull(getSession(requireSession), Session::getSequence);
    }

    /**
     * 判断当前会话是否存在手机号
     *
     * @return true/false
     */
    public static boolean hasMobile() {
        return StringUtils.nonEmpty(getMobile(false));
    }

    /**
     * 判断当前会话是否已认证
     *
     * @return true/false
     */
    public static boolean isAuthenticated() {
        return Objects.nonNull(getId(false));
    }

    /**
     * 获取缓存对象
     *
     * @param key 缓存标识
     * @param <K> 缓存键类型泛型
     * @param <V> 缓存值得类型泛型
     * @return 对象实例
     */
    public static <K, V> V getCache(@NonNull K key) {
        return getCache(key, k -> null);
    }

    /**
     * 获取缓存对象
     *
     * @param key      缓存标识
     * @param provider 对象提供方法
     * @param <K>      缓存键类型泛型
     * @param <V>      缓存值得类型泛型
     * @return 对象实例
     */
    @SuppressWarnings("unchecked")
    public static <K, V> V getCache(@NonNull K key, @NonNull Function<K, V> provider) {
        return (V) CACHE_CONTEXT.get().computeIfAbsent(key, k -> provider.apply((K) k));
    }

    /**
     * 设置缓存对象
     *
     * @param key   缓存标识
     * @param value 缓存对象
     * @param <K>   缓存键类型泛型
     * @param <V>   缓存值得类型泛型
     */
    public static <K, V> void setCache(@NonNull K key, V value) {
        CACHE_CONTEXT.get().put(key, value);
    }

    /**
     * 移除缓存对象
     *
     * @param key 缓存标识
     */
    public static void removeCache(@NonNull String key) {
        CACHE_CONTEXT.get().remove(key);
    }

    /**
     * 移除缓存对象
     *
     * @param keys 缓存标识数组
     */
    public static void removeCaches(@NonNull String... keys) {
        if (keys.length > 0) {
            Map<Object, Object> caches = CACHE_CONTEXT.get();
            for (String key : keys) {
                caches.remove(key);
            }
        }
    }

    /**
     * 移除缓存对象
     *
     * @param keys 缓存标识集合
     */
    public static void removeCaches(@NonNull Collection<String> keys) {
        if (!CollectionUtils.isEmpty(keys)) {
            Map<Object, Object> caches = CACHE_CONTEXT.get();
            keys.forEach(caches::remove);
        }
    }

    /**
     * 判断当前平台是否是指定平台
     *
     * @param platform 指定平台
     * @return true/false
     */
    public static boolean isPlatform(Platform platform) {
        return platform != null && platform == getPlatform(false);
    }

    /**
     * 判断当前终端是否是指定终端
     *
     * @param terminal 指定终端
     * @return true/false
     */
    public static boolean isTerminal(Terminal terminal) {
        return terminal != null && terminal == getTerminal(false);
    }

    /**
     * 判断当前版本是否是指定版本
     *
     * @param terminal 指定终端
     * @param version  指定版本
     * @return true/false
     */
    public static boolean isVersion(Terminal terminal, String version) {
        if (terminal == null || StringUtils.isEmpty(version)) {
            return false;
        }
        Session session = getSession(false);
        return terminal == ObjectUtils.ifNull(session, Session::getTerminal)
                && version.equals(ObjectUtils.ifNull(session, Session::getVersion));
    }

    /**
     * 判断当前版本是否和指定版本相同
     *
     * @param version 指定版本注解信息
     * @return true/false
     */
    public static boolean isVersion(com.chelvc.framework.base.annotation.Version version) {
        if (version == null) {
            return false;
        }
        Session session = getSession(false);
        String channel = ObjectUtils.ifNull(session, Session::getChannel);
        Platform platform = ObjectUtils.ifNull(session, Session::getPlatform);
        Terminal terminal = ObjectUtils.ifNull(session, Session::getTerminal);
        if ((ArrayUtils.isEmpty(version.channels()) || ArrayUtils.contains(version.channels(), channel))
                && (ArrayUtils.isEmpty(version.platforms()) || ArrayUtils.contains(version.platforms(), platform))
                && (ArrayUtils.isEmpty(version.terminals()) || ArrayUtils.contains(version.terminals(), terminal))) {
            String current = ObjectUtils.ifNull(session, Session::getVersion);
            if (StringUtils.nonEmpty(version.value())) {
                return Objects.equals(version.value(), current);
            }
            return (StringUtils.isEmpty(version.before())
                    || Version.isBefore(current, version.before(), version.contains()))
                    && (StringUtils.isEmpty(version.after())
                    || Version.isAfter(current, version.after(), version.contains()));
        }
        return false;
    }

    /**
     * 判断当前版本是否在指定版本之前
     *
     * @param terminal 指定终端
     * @param version  指定版本
     * @return true/false
     */
    public static boolean isBeforeVersion(Terminal terminal, String version) {
        return isBeforeVersion(terminal, version, false);
    }

    /**
     * 判断当前版本是否在指定版本之前
     *
     * @param terminal 指定终端
     * @param version  指定版本
     * @param contains 是否包含当前版本号
     * @return true/false
     */
    public static boolean isBeforeVersion(Terminal terminal, String version, boolean contains) {
        if (terminal == null || StringUtils.isEmpty(version)) {
            return false;
        }
        Session session = getSession(false);
        return terminal == ObjectUtils.ifNull(session, Session::getTerminal)
                && Version.isBefore(ObjectUtils.ifNull(session, Session::getVersion), version, contains);
    }

    /**
     * 判断当前版本是否在指定版本之后
     *
     * @param terminal 指定终端
     * @param version  指定版本
     * @return true/false
     */
    public static boolean isAfterVersion(Terminal terminal, String version) {
        return isAfterVersion(terminal, version, false);
    }

    /**
     * 判断当前版本是否在指定版本之后
     *
     * @param terminal 指定终端
     * @param version  指定版本
     * @param contains 是否包含当前版本号
     * @return true/false
     */
    public static boolean isAfterVersion(Terminal terminal, String version, boolean contains) {
        if (terminal == null || StringUtils.isEmpty(version)) {
            return false;
        }
        Session session = getSession(false);
        return terminal == ObjectUtils.ifNull(session, Session::getTerminal)
                && Version.isAfter(ObjectUtils.ifNull(session, Session::getVersion), version, contains);
    }

    /**
     * 判断是否是指定业务类型
     *
     * @param business 业务类型
     * @return true/false
     */
    public static boolean isBusiness(String business) {
        return StringUtils.nonEmpty(business) && Objects.equals(business, getBusiness(false));
    }

    /**
     * 判断是否是审核版本
     *
     * @return true/false
     */
    public static boolean isTesting() {
        return isTesting(null, null);
    }

    /**
     * 判断是否是审核版本
     *
     * @param platform 指定平台
     * @return true/false
     */
    public static boolean isTesting(Platform platform) {
        return isTesting(platform, null);
    }

    /**
     * 判断是否是审核版本
     *
     * @param terminal 指定终端
     * @return true/false
     */
    public static boolean isTesting(Terminal terminal) {
        return isTesting(null, terminal);
    }

    /**
     * 判断是否是审核版本
     *
     * @param platform 指定平台
     * @param terminal 指定终端
     * @return true/false
     */
    public static boolean isTesting(Platform platform, Terminal terminal) {
        Session session = getSession(false);
        if (session == null) {
            return false;
        }
        int sequence = ApplicationContextHolder.getSafeProperty("platform.test.sequence", int.class, 0);
        return (platform == null || platform == session.getPlatform())
                && (terminal == null || terminal == session.getTerminal())
                && ObjectUtils.ifNull(session.getSequence(), 0) > sequence;
    }

    /**
     * 获取当前请求对象
     *
     * @return Http请求对象
     */
    public static HttpServletRequest getRequest() {
        ServletRequestAttributes attributes = (ServletRequestAttributes) RequestContextHolder.getRequestAttributes();
        return ObjectUtils.ifNull(attributes, ServletRequestAttributes::getRequest);
    }

    /**
     * 获取当前响应对象
     *
     * @return Http响应对象
     */
    public static HttpServletResponse getResponse() {
        ServletRequestAttributes attributes = (ServletRequestAttributes) RequestContextHolder.getRequestAttributes();
        return ObjectUtils.ifNull(attributes, ServletRequestAttributes::getResponse);
    }

    /**
     * 获取请求参数
     *
     * @param name 参数名称
     * @return 参数值
     */
    public static String getParameter(@NonNull String name) {
        return getParameter(name, value -> value);
    }

    /**
     * 获取请求参数
     *
     * @param name    参数名称
     * @param adapter 参数值适配器
     * @return 参数值
     */
    public static <T> T getParameter(@NonNull String name, @NonNull Function<String, T> adapter) {
        return StringUtils.ifEmpty(ObjectUtils.ifNull(getRequest(), request -> request.getParameter(name)), adapter);
    }

    /**
     * 清空上下文信息
     */
    public static void clearSessionContext() {
        // 清空缓存
        CACHE_CONTEXT.remove();

        // 清空会话
        Deque<Session> sessions = SESSION_CONTEXT.get();
        sessions.poll();
        if (sessions.isEmpty()) {
            SESSION_CONTEXT.remove();
        }
    }

    /**
     * 获取客户端身份标识
     *
     * @param request Http请求对象
     * @return 身份标识
     */
    protected Long getId(HttpServletRequest request) {
        return StringUtils.ifEmpty(request.getHeader(HEADER_ID), Long::parseLong);
    }

    /**
     * 获取客户端主机地址
     *
     * @param request Http请求对象
     * @return 主机地址
     */
    protected String getHost(HttpServletRequest request) {
        return StringUtils.ifEmpty(HttpUtils.getRequestAddress(request), (String) null);
    }

    /**
     * 获取会话电话号码
     *
     * @param request Http请求对象
     * @return 电话号码
     */
    protected String getMobile(HttpServletRequest request) {
        return StringUtils.ifEmpty(request.getHeader(HEADER_MOBILE), (String) null);
    }

    /**
     * 获取客户端设备标识
     *
     * @param request Http请求对象
     * @return 设备标识
     */
    protected String getDevice(HttpServletRequest request) {
        return StringUtils.ifEmpty(request.getHeader(HEADER_DEVICE), (String) null);
    }

    /**
     * 获取客户端租户标识
     *
     * @param request Http请求对象
     * @return 租户标识
     */
    protected Integer getTenant(HttpServletRequest request) {
        return StringUtils.ifEmpty(request.getHeader(HEADER_TENANT), Integer::parseInt);
    }

    /**
     * 获取是否首次请求
     *
     * @param request Http请求对象
     * @return true/false
     */
    protected Boolean getInitial(HttpServletRequest request) {
        return StringUtils.ifEmpty(request.getHeader(HEADER_INITIAL), Boolean::parseBoolean);
    }

    /**
     * 获取客户端渠道来源
     *
     * @param request Http请求对象
     * @return 渠道来源
     */
    protected String getChannel(HttpServletRequest request) {
        return StringUtils.ifEmpty(request.getHeader(HEADER_CHANNEL), (String) null);
    }

    /**
     * 获取客户端平台标识
     *
     * @param request Http请求对象
     * @return 平台标识
     */
    protected Platform getPlatform(HttpServletRequest request) {
        return StringUtils.ifEmpty(request.getHeader(HEADER_PLATFORM), Platform::valueOf);
    }

    /**
     * 获取客户端终端
     *
     * @param request Http请求对象
     * @return 终端标识
     */
    protected Terminal getTerminal(HttpServletRequest request) {
        return StringUtils.ifEmpty(request.getHeader(HEADER_TERMINAL), Terminal::valueOf);
    }

    /**
     * 获取客户端版本号
     *
     * @param request Http请求对象
     * @return 版本号
     */
    protected String getVersion(HttpServletRequest request) {
        return StringUtils.ifEmpty(request.getHeader(HEADER_VERSION), (String) null);
    }

    /**
     * 获取业务类型
     *
     * @param request Http请求对象
     * @return 业务类型
     */
    protected String getBusiness(HttpServletRequest request) {
        return StringUtils.ifEmpty(request.getHeader(HEADER_BUSINESS), (String) null);
    }

    /**
     * 获取签名信息
     *
     * @param request Http请求对象
     * @return 签名信息
     */
    protected String getSignature(HttpServletRequest request) {
        return StringUtils.ifEmpty(request.getHeader(HEADER_SIGNATURE), (String) null);
    }

    /**
     * 获取客户端序号
     *
     * @param request Http请求对象
     * @return 客户端序号
     */
    protected Integer getSequence(HttpServletRequest request) {
        return StringUtils.ifEmpty(request.getHeader(HEADER_SEQUENCE), Integer::parseInt);
    }

    @Override
    public void requestInitialized(ServletRequestEvent event) {
        HttpServletRequest request = (HttpServletRequest) event.getServletRequest();
        Session session = Session.builder().id(this.getId(request)).host(this.getHost(request))
                .mobile(this.getMobile(request)).device(this.getDevice(request)).tenant(this.getTenant(request))
                .initial(this.getInitial(request)).channel(this.getChannel(request)).platform(this.getPlatform(request))
                .terminal(this.getTerminal(request)).version(this.getVersion(request))
                .business(this.getBusiness(request)).sequence(this.getSequence(request))
                .signature(this.getSignature(request)).build();
        setSession(session);
    }

    @Override
    public void requestDestroyed(ServletRequestEvent event) {
        clearSessionContext();
    }
}
