package com.chelvc.framework.base.interceptor;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.CompletionException;
import java.util.stream.Collectors;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.ValidationException;

import com.chelvc.framework.base.exception.BusinessException;
import com.chelvc.framework.base.exception.ParameterInvalidException;
import com.chelvc.framework.base.exception.ResourceInvalidException;
import com.chelvc.framework.base.exception.ResourceUnavailableException;
import com.chelvc.framework.base.model.Result;
import com.chelvc.framework.base.util.HttpUtils;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.exc.InvalidFormatException;
import com.google.common.collect.ImmutableMap;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.TypeMismatchException;
import org.springframework.boot.web.servlet.error.ErrorController;
import org.springframework.http.HttpStatus;
import org.springframework.http.converter.HttpMessageConversionException;
import org.springframework.http.converter.HttpMessageNotReadableException;
import org.springframework.stereotype.Controller;
import org.springframework.util.CollectionUtils;
import org.springframework.validation.BindException;
import org.springframework.validation.BindingResult;
import org.springframework.validation.FieldError;
import org.springframework.web.HttpMediaTypeNotAcceptableException;
import org.springframework.web.HttpMediaTypeNotSupportedException;
import org.springframework.web.HttpRequestMethodNotSupportedException;
import org.springframework.web.bind.MethodArgumentNotValidException;
import org.springframework.web.bind.ServletRequestBindingException;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestControllerAdvice;
import org.springframework.web.method.annotation.MethodArgumentTypeMismatchException;
import org.springframework.web.multipart.MultipartException;
import org.springframework.web.multipart.support.MissingServletRequestPartException;

/**
 * 异常处理拦截器
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Controller
@RestControllerAdvice
public class GlobalExceptionInterceptor implements ErrorController {
    /**
     * 参数绑定结果格式化
     *
     * @param result 参数绑定结果
     * @return 参数名称/异常消息映射表
     */
    private Map<String, String> format(BindingResult result) {
        List<FieldError> errors = ObjectUtils.ifNull(result, BindingResult::getFieldErrors);
        if (CollectionUtils.isEmpty(errors)) {
            return Collections.emptyMap();
        }
        return errors.stream().collect(Collectors.toMap(FieldError::getField,
                e -> ObjectUtils.ifNull(e.getDefaultMessage(), StringUtils.EMPTY), (o, n) -> n));
    }

    /**
     * 参数校验异常格式化
     *
     * @param e 参数校验异常
     * @return 参数名称/异常消息映射表
     */
    private Map<String, String> format(ConstraintViolationException e) {
        Set<ConstraintViolation<?>> errors =
                ObjectUtils.ifNull(e, ConstraintViolationException::getConstraintViolations);
        if (CollectionUtils.isEmpty(errors)) {
            return Collections.emptyMap();
        }
        return errors.stream().filter(Objects::nonNull).collect(Collectors.toMap(
                error -> {
                    String path = String.valueOf(error.getPropertyPath());
                    int index = path.lastIndexOf('.');
                    return index > -1 ? path.substring(index + 1) : path;
                },
                error -> ObjectUtils.ifNull(error.getMessage(), StringUtils.EMPTY)
        ));
    }

    /**
     * 参数类型异常格式化
     *
     * @param e 参数类型异常
     * @return 参数名称/异常消息映射表
     */
    private Map<String, String> format(MethodArgumentTypeMismatchException e) {
        return ImmutableMap.of(e.getName(),
                ObjectUtils.ifNull(e.getRequiredType(), Class::getSimpleName, () -> StringUtils.EMPTY));
    }

    /**
     * 参数类型异常格式化
     *
     * @param e 参数校验异常
     * @return 参数名称/异常消息映射表
     */
    private Map<String, String> format(HttpMessageNotReadableException e) {
        Throwable cause = e.getCause();
        if (cause instanceof InvalidFormatException) {
            List<JsonMappingException.Reference> references = ((InvalidFormatException) cause).getPath();
            if (!CollectionUtils.isEmpty(references)) {
                return references.stream().collect(Collectors.toMap(
                        JsonMappingException.Reference::getFieldName, r -> "Invalid format"
                ));
            }
        }
        return Collections.emptyMap();
    }

    /**
     * 将异常转换成结果对象
     *
     * @param e 异常对象
     * @return 结果对象
     */
    private Result<?> error2result(Throwable e) {
        if (e instanceof CompletionException) {
            return this.error2result(e.getCause());
        } else if (e instanceof HttpRequestMethodNotSupportedException) {
            return Result.build(HttpStatus.METHOD_NOT_ALLOWED);
        } else if (e instanceof HttpMediaTypeNotAcceptableException) {
            return Result.build(HttpStatus.NOT_ACCEPTABLE);
        } else if (e instanceof ResourceUnavailableException) {
            String code = ((ResourceUnavailableException) e).getCode();
            return Result.build(HttpStatus.NOT_ACCEPTABLE, code, e.getMessage());
        } else if (e instanceof HttpMediaTypeNotSupportedException) {
            return Result.build(HttpStatus.UNSUPPORTED_MEDIA_TYPE);
        } else if (e instanceof HttpMessageNotReadableException) {
            Map<String, String> errors = this.format((HttpMessageNotReadableException) e);
            return CollectionUtils.isEmpty(errors) ? Result.build(HttpStatus.BAD_REQUEST) :
                    Result.build(HttpStatus.PRECONDITION_FAILED, errors);
        } else if (e instanceof HttpMessageConversionException || e instanceof ServletRequestBindingException
                || e instanceof MultipartException) {
            return Result.build(HttpStatus.BAD_REQUEST);
        } else if (e instanceof ResourceInvalidException) {
            String code = ((ResourceInvalidException) e).getCode();
            return Result.build(HttpStatus.NOT_FOUND, code, e.getMessage());
        } else if (e instanceof BindException) {
            return Result.build(HttpStatus.PRECONDITION_FAILED, this.format(((BindException) e).getBindingResult()));
        } else if (e instanceof ConstraintViolationException) {
            return Result.build(HttpStatus.PRECONDITION_FAILED, this.format((ConstraintViolationException) e));
        } else if (e instanceof ValidationException) {
            return Result.build(HttpStatus.BAD_REQUEST, null, e.getMessage());
        } else if (e instanceof MethodArgumentNotValidException) {
            return Result.build(HttpStatus.PRECONDITION_FAILED,
                    this.format(((MethodArgumentNotValidException) e).getBindingResult()));
        } else if (e instanceof MethodArgumentTypeMismatchException) {
            return Result.build(HttpStatus.PRECONDITION_FAILED, this.format((MethodArgumentTypeMismatchException) e));
        } else if (e instanceof TypeMismatchException) {
            return Result.build(HttpStatus.BAD_REQUEST);
        } else if (e instanceof ParameterInvalidException) {
            return Result.build(HttpStatus.PRECONDITION_FAILED, ((ParameterInvalidException) e).getErrors());
        } else if (e instanceof MissingServletRequestPartException) {
            return Result.build(HttpStatus.PRECONDITION_FAILED, ImmutableMap.of(
                    ((MissingServletRequestPartException) e).getRequestPartName(), e.getMessage()
            ));
        } else if (e instanceof BusinessException) {
            return Result.build(((BusinessException) e).getStatus(), ((BusinessException) e).getData(), e.getMessage());
        }
        return Result.build(HttpStatus.INTERNAL_SERVER_ERROR);
    }

    /**
     * 打印异常日志
     *
     * @param request   Http请求对象
     * @param status    Http状态枚举
     * @param exception 异常对象
     */
    private void log(@NonNull HttpServletRequest request, @NonNull HttpStatus status, @NonNull Exception exception) {
        String message = exception.getMessage();
        if (StringUtils.isEmpty(message)) {
            message = status.getReasonPhrase();
        }
        message = HttpUtils.getExceptionMessage(request, status, message);
        if (status.is4xxClientError()) {
            log.warn(message);
        } else {
            log.error(message, exception);
        }
    }

    @Override
    public String getErrorPath() {
        // 该方法已过时，框架并未使用该方法，这里仅为了实现接口中的方法
        return null;
    }

    /**
     * 异常方法
     *
     * @param request   Http请求对象
     * @param response  Http响应对象
     * @param exception 异常对象
     * @return 异常结果
     */
    @RequestMapping("${server.error.path:/error}")
    public Object error(HttpServletRequest request, HttpServletResponse response, Exception exception) {
        HttpStatus status = HttpUtils.getStatus(response.getStatus());
        this.log(request, status, exception);
        HttpUtils.initializeUnifiedResponseHeader(response);
        return Result.build(status);
    }

    /**
     * 异常转换
     *
     * @param request   Http请求对象
     * @param response  Http响应对象
     * @param exception 异常对象
     * @return 请求结果
     */
    @ExceptionHandler(Exception.class)
    public Object exception(HttpServletRequest request, HttpServletResponse response, Exception exception) {
        Result<?> result = this.error2result(exception);
        HttpStatus status = HttpUtils.getStatus(result.getCode());
        this.log(request, status, exception);
        response.setStatus(status.value());
        HttpUtils.initializeUnifiedResponseHeader(response);
        return result;
    }
}