package com.chelvc.framework.base.interceptor;

import java.util.Arrays;
import java.util.Set;
import javax.servlet.http.HttpServletRequest;

import ch.qos.logback.classic.Level;
import ch.qos.logback.classic.Logger;
import ch.qos.logback.classic.LoggerContext;
import ch.qos.logback.classic.turbo.TurboFilter;
import ch.qos.logback.core.spi.FilterReply;
import com.chelvc.framework.base.context.ApplicationContextHolder;
import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.base.util.HttpUtils;
import lombok.extern.slf4j.Slf4j;
import org.aspectj.lang.JoinPoint;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.Signature;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.reflect.MethodSignature;
import org.slf4j.Marker;
import org.slf4j.impl.StaticLoggerBinder;
import org.springframework.beans.factory.SmartInitializingSingleton;
import org.springframework.boot.web.servlet.error.ErrorController;
import org.springframework.core.annotation.Order;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;
import org.springframework.web.bind.annotation.ExceptionHandler;

/**
 * 请求日志拦截器
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Aspect
@Component
@Order(Integer.MIN_VALUE)
public class RequestLoggingInterceptor implements SmartInitializingSingleton {
    /**
     * 判断是否可记录日志
     *
     * @param point 接口方法拦截点
     * @return true/false
     */
    private boolean isPrintable(JoinPoint point) {
        Object target = point.getTarget();
        if (target instanceof ErrorController) {
            return false;
        }
        Signature signature = point.getSignature();
        return signature instanceof MethodSignature
                && !((MethodSignature) signature).getMethod().isAnnotationPresent(ExceptionHandler.class);
    }

    /**
     * 接口请求日志打印
     *
     * @param point 接口方法拦截点
     * @throws Throwable 接口调用异常
     */
    @Around("@within(org.springframework.stereotype.Controller) " +
            "|| @within(org.springframework.web.bind.annotation.RestController)")
    public Object logging(ProceedingJoinPoint point) throws Throwable {
        HttpServletRequest request = SessionContextHolder.getRequest();
        boolean printable = this.isPrintable(point) && log.isDebugEnabled();
        if (printable) {
            log.debug(HttpUtils.message(request, "REQUEST", Arrays.toString(point.getArgs())));
        }
        Object value = point.proceed();
        if (printable) {
            log.debug(HttpUtils.message(request, "RESPONSE", String.valueOf(value)));
        }
        return value;
    }

    @Override
    public void afterSingletonsInstantiated() {
        // 添加自定义日志级别判断拦截器
        LoggerContext context = (LoggerContext) StaticLoggerBinder.getSingleton().getLoggerFactory();
        context.getTurboFilterList().add(new TurboFilter() {
            @Override
            public FilterReply decide(Marker marker, Logger logger, Level level, String format, Object[] params,
                                      Throwable t) {
                // 排除Spring的相关类的日志级别判断，防止出现递归
                if (logger.getName().startsWith("org.springframework")) {
                    return FilterReply.NEUTRAL;
                }

                // 根据对象名称获取日志级别
                String key = "platform.logging." + logger.getName();
                Set<?> levels = ApplicationContextHolder.getSafeProperty(key, Set.class);
                if (CollectionUtils.isEmpty(levels)
                        || levels.stream().map(Object::toString).noneMatch(level.levelStr::equalsIgnoreCase)) {
                    return FilterReply.NEUTRAL;
                }
                return FilterReply.ACCEPT;
            }
        });
    }
}
