package com.chelvc.framework.base.interceptor;

import java.util.Objects;
import javax.servlet.http.HttpServletResponse;

import com.chelvc.framework.base.annotation.UnifiedResponseBody;
import com.chelvc.framework.base.util.HttpUtils;
import lombok.NonNull;
import org.springframework.core.MethodParameter;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.context.request.NativeWebRequest;
import org.springframework.web.method.support.HandlerMethodReturnValueHandler;
import org.springframework.web.method.support.ModelAndViewContainer;
import org.springframework.web.servlet.mvc.method.annotation.RequestResponseBodyMethodProcessor;

/**
 * 响应结果统一包装拦截器，默认不包装
 *
 * @author Woody
 * @date 2023/4/5
 */
public class UnifiedResponseInterceptor implements HandlerMethodReturnValueHandler {
    private final RequestResponseBodyMethodProcessor processor;
    private final UnifiedResponseWrapper wrapper;

    public UnifiedResponseInterceptor(@NonNull RequestResponseBodyMethodProcessor processor,
                                      @NonNull UnifiedResponseWrapper wrapper) {
        this.processor = processor;
        this.wrapper = wrapper;
    }

    /**
     * 判断目标方法是否使用@ResponseBody注解
     *
     * @param method 目标方法参数对象
     * @return true/false
     */
    private boolean isResponseBodyMethod(MethodParameter method) {
        return method.hasMethodAnnotation(ResponseBody.class)
                || method.getDeclaringClass().isAnnotationPresent(ResponseBody.class)
                || method.getDeclaringClass().isAnnotationPresent(RestController.class);
    }

    /**
     * 判断目标方法是否使用{@link UnifiedResponseBody}
     *
     * @param method 目标方法参数对象
     * @return true/false
     */
    private boolean isUnifiedResponseMethod(MethodParameter method) {
        UnifiedResponseBody unified = method.getMethodAnnotation(UnifiedResponseBody.class);
        if (unified == null) {
            unified = method.getDeclaringClass().getAnnotation(UnifiedResponseBody.class);
        }
        return unified != null && unified.value();
    }

    @Override
    public boolean supportsReturnType(MethodParameter method) {
        return this.isResponseBodyMethod(method) && this.isUnifiedResponseMethod(method);
    }

    @Override
    public void handleReturnValue(Object value, MethodParameter method, ModelAndViewContainer container,
                                  NativeWebRequest request) throws Exception {
        // 设置统一封装返回对象请求头
        HttpServletResponse response = Objects.requireNonNull(request.getNativeResponse(HttpServletResponse.class));
        HttpUtils.initializeUnifiedResponseHeader(response);

        // 统一包转请求结果
        value = this.wrapper.wrap(request, method, value);

        // 处理请求结果
        this.processor.handleReturnValue(value, method, container, request);
    }
}
