package com.chelvc.framework.base.jackson;

import java.io.IOException;
import java.math.BigDecimal;

import com.chelvc.framework.base.annotation.Decimal;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.BeanProperty;
import com.fasterxml.jackson.databind.JsonSerializer;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.ser.ContextualSerializer;
import com.fasterxml.jackson.databind.ser.std.StdSerializer;
import lombok.NonNull;

/**
 * Jackson小数序列化处理器
 *
 * @author Woody
 * @date 2023/4/5
 */
public class JacksonDecimalSerializer extends StdSerializer<Number> implements ContextualSerializer {
    /**
     * 默认小数保留位数
     */
    public static final int DEFAULT_SCALE = 2;

    /**
     * 小数格式化注解实例
     */
    private Decimal decimal;

    public JacksonDecimalSerializer() {
        super(Number.class);
    }

    public JacksonDecimalSerializer(@NonNull Decimal decimal) {
        super(Number.class);
        this.decimal = decimal;
    }

    /**
     * 判断是否是小数属性
     *
     * @param property 属性对象
     * @return true/false
     */
    private boolean isDecimalProperty(BeanProperty property) {
        return Float.class.isAssignableFrom(property.getType().getRawClass())
                || Double.class.isAssignableFrom(property.getType().getRawClass())
                || BigDecimal.class.isAssignableFrom(property.getType().getRawClass());
    }

    @Override
    public void serialize(Number number, JsonGenerator generator, SerializerProvider serializer)
            throws IOException {
        if (this.decimal == null || number == null) {
            generator.writeObject(number);
        } else {
            BigDecimal decimal = number instanceof BigDecimal ? (BigDecimal) number : new BigDecimal(number.toString());
            decimal = decimal.setScale(this.decimal.scale(), this.decimal.mode());
            if (this.decimal.trimZeros()) {
                decimal = decimal.stripTrailingZeros();
            }
            generator.writeNumber(decimal.toPlainString());
        }
    }

    @Override
    public JsonSerializer<?> createContextual(SerializerProvider serializer, BeanProperty property) {
        if (property == null) {
            return serializer.getDefaultNullValueSerializer();
        } else if (!this.isDecimalProperty(property)) {
            return this;
        }
        Decimal decimal = property.getAnnotation(Decimal.class);
        return decimal == null ? this : new JacksonDecimalSerializer(decimal);
    }
}
