package com.chelvc.framework.base.jackson;

import java.io.IOException;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import com.chelvc.framework.base.config.JacksonConfigurer;
import com.chelvc.framework.base.util.StringUtils;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.BeanDescription;
import com.fasterxml.jackson.databind.JsonSerializer;
import com.fasterxml.jackson.databind.SerializationConfig;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.ser.BeanPropertyWriter;
import com.fasterxml.jackson.databind.ser.BeanSerializerModifier;

/**
 * Jackson null对象序列化处理器更新处理实现
 *
 * @author Woody
 * @date 2023/4/5
 * @see JacksonConfigurer
 */
public class JacksonNullSerializerModifier extends BeanSerializerModifier {
    /**
     * 空字符串序列化处理
     */
    private static final JsonSerializer<Object> STRING_NULL_SERIALIZER = new JsonSerializer<Object>() {
        @Override
        public void serialize(Object value, JsonGenerator generator, SerializerProvider provider) throws IOException {
            if (value == null) {
                generator.writeString(StringUtils.EMPTY);
            } else {
                generator.writeObject(value);
            }
        }
    };

    /**
     * 空字典序列化处理器
     */
    private static final JsonSerializer<Object> MAP_NULL_SERIALIZER = new JsonSerializer<Object>() {
        @Override
        public void serialize(Object value, JsonGenerator generator, SerializerProvider provider) throws IOException {
            if (value == null) {
                generator.writeStartObject();
                generator.writeEndObject();
            } else {
                generator.writeObject(value);
            }
        }
    };

    /**
     * 空集合序列化处理器
     */
    private static final JsonSerializer<Object> COLLECTION_NULL_SERIALIZER = new JsonSerializer<Object>() {
        @Override
        public void serialize(Object value, JsonGenerator generator, SerializerProvider provider) throws IOException {
            if (value == null) {
                generator.writeStartArray();
                generator.writeEndArray();
            } else {
                generator.writeObject(value);
            }
        }
    };

    @Override
    public List<BeanPropertyWriter> changeProperties(SerializationConfig config, BeanDescription bean,
                                                     List<BeanPropertyWriter> writers) {
        writers.forEach(writer -> {
            Class<?> clazz = writer.getType().getRawClass();
            if (String.class.isAssignableFrom(clazz)) {
                writer.assignNullSerializer(STRING_NULL_SERIALIZER);
            } else if (Map.class.isAssignableFrom(clazz)) {
                writer.assignNullSerializer(MAP_NULL_SERIALIZER);
            } else if (clazz.isArray() || Collection.class.isAssignableFrom(clazz)) {
                writer.assignNullSerializer(COLLECTION_NULL_SERIALIZER);
            }
        });
        return writers;
    }
}
