package com.chelvc.framework.base.jackson;

import java.io.IOException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.Date;

import com.chelvc.framework.base.config.JacksonConfigurer;
import com.chelvc.framework.base.model.Enumerable;
import com.chelvc.framework.base.model.Paging;
import com.chelvc.framework.base.model.Period;
import com.chelvc.framework.base.model.Region;
import com.chelvc.framework.base.util.DateUtils;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.core.JsonParser;
import com.fasterxml.jackson.databind.DeserializationContext;
import com.fasterxml.jackson.databind.JsonDeserializer;
import com.fasterxml.jackson.databind.JsonSerializer;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.module.SimpleModule;

/**
 * Jackson序列化处理模块
 * <p>
 * 对JS最大数字支持的处理，JS数字最大安全值为2的53次方
 * <p>
 * Xss过滤
 *
 * @author Woody
 * @date 2023/4/5
 * @see JacksonConfigurer
 */
public class JacksonSerializeModule extends SimpleModule {
    /**
     * JS最大整型数（2的53次方）
     */
    private static final BigInteger JS_MAX_INTEGER_VALUE = BigInteger.valueOf(2).pow(53);

    /**
     * JS最大小数（2的53次方）
     */
    private static final BigDecimal JS_MAX_DECIMAL_VALUE = BigDecimal.valueOf(2).pow(53);

    {
        // 设置整数序列化处理器
        this.addSerializer(new JsonSerializer<Long>() {
            @Override
            public Class<Long> handledType() {
                return Long.class;
            }

            @Override
            public void serialize(Long number, JsonGenerator generator, SerializerProvider serializer)
                    throws IOException {
                if (isSafelyNumber(number)) {
                    generator.writeNumber(number);
                } else {
                    generator.writeString(number.toString());
                }
            }
        });

        // 设置浮点数序列化处理器
        this.addSerializer(new JsonSerializer<Double>() {
            @Override
            public Class<Double> handledType() {
                return Double.class;
            }

            @Override
            public void serialize(Double number, JsonGenerator generator, SerializerProvider serializer)
                    throws IOException {
                if (isSafelyNumber(number)) {
                    generator.writeNumber(number);
                } else {
                    generator.writeString(number.toString());
                }
            }
        });

        // 设置大整数序列化处理器
        this.addSerializer(new JsonSerializer<BigInteger>() {
            @Override
            public Class<BigInteger> handledType() {
                return BigInteger.class;
            }

            @Override
            public void serialize(BigInteger number, JsonGenerator generator, SerializerProvider serializer)
                    throws IOException {
                if (isSafelyNumber(number)) {
                    generator.writeNumber(number);
                } else {
                    generator.writeString(number.toString());
                }
            }
        });

        // 设置大浮点数序列化处理器
        this.addSerializer(new JsonSerializer<BigDecimal>() {
            @Override
            public Class<BigDecimal> handledType() {
                return BigDecimal.class;
            }

            @Override
            public void serialize(BigDecimal number, JsonGenerator generator, SerializerProvider serializer)
                    throws IOException {
                if (isSafelyNumber(number)) {
                    generator.writeNumber(number);
                } else {
                    generator.writeString(number.toPlainString());
                }
            }
        });

        // 设置本地日期序列化处理器
        this.addSerializer(new JsonSerializer<LocalDate>() {
            @Override
            public Class<LocalDate> handledType() {
                return LocalDate.class;
            }

            @Override
            public void serialize(LocalDate date, JsonGenerator generator, SerializerProvider serializer)
                    throws IOException {
                generator.writeObject(DateUtils.convert(date));
            }
        });

        // 设置本地日期时间处理器
        this.addSerializer(new JsonSerializer<LocalDateTime>() {
            @Override
            public Class<LocalDateTime> handledType() {
                return LocalDateTime.class;
            }

            @Override
            public void serialize(LocalDateTime datetime, JsonGenerator generator, SerializerProvider serializer)
                    throws IOException {
                generator.writeObject(DateUtils.convert(datetime));
            }
        });

        // 设置时间周期序列化处理器
        this.addSerializer(new JsonSerializer<Period>() {
            @Override
            public Class<Period> handledType() {
                return Period.class;
            }

            @Override
            public void serialize(Period period, JsonGenerator generator, SerializerProvider serializer)
                    throws IOException {
                String begin = ObjectUtils.ifNull(period.getBegin(), DateUtils::format, () -> StringUtils.EMPTY);
                String end = ObjectUtils.ifNull(period.getEnd(), DateUtils::format, () -> StringUtils.EMPTY);
                generator.writeString(String.format("%s,%s", begin, end));
            }
        });

        // 设置数据分页序列化处理器
        this.addSerializer(new JsonSerializer<Paging>() {
            @Override
            public Class<Paging> handledType() {
                return Paging.class;
            }

            @Override
            public void serialize(Paging paging, JsonGenerator generator, SerializerProvider serializer)
                    throws IOException {
                generator.writeString(String.format("%d,%d", paging.getNumber(), paging.getSize()));
            }
        });

        // 设置地区序列化处理器
        this.addSerializer(new JsonSerializer<Region>() {
            @Override
            public Class<Region> handledType() {
                return Region.class;
            }

            @Override
            public void serialize(Region region, JsonGenerator generator, SerializerProvider serializer)
                    throws IOException {
                generator.writeNumber(region.getCode());
            }
        });

        // 设置枚举对象序列化处理器
        this.addSerializer(new JsonSerializer<Enumerable>() {
            @Override
            public Class<Enumerable> handledType() {
                return Enumerable.class;
            }

            @Override
            public void serialize(Enumerable enumerable, JsonGenerator generator, SerializerProvider serializer)
                    throws IOException {
                generator.writeObject(enumerable.structure());
            }
        });

        // 设置时间反序列化处理器
        this.addDeserializer(Date.class, new JsonDeserializer<Date>() {
            @Override
            public Class<?> handledType() {
                return Date.class;
            }

            @Override
            public Date deserialize(JsonParser parser, DeserializationContext context) throws IOException {
                return DateUtils.parse(parser.getValueAsString());
            }
        });

        // 设置时间周期反序列化处理器
        this.addDeserializer(Period.class, new JsonDeserializer<Period>() {
            @Override
            public Class<?> handledType() {
                return Period.class;
            }

            @Override
            public Period deserialize(JsonParser parser, DeserializationContext context) throws IOException {
                return Period.parse(parser.getValueAsString());
            }
        });

        // 设置数据分页反序列化处理器
        this.addDeserializer(Paging.class, new JsonDeserializer<Paging>() {
            @Override
            public Class<?> handledType() {
                return Paging.class;
            }

            @Override
            public Paging deserialize(JsonParser parser, DeserializationContext context) throws IOException {
                return Paging.parse(parser.getValueAsString());
            }
        });

        // 设置地区反序列化处理器
        this.addDeserializer(Region.class, new JsonDeserializer<Region>() {
            @Override
            public Class<?> handledType() {
                return Region.class;
            }

            @Override
            public Region deserialize(JsonParser parser, DeserializationContext context) throws IOException {
                return Region.parse(parser.getValueAsString());
            }
        });
    }

    /**
     * 判断数字是否安全
     *
     * @param number 整数
     * @return true/false
     */
    private static boolean isSafelyNumber(Long number) {
        return number == null || isSafelyNumber(BigInteger.valueOf(number));
    }

    /**
     * 判断数字是否安全
     *
     * @param number 浮点数
     * @return true/false
     */
    private static boolean isSafelyNumber(Double number) {
        return number == null || isSafelyNumber(BigDecimal.valueOf(number));
    }

    /**
     * 判断数字是否安全
     *
     * @param number 大整数
     * @return true/false
     */
    private static boolean isSafelyNumber(BigInteger number) {
        return number == null || number.compareTo(JS_MAX_INTEGER_VALUE) <= 0;
    }

    /**
     * 判断数字是否安全
     *
     * @param number 大浮点数
     * @return true/false
     */
    private static boolean isSafelyNumber(BigDecimal number) {
        return number == null || number.compareTo(JS_MAX_DECIMAL_VALUE) <= 0;
    }
}
