package com.chelvc.framework.base.model;

import java.io.Serializable;
import java.util.List;

import com.chelvc.framework.base.util.StringUtils;
import com.fasterxml.jackson.annotation.JsonIgnore;
import lombok.AllArgsConstructor;
import lombok.Builder;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.ToString;
import org.springframework.util.Assert;
import org.springframework.util.CollectionUtils;

/**
 * 分页信息
 *
 * @author Woody
 * @date 2023/4/5
 */
@Getter
@Builder
@ToString
@EqualsAndHashCode
@NoArgsConstructor
@AllArgsConstructor
public final class Paging implements Serializable {
    /**
     * 默认页码
     */
    public static final int DEFAULT_NUMBER = 1;

    /**
     * 默认页面大小
     */
    public static final int DEFAULT_SIZE = 10;

    /**
     * 页码
     */
    private int number;

    /**
     * 页面大小
     */
    private int size;

    /**
     * 是否查询总数
     */
    private boolean counting = true;

    /**
     * 构建分页信息
     *
     * @param size 页面大小
     * @return 分页实例
     */
    public static Paging of(int size) {
        return of(size, true);
    }

    /**
     * 构建分页信息
     *
     * @param size     页面大小
     * @param counting 是否查询总数
     * @return 分页实例
     */
    public static Paging of(int size, boolean counting) {
        return of(1, size, counting);
    }

    /**
     * 构建分页信息
     *
     * @param number 页码
     * @param size   页面大小
     * @return 分页实例
     */
    public static Paging of(int number, int size) {
        return of(number, size, true);
    }

    /**
     * 构建分页信息
     *
     * @param number   页码
     * @param size     页面大小
     * @param counting 是否查询总数
     * @return 分页实例
     */
    public static Paging of(int number, int size, boolean counting) {
        Assert.isTrue(number > 0, "Page's number must be greater than 0");
        Assert.isTrue(size > 0, "Page's size must be greater than 0");
        Paging paging = new Paging();
        paging.number = number;
        paging.size = size;
        paging.counting = counting;
        return paging;
    }

    /**
     * 将字符串转换成分页信息
     * <p>
     * 字符串以","号作为页码/页面大小的分隔符，如果没有分隔符则默认匹配页码
     * <p>
     * 示例：(1, 10)、(, 10)、(1, )
     *
     * @param text 分页信息字符串格式
     * @return 分页信息实例
     */
    public static Paging parse(String text) {
        return StringUtils.decompose(text, (first, second) -> first == null ? of(DEFAULT_NUMBER, second) :
                second == null ? of(first, DEFAULT_SIZE) : of(first, second), Integer::parseInt);
    }

    /**
     * 获取分页列表
     *
     * @param objects 对象列表
     * @param page    页码
     * @param size    页面大小
     * @param <T>     数据类型
     * @return 分页后的对象列表
     */
    public static <T> List<T> list(List<T> objects, int page, int size) {
        Assert.isTrue(page > 0, "page must be greater than 0");
        Assert.isTrue(size > 0, "size must be greater than 0");
        if (CollectionUtils.isEmpty(objects)) {
            return objects;
        }
        int index = (page - 1) * size;
        return objects.subList(Math.min(index, objects.size()), Math.min(index + size, objects.size()));
    }

    /**
     * 获取数据开始下标
     *
     * @return 下标数字
     */
    @JsonIgnore
    public int getIndex() {
        return (this.number - 1) * this.size;
    }

    /**
     * 获取分页列表
     *
     * @param objects 对象列表
     * @param <T>     数据类型
     * @return 分页后的对象列表
     */
    public <T> List<T> list(List<T> objects) {
        return list(objects, this.number, this.size);
    }
}
