package com.chelvc.framework.base.model;

import java.io.Serializable;
import java.time.Duration;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.Month;
import java.util.Calendar;
import java.util.Date;
import java.util.function.Function;
import java.util.regex.Pattern;

import com.chelvc.framework.base.util.DateUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.fasterxml.jackson.annotation.JsonIgnore;
import lombok.AllArgsConstructor;
import lombok.Builder;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.NonNull;
import lombok.ToString;
import org.springframework.util.Assert;

/**
 * 时间周期对象
 *
 * @author Woody
 * @date 2023/4/5
 */
@Getter
@Builder
@ToString
@EqualsAndHashCode
@NoArgsConstructor
@AllArgsConstructor
public final class Period implements Serializable {
    /**
     * 自定义时间匹配模式
     */
    private static final Pattern TIME_PATTERN = StringUtils.getPattern(String.format(
            "^(%s) *\\( *(-?[0-9]+[%s])? *\\)$", StringUtils.join(Eval.values(), "|"),
            StringUtils.join(Unit.values())
    ));

    /**
     * 开始时间
     */
    private Date begin;

    /**
     * 结束时间
     */
    private Date end;

    /**
     * 构建时间周期
     *
     * @param duration 持续时长
     * @return 时间周期实例
     */
    public static Period of(@NonNull Duration duration) {
        long timestamp = System.currentTimeMillis();
        return of(new Date(timestamp), new Date(timestamp + duration.toMillis()));
    }

    /**
     * 构建时间周期
     *
     * @param field  加减字段
     * @param amount 加减数量
     * @return 时间周期实例
     */
    public static Period of(int field, int amount) {
        Calendar calendar = Calendar.getInstance();
        Date begin = calendar.getTime();
        calendar.add(field, amount);
        return amount > 0 ? of(begin, calendar.getTime()) : of(calendar.getTime(), begin);
    }

    /**
     * 构建时间周期
     *
     * @param date   日期时间
     * @param field  加减字段
     * @param amount 加减数量
     * @return 时间周期实例
     */
    public static Period of(@NonNull Date date, int field, int amount) {
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(date);
        calendar.add(field, amount);
        return amount > 0 ? of(date, calendar.getTime()) : of(calendar.getTime(), date);
    }

    /**
     * 构建时间周期
     *
     * @param field 时间字段
     * @param begin 开始时间
     * @param end   结束时间
     * @return 时间周期实例
     */
    public static Period of(int field, int begin, int end) {
        return of(Calendar.getInstance(), field, begin, end);
    }

    /**
     * 构建时间周期
     *
     * @param date  日期实例
     * @param field 时间字段
     * @param begin 开始时间
     * @param end   结束时间
     * @return 时间周期实例
     */
    public static Period of(@NonNull Date date, int field, int begin, int end) {
        return of(date.getTime(), field, begin, end);
    }

    /**
     * 构建时间周期
     *
     * @param timestamp 时间戳
     * @param field     时间字段
     * @param begin     开始时间
     * @param end       结束时间
     * @return 时间周期实例
     */
    public static Period of(long timestamp, int field, int begin, int end) {
        Calendar calendar = Calendar.getInstance();
        calendar.setTimeInMillis(timestamp);
        return of(calendar, field, begin, end);
    }

    /**
     * 构建时间周期
     *
     * @param calendar 时间日历
     * @param field    时间字段
     * @param begin    开始时间
     * @param end      结束时间
     * @return 时间周期实例
     */
    public static Period of(@NonNull Calendar calendar, int field, int begin, int end) {
        if (field == Calendar.HOUR_OF_DAY) {
            calendar.set(Calendar.MINUTE, 0);
            calendar.set(Calendar.SECOND, 0);
            calendar.set(Calendar.MILLISECOND, 0);
        } else if (field == Calendar.MINUTE) {
            calendar.set(Calendar.SECOND, 0);
            calendar.set(Calendar.MILLISECOND, 0);
        } else if (field == Calendar.SECOND) {
            calendar.set(Calendar.MILLISECOND, 0);
        } else {
            throw new UnsupportedOperationException("Unsupported calendar field: " + field);
        }
        calendar.set(field, begin);
        Date time = calendar.getTime();
        calendar.set(field, end);
        return of(time, calendar.getTime());
    }

    /**
     * 构建时间周期
     *
     * @param begin 开始时间
     * @param end   结束时间
     * @return 时间周期实例
     */
    public static Period of(Date begin, Date end) {
        Assert.isTrue(begin != null || end != null, "Period's begin or end must not be null");
        Period period = new Period();
        if (begin != null && end != null && begin.after(end)) {
            period.begin = end;
            period.end = begin;
        } else {
            period.begin = begin;
            period.end = end;
        }
        return period;
    }

    /**
     * 获取当前小时时间周期
     *
     * @return 时间周期实例
     */
    public static Period hour() {
        Calendar calendar = Calendar.getInstance();
        calendar.set(Calendar.MINUTE, 0);
        calendar.set(Calendar.SECOND, 0);
        calendar.set(Calendar.MILLISECOND, 0);
        Date begin = calendar.getTime();
        calendar.set(Calendar.MINUTE, 59);
        calendar.set(Calendar.SECOND, 59);
        Date end = calendar.getTime();
        return of(begin, end);
    }

    /**
     * 获取当前分钟时间周期
     *
     * @return 时间周期实例
     */
    public static Period minute() {
        Calendar calendar = Calendar.getInstance();
        calendar.set(Calendar.SECOND, 0);
        calendar.set(Calendar.MILLISECOND, 0);
        Date begin = calendar.getTime();
        calendar.set(Calendar.SECOND, 59);
        Date end = calendar.getTime();
        return of(begin, end);
    }

    /**
     * 获取当天时间周期
     *
     * @return 时间周期实例
     */
    public static Period today() {
        return date(new Date());
    }

    /**
     * 获取指定时间周期
     *
     * @param date 日期时间
     * @return 时间周期实例
     */
    public static Period date(@NonNull Date date) {
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(date);
        DateUtils.clearTime(calendar);
        Date begin = calendar.getTime();
        calendar.set(Calendar.HOUR_OF_DAY, 23);
        calendar.set(Calendar.MINUTE, 59);
        calendar.set(Calendar.SECOND, 59);
        Date end = calendar.getTime();
        return of(begin, end);
    }

    /**
     * 构建明天时间周期
     *
     * @param field 时间字段
     * @param begin 开始时间
     * @param end   结束时间
     * @return 时间周期实例
     */
    public static Period tomorrow(int field, int begin, int end) {
        Calendar calendar = Calendar.getInstance();
        calendar.add(Calendar.DATE, 1);
        return of(calendar, field, begin, end);
    }

    /**
     * 获取昨天到现在时间周期
     *
     * @return 时间周期实例
     */
    public static Period yesterday() {
        return yesterday(true);
    }

    /**
     * 获取昨天到现在时间周期
     *
     * @param now 是否截止到当前时间
     * @return 时间周期实例
     */
    public static Period yesterday(boolean now) {
        Calendar calendar = Calendar.getInstance();
        calendar.add(Calendar.DATE, -1);
        DateUtils.clearTime(calendar);
        Date begin = calendar.getTime(), end;
        if (now) {
            end = new Date();
        } else {
            calendar.set(Calendar.HOUR_OF_DAY, 23);
            calendar.set(Calendar.MINUTE, 59);
            calendar.set(Calendar.SECOND, 59);
            end = calendar.getTime();
        }
        return of(begin, end);
    }

    /**
     * 获取当月时间周期（星期一为一周的第一天）
     *
     * @return 时间周期实例
     */
    public static Period week() {
        return week(Calendar.MONDAY);
    }

    /**
     * 获取当月时间周期
     *
     * @param start 一周的开始标记
     * @return 时间周期实例
     */
    public static Period week(int start) {
        Calendar calendar = Calendar.getInstance();
        calendar.set(Calendar.DAY_OF_WEEK, start);
        DateUtils.clearTime(calendar);
        Date begin = calendar.getTime();
        calendar.add(Calendar.DATE, 6);
        calendar.set(Calendar.HOUR_OF_DAY, 23);
        calendar.set(Calendar.MINUTE, 59);
        calendar.set(Calendar.SECOND, 59);
        Date end = calendar.getTime();
        return of(begin, end);
    }

    /**
     * 获取当月时间周期
     *
     * @return 时间周期实例
     */
    public static Period month() {
        Calendar calendar = Calendar.getInstance();
        calendar.set(Calendar.DATE, 1);
        DateUtils.clearTime(calendar);
        Date begin = calendar.getTime();
        calendar.add(Calendar.MONTH, 1);
        calendar.add(Calendar.DATE, -1);
        calendar.set(Calendar.HOUR_OF_DAY, 23);
        calendar.set(Calendar.MINUTE, 59);
        calendar.set(Calendar.SECOND, 59);
        Date end = calendar.getTime();
        return of(begin, end);
    }

    /**
     * 获取当前季度时间周期
     *
     * @return 时间周期实例
     */
    public static Period quarter() {
        LocalDate today = LocalDate.now();
        Month month = today.getMonth();
        Month first = month.firstMonthOfQuarter();
        Month last = Month.of(first.getValue() + 2);
        LocalDate begin = LocalDate.of(today.getYear(), first, 1);
        LocalDateTime end = LocalDateTime.of(today.getYear(), last, last.length(today.isLeapYear()), 23, 59, 59);
        return Period.of(DateUtils.convert(begin), DateUtils.convert(end));
    }

    /**
     * 获取当前上/下半年时间周期
     *
     * @return 时间周期实例
     */
    public static Period halfyear() {
        Month first, last;
        LocalDate today = LocalDate.now();
        if (today.getMonth().getValue() < 7) {
            first = Month.of(1);
            last = Month.of(6);
        } else {
            first = Month.of(7);
            last = Month.of(12);
        }
        LocalDate begin = LocalDate.of(today.getYear(), first, 1);
        LocalDateTime end = LocalDateTime.of(today.getYear(), last, last.length(today.isLeapYear()), 23, 59, 59);
        return Period.of(DateUtils.convert(begin), DateUtils.convert(end));
    }

    /**
     * 获取当年时间周期
     *
     * @return 时间周期实例
     */
    public static Period year() {
        Calendar calendar = Calendar.getInstance();
        calendar.set(Calendar.MONTH, 0);
        calendar.set(Calendar.DATE, 1);
        DateUtils.clearTime(calendar);
        Date begin = calendar.getTime();
        calendar.add(Calendar.YEAR, 1);
        calendar.add(Calendar.DATE, -1);
        calendar.set(Calendar.HOUR_OF_DAY, 23);
        calendar.set(Calendar.MINUTE, 59);
        calendar.set(Calendar.SECOND, 59);
        Date end = calendar.getTime();
        return of(begin, end);
    }

    /**
     * 获取最近一段时间的时间周期
     *
     * @param field  加减字段
     * @param amount 加减数量
     * @return 时间周期实例
     */
    public static Period lately(int field, int amount) {
        Calendar calendar = Calendar.getInstance();
        Date begin = calendar.getTime();
        calendar.add(field, amount);
        Date end = calendar.getTime();
        return of(begin, end);
    }

    /**
     * 时间周期解析
     *
     * @param text  时间周期文本
     * @param begin 开始时间
     * @return 日期时间
     */
    private static Date parse(String text, Date begin) {
        if (TIME_PATTERN.matcher(text).matches()) {
            int m = text.indexOf('('), n = text.indexOf(')');
            Eval eval = Eval.valueOf(text.substring(0, m).trim());
            String input = text.substring(m + 1, n).trim();
            Integer value = input.isEmpty() ? null : Integer.parseInt(input.substring(0, input.length() - 1));
            Unit unit = input.isEmpty() ? null : Unit.valueOf(input.substring(input.length() - 1));
            return eval.datetime(unit, value, begin);
        }
        return DateUtils.parse(text);
    }

    /**
     * 将字符串转换成时间周期
     * <p>
     * 字符串以","号作为开始/结束时间的分隔符，如果没有分隔符则默认匹配开始时间
     * <p>
     * 示例：(2021-09-12, 2021-09-14)、(, 2021-09-14)、(2021-09-12, )、(timer(), duration(3h))、(timer(3d), duration(3h))
     *
     * @param text 时间周期字符串格式
     * @return 时间周期实例
     */
    public static Period parse(String text) {
        return StringUtils.decompose(text, Period::of, string -> parse(string, null), Period::parse);
    }

    /**
     * 获取周期时长
     *
     * @return 周期时长
     */
    @JsonIgnore
    public Duration getDuration() {
        return this.begin == null || this.end == null ? null :
                Duration.ofMillis(this.end.getTime() - this.getBegin().getTime());
    }

    /**
     * 根据数量获取当前时间周期内的延时周期
     *
     * @param number 数量
     * @return 延时周期
     */
    public Duration delay(int number) {
        long begin = this.begin.getTime();
        long offset = Math.abs(number) % ((this.end.getTime() - begin) / 1000);
        return Duration.ofMillis((begin + offset * 1000) - System.currentTimeMillis());
    }

    /**
     * 判断指定日期是否包含在当前时间周期内
     *
     * @param date 日期对象
     * @return true/false
     */
    public boolean contains(Date date) {
        return date != null && this.begin != null && this.end != null
                && !this.begin.after(date) && !this.end.before(date);
    }

    /**
     * 时间周期偏移
     *
     * @param field  时间字段
     * @param amount 偏移量
     * @return 时间周期实例
     */
    public Period offset(int field, int amount) {
        return this.offset(begin -> DateUtils.add(begin, field, amount), end -> DateUtils.add(end, field, amount));
    }

    /**
     * 时间周期偏移
     *
     * @param begin 开始时间偏移方法
     * @param end   结束时间偏移方法
     * @return 时间周期实例
     */
    public Period offset(@NonNull Function<Date, Date> begin, @NonNull Function<Date, Date> end) {
        return of(begin.apply(this.begin), end.apply(this.end));
    }

    /**
     * 时间周期函数
     */
    public enum Eval {
        /**
         * 时间
         */
        time,

        /**
         * 时长
         */
        duration;

        /**
         * 日期时间转化
         *
         * @param unit  时间单位
         * @param value 时间值
         * @param begin 开始时间
         * @return 日期时间
         */
        public Date datetime(Unit unit, Integer value, Date begin) {
            Calendar calendar = Calendar.getInstance();
            if (begin != null) {
                calendar.setTime(begin);
            }
            if (unit == null || value == null) {
                return calendar.getTime();
            }
            if (this == time) {
                switch (unit) {
                    case y:
                        calendar.set(Calendar.MONTH, 0);
                        calendar.set(Calendar.DATE, 1);
                    case d:
                        calendar.set(Calendar.HOUR_OF_DAY, 0);
                    case h:
                        calendar.set(Calendar.MINUTE, 0);
                    case m:
                        calendar.set(Calendar.SECOND, 0);
                    case s:
                        calendar.set(Calendar.MILLISECOND, 0);
                }
                calendar.set(unit.field(), value);
            } else {
                calendar.add(unit.field(), value);
            }
            return calendar.getTime();
        }
    }

    /**
     * 时间单位枚举
     */
    public enum Unit {
        /**
         * 年
         */
        y,

        /**
         * 日
         */
        d,

        /**
         * 时
         */
        h,

        /**
         * 分
         */
        m,

        /**
         * 秒
         */
        s;

        /**
         * 将之间单位转换城字段标识
         *
         * @return 字段标识
         */
        public int field() {
            if (this == y) {
                return Calendar.YEAR;
            } else if (this == d) {
                return Calendar.DATE;
            } else if (this == h) {
                return Calendar.HOUR_OF_DAY;
            } else if (this == m) {
                return Calendar.MINUTE;
            }
            return Calendar.SECOND;
        }
    }

    /**
     * 固定周期枚举
     */
    public enum Fixed {
        /**
         * 当天
         */
        TODAY,

        /**
         * 昨天
         */
        YESTERDAY,

        /**
         * 当周
         */
        THIS_WEEK,

        /**
         * 上周
         */
        LAST_WEEK,

        /**
         * 当月
         */
        THIS_MONTH,

        /**
         * 上月
         */
        LAST_MONTH,

        /**
         * 当年
         */
        THIS_YEAR,

        /**
         * 去年同月
         */
        LAST_YEAR_MONTH,

        /**
         * 最近一周
         */
        RECENT_WEEK,

        /**
         * 最近一月
         */
        RECENT_MONTH,

        /**
         * 最近半年
         */
        RECENT_HALF_YEAR;

        /**
         * 获取周期开始时间
         *
         * @return 开始时间
         */
        public Date getBegin() {
            Calendar calendar = Calendar.getInstance();
            DateUtils.clearTime(calendar);
            int week = calendar.get(Calendar.DAY_OF_WEEK);
            switch (this) {
                case TODAY:
                    break;
                case YESTERDAY:
                    calendar.add(Calendar.DATE, -1);
                    break;
                case THIS_WEEK:
                    calendar.add(Calendar.DATE, week == 1 ? -6 : 2 - week);
                    break;
                case LAST_WEEK:
                    calendar.add(Calendar.DATE, week == 1 ? -13 : -5 - week);
                    break;
                case THIS_MONTH:
                    calendar.set(Calendar.DAY_OF_MONTH, 1);
                    break;
                case LAST_MONTH:
                    calendar.add(Calendar.MONTH, -1);
                    calendar.set(Calendar.DAY_OF_MONTH, 1);
                    break;
                case THIS_YEAR:
                    calendar.set(Calendar.MONTH, 0);
                    calendar.set(Calendar.DAY_OF_MONTH, 1);
                    break;
                case LAST_YEAR_MONTH:
                    calendar.add(Calendar.YEAR, -1);
                    calendar.set(Calendar.DAY_OF_MONTH, 1);
                    break;
                case RECENT_WEEK:
                    calendar.add(Calendar.DATE, -7);
                    break;
                case RECENT_MONTH:
                    calendar.add(Calendar.DATE, -30);
                    break;
                case RECENT_HALF_YEAR:
                    calendar.add(Calendar.DATE, -180);
                    break;
            }
            return calendar.getTime();
        }

        /**
         * 获取周期结束时间
         *
         * @return 结束时间
         */
        public Date getEnd() {
            Calendar calendar = Calendar.getInstance();
            switch (this) {
                case YESTERDAY:
                    calendar.add(Calendar.DATE, -1);
                    break;
                case LAST_WEEK:
                    calendar.add(Calendar.DATE, -7);
                    break;
                case LAST_MONTH:
                    calendar.add(Calendar.MONTH, -1);
                    break;
                case LAST_YEAR_MONTH:
                    calendar.add(Calendar.YEAR, -1);
                    break;
            }
            return calendar.getTime();
        }

        /**
         * 构建时间周期实例
         *
         * @return 时间周期实例
         */
        public Period period() {
            return of(this.getBegin(), this.getEnd());
        }
    }
}
