package com.chelvc.framework.base.model;

import java.io.Serializable;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.fasterxml.jackson.annotation.JsonIgnore;
import lombok.AllArgsConstructor;
import lombok.Builder;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.NonNull;
import lombok.ToString;
import org.springframework.util.Assert;

/**
 * 地区编码
 *
 * @author Woody
 * @date 2023/4/5
 */
@Getter
@Builder
@ToString
@EqualsAndHashCode
@NoArgsConstructor
@AllArgsConstructor
public final class Region implements Serializable {
    /**
     * 地区类型枚举
     */
    public enum Type {
        /**
         * 省
         */
        PROVINCE,

        /**
         * 市
         */
        CITY,

        /**
         * 区
         */
        DISTRICT
    }

    /**
     * 地区编码最小值
     */
    private static final int MIN_VALUE = 100000;

    /**
     * 地区编码最大值
     */
    private static final int MAX_VALUE = 999999;

    /**
     * 地区类型
     */
    private Type type;

    /**
     * 地区编码
     */
    private int code;

    /**
     * 地区边界编码
     */
    private int boundary;

    /**
     * 判断是否是地区编码
     *
     * @param code 地区编码
     * @return true/false
     */
    public static boolean isRegion(Number code) {
        Integer value = ObjectUtils.ifNull(code, Number::intValue);
        return code != null && value >= MIN_VALUE && value <= MAX_VALUE;
    }

    /**
     * 获取可用的地区编码
     *
     * @param code 地区编码
     * @return 地区编码
     */
    public static Integer getActiveCode(Number code) {
        Integer value = ObjectUtils.ifNull(code, Number::intValue);
        if (value == null || value < MIN_VALUE) {
            return null;
        }
        return value <= MAX_VALUE ? value : (int) (value / Math.pow(10, (int) Math.log10(value) - 5));
    }

    /**
     * 判断是否是同一地区
     *
     * @param code1 地区编码
     * @param code2 地区编码
     * @return true/false
     */
    public static boolean isSame(Number code1, Number code2) {
        Integer value1 = ObjectUtils.ifNull(code1, Number::intValue);
        Integer value2 = ObjectUtils.ifNull(code2, Number::intValue);
        return isRegion(value1) && isRegion(value2)
                && (Objects.equals(value1, value2) || Objects.equals(value1, value2 / 100 * 100)
                || Objects.equals(value1, value2 / 10000 * 10000)
                || Objects.equals(value1 / 100 * 100, value2) || Objects.equals(value1 / 10000 * 10000, value2));
    }

    /**
     * 判断地区编码是否是市级编码
     *
     * @param code 地区编码
     * @return true/false
     */
    public static boolean isCity(Number code) {
        Integer value = ObjectUtils.ifNull(code, Number::intValue);
        return isRegion(value) && (value % 10000 > 0) && (value % 100 == 0);
    }

    /**
     * 判断地区编码是否是省级编码
     *
     * @param code 地区编码
     * @return true/false
     */
    public static boolean isProvince(Number code) {
        return isRegion(code) && (code.intValue() % 10000 == 0);
    }

    /**
     * 判断地区编码是否是区级编码
     *
     * @param code 地区编码
     * @return true/false
     */
    public static boolean isDistrict(Number code) {
        return isRegion(code) && (code.intValue() % 100 > 0);
    }

    /**
     * 构建地区编码
     *
     * @param type     地区类型
     * @param code     地区编码
     * @param boundary 地区边界编码
     * @return 地区实例
     */
    public static Region of(Type type, int code, int boundary) {
        Assert.isTrue(code > 0, "Region's code must be greater than 0");
        Assert.isTrue(boundary >= 0, "Region's boundary must be greater than or equal 0");
        Region region = new Region();
        region.type = type;
        region.code = code;
        region.boundary = boundary;
        return region;
    }

    /**
     * 构建地区编码
     *
     * @param code 地区编码
     * @return 地区实例
     */
    public static Region of(Number code) {
        Integer value = ObjectUtils.ifNull(code, Number::intValue);
        if (isProvince(value)) {
            return of(Type.PROVINCE, value, value + 9999);
        } else if (isCity(value)) {
            return of(Type.CITY, value, value + 99);
        } else if (isDistrict(value)) {
            return of(Type.DISTRICT, value, 0);
        }
        return null;
    }

    /**
     * 将字符串转换成地区编码
     *
     * @param text 地区编码字符串格式
     * @return 地区实例
     */
    public static Region parse(String text) {
        return text == null || (text = text.trim()).isEmpty() ? null : of(Integer.parseInt(text));
    }

    /**
     * 将地区编码转换成市级编码
     *
     * @param code 地区编码
     * @return 地区实例
     */
    public static Region code2city(Number code) {
        Integer value = ObjectUtils.ifNull(code, Number::intValue);
        return !isRegion(value) || isProvince(value) ? null : of(Type.CITY, value = value / 100 * 100, value + 99);
    }

    /**
     * 将地区编码转换成省级编码
     *
     * @param code 地区编码
     * @return 地区实例
     */
    public static Region code2province(Number code) {
        Integer value = ObjectUtils.ifNull(code, Number::intValue);
        return !isRegion(value) ? null : of(Type.PROVINCE, value = value / 10000 * 10000, value + 9999);
    }

    /**
     * 获取市级地区
     *
     * @param code 地区编码
     * @return 地区实例
     */
    public static Region urbanize(Number code) {
        return isRegion(code) ? of(code.intValue() / 100 * 100) : null;
    }

    /**
     * 判断地区是否属于敏感地区
     *
     * @param code 地区编码
     * @return true/false
     */
    public static boolean isSensitive(Number code) {
        if (code == null) {
            return false;
        }
        int province = (code.intValue() / 10000);
        return province == 11 || province == 65 || province == 54;
    }

    /**
     * 获取地区简称
     *
     * @param name 地区名称
     * @return 地区简称
     */
    public static String shortname(String name) {
        if (StringUtils.isEmpty(name)) {
            return name;
        }
        if (name.endsWith("特别行政区")) {
            name = name.substring(0, name.length() - 5);
        }
        if (name.endsWith("地区")) {
            name = name.substring(0, name.length() - 2);
        }
        if (name.endsWith("自治县") || name.endsWith("自治区") || name.endsWith("自治州")
                || name.endsWith("直辖市") || name.endsWith("市辖区")) {
            name = name.substring(0, name.length() - 3);
        }
        if (name.endsWith("省") || name.endsWith("市") || name.endsWith("县")) {
            name = name.substring(0, name.length() - 1);
        }
        return name;
    }

    /**
     * 获取地区简称
     *
     * @param name      地区名称
     * @param delimiter 地区分割符
     * @return 地区简称
     */
    public static String shortname(String name, @NonNull String delimiter) {
        if (StringUtils.isEmpty(name)) {
            return name;
        }
        return Stream.of(name.split(delimiter)).map(Region::shortname).collect(Collectors.joining(StringUtils.EMPTY));
    }

    /**
     * 获取地区上级地区
     *
     * @return 地区信息
     */
    @JsonIgnore
    public Region getParent() {
        return this.type == Type.DISTRICT ? code2city(this.code) :
                this.type == Type.CITY ? code2province(this.code) : null;
    }

    /**
     * 获取当前地区所在省
     *
     * @return 省
     */
    @JsonIgnore
    public Region getProvince() {
        return this.type == Type.PROVINCE ? this : code2province(this.code);
    }

    /**
     * 获取当前地区所在市
     *
     * @return 市
     */
    @JsonIgnore
    public Region getCity() {
        return this.type == Type.CITY ? this : this.type == Type.PROVINCE ? null : code2city(this.code);
    }

    /**
     * 获取当前地区所在区
     *
     * @return 区
     */
    @JsonIgnore
    public Region getDistrict() {
        return this.type == Type.DISTRICT ? this : null;
    }

    /**
     * 判断是否是同一地区
     *
     * @param code 地区编码
     * @return true/false
     */
    @JsonIgnore
    public boolean isSame(Number code) {
        return isSame(this.code, code);
    }

    /**
     * 判断是否是同一地区
     *
     * @param region 地区实例
     * @return true/false
     */
    @JsonIgnore
    public boolean isSame(Region region) {
        return region != null && isSame(this.code, region.getCode());
    }

    /**
     * 判断是否是直辖市
     *
     * @return true/false
     */
    @JsonIgnore
    public boolean isDirectly() {
        int code = this.code / 10000;
        return code == 11 || code == 12 || code == 31 || code == 50;
    }
}
