package com.chelvc.framework.base.model;

import java.io.Serializable;
import java.util.Objects;

import com.chelvc.framework.base.util.ObjectUtils;
import lombok.AllArgsConstructor;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.NonNull;
import lombok.ToString;
import org.springframework.http.HttpStatus;

/**
 * 请求结果对象
 *
 * @author Woody
 * @date 2023/4/5
 */
@Getter
@ToString
@EqualsAndHashCode
@NoArgsConstructor
@AllArgsConstructor
public class Result<T> implements Serializable {
    /**
     * 时间戳
     */
    private final Long timestamp = System.currentTimeMillis();
    /**
     * 结果数据
     */
    private T data;

    /**
     * 结果码（Http状态码）
     *
     * @example 200
     */
    private Integer code;

    /**
     * 结果信息
     *
     * @example Success
     */
    private String message;

    /**
     * 构建成功结果
     *
     * @return 结果对象
     */
    public static <M> Result<M> success() {
        return success(null);
    }

    /**
     * 构建成功结果
     *
     * @param data 结果数据
     * @return 结果对象
     */
    public static <M> Result<M> success(M data) {
        return success(data, HttpStatus.OK.getReasonPhrase());
    }

    /**
     * 构建成功结果
     *
     * @param data    结果数据
     * @param message 结果信息
     * @return 结果对象
     */
    public static <M> Result<M> success(M data, String message) {
        return build(HttpStatus.OK, data, message);
    }

    /**
     * 构建异常结果
     *
     * @return 结果对象
     */
    public static <M> Result<M> error() {
        return error(null);
    }

    /**
     * 构建异常结果
     *
     * @param data 结果数据
     * @return 结果对象
     */
    public static <M> Result<M> error(M data) {
        return error(data, HttpStatus.INTERNAL_SERVER_ERROR.getReasonPhrase());
    }

    /**
     * 构建异常结果
     *
     * @param data    结果数据
     * @param message 结果信息
     * @return 结果对象
     */
    public static <M> Result<M> error(M data, String message) {
        return build(HttpStatus.INTERNAL_SERVER_ERROR, data, message);
    }

    /**
     * 构建请求结果
     *
     * @param status Http状态
     * @return 结果对象
     */
    public static <M> Result<M> build(HttpStatus status) {
        return build(status, null);
    }

    /**
     * 构建请求结果
     *
     * @param status Http状态
     * @param data   结果数据
     * @return 结果对象
     */
    public static <M> Result<M> build(HttpStatus status, M data) {
        return build(status, data, ObjectUtils.ifNull(status, HttpStatus::getReasonPhrase));
    }

    /**
     * 构建请求结果
     *
     * @param status  Http状态
     * @param data    结果数据
     * @param message 结果信息
     * @return 结果对象
     */
    public static <M> Result<M> build(@NonNull HttpStatus status, M data, String message) {
        Result<M> result = new Result<>();
        result.code = status.value();
        result.data = data;
        result.message = message;
        return result;
    }

    /**
     * 判断是否为成功的请求结果
     *
     * @return true/false
     */
    public boolean ok() {
        return Objects.equals(this.code, HttpStatus.OK.value());
    }
}
