package com.chelvc.framework.base.model;

import java.io.Serializable;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.function.Consumer;
import java.util.stream.Collectors;

import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.chelvc.framework.base.util.TreeUtils;
import com.google.common.collect.Maps;
import lombok.NonNull;
import org.apache.commons.lang3.tuple.Pair;
import org.springframework.util.CollectionUtils;

/**
 * 版本对象
 *
 * @author Woody
 * @date 2023/4/5
 */
public class Version extends Tree.Simple<String, Version> implements Serializable, Comparable<Version> {
    /**
     * 版本号部分解析，获取大版本号数字及后续版本号信息
     *
     * @param version 版本号
     * @return 版本号部分信息
     */
    private static Pair<Integer, String> pair(String version) {
        if (StringUtils.isEmpty(version)) {
            return Pair.of(0, null);
        }
        int index = version.indexOf('.');
        if (index < 0) {
            return Pair.of(Integer.parseInt(version), null);
        } else if (index == 0) {
            return Pair.of(0, version.substring(index + 1));
        }
        return Pair.of(Integer.parseInt(version.substring(0, index)), version.substring(index + 1));
    }

    /**
     * 版本号比较，仅支持数字版本号，如：1.7.4
     *
     * @param version1 版本号
     * @param version2 版本号
     * @return 比较数字
     */
    public static int compare(String version1, String version2) {
        if (Objects.equals(version1, version2)) {
            return 0;
        } else if (StringUtils.isEmpty(version1)) {
            return -1;
        } else if (StringUtils.isEmpty(version2)) {
            return 1;
        }
        Pair<Integer, String> pair1 = pair(version1), pair2 = pair(version2);
        int difference = pair1.getLeft() - pair2.getLeft();
        return difference == 0 ? compare(pair1.getRight(), pair2.getRight()) : difference;
    }

    /**
     * 判断版本号是否在指定版本号之前
     *
     * @param version1 版本号
     * @param version2 版本号
     * @return true/false
     */
    public static boolean isBefore(String version1, String version2) {
        return isBefore(version1, version2, false);
    }

    /**
     * 判断版本号是否在指定版本号之前
     *
     * @param version1 版本号
     * @param version2 版本号
     * @param contains 是否包含指定版本号
     * @return true/false
     */
    public static boolean isBefore(String version1, String version2, boolean contains) {
        int compared = compare(version1, version2);
        return contains ? compared <= 0 : compared < 0;
    }

    /**
     * 判断版本号是否在指定版本号之后
     *
     * @param version1 版本号
     * @param version2 版本号
     * @return true/false
     */
    public static boolean isAfter(String version1, String version2) {
        return isAfter(version1, version2, false);
    }

    /**
     * 判断版本号是否在指定版本号之后
     *
     * @param version1 版本号
     * @param version2 版本号
     * @param contains 是否包含指定版本号
     * @return true/false
     */
    public static boolean isAfter(String version1, String version2, boolean contains) {
        int compared = compare(version1, version2);
        return contains ? compared >= 0 : compared > 0;
    }

    /**
     * 构建版本信息
     *
     * @param id     唯一标识
     * @param name   版本名称
     * @param parent 上级版本
     * @return 版本信息
     */
    public static Version build(@NonNull String id, @NonNull String name, Version parent) {
        Version version = new Version();
        version.setId(id);
        version.setName(name);
        version.setParentId(ObjectUtils.ifNull(parent, Version::getId));
        version.setLevel(ObjectUtils.ifNull(ObjectUtils.ifNull(parent, Version::getLevel), 0) + 1);
        return version;
    }

    /**
     * 将版本号转换成树列表
     *
     * @param versions 版本号集合
     * @return 版本树列表
     */
    public static List<Version> trees(Collection<String> versions) {
        return trees(versions, null);
    }

    /**
     * 将版本号转换成树列表
     *
     * @param versions 版本号集合
     * @param consumer 版本回调方法
     * @return 版本树列表
     */
    public static List<Version> trees(Collection<String> versions, Consumer<Version> consumer) {
        if (CollectionUtils.isEmpty(versions)) {
            return Collections.emptyList();
        }
        Map<String, Version> caches = Maps.newHashMap();
        versions.forEach(code -> {
            int i = 0;
            Version parent = null;
            do {
                i = code.indexOf('.', i + 1);
                if (i > 0) {
                    Version p = parent;
                    parent = caches.computeIfAbsent(code.substring(0, i), id -> {
                        Version version = build(id, id + ".x", p);
                        if (consumer != null) {
                            consumer.accept(version);
                        }
                        return version;
                    });
                } else {
                    Version version = build(code, code, parent);
                    if (consumer != null) {
                        consumer.accept(version);
                    }
                    caches.put(code, version);
                }
            } while (i > 0);
        });
        Collection<Version> values = caches.values().stream()
                .sorted(Comparator.comparing(Version::getId, Version::compare)).collect(Collectors.toList());
        return TreeUtils.assemble(values);
    }

    @Override
    public int compareTo(Version version) {
        return compare(this.getId(), ObjectUtils.ifNull(version, Version::getId));
    }
}
