package com.chelvc.framework.base.util;

import java.security.AlgorithmParameters;
import java.security.GeneralSecurityException;
import java.security.Key;
import java.security.NoSuchAlgorithmException;
import java.security.Security;
import javax.crypto.Cipher;
import javax.crypto.KeyGenerator;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;

import lombok.NonNull;
import org.apache.commons.codec.binary.Base64;
import org.bouncycastle.jce.provider.BouncyCastleProvider;

/**
 * 数据加解密处理工具类
 *
 * @author woody
 * @date 2023/4/5
 */
public final class AESUtils {
    /**
     * 密钥长度
     */
    public static final int KEY_SIZE = 128;

    /**
     * AES算法名称
     */
    public static final String ALGORITHM = "AES";

    /**
     * 密码处理器算法
     */
    public static final String CIPHER_ALGORITHM = "AES/CBC/PKCS5Padding";

    static {
        Security.addProvider(new BouncyCastleProvider());
        try {
            KeyGenerator.getInstance(ALGORITHM).init(KEY_SIZE);
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException(e);
        }
    }

    private AESUtils() {
    }

    /**
     * 修正密钥
     *
     * @param key 密钥标识
     * @return 修正后的密钥
     */
    public static String revise(String key) {
        if (key == null) {
            return null;
        }

        // 如果小于16位则左边补0到16位
        if (key.length() < 16) {
            return StringUtils.rjust(key, 16, '0');
        }

        // 如果长度不是8的倍数则左边补0到8的倍数
        int offset = 8 - key.length() % 8;
        return offset == 8 ? key : StringUtils.rjust(key, key.length() + offset, '0');
    }

    /**
     * 获取密码处理器
     *
     * @param mode   加解密模式
     * @param secret 密钥
     * @param iv     偏移量
     * @return 密码处理器实例
     */
    public static Cipher getCipher(int mode, @NonNull String secret, @NonNull String iv) {
        try {
            Cipher cipher = Cipher.getInstance(CIPHER_ALGORITHM);
            Key key = new SecretKeySpec(Base64.decodeBase64(secret), ALGORITHM);
            AlgorithmParameters parameters = AlgorithmParameters.getInstance(ALGORITHM);
            parameters.init(new IvParameterSpec(Base64.decodeBase64(iv)));
            cipher.init(mode, key, parameters);
            return cipher;
        } catch (GeneralSecurityException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * AES加密
     *
     * @param cipher   密码处理器实例
     * @param original 明文
     * @return 密文（base64）
     */
    public static String encode(@NonNull Cipher cipher, String original) {
        if (original == null) {
            return null;
        }
        try {
            return Base64.encodeBase64String(cipher.doFinal(original.getBytes()));
        } catch (GeneralSecurityException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * AES加密
     *
     * @param original 明文
     * @param secret   密钥
     * @param iv       偏移量
     * @return 密文（base64）
     */
    public static String encode(String original, @NonNull String secret, @NonNull String iv) {
        return original == null ? null : encode(getCipher(Cipher.ENCRYPT_MODE, secret, iv), original);
    }

    /**
     * AES解密
     *
     * @param cipher     密码处理器实例
     * @param ciphertext 密文（base64）
     * @return 明文
     */
    public static String decode(@NonNull Cipher cipher, String ciphertext) {
        if (ciphertext == null) {
            return null;
        }
        try {
            return new String(cipher.doFinal(Base64.decodeBase64(ciphertext)));
        } catch (GeneralSecurityException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * AES解密
     *
     * @param ciphertext 密文（base64）
     * @param secret     密钥
     * @param iv         偏移量
     * @return 明文
     */
    public static String decode(String ciphertext, String secret, String iv) {
        return ciphertext == null ? null : decode(getCipher(Cipher.DECRYPT_MODE, secret, iv), ciphertext);
    }
}
