package com.chelvc.framework.base.util;

import java.time.Duration;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.Calendar;
import java.util.Date;
import java.util.Locale;
import java.util.Map;
import java.util.function.Function;
import java.util.regex.Pattern;

import com.google.common.collect.ImmutableMap;
import lombok.NonNull;
import org.springframework.context.i18n.LocaleContextHolder;

/**
 * 日期处理工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
public final class DateUtils {
    /**
     * 默认日期格式化处理器
     */
    public static final DateTimeFormatter DEFAULT_DATE_FORMATTER = DateTimeFormatter.ofPattern("yyyy-MM-dd");

    /**
     * 默认日期时间格式化处理器
     */
    public static final DateTimeFormatter DEFAULT_DATETIME_FORMATTER =
            DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss");

    /**
     * 年份数字格式化处理器
     */
    public static final DateTimeFormatter YEAR_NUMBER_FORMATTER = DateTimeFormatter.ofPattern("yyyy");

    /**
     * 月份数字格式化处理器
     */
    public static final DateTimeFormatter MONTH_NUMBER_FORMATTER = DateTimeFormatter.ofPattern("yyyyMM");

    /**
     * 日期数字格式化处理器
     */
    public static final DateTimeFormatter DATE_NUMBER_FORMATTER = DateTimeFormatter.ofPattern("yyyyMMdd");

    /**
     * 日期时间数字格式化处理器
     */
    public static final DateTimeFormatter DATETIME_NUMBER_FORMATTER = DateTimeFormatter.ofPattern("yyyyMMddHHmmss");

    /**
     * 日期cron表达式格式化处理器
     */
    public static final DateTimeFormatter DATETIME_CRON_FORMATTER = DateTimeFormatter.ofPattern("ss mm HH dd MM ? " +
            "yyyy");

    /**
     * 日期格式匹配模式/日期文本解析器映射表
     */
    private static final Map<Pattern, Parser> PATTERN_PARSER_MAPPING = ImmutableMap.<Pattern, Parser>builder()
            .put(
                    StringUtils.getPattern("^-?[0-9]\\d{7,12}$"),
                    text -> new Date(Long.parseLong(text))
            ).put(
                    StringUtils.getPattern("^\\d{4}-\\d{1,2}-\\d{1,2}$"),
                    new DateParser(DateTimeFormatter.ofPattern("yyyy-M-d"))
            ).put(
                    StringUtils.getPattern("^\\d{4}/\\d{1,2}/\\d{1,2}$"),
                    new DateParser(DateTimeFormatter.ofPattern("yyyy/M/d"))
            ).put(
                    StringUtils.getPattern("^\\d{4}-\\d{1,2}-\\d{1,2} \\d{1,2}$"),
                    new DatetimeParser(DateTimeFormatter.ofPattern("yyyy-M-d H"))
            ).put(
                    StringUtils.getPattern("^\\d{4}/\\d{1,2}/\\d{1,2} \\d{1,2}$"),
                    new DatetimeParser(DateTimeFormatter.ofPattern("yyyy/M/d H"))
            ).put(
                    StringUtils.getPattern("^\\d{4}-\\d{1,2}-\\d{1,2} \\d{1,2}:\\d{1,2}$"),
                    new DatetimeParser(DateTimeFormatter.ofPattern("yyyy-M-d H:m"))
            ).put(
                    StringUtils.getPattern("^\\d{4}/\\d{1,2}/\\d{1,2} \\d{1,2}:\\d{1,2}$"),
                    new DatetimeParser(DateTimeFormatter.ofPattern("yyyy/M/d H:m"))
            ).put(
                    StringUtils.getPattern("^\\d{4}-\\d{1,2}-\\d{1,2} \\d{1,2}:\\d{1,2}:\\d{1,2}$"),
                    new DatetimeParser(DateTimeFormatter.ofPattern("yyyy-M-d H:m:s"))
            ).put(
                    StringUtils.getPattern("^\\d{4}/\\d{1,2}/\\d{1,2} \\d{1,2}:\\d{1,2}:\\d{1,2}$"),
                    new DatetimeParser(DateTimeFormatter.ofPattern("yyyy/M/d H:m:s"))
            ).put(
                    StringUtils.getPattern("^\\d{4}-\\d{1,2}-\\d{1,2} \\d{1,2}:\\d{1,2}:\\d{1,2}\\.\\d{1,3}$"),
                    new DatetimeParser(DateTimeFormatter.ofPattern("yyyy-M-d H:m:s.SSS"))
            ).put(
                    StringUtils.getPattern("^\\d{4}/\\d{1,2}/\\d{1,2} \\d{1,2}:\\d{1,2}:\\d{1,2}\\.\\d{1,3}$"),
                    new DatetimeParser(DateTimeFormatter.ofPattern("yyyy-M-d H:m:s.SSS"))
            ).put(
                    StringUtils.getPattern(
                            "^[a-zA-Z]{3} [a-zA-Z]{3} [0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2} CST [0-9]{4}$"
                    ),
                    new DatetimeParser(DateTimeFormatter.ofPattern("EEE MMM dd HH:mm:ss zzz yyyy", Locale.US))
            ).put(
                    StringUtils.getPattern(
                            "^[a-zA-Z]{3} [a-zA-Z]{3} [0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2} UTC [0-9]{4}$"
                    ),
                    new DatetimeParser(DateTimeFormatter.ofPattern("EEE MMM dd HH:mm:ss zzz yyyy", Locale.US))
            ).put(
                    StringUtils.getPattern(
                            "^[a-zA-Z]{3} [a-zA-Z]{3} [0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2} GMT[\\+-:0-9]{6} [0-9]{4}$"
                    ),
                    new DatetimeParser(DateTimeFormatter.ofPattern("EEE MMM dd HH:mm:ss z yyyy", Locale.US))
            ).build();

    private DateUtils() {
    }

    /**
     * 日期文本解析器
     */
    @FunctionalInterface
    public interface Parser {
        /**
         * 日期文本解析
         *
         * @param text 日期文本
         * @return 日期对象实例
         */
        Date parse(String text);
    }

    /**
     * 日期文本解析实现
     */
    public static class DateParser implements Parser {
        private final DateTimeFormatter formatter;

        public DateParser(@NonNull DateTimeFormatter formatter) {
            this.formatter = formatter;
        }

        @Override
        public Date parse(String text) {
            return convert(LocalDate.parse(text, this.formatter));
        }
    }

    /**
     * 日期时间文本解析实现
     */
    public static class DatetimeParser implements Parser {
        private final DateTimeFormatter formatter;

        public DatetimeParser(@NonNull DateTimeFormatter formatter) {
            this.formatter = formatter;
        }

        @Override
        public Date parse(String text) {
            return convert(LocalDateTime.parse(text, this.formatter));
        }
    }

    /**
     * 查找日期文本对应的解析器
     *
     * @param text 日期文本
     * @return 日期文本解析器
     */
    private static Parser lookupDatetimeParser(String text) {
        return StringUtils.isEmpty(text) ? null : PATTERN_PARSER_MAPPING.entrySet().stream()
                .filter(entry -> entry.getKey().matcher(text).matches()).map(Map.Entry::getValue).findAny().orElse(null);
    }

    /**
     * 判断是否是日期格式
     *
     * @param text 日期文本
     * @return true/false
     */
    public static boolean isDate(String text) {
        return lookupDatetimeParser(text) != null;
    }

    /**
     * 判断日期时间对象实例是否包含时间信息
     *
     * @param date 日期时间对象实例
     * @return true/false
     */
    public static boolean isTemporal(Date date) {
        if (date == null) {
            return false;
        }
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(date);
        return calendar.get(Calendar.HOUR) > 0 || calendar.get(Calendar.MINUTE) > 0
                || calendar.get(Calendar.SECOND) > 0 || calendar.get(Calendar.MILLISECOND) > 0;
    }

    /**
     * 判断日期时间对象实例是否包含时间信息
     *
     * @param datetime 日期时间对象实例
     * @return true/false
     */
    public static boolean isTemporal(LocalDateTime datetime) {
        return datetime != null && (datetime.getHour() > 0 || datetime.getMinute() > 0 || datetime.getSecond() > 0
                || datetime.getNano() > 0);
    }

    /**
     * 清空日历时间
     *
     * @param calendar 日历实例
     */
    public static void clearTime(Calendar calendar) {
        if (calendar != null) {
            calendar.set(Calendar.HOUR_OF_DAY, 0);
            calendar.set(Calendar.MINUTE, 0);
            calendar.set(Calendar.SECOND, 0);
            calendar.set(Calendar.MILLISECOND, 0);
        }
    }

    /**
     * 获取当前时区标识
     *
     * @return 时区标识
     */
    public static ZoneId getCurrentZoneId() {
        return LocaleContextHolder.getTimeZone().toZoneId();
    }

    /**
     * 获取当天日期
     *
     * @return 日期实例
     */
    public static Date today() {
        Calendar calendar = Calendar.getInstance();
        clearTime(calendar);
        return calendar.getTime();
    }

    /**
     * 获取明天日期
     *
     * @return 日期实例
     */
    public static Date tomorrow() {
        return tomorrow(false);
    }

    /**
     * 获取明天日期
     *
     * @param temporal 是否包含时间
     * @return 日期实例
     */
    public static Date tomorrow(boolean temporal) {
        Calendar calendar = Calendar.getInstance();
        calendar.add(Calendar.DATE, 1);
        if (!temporal) {
            clearTime(calendar);
        }
        return calendar.getTime();
    }

    /**
     * 获取明天日期
     *
     * @param datetime 相对日期时间
     * @return 日期实例
     */
    public static Date tomorrow(Date datetime) {
        return tomorrow(datetime, false);
    }

    /**
     * 获取明天日期
     *
     * @param datetime 相对日期时间
     * @param temporal 是否包含时间
     * @return 日期实例
     */
    public static Date tomorrow(Date datetime, boolean temporal) {
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(datetime);
        calendar.add(Calendar.DATE, 1);
        if (!temporal) {
            clearTime(calendar);
        }
        return calendar.getTime();
    }

    /**
     * 获取昨天日期
     *
     * @return 日期实例
     */
    public static Date yesterday() {
        return yesterday(false);
    }

    /**
     * 获取昨天日期
     *
     * @param temporal 是否包含时间
     * @return 日期实例
     */
    public static Date yesterday(boolean temporal) {
        Calendar calendar = Calendar.getInstance();
        calendar.add(Calendar.DATE, -1);
        if (!temporal) {
            clearTime(calendar);
        }
        return calendar.getTime();
    }

    /**
     * 获取昨天日期
     *
     * @param datetime 相对日期时间
     * @return 日期实例
     */
    public static Date yesterday(Date datetime) {
        return yesterday(datetime, false);
    }

    /**
     * 获取昨天日期
     *
     * @param datetime 相对日期时间
     * @param temporal 是否包含时间
     * @return 日期实例
     */
    public static Date yesterday(Date datetime, boolean temporal) {
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(datetime);
        calendar.add(Calendar.DATE, -1);
        if (!temporal) {
            clearTime(calendar);
        }
        return calendar.getTime();
    }

    /**
     * 获取日期
     *
     * @param date 日期时间实例
     * @return 日期实例
     */
    public static Date date(Date date) {
        if (date == null) {
            return null;
        }
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(date);
        clearTime(calendar);
        return calendar.getTime();
    }

    /**
     * 获取日期
     *
     * @param timestamp 时间戳
     * @return 日期实例
     */
    public static Date date(Long timestamp) {
        if (timestamp == null) {
            return null;
        }
        Calendar calendar = Calendar.getInstance();
        calendar.setTimeInMillis(timestamp);
        clearTime(calendar);
        return calendar.getTime();
    }

    /**
     * 获取日期
     *
     * @param date 日期实例
     * @return 日期实例
     */
    public static Date date(LocalDate date) {
        return convert(date);
    }

    /**
     * 获取日期
     *
     * @param datetime 日期时间实例
     * @return 日期实例
     */
    public static Date date(LocalDateTime datetime) {
        return datetime == null ? null : date(datetime.toLocalDate());
    }

    /**
     * 将java.util.Date转换成java.time.LocalDateTime
     *
     * @param date java.util.Date实例
     * @return java.time.LocalDateTime实例
     */
    public static LocalDateTime convert(Date date) {
        return date == null ? null : LocalDateTime.ofInstant(date.toInstant(), getCurrentZoneId());
    }

    /**
     * 将java.time.LocalDate转换成java.util.Date
     *
     * @param date java.time.LocalDate实例
     * @return java.util.Date实例
     */
    public static Date convert(LocalDate date) {
        return date == null ? null : Date.from(date.atStartOfDay().atZone(getCurrentZoneId()).toInstant());
    }

    /**
     * 将java.time.LocalDateTime转换成java.util.Date
     *
     * @param datetime java.time.LocalDateTime实例
     * @return java.util.Date实例
     */
    public static Date convert(LocalDateTime datetime) {
        return datetime == null ? null : Date.from(datetime.atZone(getCurrentZoneId()).toInstant());
    }

    /**
     * 将日期文本转换成日期对象
     *
     * @param text 日期文本
     * @return 日期对象实例
     */
    public static Date parse(String text) {
        if (StringUtils.isEmpty(text) || (text = text.trim()).isEmpty()) {
            return null;
        }
        Parser parser = lookupDatetimeParser(text);
        if (parser != null) {
            return parser.parse(text);
        } else if (!StringUtils.isEmpty(text)) {
            throw new IllegalArgumentException("Invalid date: " + text);
        }
        return null;
    }

    /**
     * 将日期文本转换成日期对象
     *
     * @param text      日期文本
     * @param formatter 日期时间格式处理器
     * @return 日期对象实例
     */
    public static Date parse(String text, @NonNull DateTimeFormatter formatter) {
        if (StringUtils.isEmpty(text) || (text = text.trim()).isEmpty()) {
            return null;
        }
        return convert(LocalDateTime.parse(text, formatter));
    }

    /**
     * 日期格式化
     *
     * @param date 日期对象实例
     * @return 日期文本
     */
    public static String format(Date date) {
        return date == null ? null : format(date, isTemporal(date));
    }

    /**
     * 日期格式化
     *
     * @param date     日期对象实例
     * @param temporal 是否格式化时间
     * @return 日期文本
     */
    public static String format(Date date, boolean temporal) {
        LocalDateTime datetime = convert(date);
        return datetime == null ? null : temporal ? format(datetime) : format(datetime.toLocalDate());
    }

    /**
     * 日期格式化
     *
     * @param date 日期对象实例
     * @return 日期文本
     */
    public static String format(LocalDate date) {
        return date == null ? null : DEFAULT_DATE_FORMATTER.format(date);
    }

    /**
     * 日期时间格式化
     *
     * @param datetime 日期时间对象实例
     * @return 日期文本
     */
    public static String format(LocalDateTime datetime) {
        return datetime == null ? null : DEFAULT_DATETIME_FORMATTER.format(datetime);
    }

    /**
     * 将年份转换成数字
     *
     * @param date 日期对象
     * @return 年份数字
     */
    public static String year2number(Date date) {
        return year2number(date, number -> number);
    }

    /**
     * 将年份转换成数字
     *
     * @param date      日期对象
     * @param converter 字符串数字转换器
     * @param <T>       目标数据类型
     * @return 年份数字
     */
    public static <T> T year2number(Date date, @NonNull Function<String, T> converter) {
        return date == null ? null : converter.apply(YEAR_NUMBER_FORMATTER.format(convert(date)));
    }

    /**
     * 将年份转换成数字
     *
     * @param date 日期对象
     * @return 年份数字
     */
    public static String year2number(LocalDate date) {
        return year2number(date, number -> number);
    }

    /**
     * 将年份转换成数字
     *
     * @param date      日期对象
     * @param converter 字符串数字转换器
     * @param <T>       目标数据类型
     * @return 年份数字
     */
    public static <T> T year2number(LocalDate date, @NonNull Function<String, T> converter) {
        return date == null ? null : converter.apply(YEAR_NUMBER_FORMATTER.format(date));
    }

    /**
     * 将年份转换成数字
     *
     * @param datetime 日期时间对象
     * @return 年份数字
     */
    public static String year2number(LocalDateTime datetime) {
        return year2number(datetime, number -> number);
    }

    /**
     * 将年份转换成数字
     *
     * @param datetime  日期时间对象
     * @param converter 字符串数字转换器
     * @param <T>       目标数据类型
     * @return 年份数字
     */
    public static <T> T year2number(LocalDateTime datetime, @NonNull Function<String, T> converter) {
        return datetime == null ? null : converter.apply(YEAR_NUMBER_FORMATTER.format(datetime));
    }

    /**
     * 将月份转换成数字
     *
     * @param date 日期对象
     * @return 月份数字
     */
    public static String month2number(Date date) {
        return month2number(date, number -> number);
    }

    /**
     * 将月份转换成数字
     *
     * @param date      日期对象
     * @param converter 字符串数字转换器
     * @param <T>       目标数据类型
     * @return 月份数字
     */
    public static <T> T month2number(Date date, @NonNull Function<String, T> converter) {
        return date == null ? null : converter.apply(MONTH_NUMBER_FORMATTER.format(convert(date)));
    }

    /**
     * 将月份转换成数字
     *
     * @param date 日期对象
     * @return 月份数字
     */
    public static String month2number(LocalDate date) {
        return month2number(date, number -> number);
    }

    /**
     * 将月份转换成数字
     *
     * @param date      日期对象
     * @param converter 字符串数字转换器
     * @param <T>       目标数据类型
     * @return 月份数字
     */
    public static <T> T month2number(LocalDate date, @NonNull Function<String, T> converter) {
        return date == null ? null : converter.apply(MONTH_NUMBER_FORMATTER.format(date));
    }

    /**
     * 将月份转换成数字
     *
     * @param datetime 日期时间对象
     * @return 月份数字
     */
    public static String month2number(LocalDateTime datetime) {
        return month2number(datetime, number -> number);
    }

    /**
     * 将月份转换成数字
     *
     * @param datetime  日期时间对象
     * @param converter 字符串数字转换器
     * @param <T>       目标数据类型
     * @return 月份数字
     */
    public static <T> T month2number(LocalDateTime datetime, @NonNull Function<String, T> converter) {
        return datetime == null ? null : converter.apply(MONTH_NUMBER_FORMATTER.format(datetime));
    }

    /**
     * 将日期转换成数字
     *
     * @param date 日期对象
     * @return 日期数字
     */
    public static String date2number(Date date) {
        return date2number(date, number -> number);
    }

    /**
     * 将日期转换成数字
     *
     * @param date      日期对象
     * @param converter 字符串数字转换器
     * @param <T>       目标数据类型
     * @return 日期数字
     */
    public static <T> T date2number(Date date, @NonNull Function<String, T> converter) {
        return date == null ? null : converter.apply(DATE_NUMBER_FORMATTER.format(convert(date)));
    }

    /**
     * 将日期转换成数字
     *
     * @param date 日期对象
     * @return 日期数字
     */
    public static String date2number(LocalDate date) {
        return date2number(date, number -> number);
    }

    /**
     * 将日期转换成数字
     *
     * @param date      日期对象
     * @param converter 字符串数字转换器
     * @param <T>       目标数据类型
     * @return 日期数字
     */
    public static <T> T date2number(LocalDate date, @NonNull Function<String, T> converter) {
        return date == null ? null : converter.apply(DATE_NUMBER_FORMATTER.format(date));
    }

    /**
     * 将日期转换成数字
     *
     * @param datetime 日期时间对象
     * @return 日期数字
     */
    public static String date2number(LocalDateTime datetime) {
        return date2number(datetime, number -> number);
    }

    /**
     * 将日期转换成数字
     *
     * @param datetime  日期时间对象
     * @param converter 字符串数字转换器
     * @param <T>       目标数据类型
     * @return 日期数字
     */
    public static <T> T date2number(LocalDateTime datetime, @NonNull Function<String, T> converter) {
        return datetime == null ? null : converter.apply(DATE_NUMBER_FORMATTER.format(datetime));
    }

    /**
     * 将字符串形式数字日期转换成日期对象
     *
     * @param number 字符串形式数字日期
     * @return 日期对象
     */
    public static Date number2date(String number) {
        return StringUtils.isEmpty(number) ? null : convert(LocalDate.parse(number, DATE_NUMBER_FORMATTER));
    }

    /**
     * 将数字形式日期转换成日期对象
     *
     * @param number 数字形式日期
     * @return 日期对象
     */
    public static Date number2date(Integer number) {
        return number == null ? null : convert(LocalDate.parse(String.valueOf(number), DATE_NUMBER_FORMATTER));
    }

    /**
     * 将日期时间转换成cron表达式
     *
     * @param date 日期对象
     * @return cron表达式
     */
    public static String date2cron(Date date) {
        return date2cron(date, number -> number);
    }

    /**
     * 将日期时间转换成cron表达式
     *
     * @param date      日期对象
     * @param converter 字符串数字转换器
     * @param <T>       目标数据类型
     * @return cron表达式
     */
    public static <T> T date2cron(Date date, @NonNull Function<String, T> converter) {
        return date == null ? null : converter.apply(DATETIME_CRON_FORMATTER.format(convert(date)));
    }

    /**
     * 将日期时间转换成数字
     *
     * @param date 日期对象
     * @return 日期时间数字
     */
    public static String datetime2number(Date date) {
        return datetime2number(date, number -> number);
    }

    /**
     * 将日期时间转换成数字
     *
     * @param date      日期对象
     * @param converter 字符串数字转换器
     * @param <T>       目标数据类型
     * @return 日期时间数字
     */
    public static <T> T datetime2number(Date date, @NonNull Function<String, T> converter) {
        return date == null ? null : converter.apply(DATETIME_NUMBER_FORMATTER.format(convert(date)));
    }

    /**
     * 将日期时间转换成数字
     *
     * @param date 日期对象
     * @return 日期时间数字
     */
    public static String datetime2number(LocalDate date) {
        return datetime2number(date, number -> number);
    }

    /**
     * 将日期时间转换成数字
     *
     * @param date      日期对象
     * @param converter 字符串数字转换器
     * @param <T>       目标数据类型
     * @return 日期时间数字
     */
    public static <T> T datetime2number(LocalDate date, @NonNull Function<String, T> converter) {
        return date == null ? null : converter.apply(DATETIME_NUMBER_FORMATTER.format(date));
    }

    /**
     * 将日期时间转换成数字
     *
     * @param datetime 日期时间对象
     * @return 日期时间数字
     */
    public static String datetime2number(LocalDateTime datetime) {
        return datetime2number(datetime, number -> number);
    }

    /**
     * 将日期时间转换成数字
     *
     * @param datetime  日期时间对象
     * @param converter 字符串数字转换器
     * @param <T>       目标数据类型
     * @return 日期时间数字
     */
    public static <T> T datetime2number(LocalDateTime datetime, @NonNull Function<String, T> converter) {
        return datetime == null ? null : converter.apply(DATETIME_NUMBER_FORMATTER.format(datetime));
    }

    /**
     * 将字符串形式数字日期时间转换成日期对象
     *
     * @param number 字符串形式数字日期时间
     * @return 日期对象
     */
    public static Date number2datetime(String number) {
        return StringUtils.isEmpty(number) ? null : convert(LocalDate.parse(number, DATETIME_NUMBER_FORMATTER));
    }

    /**
     * 将数字形式日期时间转换成日期对象
     *
     * @param number 数字形式日期时间
     * @return 日期对象
     */
    public static Date number2datetime(Integer number) {
        return number == null ? null : convert(LocalDate.parse(String.valueOf(number), DATETIME_NUMBER_FORMATTER));
    }

    /**
     * 获取当前日期在当周的位置
     *
     * @return 当周位置
     */
    public static int week() {
        return Calendar.getInstance().get(Calendar.DAY_OF_WEEK);
    }

    /**
     * 获取当前日期在当周的位置
     *
     * @param date 目标日期
     * @return 当周位置
     */
    public static int week(Date date) {
        if (date == null) {
            return 0;
        }
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(date);
        return calendar.get(Calendar.DAY_OF_WEEK);
    }

    /**
     * 日期加减
     *
     * @param duration 有效时间
     * @return 运算后的日期对象实例
     */
    public static Date add(Duration duration) {
        return add(duration, true);
    }

    /**
     * 日期加减
     *
     * @param duration 有效时间
     * @param temporal 是否包含时间
     * @return 运算后的日期对象实例
     */
    public static Date add(Duration duration, boolean temporal) {
        return duration == null ? null : add(Calendar.SECOND, (int) duration.getSeconds(), temporal);
    }

    /**
     * 日期加减
     *
     * @param date     原始日期
     * @param duration 有效时间
     * @return 运算后的日期对象实例
     */
    public static Date add(Date date, Duration duration) {
        return add(date, duration, true);
    }

    /**
     * 日期加减
     *
     * @param date     原始日期
     * @param duration 有效时间
     * @param temporal 是否包含时间
     * @return 运算后的日期对象实例
     */
    public static Date add(Date date, Duration duration, boolean temporal) {
        if (date == null || duration == null) {
            return null;
        }
        return add(date, Calendar.SECOND, (int) duration.getSeconds(), temporal);
    }

    /**
     * 日期加减
     *
     * @param field  加减字段
     * @param amount 加减数量
     * @return 运算后的日期对象实例
     */
    public static Date add(int field, int amount) {
        return add(field, amount, true);
    }

    /**
     * 日期加减
     *
     * @param field    加减字段
     * @param amount   加减数量
     * @param temporal 是否包含时间
     * @return 运算后的日期对象实例
     */
    public static Date add(int field, int amount, boolean temporal) {
        Calendar calendar = Calendar.getInstance();
        calendar.add(field, amount);
        if (!temporal) {
            clearTime(calendar);
        }
        return calendar.getTime();
    }

    /**
     * 日期加减
     *
     * @param date   原始日期
     * @param field  加减字段
     * @param amount 加减数量
     * @return 运算后的日期对象实例
     */
    public static Date add(Date date, int field, int amount) {
        return add(date, field, amount, true);
    }

    /**
     * 日期加减
     *
     * @param date     原始日期
     * @param field    加减字段
     * @param amount   加减数量
     * @param temporal 是否包含时间
     * @return 运算后的日期对象实例
     */
    public static Date add(Date date, int field, int amount, boolean temporal) {
        if (date == null) {
            return null;
        }
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(date);
        calendar.add(field, amount);
        if (!temporal) {
            clearTime(calendar);
        }
        return calendar.getTime();
    }

    /**
     * 根据生日获取年龄
     *
     * @param birthday 生日
     * @return 年龄
     */
    public static Integer age(Date birthday) {
        if (birthday == null) {
            return null;
        }
        Calendar calendar = Calendar.getInstance();
        int currentYear = calendar.get(Calendar.YEAR);
        int currentMonth = calendar.get(Calendar.MONTH);
        int currentDay = calendar.get(Calendar.DAY_OF_MONTH);

        calendar.setTime(birthday);
        int birthdayYear = calendar.get(Calendar.YEAR);
        int birthdayMonth = calendar.get(Calendar.MONTH);
        int birthdayDay = calendar.get(Calendar.DAY_OF_MONTH);
        int age = currentYear - birthdayYear;
        if (currentMonth < birthdayMonth || (currentMonth == birthdayMonth && currentDay < birthdayDay)) {
            age--;
        }
        return age;
    }

    /**
     * 判断目标日期时间是否是今天
     *
     * @param datetime 目标日期时间
     * @return true/false
     */
    public static boolean isToday(Date datetime) {
        return datetime != null && isSameDay(new Date(), datetime);
    }

    /**
     * 判断两个日期是否是同一天
     *
     * @param self  比较日期
     * @param other 被比较日期
     * @return true/false
     */
    public static boolean isSameDay(Date self, Date other) {
        if (self == null || other == null) {
            return false;
        }
        Calendar current = Calendar.getInstance();
        current.setTime(self);
        Calendar target = Calendar.getInstance();
        target.setTime(other);
        return current.get(Calendar.YEAR) == target.get(Calendar.YEAR)
                && current.get(Calendar.MONTH) == target.get(Calendar.MONTH)
                && current.get(Calendar.DATE) == target.get(Calendar.DATE);
    }

    /**
     * 判断日期时间是否在置顶时间范围
     *
     * @param datetime 日期时间
     * @param begin    开始时间
     * @param end      结束时间
     * @return true/false
     */
    public static boolean between(Date datetime, Date begin, Date end) {
        return datetime != null && begin != null && end != null && !datetime.before(begin) && !datetime.after(end);
    }

    /**
     * 获取目标日期时间倒计时（当前时间距置顶日期时间时长）
     *
     * @param datetime 目标日期时间
     * @return 倒计时长
     */
    public static Duration countdown(Date datetime) {
        return datetime == null ? null : Duration.ofMillis(datetime.getTime() - System.currentTimeMillis());
    }

    /**
     * 判断日期时间是否在某一时间之前
     *
     * @param datetime 日期时间
     * @param when     某一时间
     * @return true/false
     */
    public static boolean before(Date datetime, Date when) {
        return datetime != null && when != null && datetime.before(when);
    }

    /**
     * 判断日期时间是否在某一时间之后
     *
     * @param datetime 日期时间
     * @param when     某一时间
     * @return true/false
     */
    public static boolean after(Date datetime, Date when) {
        return datetime != null && when != null && datetime.after(when);
    }

    /**
     * 根据指定过期时间判断是否已过期
     *
     * @param expiration 过期时间
     * @return true/false
     */
    public static boolean expired(Date expiration) {
        return expiration == null || expired(new Date(), expiration);
    }

    /**
     * 根据过期时间判断指定日期时间是否已过期
     *
     * @param datetime   日期时间
     * @param expiration 过期时间
     * @return true/false
     */
    public static boolean expired(Date datetime, Date expiration) {
        return datetime == null || expiration == null || datetime.after(expiration);
    }

    /**
     * 判断当天是否是工作日（周一至周五）
     *
     * @return true/false
     */
    public static boolean isWeekday() {
        int week = Calendar.getInstance().get(Calendar.DAY_OF_WEEK);
        return week > Calendar.SUNDAY && week < Calendar.SATURDAY;
    }

    /**
     * 判断指定日期是否是工作日（周一至周五）
     *
     * @param date 日期对象
     * @return true/false
     */
    public static boolean isWeekday(Date date) {
        if (date == null) {
            return false;
        }
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(date);
        int week = calendar.get(Calendar.DAY_OF_WEEK);
        return week > Calendar.SUNDAY && week < Calendar.SATURDAY;
    }

    /**
     * 判断当前时间是否是工作时间（09:00至18:00）
     *
     * @return true/false
     */
    public static boolean isWorktime() {
        int hour = Calendar.getInstance().get(Calendar.HOUR_OF_DAY);
        return hour >= 9 && hour < 18;
    }

    /**
     * 判断指定日期时间是否是工作时间（09:00至18:00）
     *
     * @param date 日期时间对象
     * @return true/false
     */
    public static boolean isWorktime(Date date) {
        if (date == null) {
            return false;
        }
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(date);
        int hour = calendar.get(Calendar.HOUR_OF_DAY);
        return hour >= 9 && hour < 18;
    }

    /**
     * 获取当前年份
     *
     * @return 年份
     */
    public static int year() {
        return get(Calendar.YEAR);
    }

    /**
     * 获取当前月份
     *
     * @return 月份
     */
    public static int month() {
        return get(Calendar.MONTH) + 1;
    }

    /**
     * 获取当前日期
     *
     * @return 日期
     */
    public static int date() {
        return get(Calendar.DATE);
    }

    /**
     * 获取当前小时
     *
     * @return 小时
     */
    public static int hour() {
        return get(Calendar.HOUR_OF_DAY);
    }

    /**
     * 获取当前分钟
     *
     * @return 分钟
     */
    public static int minute() {
        return get(Calendar.MINUTE);
    }

    /**
     * 获取当前秒钟
     *
     * @return 秒钟
     */
    public static int second() {
        return get(Calendar.SECOND);
    }

    /**
     * 获取当前毫秒
     *
     * @return 毫秒
     */
    public static int millis() {
        return get(Calendar.MILLISECOND);
    }

    /**
     * 获取当前时间信息
     *
     * @param field 时间信息字段
     * @return 时间信息
     */
    public static int get(int field) {
        Calendar calendar = Calendar.getInstance();
        return calendar.get(field);
    }

    /**
     * 叠加日期
     *
     * @param date  日期对象
     * @param other 日期对象
     * @return 日期对象
     */
    public static Date superpose(Date date, Date other) {
        if (date == null) {
            return other;
        } else if (other == null) {
            return date;
        }

        long t1 = date.getTime(), t2 = other.getTime(), now = System.currentTimeMillis();
        if (t1 <= now && t2 <= now) {
            return t1 < t2 ? other : date;
        } else if (t1 > now && t2 > now) {
            return new Date(t1 + t2 - now);
        }
        return t1 > now ? date : other;
    }

    /**
     * 获取当前时间戳
     *
     * @return 时间戳
     */
    public static long timestamp() {
        return timestamp(true);
    }

    /**
     * 获取当前时间戳
     *
     * @param millis 是否保留毫秒数
     * @return 时间戳
     */
    public static long timestamp(boolean millis) {
        return timestamp(System.currentTimeMillis(), millis);
    }

    /**
     * 获取日期时间戳
     *
     * @param date 日期对象
     * @return 时间戳
     */
    public static long timestamp(@NonNull Date date) {
        return timestamp(date, true);
    }

    /**
     * 获取日期时间戳
     *
     * @param date   日期对象
     * @param millis 是否保留毫秒数
     * @return 时间戳
     */
    public static long timestamp(@NonNull Date date, boolean millis) {
        return timestamp(date.getTime(), millis);
    }

    /**
     * 获取时间戳
     *
     * @param timestamp 原始时间戳
     * @param millis    是否保留毫秒数
     * @return 时间戳
     */
    public static long timestamp(long timestamp, boolean millis) {
        if (millis) {
            return timestamp;
        }
        int millisecond = (int) (timestamp % 1000);
        timestamp = timestamp / 1000;
        if (millisecond >= 500) {
            timestamp++;
        }
        return timestamp * 1000;
    }
}
