package com.chelvc.framework.base.util;

import java.math.BigDecimal;
import java.math.RoundingMode;

import lombok.NonNull;

/**
 * 小数工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
public final class DecimalUtils {
    /**
     * 默认小数保留位数
     */
    public static final int DEFAULT_SCALE = 2;

    /**
     * 100
     */
    public static final BigDecimal HUNDRED = new BigDecimal(100);

    /**
     * 默认四舍五入模式
     */
    public static final RoundingMode DEFAULT_MODE = RoundingMode.HALF_UP;

    private DecimalUtils() {
    }

    /**
     * 除法
     *
     * @param dividend 被除数
     * @param divisor  除数
     * @return 结果
     */
    public static BigDecimal divide(Long dividend, Long divisor) {
        return divide(dividend, divisor, DEFAULT_SCALE, DEFAULT_MODE);
    }

    /**
     * 除法
     *
     * @param dividend 被除数
     * @param divisor  除数
     * @param scale    小数位数
     * @param mode     四舍五入模式
     * @return 结果
     */
    public static BigDecimal divide(Long dividend, Long divisor, int scale, RoundingMode mode) {
        BigDecimal zero = BigDecimal.ZERO.setScale(scale, mode);
        return dividend == null || dividend == 0 || divisor == null || divisor == 0 ? zero :
                BigDecimal.valueOf(dividend).divide(BigDecimal.valueOf(divisor), scale, mode);
    }

    /**
     * 除法
     *
     * @param dividend 被除数
     * @param divisor  除数
     * @return 结果
     */
    public static BigDecimal divide(BigDecimal dividend, BigDecimal divisor) {
        return divide(dividend, divisor, DEFAULT_SCALE, DEFAULT_MODE);
    }

    /**
     * 除法
     *
     * @param dividend 被除数
     * @param divisor  除数
     * @param scale    小数位数
     * @param mode     四舍五入模式
     * @return 结果
     */
    public static BigDecimal divide(BigDecimal dividend, BigDecimal divisor, int scale, RoundingMode mode) {
        return dividend == null || dividend.equals(BigDecimal.ZERO) ||
                divisor == null || divisor.equals(BigDecimal.ZERO) ?
                BigDecimal.ZERO.setScale(scale, mode) : dividend.divide(divisor, scale, mode);
    }

    /**
     * 除法
     *
     * @param dividend 被除数
     * @param divisor  除数
     * @return 结果
     */
    public static BigDecimal divide(BigDecimal dividend, Long divisor) {
        return divide(dividend, divisor, DEFAULT_SCALE, DEFAULT_MODE);
    }

    /**
     * 除法
     *
     * @param dividend 被除数
     * @param divisor  除数
     * @param scale    小数位数
     * @param mode     四舍五入模式
     * @return 结果
     */
    public static BigDecimal divide(BigDecimal dividend, Long divisor, int scale, RoundingMode mode) {
        return dividend == null || dividend.equals(BigDecimal.ZERO) || divisor == null || divisor == 0 ?
                BigDecimal.ZERO.setScale(scale, mode) : dividend.divide(BigDecimal.valueOf(divisor), scale, mode);
    }

    /**
     * 除法
     *
     * @param dividend 被除数
     * @param divisor  除数
     * @return 结果
     */
    public static BigDecimal divide(Long dividend, BigDecimal divisor) {
        return divide(dividend, divisor, DEFAULT_SCALE, DEFAULT_MODE);
    }

    /**
     * 除法
     *
     * @param dividend 被除数
     * @param divisor  除数
     * @param scale    小数位数
     * @param mode     四舍五入模式
     * @return 结果
     */
    public static BigDecimal divide(Long dividend, BigDecimal divisor, int scale, RoundingMode mode) {
        return dividend == null || dividend == 0 || divisor == null || divisor.equals(BigDecimal.ZERO) ?
                BigDecimal.ZERO.setScale(scale, mode) : BigDecimal.valueOf(dividend).divide(divisor, scale, mode);
    }

    /**
     * 格式化小数
     *
     * @param decimal 小数
     * @return 小数字符串
     */
    public static String format(BigDecimal decimal) {
        return format(decimal, false);
    }

    /**
     * 格式化小数
     *
     * @param decimal   小数
     * @param trimZeros 是否去掉多余的0
     * @return 小数字符串
     */
    public static String format(BigDecimal decimal, boolean trimZeros) {
        return format(decimal, DEFAULT_SCALE, RoundingMode.HALF_UP, trimZeros);
    }

    /**
     * 格式化小数
     *
     * @param decimal 小数
     * @param scale   小数保留位数
     * @param mode    小数舍弃模式
     * @return 小数字符串
     */
    public static String format(BigDecimal decimal, int scale, @NonNull RoundingMode mode) {
        return format(decimal, scale, mode, false);
    }

    /**
     * 格式化小数
     *
     * @param decimal   小数
     * @param scale     小数保留位数
     * @param mode      小数舍弃模式
     * @param trimZeros 是否去掉多余的0
     * @return 小数字符串
     */
    public static String format(BigDecimal decimal, int scale, @NonNull RoundingMode mode, boolean trimZeros) {
        if (decimal == null) {
            return null;
        }
        decimal = decimal.setScale(scale, mode);
        if (trimZeros) {
            decimal = decimal.stripTrailingZeros();
        }
        return decimal.toPlainString();
    }
}
