package com.chelvc.framework.base.util;

import java.util.Arrays;
import java.util.Collection;
import java.util.Map;
import java.util.Objects;

import com.chelvc.framework.base.exception.ParameterInvalidException;
import com.chelvc.framework.base.exception.ResourceInvalidException;
import com.chelvc.framework.base.exception.ResourceUnavailableException;
import lombok.NonNull;
import org.springframework.util.CollectionUtils;

/**
 * 异常工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
public final class ErrorUtils {
    private ErrorUtils() {
    }

    /**
     * 查找异常栈中根异常
     *
     * @param throwable 异常栈
     * @return 根异常
     */
    public static Throwable root(Throwable throwable) {
        while (throwable != null && throwable.getCause() != null) {
            throwable = throwable.getCause();
        }
        return throwable;
    }

    /**
     * 判断异常是否包含任意指定异常
     *
     * @param throwable 异常栈
     * @param types     异常类型数组
     * @return true/false
     */
    public static boolean isError(Throwable throwable, Class<? extends Throwable>[] types) {
        return types != null && types.length > 0 && isError(throwable, Arrays.asList(types));
    }

    /**
     * 判断异常是否包含任意指定异常
     *
     * @param throwable 异常栈
     * @param types     异常类型数组
     * @return true/false
     */
    public static boolean isError(Throwable throwable, Collection<Class<? extends Throwable>> types) {
        return !CollectionUtils.isEmpty(types)
                && types.stream().filter(Objects::nonNull).anyMatch(type -> contains(throwable, type));
    }

    /**
     * 判断异常栈是否包含制定类型的异常
     *
     * @param throwable 异常栈
     * @param type      异常类型
     * @return true/false
     */
    public static boolean contains(Throwable throwable, Class<? extends Throwable> type) {
        return lookup(throwable, type) != null;
    }

    /**
     * 从异常栈中查找制定类型的异常
     *
     * @param throwable 异常栈
     * @param type      异常类型
     * @param <T>       异常类型泛型
     * @return 异常对象
     */
    @SuppressWarnings("unchecked")
    public static <T extends Throwable> T lookup(Throwable throwable, @NonNull Class<T> type) {
        for (; throwable != null && !type.isAssignableFrom(throwable.getClass()); throwable = throwable.getCause()) ;
        return (T) throwable;
    }

    /**
     * 非空资源校验
     *
     * @param resource 资源对象
     */
    public static void requireResource(Object resource) {
        requireResource(resource, "Invalid resource");
    }

    /**
     * 非空资源校验
     *
     * @param resource 资源对象
     * @param message  异常信息
     */
    public static void requireResource(Object resource, String message) {
        requireResource(Objects.nonNull(resource), message);
    }

    /**
     * 非空资源校验
     *
     * @param expression 验证表达式
     * @param message    异常信息
     */
    public static void requireResource(boolean expression, String message) {
        if (!expression) {
            throw new ResourceInvalidException(message);
        }
    }

    /**
     * 可用资源校验
     *
     * @param expression 验证表达式
     */
    public static void availableResource(boolean expression) {
        availableResource(expression, "Unavailable resource");
    }

    /**
     * 可用资源校验
     *
     * @param expression 验证表达式
     * @param message    异常信息
     */
    public static void availableResource(boolean expression, String message) {
        availableResource(expression, null, message);
    }

    /**
     * 可用资源校验
     *
     * @param expression 验证表达式
     * @param code       异常编码
     * @param message    异常信息
     */
    public static void availableResource(boolean expression, String code, String message) {
        if (!expression) {
            throw new ResourceUnavailableException(code, message);
        }
    }

    /**
     * 无效参数
     *
     * @param name    参数名称
     * @param message 异常信息
     */
    public static void invalidParameter(String name, String message) {
        throw new ParameterInvalidException(name, message);
    }

    /**
     * 无效参数
     *
     * @param errors 参数名称/异常信息映射表
     */
    public static void invalidParameter(Map<String, String> errors) {
        throw new ParameterInvalidException(errors);
    }
}
