package com.chelvc.framework.base.util;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.time.LocalDateTime;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeSet;
import java.util.UUID;
import java.util.function.BiFunction;

import lombok.NonNull;
import org.apache.poi.openxml4j.exceptions.OpenXML4JException;
import org.apache.poi.openxml4j.opc.OPCPackage;
import org.apache.poi.ss.formula.FormulaParseException;
import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.CellStyle;
import org.apache.poi.ss.usermodel.CellType;
import org.apache.poi.ss.usermodel.Comment;
import org.apache.poi.ss.usermodel.DateUtil;
import org.apache.poi.ss.usermodel.FormulaError;
import org.apache.poi.ss.usermodel.Hyperlink;
import org.apache.poi.ss.usermodel.RichTextString;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.ss.util.CellAddress;
import org.apache.poi.ss.util.CellRangeAddress;
import org.apache.poi.xssf.eventusermodel.XSSFReader;
import org.apache.poi.xssf.model.SharedStringsTable;
import org.apache.poi.xssf.streaming.SXSSFWorkbook;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.springframework.util.Assert;
import org.springframework.util.CollectionUtils;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;

/**
 * Excel工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
public final class ExcelUtils {
    /**
     * 批量抓取数据量
     */
    public static final int BATCH_FETCH_SIZE = 1000;

    /**
     * XLS工作表容量大小
     */
    public static final int XLS_SHEET_VOLUME = 50000;

    /**
     * XLSX工作表容量大小
     */
    public static final int XLSX_SHEET_VOLUME = 1000000;

    private ExcelUtils() {
    }

    /**
     * 将Excel列字母字符串转换成下标
     *
     * @param column 列字母字符串
     * @return 列下标
     */
    public static int column2number(@NonNull String column) {
        if (!StringUtils.isLetter(column)) {
            throw new IllegalArgumentException("Invalid column: " + column);
        }
        int basic = 1, number = 0;
        for (int i = column.toUpperCase().length() - 1; i >= 0; i--) {
            number += (Character.toUpperCase(column.charAt(i)) - 'A' + 1) * basic;
            basic *= 26;
        }
        return number - 1;
    }

    /**
     * 将单元格地址字符串转换成地址对象
     *
     * @param address 单元格地址字符串
     * @return 单元格地址对象
     */
    public static CellAddress string2address(@NonNull String address) {
        for (int i = 0; i < address.length(); i++) {
            char c = address.charAt(i);
            if (c >= '1' && c <= '9') {
                int row = Integer.parseInt(address.substring(i)) - 1;
                int column = column2number(address.substring(0, i));
                return new CellAddress(row, column);
            }
        }
        throw new IllegalArgumentException("Invalid address: " + address);
    }

    /**
     * 判断Excel数据行是否为空
     *
     * @param row Excel数据行对象
     * @return true/false
     */
    public static boolean isEmpty(Row row) {
        if (row != null) {
            for (int i = row.getFirstCellNum(), len = row.getLastCellNum(); i < len; i++) {
                Object value = getValue(row.getCell(i));
                if (value != null && (!(value instanceof CharSequence) || !StringUtils.isBlank((CharSequence) value))) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * 获取Excel单元格值
     *
     * @param cell Excel单元格对象
     * @return 值
     */
    public static Object getValue(Cell cell) {
        if (cell == null) {
            return null;
        } else if (cell instanceof XMLCell) {
            Object value = ((XMLCell) cell).value;
            return value instanceof String ? StringUtils.trim((String) value) : value;
        }
        CellType type = cell.getCellType();
        if (type == CellType.BLANK) {
            return null;
        } else if (type == CellType.BOOLEAN) {
            return cell.getBooleanCellValue();
        } else if (type == CellType.NUMERIC) {
            return DateUtil.isCellDateFormatted(cell) ? cell.getDateCellValue() : cell.getNumericCellValue();
        } else if (type == CellType.ERROR) {
            return FormulaError.forInt(cell.getErrorCellValue()).getString();
        }
        return StringUtils.trim(cell.getStringCellValue());
    }

    /**
     * 通过解析XML的方式读取Excel数据
     *
     * @param pkg    文件包
     * @param reader Excel读接口
     * @return 读取数量
     */
    public static int read(OPCPackage pkg, Reader reader) {
        return read(pkg, 0, reader);
    }

    /**
     * 通过解析XML的方式读取Excel数据
     *
     * @param pkg    文件包
     * @param index  开始数据行下标（从0开始）
     * @param reader Excel读接口
     * @return 读取数量
     */
    public static int read(@NonNull OPCPackage pkg, int index, @NonNull Reader reader) {
        return new AbstractExcelReader(pkg, index) {
            @Override
            public void read(Row row, int count) {
                reader.read(row, count);
            }
        }.process();
    }

    /**
     * 初始化工作表
     *
     * @param workbook 工作薄实例
     * @param titles   表头数组
     * @return 工作表实例
     */
    public static Sheet initializeSheet(@NonNull Workbook workbook, @NonNull String... titles) {
        return initializeSheet(workbook, null, null, titles);
    }

    /**
     * 初始化工作表
     *
     * @param workbook 工作薄实例
     * @param name     工作表名称
     * @param titles   表头数组
     * @return 工作表实例
     */
    public static Sheet initializeSheet(@NonNull Workbook workbook, String name, @NonNull String... titles) {
        return initializeSheet(workbook, name, null, titles);
    }

    /**
     * 初始化工作表
     *
     * @param workbook 工作薄实例
     * @param style    表头样式
     * @param titles   表头数组
     * @return 工作表实例
     */
    public static Sheet initializeSheet(@NonNull Workbook workbook, CellStyle style, @NonNull String... titles) {
        return initializeSheet(workbook, null, style, titles);
    }

    /**
     * 初始化工作表
     *
     * @param workbook 工作薄实例
     * @param name     工作表名称
     * @param style    表头样式
     * @param titles   表头数组
     * @return 工作表实例
     */
    public static Sheet initializeSheet(@NonNull Workbook workbook, String name, CellStyle style,
                                        @NonNull String... titles) {
        Sheet sheet = StringUtils.isEmpty(name) ? workbook.createSheet() : workbook.createSheet(name);
        if (titles.length == 0) {
            return sheet;
        }
        Row head = sheet.createRow(0);
        if (style != null) {
            head.setRowStyle(style);
        }
        for (int i = 0; i < titles.length; i++) {
            Cell cell = head.createCell(i);
            if (style != null) {
                cell.setCellStyle(style);
            }
            cell.setCellValue(titles[i]);
        }
        return sheet;
    }

    /**
     * 将对象实例写入到Excel中
     *
     * @param sheet    Excel工作表
     * @param provider 对象实例列表分页提供方法
     * @param writer   Excel对象实例写入接口
     * @param <T>      数据类型
     * @return 写入数量
     */
    public static <T> int write(@NonNull Sheet sheet, @NonNull BiFunction<Integer, Integer, Collection<T>> provider,
                                @NonNull Writer<T> writer) {
        return write(sheet, 0, provider, writer);
    }

    /**
     * 将对象实例写入到Excel中
     *
     * @param sheet    Excel工作表
     * @param index    开始行下标
     * @param provider 对象实例列表分页提供方法
     * @param writer   Excel对象实例写入接口
     * @param <T>      数据类型
     * @return 写入数量
     */
    public static <T> int write(@NonNull Sheet sheet, int index,
                                @NonNull BiFunction<Integer, Integer, Collection<T>> provider,
                                @NonNull Writer<T> writer) {
        int page = 1, size = BATCH_FETCH_SIZE, count = 0;
        while (true) {
            Collection<T> objects = provider.apply(page++, size);
            if (CollectionUtils.isEmpty(objects)) {
                break;
            }
            for (T object : objects) {
                writer.write(sheet.createRow(index++), object, ++count);
            }
            if (objects.size() < size) {
                break;
            }
        }
        return count;
    }

    /**
     * 将对象实例写入到Excel中
     *
     * @param workbook Excel工作薄
     * @param provider 对象实例列表分页提供方法
     * @param writer   Excel对象实例写入接口
     * @param titles   表头数组
     * @param <T>      数据类型
     * @return 写入数量
     */
    public static <T> int write(@NonNull Workbook workbook,
                                @NonNull BiFunction<Integer, Integer, Collection<T>> provider,
                                @NonNull Writer<T> writer, @NonNull String... titles) {
        int page = 1, size = BATCH_FETCH_SIZE, index = 0, origin = titles.length == 0 ? 0 : 1, row = origin;
        int volume = workbook instanceof SXSSFWorkbook || workbook instanceof XSSFWorkbook ?
                XLSX_SHEET_VOLUME : XLS_SHEET_VOLUME;
        Sheet sheet = initializeSheet(workbook, titles);
        while (true) {
            Collection<T> objects = provider.apply(page++, size);
            if (CollectionUtils.isEmpty(objects)) {
                break;
            }
            for (T object : objects) {
                if (index++ > volume) {
                    sheet = initializeSheet(workbook, titles);
                    row = origin;
                }
                writer.write(sheet.createRow(row++), object, index);
            }
            if (objects.size() < size) {
                break;
            }
        }
        return index;
    }

    /**
     * 保存工作薄到文件
     *
     * @param workbook 工作薄
     * @return 文件
     */
    public static File save(@NonNull Workbook workbook) {
        return save(workbook, FileUtils.EXPORT);
    }

    /**
     * 保存工作薄到文件
     *
     * @param workbook 工作薄
     * @param path     文件目录
     * @return 文件
     */
    public static File save(@NonNull Workbook workbook, @NonNull String path) {
        File file = new File(path, UUID.randomUUID().toString() + ".xlsx");
        try (OutputStream output = new FileOutputStream(file)) {
            workbook.write(output);
            workbook.close();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
        return file;
    }

    /**
     * Excel读接口
     */
    public interface Reader {
        /**
         * 读取Excel数据行并转换成对象实例
         *
         * @param row   数据行对象
         * @param count 当前记录数（从1开始）
         */
        void read(Row row, int count);
    }

    /**
     * Excel写接口
     *
     * @param <T> 数据模型
     */
    public interface Writer<T> {
        /**
         * 将对象实例写入到Excel数据行
         *
         * @param row    数据行对象
         * @param object 对象实例
         * @param count  当前记录数（从1开始）
         */
        void write(Row row, T object, int count);
    }

    /**
     * 基于XML的Excel行实现
     */
    public static class XMLRow implements Row {
        /**
         * 单元格列下标列表（从小到大排序）
         */
        private final TreeSet<Integer> columns = new TreeSet<>();
        /**
         * 单元格列下标/单元格映射表
         */
        private final Map<Integer, Cell> cells = new HashMap<>();
        /**
         * 行下标（从0开始）
         */
        private int index;

        @Override
        public Cell createCell(int i) {
            return this.createCell(i, CellType.BLANK);
        }

        @Override
        public Cell createCell(int i, CellType type) {
            Cell cell = new XMLCell(this, i, type);
            this.cells.put(i, cell);
            this.columns.add(i);
            return cell;
        }

        @Override
        public void removeCell(Cell cell) {
            if (cell != null) {
                this.cells.remove(cell.getColumnIndex());
            }
        }

        @Override
        public int getRowNum() {
            return this.index;
        }

        @Override
        public void setRowNum(int i) {
            this.index = i;
        }

        @Override
        public Cell getCell(int i) {
            return this.getCell(i, MissingCellPolicy.RETURN_NULL_AND_BLANK);
        }

        @Override
        public Cell getCell(int i, MissingCellPolicy missingCellPolicy) {
            return this.cells.get(i);
        }

        @Override
        public short getFirstCellNum() {
            return this.columns.isEmpty() ? -1 : this.columns.first().shortValue();
        }

        @Override
        public short getLastCellNum() {
            return this.columns.isEmpty() ? -1 : (short) (this.columns.last().shortValue() + 1);
        }

        @Override
        public int getPhysicalNumberOfCells() {
            return this.columns.size();
        }

        @Override
        public void shiftCellsRight(int i, int i1, int i2) {
            throw new UnsupportedOperationException();
        }

        @Override
        public void shiftCellsLeft(int i, int i1, int i2) {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean getZeroHeight() {
            throw new UnsupportedOperationException();
        }

        @Override
        public void setZeroHeight(boolean b) {
            throw new UnsupportedOperationException();
        }

        @Override
        public short getHeight() {
            throw new UnsupportedOperationException();
        }

        @Override
        public void setHeight(short i) {
            throw new UnsupportedOperationException();
        }

        @Override
        public float getHeightInPoints() {
            throw new UnsupportedOperationException();
        }

        @Override
        public void setHeightInPoints(float v) {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean isFormatted() {
            throw new UnsupportedOperationException();
        }

        @Override
        public CellStyle getRowStyle() {
            throw new UnsupportedOperationException();
        }

        @Override
        public void setRowStyle(CellStyle cellStyle) {
            throw new UnsupportedOperationException();
        }

        @Override
        public Iterator<Cell> cellIterator() {
            return new Iterator<Cell>() {
                private final Iterator<Integer> iterator = columns.iterator();

                @Override
                public boolean hasNext() {
                    return this.iterator.hasNext();
                }

                @Override
                public Cell next() {
                    return cells.get(this.iterator.next());
                }
            };
        }

        @Override
        public Sheet getSheet() {
            throw new UnsupportedOperationException();
        }

        @Override
        public int getOutlineLevel() {
            throw new UnsupportedOperationException();
        }

        @Override
        public Iterator<Cell> iterator() {
            return this.cellIterator();
        }

        @Override
        public String toString() {
            Integer last = -1;
            List<Cell> cells = new LinkedList<>();
            for (Integer column : this.columns) {
                int differ = column - last;
                if (differ > 1) {
                    for (int i = 0, len = differ - 1; i < len; i++) {
                        cells.add(null);
                    }
                }
                cells.add(this.cells.get(column));
                last = column;
            }
            return cells.toString();
        }
    }

    /**
     * 基于XML的Excel单元格实现
     */
    public static class XMLCell implements Cell {
        private final Row row;
        private final int column;
        private CellType type;
        private Object value;

        public XMLCell(@NonNull Row row, int column, CellType type) {
            Assert.isTrue(column > -1, "column must be greater than -1");
            this.row = row;
            this.column = column;
            this.type = type;
        }

        @Override
        public int getColumnIndex() {
            return this.column;
        }

        @Override
        public int getRowIndex() {
            return this.row.getRowNum();
        }

        @Override
        public Sheet getSheet() {
            return this.row.getSheet();
        }

        @Override
        public Row getRow() {
            return this.row;
        }

        @Override
        public CellType getCellType() {
            return this.type;
        }

        @Override
        public void setCellType(CellType type) {
            this.type = type;
        }

        @Override
        public CellType getCachedFormulaResultType() {
            throw new UnsupportedOperationException();
        }

        @Override
        public void setBlank() {
            this.value = StringUtils.EMPTY;
        }

        @Override
        public void setCellValue(LocalDateTime value) {
            this.value = value;
        }

        @Override
        public void setCellValue(double value) {
            this.value = value;
        }

        @Override
        public void setCellValue(Date value) {
            this.value = value;
        }

        @Override
        public void setCellValue(Calendar value) {
            this.value = value;
        }

        @Override
        public void setCellValue(RichTextString value) {
            this.value = value;
        }

        @Override
        public void setCellValue(String value) {
            this.value = value;
        }

        @Override
        public void setCellValue(boolean value) {
            this.value = value;
        }

        @Override
        public void setCellErrorValue(byte value) {
            this.value = value;
        }

        @Override
        public String getCellFormula() {
            throw new UnsupportedOperationException();
        }

        @Override
        public void setCellFormula(String s) throws FormulaParseException {
            throw new UnsupportedOperationException();
        }

        @Override
        public double getNumericCellValue() {
            return this.type == CellType.BLANK ? 0.0D : (double) this.value;
        }

        @Override
        public Date getDateCellValue() {
            return (Date) this.value;
        }

        @Override
        public LocalDateTime getLocalDateTimeCellValue() {
            return (LocalDateTime) this.value;
        }

        @Override
        public RichTextString getRichStringCellValue() {
            return (RichTextString) this.value;
        }

        @Override
        public String getStringCellValue() {
            return (String) this.value;
        }

        @Override
        public boolean getBooleanCellValue() {
            return this.type != CellType.BLANK && (boolean) this.value;
        }

        @Override
        public byte getErrorCellValue() {
            throw new UnsupportedOperationException();
        }

        @Override
        public CellStyle getCellStyle() {
            throw new UnsupportedOperationException();
        }

        @Override
        public void setCellStyle(CellStyle cellStyle) {
            throw new UnsupportedOperationException();
        }

        @Override
        public void setAsActiveCell() {
            throw new UnsupportedOperationException();
        }

        @Override
        public CellAddress getAddress() {
            throw new UnsupportedOperationException();
        }

        @Override
        public Comment getCellComment() {
            throw new UnsupportedOperationException();
        }

        @Override
        public void setCellComment(Comment comment) {
            throw new UnsupportedOperationException();
        }

        @Override
        public void removeFormula() throws IllegalStateException {
            throw new UnsupportedOperationException();
        }

        @Override
        public void removeCellComment() {
            throw new UnsupportedOperationException();
        }

        @Override
        public Hyperlink getHyperlink() {
            throw new UnsupportedOperationException();
        }

        @Override
        public void setHyperlink(Hyperlink hyperlink) {
            throw new UnsupportedOperationException();
        }

        @Override
        public void removeHyperlink() {
            throw new UnsupportedOperationException();
        }

        @Override
        public CellRangeAddress getArrayFormulaRange() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean isPartOfArrayFormulaGroup() {
            throw new UnsupportedOperationException();
        }

        @Override
        public String toString() {
            return this.value == null ? null : this.value.toString();
        }
    }

    /**
     * Excel数据读取处理器抽象实现
     */
    public static abstract class AbstractExcelReader extends DefaultHandler implements Reader {
        /**
         * 开始行下标
         */
        private final int index;
        /**
         * 值是否为日期
         */
        protected boolean datable;
        /**
         * 数据是否关联
         */
        protected boolean related;
        /**
         * 文件包
         */
        protected OPCPackage pkg;
        /**
         * 共享字符串表
         */
        protected SharedStringsTable shared;
        /**
         * 当前行
         */
        private Row row;
        /**
         * 迭代数量
         */
        private int count;
        /**
         * 当前列下标
         */
        private int column;
        /**
         * 当前单元格值
         */
        private String value;

        public AbstractExcelReader(@NonNull OPCPackage pkg, int index) {
            Assert.isTrue(index > -1, "index must be greater than -1");
            this.pkg = pkg;
            this.index = index;
            this.value = StringUtils.EMPTY;
        }

        /**
         * 构建XML数据读取处理器
         *
         * @return XML数据读取处理器
         * @throws SAXException 构建异常
         */
        protected XMLReader buildXMLReader() throws SAXException {
            return XMLReaderFactory.createXMLReader("org.apache.xerces.parsers.SAXParser");
        }

        /**
         * 根据单元格标签属性解析单元格类型，解析单元格开始标签时调用
         *
         * @param attributes 标签属性对象
         * @return 单元格类型
         */
        protected CellType analyseCellType(Attributes attributes) {
            String t = attributes.getValue("t");
            String s = attributes.getValue("s");
            this.datable = "1".equals(s) || "2".equals(s);
            if ((this.related = "s".equals(t)) || "inlineStr".equals(t)) {
                return CellType.STRING;
            } else if ("b".equals(t)) {
                return CellType.BOOLEAN;
            } else if (this.datable || StringUtils.isEmpty(t)) {
                return CellType.NUMERIC;
            }
            return CellType.STRING;
        }

        /**
         * 根据单元格标签属性解析单元格地址，解析单元格开始标签时调用
         *
         * @param attributes 标签属性对象
         * @return 单元格地址
         */
        protected CellAddress analyseCellAddress(Attributes attributes) {
            return string2address(attributes.getValue("r"));
        }

        /**
         * 初始化单元格值，解析单元格值结束标签时调用
         *
         * @param cell  单元格对象
         * @param value 原始值
         */
        protected void initializeCellValue(Cell cell, String value) {
            if (this.related) {
                // 如果为字符串，则从共享关联表中取数据
                cell.setCellValue(this.shared.getItemAt(Integer.parseInt(value)).getString());
            } else if (cell.getCellType() == CellType.BOOLEAN) {
                cell.setCellValue(Integer.parseInt(value) > 0);
            } else if (cell.getCellType() == CellType.NUMERIC) {
                double number = Double.parseDouble(value);
                if (this.datable) {
                    cell.setCellValue(DateUtil.getJavaDate(number));
                } else {
                    cell.setCellValue(number);
                }
            } else {
                cell.setCellValue(value);
            }
        }

        /**
         * Excel读操作
         *
         * @return 数据行总数
         */
        public int process() {
            try {
                XMLReader parser = this.buildXMLReader();
                parser.setContentHandler(this);
                XSSFReader reader = new XSSFReader(this.pkg);
                this.shared = reader.getSharedStringsTable();
                Iterator<InputStream> sheets = reader.getSheetsData();
                while (sheets.hasNext()) {
                    try (InputStream sheet = sheets.next()) {
                        parser.parse(new InputSource(sheet));
                    }
                }
            } catch (IOException | SAXException | OpenXML4JException e) {
                throw new RuntimeException(e);
            }
            return this.count;
        }

        @Override
        public void characters(char[] ch, int start, int length) {
            if (this.value.isEmpty()) {
                this.value = new String(ch, start, length);
            } else {
                this.value += new String(ch, start, length);
            }
        }

        @Override
        public void startElement(String uri, String localName, String name, Attributes attributes) {
            if ("row".equals(name)) { // 行开始标签
                this.row = null; // 清空当前行
                int number = Integer.parseInt(attributes.getValue("r")) - 1; // 当前行下标
                if (number >= this.index) {
                    this.row = new XMLRow();
                    this.row.setRowNum(number);
                }
            } else if (this.row != null && this.row.getRowNum() >= this.index) {
                this.value = StringUtils.EMPTY; // 清空过程数据
                if ("c".equals(name)) { // 单元格开始标签
                    // 解析单元格所在列下标
                    this.column = this.analyseCellAddress(attributes).getColumn();
                    // 构建单元格并初始化单元格值类型
                    this.row.createCell(this.column, this.analyseCellType(attributes));
                }
            }
        }

        @Override
        public void endElement(String uri, String localName, String name) {
            if (this.row != null && this.row.getRowNum() >= this.index) {
                if ("row".equals(name) && !isEmpty(this.row)) { // 行解析完成
                    this.read(this.row, ++this.count);
                } else if (!this.value.isEmpty() && ("v".equals(name) || "t".equals(name))) { // 值解析完成
                    this.initializeCellValue(this.row.getCell(this.column), this.value);
                }
            }
        }
    }
}
