package com.chelvc.framework.base.util;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.io.Writer;
import java.util.function.Consumer;

import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;

/**
 * 文件工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
public final class FileUtils {
    /**
     * 临时目录
     */
    public static final String TEMPORARY = System.getProperty("java.io.tmpdir");

    /**
     * 文件上传目录
     */
    public static final String UPLOAD = new File(TEMPORARY, "upload").getPath();

    /**
     * 文件导出目录
     */
    public static final String EXPORT = new File(TEMPORARY, "export").getPath();

    /**
     * 文件下载目录
     */
    public static final String DOWNLOAD = new File(TEMPORARY, "download").getPath();

    static {
        // 初始化文件上传目录
        File upload = new File(UPLOAD);
        if (!upload.exists() && !upload.mkdirs()) {
            log.warn("Make upload dir failed: {}", UPLOAD);
        }

        // 初始化文件导出目录
        File export = new File(EXPORT);
        if (!export.exists() && !export.mkdirs()) {
            log.warn("Make export dir failed: {}", EXPORT);
        }

        // 初始化文件下载目录
        File download = new File(DOWNLOAD);
        if (!download.exists() && !download.mkdirs()) {
            log.warn("Make download dir failed: {}", DOWNLOAD);
        }
    }

    private FileUtils() {
    }

    /**
     * 获取文件上传目录
     *
     * @return 文件目录
     */
    public static File getUploadDirectory() {
        return new File(UPLOAD);
    }

    /**
     * 获取数据导出目录
     *
     * @return 文件目录
     */
    public static File getExportDirectory() {
        return new File(EXPORT);
    }

    /**
     * 获取文件下载目录
     *
     * @return 文件目录
     */
    public static File getDownloadDirectory() {
        return new File(DOWNLOAD);
    }

    /**
     * 获取文件后缀名
     *
     * @param file 文件对象
     * @return 后缀名
     */
    public static String getSuffix(@NonNull File file) {
        return getSuffix(file.getName());
    }

    /**
     * 获取文件后缀名
     *
     * @param filename 文件名称
     * @return 后缀名
     */
    public static String getSuffix(@NonNull String filename) {
        int separator = filename.indexOf('.');
        return separator < 0 ? null : StringUtils.ifEmpty(filename.substring(separator + 1), (String) null);
    }

    /**
     * 图片压缩
     *
     * @param image 图片对象
     * @return 图片对象
     */
    public static BufferedImage compress(@NonNull BufferedImage image) {
        int width = image.getWidth(null);
        int height = image.getHeight(null);
        BufferedImage compress = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
        compress.getGraphics().drawImage(image.getScaledInstance(width, height, Image.SCALE_SMOOTH), 0, 0, null);
        return compress;
    }

    /**
     * 读取文件内容
     *
     * @param file 文件对象
     * @return 文件内容
     */
    public static String read(@NonNull File file) {
        try {
            return read(new FileReader(file));
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 读取文件内容
     *
     * @param file     文件对象
     * @param consumer 文件内容消费者
     */
    public static void read(@NonNull File file, @NonNull Consumer<String> consumer) {
        try {
            read(new FileReader(file), consumer);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 读取文件内容
     *
     * @param file 文件路径
     * @return 文件内容
     */
    public static String read(@NonNull String file) {
        try {
            return read(new FileReader(file));
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 读取文件内容
     *
     * @param file     文件路径
     * @param consumer 文件内容消费者
     */
    public static void read(@NonNull String file, @NonNull Consumer<String> consumer) {
        try {
            read(new FileReader(file), consumer);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 读取文件内容
     *
     * @param in 文件读取器
     * @return 文件内容
     */
    public static String read(@NonNull Reader in) {
        StringBuilder buffer = new StringBuilder();
        read(in, line -> {
            if (buffer.length() > 0) {
                buffer.append(StringUtils.ENTER);
            }
            buffer.append(line);
        });
        return buffer.toString();
    }

    /**
     * 读取文件内容
     *
     * @param in       文件读取处理器
     * @param consumer 文件内容消费者
     */
    public static void read(@NonNull Reader in, @NonNull Consumer<String> consumer) {
        try (BufferedReader reader = new BufferedReader(in)) {
            String line;
            while ((line = reader.readLine()) != null) {
                consumer.accept(line);
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 写文件
     *
     * @param file    目标文件
     * @param content 文件内容
     */
    public static void write(@NonNull File file, @NonNull String content) {
        try {
            write(new FileWriter(file), content);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 写文件
     *
     * @param file    目标文件路径
     * @param content 文件内容
     */
    public static void write(@NonNull String file, @NonNull String content) {
        try {
            write(new FileWriter(file), content);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 写文件
     *
     * @param out     文件写入处理器
     * @param content 文件内容
     */
    public static void write(@NonNull Writer out, @NonNull String content) {
        try (BufferedWriter writer = new BufferedWriter(out)) {
            writer.write(content);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 将文件写入文件
     *
     * @param out 输出文件
     * @param in  输入文件
     */
    public static void write(@NonNull File out, @NonNull File in) {
        try {
            try (OutputStream output = new FileOutputStream(out)) {
                write(output, in);
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 将输入流写入文件
     *
     * @param file  文件对象
     * @param input 输入流
     */
    public static void write(@NonNull File file, @NonNull InputStream input) {
        try {
            try (OutputStream output = new FileOutputStream(file)) {
                write(output, input);
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 将文件写入输入流
     *
     * @param output 输出流
     * @param file   文件对象
     */
    public static void write(@NonNull OutputStream output, @NonNull File file) {
        try {
            try (InputStream input = new FileInputStream(file)) {
                write(output, input);
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 将输入流写入输出流
     *
     * @param output 输出流
     * @param input  输入流
     */
    public static void write(@NonNull OutputStream output, @NonNull InputStream input) {
        int len;
        byte[] buffer = new byte[1024];
        try {
            while ((len = input.read(buffer)) > -1) {
                output.write(buffer, 0, len);
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 删除文件
     *
     * @param file 文件对象
     */
    public static void delete(@NonNull File file) {
        if (!file.delete()) {
            log.warn("File delete failed: {}", file.getPath());
        }
    }

    /**
     * 获取文件数组
     *
     * @param path 文件目录
     * @return 文件数组
     */
    public static File[] files(@NonNull File path) {
        if (!path.isDirectory()) {
            return new File[0];
        }
        return path.listFiles(file -> file != null && file.isFile());
    }

    /**
     * 获取文件
     *
     * @param path     文件目录
     * @param consumer 文件消费者
     */
    public static void files(@NonNull File path, @NonNull Consumer<File> consumer) {
        if (!path.isDirectory()) {
            return;
        }
        path.listFiles(file -> {
            if (file != null && file.isFile()) {
                consumer.accept(file);
            }
            return false;
        });
    }

    /**
     * 获取文件数组
     *
     * @param path 文件目录
     * @return 文件数组
     */
    public static File[] files(@NonNull String path) {
        return files(new File(path));
    }

    /**
     * 获取文件
     *
     * @param path     文件目录
     * @param consumer 文件消费者
     */
    public static void files(@NonNull String path, @NonNull Consumer<File> consumer) {
        files(new File(path), consumer);
    }

    /**
     * 获取文件目录数组
     *
     * @param path 文件目录
     * @return 文件目录数组
     */
    public static File[] directories(@NonNull File path) {
        if (!path.isDirectory()) {
            return new File[0];
        }
        return path.listFiles(file -> file != null && file.isFile());
    }

    /**
     * 获取文件目录
     *
     * @param path     文件目录
     * @param consumer 文件目录消费者
     */
    public static void directories(@NonNull File path, @NonNull Consumer<File> consumer) {
        if (!path.isDirectory()) {
            return;
        }
        path.listFiles(file -> {
            if (file != null && file.isFile()) {
                consumer.accept(file);
            }
            return false;
        });
    }

    /**
     * 获取文件目录数组
     *
     * @param path 文件目录
     * @return 文件目录数组
     */
    public static File[] directories(@NonNull String path) {
        return directories(new File(path));
    }

    /**
     * 获取文件目录
     *
     * @param path     文件目录
     * @param consumer 文件目录消费者
     */
    public static void directories(@NonNull String path, @NonNull Consumer<File> consumer) {
        directories(new File(path), consumer);
    }
}
