package com.chelvc.framework.base.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.chelvc.framework.base.context.JacksonContextHolder;
import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.base.model.Platform;
import com.chelvc.framework.base.model.Result;
import com.chelvc.framework.base.model.Session;
import com.chelvc.framework.base.model.Terminal;
import com.fasterxml.jackson.core.JsonEncoding;
import lombok.NonNull;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;

/**
 * Http工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
public final class HttpUtils {
    /**
     * 请求结果统一封装响应头
     */
    public static final String UNIFIED_RESPONSE_HEADER = "unified-response-body";

    /**
     * 资源分隔符匹配模式
     */
    private static final Pattern URI_SEPARATOR_PATTERN = Pattern.compile("/+");

    /**
     * 客户端主机地址请求头数组
     */
    private static final String[] REQUEST_ADDRESS_HEADER_ARRAY = new String[]{
            "X-Real-IP", "X-Forwarded-For", "Proxy-Client-IP", "WL-Proxy-Client-IP"
    };

    private HttpUtils() {
    }

    /**
     * 获取Http状态
     *
     * @param code Http状态码
     * @return Http状态枚举
     */
    public static HttpStatus getStatus(Integer code) {
        if (code == null) {
            return HttpStatus.INTERNAL_SERVER_ERROR;
        }
        try {
            return HttpStatus.valueOf(code);
        } catch (Exception e) {
            return HttpStatus.INTERNAL_SERVER_ERROR;
        }
    }

    /**
     * 获取请求资源
     *
     * @param request Http请求对象
     * @return 资源地址
     */
    public static String getRequestURI(@NonNull HttpServletRequest request) {
        String uri = request.getRequestURI();
        if (StringUtils.isEmpty(uri)) {
            return StringUtils.URI_SEPARATOR;
        } else if (uri.charAt(0) != StringUtils.URI_SEPARATOR.charAt(0)) {
            return StringUtils.URI_SEPARATOR.concat(uri);
        }
        return uri;
    }

    /**
     * 获取请求域名
     *
     * @param request Http请求对象
     * @return 域名
     */
    public static String getRequestDomain(@NonNull HttpServletRequest request) {
        return request.getScheme() + "://" + request.getServerName() +
                ":" + request.getServerPort() + request.getContextPath();
    }

    /**
     * 获取请求客户端主机地址
     *
     * @param request Http请求对象
     * @return 请求客户端主机地址
     */
    public static String getRequestAddress(@NonNull HttpServletRequest request) {
        // 获取请求主机地址
        String ip = null;
        for (String header : REQUEST_ADDRESS_HEADER_ARRAY) {
            if (!StringUtils.isEmpty(ip = request.getHeader(header)) && !HostUtils.UNKNOWN_ADDRESS.equals(ip)) {
                break;
            }
        }
        if (StringUtils.isEmpty(ip)) {
            ip = request.getRemoteAddr();
        }

        // 如果地址存在多个则获取第一个
        int separatorIndex = ip.indexOf(',');
        if (separatorIndex > -1) {
            ip = ip.substring(0, separatorIndex);
        }

        return HostUtils.DEFAULT_LOCAL_ADDRESS_IPV6.equals(ip) ? HostUtils.LOCAL_ADDRESS : ip;
    }

    /**
     * 初始化请求结果统一封装响应头
     *
     * @param response Http响应对象
     */
    public static void initializeUnifiedResponseHeader(@NonNull HttpServletResponse response) {
        response.setHeader(UNIFIED_RESPONSE_HEADER, String.valueOf(true));
    }

    /**
     * 响应文件流
     *
     * @param response Http响应对象
     * @param file     文件对象
     * @throws IOException I/O异常
     */
    public static void writeFileResponse(@NonNull HttpServletResponse response, @NonNull File file) throws IOException {
        writeFileResponse(response, file, file.getName());
    }

    /**
     * 响应文件流
     *
     * @param response Http响应对象
     * @param file     文件对象
     * @param filename 文件名称
     * @throws IOException I/O异常
     */
    public static void writeFileResponse(@NonNull HttpServletResponse response, @NonNull File file,
                                         @NonNull String filename) throws IOException {
        // 设置响应头
        response.reset();
        response.addHeader("Content-Length", String.valueOf(file.length()));
        response.setHeader("Content-Disposition", "attachment;filename=" + URLEncoder.encode(filename, "utf-8"));
        response.setHeader("Access-Control-Expose-Headers", "Content-Disposition");
        response.addHeader("Access-Control-Allow-Origin", "*");
        response.setContentType(MediaType.APPLICATION_OCTET_STREAM_VALUE);

        // 将文件写入响应流
        int length;
        byte[] buffer = new byte[2048];
        try (InputStream input = new FileInputStream(file); OutputStream output = response.getOutputStream()) {
            while ((length = input.read(buffer)) > 0) {
                output.write(buffer, 0, length);
            }
            output.flush();
        }
    }

    /**
     * 返回统一封装请求结果
     *
     * @param response Http响应对象
     * @param result   请求结果
     * @throws IOException I/O异常
     */
    public static void writeUnifiedResponse(@NonNull HttpServletResponse response, @NonNull Result<?> result)
            throws IOException {
        response.setStatus(result.getCode());
        response.setContentType(MediaType.APPLICATION_JSON_VALUE);
        response.setCharacterEncoding(JsonEncoding.UTF8.getJavaName());
        initializeUnifiedResponseHeader(response);
        OutputStream outputStream = response.getOutputStream();
        JacksonContextHolder.getGlobalSerializer().writeValue(outputStream, result);
        outputStream.flush();
    }

    /**
     * 对象URL编码
     *
     * @param object 编码对象
     * @return 对象编码结果
     */
    public static String encode(Object object) {
        String serialize = JacksonContextHolder.serialize(object);
        if (StringUtils.isEmpty(serialize)) {
            return StringUtils.EMPTY;
        }
        try {
            return URLEncoder.encode(serialize, "UTF-8");
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 对象URL解码
     *
     * @param encode 编码信息
     * @return 对象编码结果
     */
    public static String decode(String encode) {
        if (StringUtils.isEmpty(encode)) {
            return encode;
        }
        try {
            return URLDecoder.decode(encode, "UTF-8");
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * URL地址拼接
     *
     * @param domain 域名
     * @param uris   资源地址数组
     * @return URL地址
     */
    public static String url(@NonNull String domain, @NonNull String... uris) {
        if (uris == null || uris.length == 0) {
            return domain;
        }
        Matcher matcher = URI_SEPARATOR_PATTERN.matcher(StringUtils.join(uris, StringUtils.URI_SEPARATOR));
        String uri = matcher.replaceAll(StringUtils.URI_SEPARATOR);
        if (domain.charAt(domain.length() - 1) == StringUtils.URI_SEPARATOR.charAt(0)) {
            if (uri.charAt(0) == StringUtils.URI_SEPARATOR.charAt(0)) {
                return domain.concat(uri.substring(1));
            }
            return domain.concat(uri);
        } else if (uri.charAt(0) == StringUtils.URI_SEPARATOR.charAt(0)) {
            return domain.concat(uri);
        }
        return domain.concat(StringUtils.URI_SEPARATOR).concat(uri);
    }

    /**
     * 构建Http请求消息
     *
     * @param request Http请求对象
     * @param type    消息类型
     * @param message 消息内容
     * @return 消息
     */
    public static String message(HttpServletRequest request, Object type, String message) {
        Session session = SessionContextHolder.getSession(false);
        Long id = ObjectUtils.ifNull(session, Session::getId);
        String host = ObjectUtils.ifNull(session, Session::getHost);
        Platform platform = ObjectUtils.ifNull(session, Session::getPlatform);
        Terminal terminal = ObjectUtils.ifNull(session, Session::getTerminal);
        String version = ObjectUtils.ifNull(session, Session::getVersion);
        String uri = ObjectUtils.ifNull(request, HttpServletRequest::getRequestURI);
        String method = ObjectUtils.ifNull(request, HttpServletRequest::getMethod);
        return String.format("[%s] [%s] [%s] [%s] [%d] [%s %s] [%s]: %s",
                host, platform, terminal, version, id, method, uri, type, message
        );
    }

    /**
     * 获取异常消息
     *
     * @param request   Http请求对象
     * @param status    Http状态码
     * @param exception 异常信息
     * @return 异常消息
     */
    public static String getExceptionMessage(@NonNull HttpServletRequest request, @NonNull HttpStatus status,
                                             @NonNull Exception exception) {
        return getExceptionMessage(request, status, exception.getMessage());
    }

    /**
     * 获取异常消息
     *
     * @param request   Http请求对象
     * @param status    Http状态码
     * @param exception 异常信息
     * @return 异常消息
     */
    public static String getExceptionMessage(@NonNull HttpServletRequest request, @NonNull HttpStatus status,
                                             String exception) {
        return message(request, status.value(), exception);
    }
}
