package com.chelvc.framework.base.util;

import java.util.Random;
import java.util.concurrent.ThreadLocalRandom;

import lombok.extern.slf4j.Slf4j;

/**
 * 唯一ID工具类（基于雪花算法）
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
public final class IdentityUtils {
    /**
     * 时间起始标记点，作为基准，一般取系统的最近时间（一旦确定不能变动）
     */
    private static final long BENCHMARK = 1288834974657L;

    /**
     * 机器标识位数
     */
    private static final long WORKER_ID_BITS = 5L;
    private static final long DATACENTER_ID_BITS = 5L;

    /**
     * 毫秒内自增位
     */
    private static final long SEQUENCE_BITS = 12L;
    private static final long WORKER_ID_SHIFT = SEQUENCE_BITS;
    private static final long DATACENTER_ID_SHIFT = SEQUENCE_BITS + WORKER_ID_BITS;

    /**
     * 时间戳左移动位
     */
    private static final long TIMESTAMP_LEFT_SHIFT = SEQUENCE_BITS + WORKER_ID_BITS + DATACENTER_ID_BITS;
    private static final long SEQUENCE_MASK = ~(-1L << SEQUENCE_BITS);

    /**
     * 时间回拨偏移量（毫秒）
     */
    private static final long CLOCK_MOVED_BACKWARD_OFFSET = 5;

    /**
     * 机器标识
     */
    private static final long WORKER_ID = new Random().nextInt(32);

    /**
     * 数据标识 ID 部分
     */
    private static final long DATACENTER_ID = new Random().nextInt(32);

    /**
     * 并发控制
     */
    private long sequence = 0L;

    /**
     * 上次生产 ID 时间戳
     */
    private long lastTimestamp = -1L;

    private IdentityUtils() {
    }

    /**
     * 获取当前时间戳
     *
     * @return 时间戳
     */
    private static long now() {
        return System.currentTimeMillis();
    }

    /**
     * 直到下一毫秒
     *
     * @param timestamp 下一目标时间戳
     * @return 当前时间戳
     */
    protected static long untilNextMillis(long timestamp) {
        long now = now();
        while (now <= timestamp) {
            now = now();
        }
        return now;
    }

    /**
     * 生成唯一ID
     *
     * @return 唯一ID
     */
    public static long generate() {
        return Builder.INSTANCE.next();
    }

    /**
     * 获取下一个 ID
     *
     * @return 下一个 ID
     */
    private synchronized long next() {
        long timestamp = now();
        //闰秒
        if (timestamp < this.lastTimestamp) {
            // 校验时间偏移回拨量
            long offset = this.lastTimestamp - timestamp;
            if (offset > CLOCK_MOVED_BACKWARD_OFFSET) {
                throw new RuntimeException("Clock moved backwards, refusing to generate id for [" + offset + "ms]");
            }

            // 时间回退timeOffset毫秒内，则允许等待2倍的偏移量后重新获取，解决小范围的时间回拨问题
            try {
                wait(offset << 1);
            } catch (InterruptedException e) {
                throw new RuntimeException(e);
            }

            // 再次获取并校验
            if ((timestamp = now()) < this.lastTimestamp) {
                throw new RuntimeException("Clock moved backwards, refusing to generate id for [" + offset + "ms]");
            }
        }

        if (this.lastTimestamp == timestamp) {
            // 相同毫秒内，序列号自增
            this.sequence = (this.sequence + 1) & SEQUENCE_MASK;
            if (this.sequence == 0) {
                // 同一毫秒的序列数已经达到最大
                timestamp = untilNextMillis(this.lastTimestamp);
            }
        } else {
            // 不同毫秒内，序列号置为 1 - 3 随机数
            this.sequence = ThreadLocalRandom.current().nextLong(1, 3);
        }

        this.lastTimestamp = timestamp;

        // 时间戳部分 | 数据中心部分 | 机器标识部分 | 序列号部分
        return ((timestamp - BENCHMARK) << TIMESTAMP_LEFT_SHIFT)
                | (DATACENTER_ID << DATACENTER_ID_SHIFT)
                | (WORKER_ID << WORKER_ID_SHIFT)
                | this.sequence;
    }

    /**
     * 唯一ID工具实例构造类
     */
    private static final class Builder {
        /**
         * 工具实例
         */
        private static final IdentityUtils INSTANCE = new IdentityUtils();
    }
}
