package com.chelvc.framework.base.util;

import java.util.Locale;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.google.i18n.phonenumbers.NumberParseException;
import com.google.i18n.phonenumbers.PhoneNumberUtil;
import com.google.i18n.phonenumbers.Phonenumber;
import com.google.i18n.phonenumbers.geocoding.PhoneNumberOfflineGeocoder;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;

/**
 * 手机号工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
public final class MobileUtils {
    /**
     * 手机号处理工具实例
     */
    private static final PhoneNumberUtil UTIL = PhoneNumberUtil.getInstance();

    /**
     * 地理编码器实例
     */
    private static final PhoneNumberOfflineGeocoder GEOCODER = PhoneNumberOfflineGeocoder.getInstance();

    /**
     * 手机号脱敏匹配模式
     */
    private static final Pattern MOBILE_DESENSITIZE_PATTERN
            = Pattern.compile("1[\\.\\-—~～_ ]*[3-9]([\\.\\-—~～_ ]*[0-9]){9}");

    private MobileUtils() {
    }

    /**
     * 解析手机号
     *
     * @param mobile 手机号文本
     * @return 手机号对象实例
     */
    public static Phonenumber.PhoneNumber parse(String mobile) {
        return parse(mobile, Locale.CHINA);
    }

    /**
     * 解析手机号
     *
     * @param mobile 手机号文本
     * @param locale 地区信息
     * @return 手机号对象实例
     */
    public static Phonenumber.PhoneNumber parse(String mobile, @NonNull Locale locale) {
        try {
            return StringUtils.isEmpty(mobile) ? null : UTIL.parse(mobile, locale.getCountry());
        } catch (NumberParseException e) {
            log.warn("Invalid mobile[{}]: {}", locale.getCountry(), mobile);
        }
        return null;
    }

    /**
     * 判断是否是有效的手机号
     *
     * @param number 手机号对象
     * @return true/false
     */
    public static boolean isMobile(Phonenumber.PhoneNumber number) {
        return number != null && UTIL.isValidNumber(number);
    }

    /**
     * 判断是否是有效的手机号
     *
     * @param mobile 手机号文本
     * @return true/false
     */
    public static boolean isMobile(String mobile) {
        return isMobile(mobile, Locale.CHINA);
    }

    /**
     * 判断是否是有效的手机号
     *
     * @param mobile 手机号对象
     * @param locale 地区信息
     * @return true/false
     */
    public static boolean isMobile(String mobile, @NonNull Locale locale) {
        return isMobile(parse(mobile, locale));
    }

    /**
     * 获取手机号地址信息
     *
     * @param number 手机号对象
     * @return 地址信息
     */
    public static String mobile2address(Phonenumber.PhoneNumber number) {
        return mobile2address(number, Locale.CHINA);
    }

    /**
     * 获取手机号地址信息
     *
     * @param number 手机号对象
     * @param locale 地区信息
     * @return 地址信息
     */
    public static String mobile2address(Phonenumber.PhoneNumber number, @NonNull Locale locale) {
        return number == null ? null : GEOCODER.getDescriptionForNumber(number, locale);
    }

    /**
     * 获取手机号地址信息
     *
     * @param mobile 手机号文本
     * @return 地址信息
     */
    public static String mobile2address(String mobile) {
        return mobile2address(mobile, Locale.CHINA);
    }

    /**
     * 获取手机号地址信息
     *
     * @param mobile 手机号文本
     * @param locale 地区信息
     * @return 地址信息
     */
    public static String mobile2address(String mobile, @NonNull Locale locale) {
        return mobile2address(parse(mobile, locale), locale);
    }

    /**
     * 脱敏手机号
     *
     * @param text 文本内容
     * @return 文本内容
     */
    public static String desensitize(String text) {
        if (StringUtils.isEmpty(text)) {
            return text;
        }

        // 脱敏手机号
        Matcher matcher = MOBILE_DESENSITIZE_PATTERN.matcher(text);
        while (matcher.find()) {
            String group = matcher.group();
            String clean = StringUtils.clear(group);
            text = text.replace(group, clean.substring(0, 3) + "****" + clean.substring(7, 11));
        }
        return text;
    }
}
