package com.chelvc.framework.base.util;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.io.IOException;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.BiConsumer;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import com.chelvc.framework.base.model.Modification;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import io.protostuff.LinkedBuffer;
import io.protostuff.ProtobufIOUtil;
import io.protostuff.Schema;
import io.protostuff.runtime.RuntimeSchema;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import ma.glasnost.orika.MapperFactory;
import ma.glasnost.orika.MappingContext;
import ma.glasnost.orika.converter.BidirectionalConverter;
import ma.glasnost.orika.impl.DefaultMapperFactory;
import ma.glasnost.orika.metadata.Type;
import org.springframework.beans.BeanUtils;
import org.springframework.core.io.Resource;
import org.springframework.core.io.support.PathMatchingResourcePatternResolver;
import org.springframework.core.io.support.ResourcePatternResolver;
import org.springframework.core.type.classreading.CachingMetadataReaderFactory;
import org.springframework.core.type.classreading.MetadataReader;
import org.springframework.core.type.classreading.MetadataReaderFactory;
import org.springframework.util.ClassUtils;
import org.springframework.util.CollectionUtils;

/**
 * 对象工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
public final class ObjectUtils {
    /**
     * 空字节数组
     */
    public static final byte[] EMPTY_BYTES = new byte[0];

    /**
     * CPU数量
     */
    public static final int CPU_COUNT = Runtime.getRuntime().availableProcessors();

    /**
     * 对象转换映射工厂
     */
    private static final MapperFactory MAPPER_FACTORY = new DefaultMapperFactory.Builder().build();

    /**
     * 对象字段映射表
     */
    private static final Map<Class<?>, Field[]> CLASS_FIELD_MAPPING = new ConcurrentHashMap<>();

    /**
     * Protostuff对象类型/Schema映射表
     */
    private static final Map<Class<?>, Schema<?>> PROTOSTUFF_CLASS_SCHEMA_MAPPING = new ConcurrentHashMap<>();

    /**
     * 类对象/属性描述映射表
     */
    private static final Map<Class<?>, List<PropertyDescriptor>> CLASS_PROPERTY_DESCRIPTOR_MAPPING =
            new ConcurrentHashMap<>();

    static {
        // 注册简单的对象拷贝转换器
        MAPPER_FACTORY.getConverterFactory().registerConverter("simple",
                new BidirectionalConverter<Object, Object>() {
                    @Override
                    public Object convertTo(Object object, Type type, MappingContext mappingContext) {
                        return object;
                    }

                    @Override
                    public Object convertFrom(Object object, Type type, MappingContext mappingContext) {
                        return object;
                    }
                });

        // 注册Modification属性转换器
        MAPPER_FACTORY.classMap(Modification.class, Modification.class)
                .fieldMap("before", "before").converter("simple").add()
                .fieldMap("after", "after").converter("simple").add().byDefault().register();
    }

    private ObjectUtils() {
    }

    /**
     * 返回不为空的对象（如果第一个对象为空，则返回第二个对象）
     *
     * @param object 目标对象
     * @param value  替换对象
     * @param <T>    对象类型泛型
     * @return 返回对象
     */
    public static <T> T ifNull(T object, T value) {
        return object == null ? value : object;
    }

    /**
     * 返回不为空的对象（如果第一个对象为空，则返回第二个对象）
     *
     * @param object   目标对象
     * @param supplier 替换对象提供方法
     * @param <T>      对象类型泛型
     * @return 返回对象
     */
    public static <T> T ifNull(T object, @NonNull Supplier<T> supplier) {
        return object == null ? supplier.get() : object;
    }

    /**
     * 返回不为空的对象（如果第一个对象为空，则返回第二个对象）
     *
     * @param object   目标对象
     * @param function 目标对象方法
     * @param <T>      目标对象类型泛型
     * @param <R>      返回对象类型泛型
     * @return 返回对象
     */
    public static <T, R> R ifNull(T object, @NonNull Function<T, R> function) {
        return object == null ? null : function.apply(object);
    }

    /**
     * 返回不为空的对象（如果第一个对象为空，则返回第二个对象）
     *
     * @param object   目标对象
     * @param function 目标对象方法
     * @param supplier 替换对象提供方法
     * @param <T>      目标对象类型泛型
     * @param <R>      返回对象类型泛型
     * @return 返回对象
     */
    public static <T, R> R ifNull(T object, @NonNull Function<T, R> function, @NonNull Supplier<R> supplier) {
        return object == null ? supplier.get() : ifNull(function.apply(object), supplier);
    }

    /**
     * 获取对象字段值
     *
     * @param object 对象实例
     * @param field  字段对象
     * @return 值对象
     */
    public static Object getValue(Object object, @NonNull Field field) {
        if (object == null) {
            return null;
        } else if (!field.isAccessible()) {
            field.setAccessible(true);
        }
        try {
            return field.get(object);
        } catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 设置对象字段值
     *
     * @param object 对象实例
     * @param field  属性字段对象
     * @param value  属性值
     */
    public static void setValue(Object object, @NonNull Field field, Object value) {
        if (object == null) {
            return;
        } else if (!field.isAccessible()) {
            field.setAccessible(true);
        }
        try {
            field.set(object, value);
        } catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 类及父类实例字段访问
     *
     * @param clazz    对象类型
     * @param consumer 字段接收器
     */
    public static void iterateFields(@NonNull Class<?> clazz, @NonNull BiConsumer<Integer, Field> consumer) {
        iterateFields(clazz, (index, field) -> {
            consumer.accept(index, field);
            return true;
        });
    }

    /**
     * 类及父类实例字段访问
     *
     * @param clazz    对象类型
     * @param function 字段接收器（返回false则中断循环）
     */
    public static void iterateFields(@NonNull Class<?> clazz, @NonNull BiFunction<Integer, Field, Boolean> function) {
        int index = 0;
        boolean continuing = true;
        do {
            for (Field field : CLASS_FIELD_MAPPING.computeIfAbsent(clazz, Class::getDeclaredFields)) {
                if (!Modifier.isStatic(field.getModifiers()) && !field.getName().startsWith("this$")) {
                    if (!(continuing = Boolean.TRUE.equals(function.apply(index++, field)))) {
                        break;
                    }
                }
            }
        } while (continuing && (clazz = clazz.getSuperclass()) != Object.class);
    }

    /**
     * 查找工程内部的对象类型
     *
     * @param pack 包名
     * @return 类对象列表
     */
    public static List<Class<?>> lookupClasses(@NonNull String pack) {
        return lookupClasses(pack, clazz -> true);
    }

    /**
     * 查找工程内部的对象类型
     *
     * @param pack      包名
     * @param condition 类对象条件判断
     * @return 类对象列表
     */
    public static List<Class<?>> lookupClasses(@NonNull String pack, @NonNull Predicate<Class<?>> condition) {
        ResourcePatternResolver resourcePatternResolver = new PathMatchingResourcePatternResolver();
        String pattern = (pack = pack.trim()).isEmpty() ? "**/*.class" : (
                ResourcePatternResolver.CLASSPATH_ALL_URL_PREFIX
                        + ClassUtils.convertClassNameToResourcePath(pack)
                        + "/**/*.class"
        );

        // 获取对象资源文件
        Resource[] resources;
        try {
            resources = resourcePatternResolver.getResources(pattern);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }

        // 将资源文件转换为类对象
        Stream.Builder<Class<?>> classes = Stream.builder();
        MetadataReaderFactory readerFactory = new CachingMetadataReaderFactory(resourcePatternResolver);
        ClassLoader classLoader = Objects.requireNonNull(ClassUtils.getDefaultClassLoader(), "Class loader not found");
        for (Resource resource : resources) {
            try {
                // 加载类对象
                MetadataReader reader = readerFactory.getMetadataReader(resource);
                Class<?> clazz = classLoader.loadClass(reader.getClassMetadata().getClassName());
                if (clazz != null && condition.test(clazz)) {
                    classes.add(clazz);
                }
            } catch (IOException | ClassNotFoundException e) {
                throw new RuntimeException();
            }
        }
        return classes.build().distinct().collect(Collectors.toList());
    }

    /**
     * 获取对象属性复制处理器工厂
     *
     * @return 对象属性复制处理器工厂
     */
    public static MapperFactory getCopierBuilder() {
        return MAPPER_FACTORY;
    }

    /**
     * 对象属性复制
     *
     * @param source 源对象实例
     * @param target 目标对象实例
     * @param <T>    源对象类型
     * @param <R>    目标对象类型
     */
    public static <T, R> void copying(T source, R target) {
        getCopierBuilder().getMapperFacade().map(source, target);
    }

    /**
     * 对象属性复制
     *
     * @param source 源对象实例
     * @param target 目标对象
     * @param <T>    源对象类型
     * @param <R>    目标对象类型
     * @return 目标对象实例
     */
    public static <T, R> R copying(T source, Class<R> target) {
        return getCopierBuilder().getMapperFacade().map(source, target);
    }

    /**
     * 对象属性复制
     *
     * @param sources 源对象实例数组
     * @param target  目标对象
     * @param <T>     源对象类型
     * @param <R>     目标对象类型
     * @return 目标对象实例列表
     */
    public static <T, R> List<R> copying(T[] sources, @NonNull Class<R> target) {
        return sources == null || sources.length == 0 ? Collections.emptyList() :
                getCopierBuilder().getMapperFacade().mapAsList(sources, target);
    }

    /**
     * 对象属性复制
     *
     * @param sources 源对象实例列表
     * @param target  目标对象
     * @param <T>     源对象类型
     * @param <R>     目标对象类型
     * @return 目标对象实例列表
     */
    public static <T, R> List<R> copying(Collection<T> sources, @NonNull Class<R> target) {
        return CollectionUtils.isEmpty(sources) ? Collections.emptyList() :
                getCopierBuilder().getMapperFacade().mapAsList(sources, target);
    }

    /**
     * 判断集合是否为空，如果集合为空则调用适配函数并返回
     *
     * @param collection 对象集合
     * @param supplier   对象集合适配函数
     * @param <T>        对象类型
     * @param <C>        集合类型
     * @return 对象集合
     */
    public static <T, C extends Collection<T>> C ifEmpty(C collection, @NonNull Supplier<C> supplier) {
        return CollectionUtils.isEmpty(collection) ? supplier.get() : collection;
    }

    /**
     * 判断集合是否为空，如果集合非空则使用适配函数
     *
     * @param collection 对象集合
     * @param adapter    对象集合适配函数
     * @param <T>        对象类型
     * @param <R>        目标类型
     * @param <C>        集合类型
     * @return 对象集合
     */
    public static <T, R, C extends Collection<T>> List<R> ifEmpty(C collection, @NonNull Function<C, List<R>> adapter) {
        return CollectionUtils.isEmpty(collection) ? Collections.emptyList() : adapter.apply(collection);
    }

    /**
     * 合并集合
     *
     * @param collections 集合数组
     * @param <T>         对象类型
     * @return 对象列表
     */
    @SafeVarargs
    public static <T> List<T> merge(Collection<T>... collections) {
        if (collections == null || collections.length == 0) {
            return Collections.emptyList();
        }
        return Stream.of(collections).filter(Objects::nonNull).flatMap(Collection::stream).filter(Objects::nonNull)
                .distinct().collect(Collectors.toList());
    }

    /**
     * 获取Protostuff schema
     *
     * @param clazz 目标对象
     * @param <T>   目标对象类型
     * @return Schema对象实例
     */
    @SuppressWarnings("unchecked")
    private static <T> Schema<T> getProtostuffSchema(Class<T> clazz) {
        return (Schema<T>) PROTOSTUFF_CLASS_SCHEMA_MAPPING.computeIfAbsent(clazz, k -> RuntimeSchema.getSchema(clazz));
    }

    /**
     * 序列化方法，把指定对象序列化成字节数组
     *
     * @param object 被序列化对象实例
     * @return 字节数组
     */
    @SuppressWarnings("unchecked")
    public static <T> byte[] serialize(T object) {
        if (object == null) {
            return null;
        }
        Schema<T> schema = getProtostuffSchema((Class<T>) object.getClass());
        LinkedBuffer buffer = LinkedBuffer.allocate(LinkedBuffer.DEFAULT_BUFFER_SIZE);
        try {
            return ProtobufIOUtil.toByteArray(object, schema, buffer);
        } finally {
            buffer.clear();
        }
    }

    /**
     * 反序列化方法，将字节数组反序列化成指定Class类型
     *
     * @param bytes 字节数组
     * @param clazz 目标对象类型
     * @return 目标对象实例
     */
    public static <T> T deserialize(byte[] bytes, @NonNull Class<T> clazz) {
        if (bytes == null) {
            return null;
        }
        Schema<T> schema = getProtostuffSchema(clazz);
        T object = schema.newMessage();
        ProtobufIOUtil.mergeFrom(bytes, object, schema);
        return object;
    }

    /**
     * 将对象转换成键/值对
     *
     * @param object   对象实例
     * @param excludes 需要排除的属性数组
     * @return 键/值对
     */
    public static Map<String, Object> mapping(Object object, String... excludes) {
        if (object == null) {
            return Collections.emptyMap();
        }
        PropertyDescriptor[] properties = BeanUtils.getPropertyDescriptors(object.getClass());
        if (properties.length == 0) {
            return Collections.emptyMap();
        }
        Set<String> excludeProperties = excludes.length == 0 ? Collections.emptySet() : Sets.newHashSet(excludes);
        Map<String, Object> propertyValues = Maps.newHashMapWithExpectedSize(properties.length - excludes.length);
        iterateFields(object.getClass(), (index, field) -> {
            if (!excludeProperties.contains(field.getName())) {
                field.setAccessible(true);
                try {
                    propertyValues.put(field.getName(), field.get(object));
                } catch (ReflectiveOperationException e) {
                    throw new RuntimeException(e);
                }
            }
        });
        return propertyValues;
    }

    /**
     * 获取对象属性描述列表
     *
     * @param clazz 目标对象
     * @return 属性描述列表
     */
    public static List<PropertyDescriptor> getPropertyDescriptors(@NonNull Class<?> clazz) {
        return CLASS_PROPERTY_DESCRIPTOR_MAPPING.computeIfAbsent(clazz, c -> {
            BeanInfo bean;
            try {
                bean = Introspector.getBeanInfo(c, Object.class);
            } catch (IntrospectionException e) {
                throw new RuntimeException(e);
            }
            return Collections.unmodifiableList(Arrays.asList(bean.getPropertyDescriptors()));
        });
    }

    /**
     * 查找属性对应的Getter方法
     *
     * @param clazz    目标对象
     * @param property 属性名称
     * @return 方法对象
     */
    public static Method lookupPropertyGetter(@NonNull Class<?> clazz, @NonNull String property) {
        return getPropertyDescriptors(clazz).stream().filter(descriptor -> descriptor.getName().equals(property))
                .findAny().map(PropertyDescriptor::getReadMethod).orElse(null);
    }

    /**
     * 将对象转换成长整形数字
     *
     * @param object 目标对象
     * @return 长整形数字
     */
    public static Long object2long(Object object) {
        if (object == null || StringUtils.isEmpty(object)) {
            return null;
        } else if (object instanceof Byte || object instanceof Short || object instanceof Integer
                || object instanceof Long) {
            return ((Number) object).longValue();
        }
        return Long.parseLong(object.toString());
    }

    /**
     * 判断对象是否相同
     *
     * @param object 对象实例
     * @param other  对象实例
     * @return true/false
     */
    public static boolean equals(Object object, Object other) {
        if (Objects.equals(object, other)
                || (object instanceof Map && ((Map<?, ?>) object).isEmpty() && Objects.isNull(other))
                || (other instanceof Map && ((Map<?, ?>) other).isEmpty() && Objects.isNull(object))
                || (object instanceof Collection && ((Collection<?>) object).isEmpty() && Objects.isNull(other))
                || (other instanceof Collection && ((Collection<?>) other).isEmpty() && Objects.isNull(object))) {
            return true;
        } else if (object instanceof Map && other instanceof Map) {
            for (Map.Entry<?, ?> entry : ((Map<?, ?>) object).entrySet()) {
                if (!equals(entry.getValue(), ((Map<?, ?>) other).get(entry.getKey()))) {
                    return false;
                }
            }
            for (Map.Entry<?, ?> entry : ((Map<?, ?>) other).entrySet()) {
                if (!equals(entry.getValue(), ((Map<?, ?>) object).get(entry.getKey()))) {
                    return false;
                }
            }
            return true;
        } else if (object instanceof Collection && other instanceof Collection) {
            for (Object value : (Collection<?>) object) {
                if (((Collection<?>) other).stream().noneMatch(o -> equals(o, value))) {
                    return false;
                }
            }
            for (Object value : (Collection<?>) other) {
                if (((Collection<?>) object).stream().noneMatch(o -> equals(o, value))) {
                    return false;
                }
            }
            return true;
        }
        return false;
    }

    /**
     * 获取两个对象实例不同属性值
     *
     * @param <T>      数据类型
     * @param object   对象实例
     * @param other    对象实例
     * @param excludes 需要排除的属性名称集合
     * @return 不同属性值
     */
    public static <T> List<Modification> compare(T object, T other, @NonNull String... excludes) {
        if (object == null || other == null) {
            return Collections.emptyList();
        }
        Set<String> excludeFields = excludes.length == 0 ? Collections.emptySet() : Sets.newHashSet(excludes);
        List<Modification> modifications = Lists.newLinkedList();
        iterateFields(object.getClass(), (index, field) -> {
            if (excludeFields.isEmpty() || !excludeFields.contains(field.getName())) {
                Object before = getValue(object, field);
                Object after = getValue(other, field);
                if (!equals(before, after)) {
                    modifications.add(Modification.builder().name(field.getName()).before(before).after(after).build());
                }
            }
        });
        return modifications;
    }

    /**
     * 循环获取对象
     *
     * @param supplier 对象提供函数
     * @param function 对象处理函数（返回false则中断循环）
     * @param <T>      对象类型
     * @return 对象实例
     */
    public static <T> T loop(@NonNull Supplier<T> supplier, @NonNull BiFunction<Integer, T, Boolean> function) {
        T t;
        int count = 1;
        do {
            t = supplier.get();
        } while (Boolean.TRUE.equals(function.apply(count++, t)));
        return t;
    }

    /**
     * 数字分组
     *
     * @param total 总数
     * @param size  分组大小
     * @return 分组数量
     */
    public static int group(int total, int size) {
        if (size < 1 || size >= total) {
            return 1;
        }
        return total % size == 0 ? total / size : total / size + 1;
    }

    /**
     * 对象列表拆分
     *
     * @param objects 对象列表
     * @param size    分组大小
     * @param <T>     对象类型
     * @return 对象分组列表
     */
    public static <T> List<List<T>> split(List<T> objects, int size) {
        return CollectionUtils.isEmpty(objects) ? Collections.emptyList() : group(objects, group(objects.size(), size));
    }

    /**
     * 对象列表拆分
     *
     * @param objects  对象列表
     * @param size     分组大小
     * @param consumer 对象分组处理器
     * @param <T>      对象类型
     */
    public static <T> void split(List<T> objects, int size, @NonNull BiConsumer<Integer, List<T>> consumer) {
        if (!CollectionUtils.isEmpty(objects)) {
            group(objects, group(objects.size(), size), consumer);
        }
    }

    /**
     * 对象列表分组
     *
     * @param objects 对象列表
     * @param group   分组数量
     * @param <T>     对象类型
     * @return 对象分组列表
     */
    public static <T> List<List<T>> group(List<T> objects, int group) {
        if (CollectionUtils.isEmpty(objects)) {
            return Collections.emptyList();
        } else if (group < 2) {
            return Collections.singletonList(objects);
        }
        List<List<T>> groups = Lists.newArrayListWithCapacity(group);
        group(objects, group, (index, list) -> groups.add(list));
        return groups;
    }

    /**
     * 对象列表分组
     *
     * @param objects  对象列表
     * @param group    分组数量
     * @param consumer 对象分组处理器
     * @param <T>      对象类型
     */
    public static <T> void group(List<T> objects, int group, @NonNull BiConsumer<Integer, List<T>> consumer) {
        if (CollectionUtils.isEmpty(objects)) {
            return;
        }
        if (group < 2) {
            consumer.accept(0, objects);
        } else {
            int total = objects.size(), size = group(total, group);
            for (int i = 0; i < group; i++) {
                consumer.accept(i, objects.subList(i * size, Math.min((i + 1) * size, total)));
            }
        }
    }
}