package com.chelvc.framework.database.context;

import java.io.Serializable;
import java.lang.reflect.Field;
import java.sql.Connection;
import java.sql.Statement;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;
import java.util.function.Supplier;
import javax.crypto.Cipher;

import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.metadata.TableInfo;
import com.baomidou.mybatisplus.core.metadata.TableInfoHelper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.service.IService;
import com.chelvc.framework.base.context.ApplicationContextHolder;
import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.base.function.Executor;
import com.chelvc.framework.base.function.Handler;
import com.chelvc.framework.base.function.Provider;
import com.chelvc.framework.base.util.AESUtils;
import com.chelvc.framework.base.util.ErrorUtils;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.chelvc.framework.database.annotation.Confidential;
import com.chelvc.framework.database.config.DatabaseSecurityProperties;
import com.chelvc.framework.database.entity.Creatable;
import com.chelvc.framework.database.entity.Deletable;
import com.chelvc.framework.database.entity.Entity;
import com.chelvc.framework.database.entity.Isolatable;
import com.chelvc.framework.database.entity.TreeEntity;
import com.chelvc.framework.database.entity.Updatable;
import com.chelvc.framework.database.util.EntityUtils;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Base64;
import org.apache.ibatis.mapping.SqlCommandType;
import org.mybatis.spring.SqlSessionTemplate;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.annotation.Order;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.CollectionUtils;

/**
 * 数据库操作相关上下文工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Configuration
@Order(Integer.MIN_VALUE)
public class DatabaseContextHolder implements ApplicationContextAware {
    /**
     * 类对象/业务操作实例映射表
     */
    private static final Map<Class<?>, IService<Object>> CLASS_SERVICE_MAPPING = new ConcurrentHashMap<>();

    /**
     * 对象机密数据字段映射表
     */
    private static final Map<Class<?>, List<Field>> CONFIDENTIAL_CLASS_FIELD_MAPPING = new ConcurrentHashMap<>();

    /**
     * 事务处理器
     */
    private static Transactor TRANSACTOR = null;

    /**
     * 全局SqlSessionTemplate实例
     */
    private static SqlSessionTemplate SESSION_TEMPLATE = null;

    /**
     * 数据安全配置属性
     */
    private static DatabaseSecurityProperties.Codec CODEC = null;

    /**
     * 数据加密密码处理器
     */
    private static final ThreadLocal<Cipher> ENCRYPT_CIPHER =
            ThreadLocal.withInitial(() -> getCipher(Cipher.ENCRYPT_MODE));

    /**
     * 数据解密密码处理器
     */
    private static final ThreadLocal<Cipher> DECRYPT_CIPHER =
            ThreadLocal.withInitial(() -> getCipher(Cipher.DECRYPT_MODE));

    /**
     * 获取事务处理器
     *
     * @return 事务处理器
     */
    public static Transactor getTransactor() {
        return Objects.requireNonNull(TRANSACTOR, "Transactor has not been initialized");
    }

    /**
     * 获取SqlSessionTemplate
     *
     * @return SqlSessionTemplate实例
     */
    public static SqlSessionTemplate getSessionTemplate() {
        return Objects.requireNonNull(SESSION_TEMPLATE, "Session template has not been initialized");
    }

    /**
     * 获取数据编解码配置
     *
     * @return 数据编解码配置
     */
    public static DatabaseSecurityProperties.Codec getCodec() {
        return Objects.requireNonNull(CODEC, "Database codec has not been initialized");
    }

    /**
     * 获取原生的数据库连接
     *
     * @return 数据库连接对象
     */
    public static Connection getConnection() {
        return getSessionTemplate().getConnection();
    }

    /**
     * 执行原生SQL
     *
     * @param sql SQL语句
     * @return true/false
     * @throws Exception 执行异常
     */
    public static boolean executeNative(String sql) throws Exception {
        try (Statement statement = getConnection().createStatement()) {
            return statement.execute(sql);
        }
    }

    /**
     * 事务执行方法
     *
     * @param executor 回调方法
     */
    public static void transactional(@NonNull Executor executor) {
        getTransactor().execute(executor);
    }

    /**
     * 事务执行方法
     *
     * @param executor 回调方法
     * @param <T>      对象类型
     * @return 执行结果
     */
    public static <T> T transactional(@NonNull Supplier<T> executor) {
        return getTransactor().execute(executor);
    }

    /**
     * 事务执行方法
     *
     * @param executor 回调方法
     */
    public static void transactional(@NonNull Handler executor) throws Exception {
        getTransactor().execute(executor);
    }

    /**
     * 事务执行方法
     *
     * @param executor 回调方法
     * @param <T>      对象类型
     * @return 执行结果
     * @throws Exception 执行异常
     */
    public static <T> T transactional(@NonNull Provider<T> executor) throws Exception {
        return getTransactor().execute(executor);
    }

    /**
     * 根据表名获取表信息
     *
     * @param name 表名称
     * @return 表信息
     */
    public static TableInfo getTable(String name) {
        if (StringUtils.isEmpty(name)) {
            return null;
        }
        TableInfo table = TableInfoHelper.getTableInfo(name);
        if (table == null) {
            if (name.charAt(0) == '`' && name.charAt(name.length() - 1) == '`') {
                table = TableInfoHelper.getTableInfo(name.substring(1, name.length() - 1));
            } else {
                table = TableInfoHelper.getTableInfo(String.format("`%s`", name));
            }
        }
        return table;
    }

    /**
     * 获取对象的机密字段列表
     *
     * @param clazz 对象类型
     * @return 字段列表
     */
    private static List<Field> getConfidentialFields(@NonNull Class<?> clazz) {
        return CONFIDENTIAL_CLASS_FIELD_MAPPING.computeIfAbsent(clazz, c -> {
            List<Field> fields = new LinkedList<>();
            Confidential confidential = clazz.getAnnotation(Confidential.class);
            ObjectUtils.iterateFields(clazz, (index, field) -> {
                Confidential _confidential;
                if (field.getType() == String.class
                        && ((_confidential = field.getAnnotation(Confidential.class)) != null
                        || (_confidential = confidential) != null) && _confidential.value()) {
                    fields.add(field);
                }
            });
            return fields;
        });
    }

    /**
     * 获取指定对象对应的业务操作实例
     *
     * @param clazz 数据实体对象
     * @param <T>   数据对象类型
     * @return 业务操作实例
     */
    @SuppressWarnings("unchecked")
    public static <T> IService<T> getService(Class<T> clazz) {
        return (IService<T>) CLASS_SERVICE_MAPPING.computeIfAbsent(clazz, c -> {
            ApplicationContext applicationContext = ApplicationContextHolder.getApplicationContext();
            IService<Object> service = applicationContext.getBeansOfType(IService.class).values().stream()
                    .filter(s -> Objects.equals(s.getEntityClass(), clazz)).findAny().orElse(null);
            return Objects.requireNonNull(service, "Service not found for class: " + clazz.getName());
        });
    }

    /**
     * 初始化树形实体信息
     *
     * @param entity    数据实体
     * @param operation 操作类型
     */
    public static void initializeTreeProperties(Object entity, @NonNull SqlCommandType operation) {
        if (!(entity instanceof TreeEntity) || operation != SqlCommandType.INSERT) {
            return;
        }
        // 标记层级字段，用于参数占位
        ((TreeEntity<?, ?>) entity).setLevel(1);
        // 标记序列号字段，用于参数占位
        ((TreeEntity<?, ?>) entity).setSequence(StringUtils.EMPTY);
    }

    /**
     * 初始化创建信息
     *
     * @param entity    数据实体
     * @param operation 操作类型
     */
    public static void initializeCreatableProperties(Object entity, @NonNull SqlCommandType operation) {
        if (!(entity instanceof Creatable)
                || (operation != SqlCommandType.INSERT && operation != SqlCommandType.UPDATE)) {
            return;
        }
        Creatable<?> creatable = (Creatable<?>) entity;
        if (creatable.getCreator() == null) {
            creatable.setCreator(SessionContextHolder.getId(false));
        }
        if (creatable.getCreateTime() == null) {
            creatable.setCreateTime(new Date());
        }
    }

    /**
     * 初始化更新信息
     *
     * @param entity    数据实体
     * @param operation 操作类型
     */
    public static void initializeUpdatableProperties(Object entity, @NonNull SqlCommandType operation) {
        if (!(entity instanceof Updatable)
                || (operation != SqlCommandType.INSERT && operation != SqlCommandType.UPDATE)) {
            return;
        }
        Updatable<?> updatable = (Updatable<?>) entity;
        Long operator = SessionContextHolder.getId(false);
        if (operator != null) {
            updatable.setUpdater(operator);
        }
        updatable.setUpdateTime(new Date());
    }

    /**
     * 初始化租户信息
     *
     * @param entity    数据实体
     * @param operation 操作类型
     */
    public static void initializeIsolatableProperties(Object entity, @NonNull SqlCommandType operation) {
        if (!(entity instanceof Isolatable)
                || (operation != SqlCommandType.INSERT && operation != SqlCommandType.UPDATE)) {
            return;
        }
        Isolatable<?> isolatable = (Isolatable<?>) entity;
        isolatable.setTenant(SessionContextHolder.getTenant(false));
    }

    /**
     * 初始化删除标记
     *
     * @param entity    数据实体
     * @param operation 操作类型
     */
    public static void initializeDeletableProperties(Object entity, @NonNull SqlCommandType operation) {
        if (!(entity instanceof Deletable)
                || (operation != SqlCommandType.INSERT && operation != SqlCommandType.UPDATE)) {
            return;
        }
        Deletable<?> deletable = (Deletable<?>) entity;
        if (deletable.getDeleted() == null) {
            deletable.setDeleted(false);
        }
    }

    /**
     * 根据加解密模式获取密码处理器
     *
     * @param mode 加解密模式
     * @return 密码处理器
     */
    public static Cipher getCipher(int mode) {
        DatabaseSecurityProperties.Codec codec = getCodec();
        return StringUtils.isEmpty(codec.getSecret()) || StringUtils.isEmpty(codec.getIv()) ? null :
                AESUtils.getCipher(mode, codec.getSecret(), codec.getIv());
    }

    /**
     * 编码数据
     *
     * @param original 原始数据
     * @return 数据密文
     */
    public static String encode(String original) {
        return AESUtils.encode(getCipher(Cipher.ENCRYPT_MODE), original);
    }

    /**
     * 解密数据
     *
     * @param ciphertext 数据密文
     * @return 数据明文
     */
    public static String decode(String ciphertext) {
        return AESUtils.decode(getCipher(Cipher.DECRYPT_MODE), ciphertext);
    }

    /**
     * 编码敏感数据属性
     *
     * @param entity 数据实体
     * @param <T>    数据类型
     */
    public static <T> void encodeConfidentialProperties(T entity) {
        try {
            updateConfidentialProperties(entity, original -> AESUtils.encode(ENCRYPT_CIPHER.get(), original));
        } finally {
            ENCRYPT_CIPHER.remove();
        }
    }

    /**
     * 编码敏感数据属性
     *
     * @param entities 数据实体集合
     * @param <T>      数据类型
     */
    public static <T> void encodeConfidentialProperties(Collection<T> entities) {
        try {
            updateConfidentialProperties(entities, original -> AESUtils.encode(ENCRYPT_CIPHER.get(), original));
        } finally {
            ENCRYPT_CIPHER.remove();
        }
    }

    /**
     * 解码敏感数据属性
     *
     * @param entity 数据实体
     * @param <T>    数据类型
     */
    public static <T> void decodeConfidentialProperties(T entity) {
        try {
            updateConfidentialProperties(entity, ciphertext -> AESUtils.decode(DECRYPT_CIPHER.get(), ciphertext));
        } finally {
            DECRYPT_CIPHER.remove();
        }
    }

    /**
     * 解码敏感数据属性
     *
     * @param entities 数据实体集合
     * @param <T>      数据类型
     */
    public static <T> void decodeConfidentialProperties(Collection<T> entities) {
        try {
            updateConfidentialProperties(entities, ciphertext -> AESUtils.decode(DECRYPT_CIPHER.get(), ciphertext));
        } finally {
            DECRYPT_CIPHER.remove();
        }
    }

    /**
     * 更新敏感属性数据
     *
     * @param entities 数据实体集合
     * @param codec    敏感数据编解码回调方法
     * @param <T>      数据类型
     */
    public static <T> void updateConfidentialProperties(Collection<T> entities,
                                                        @NonNull Function<String, String> codec) {
        if (CollectionUtils.isEmpty(entities)) {
            return;
        }
        Object object = entities.iterator().next();
        if (object == null) {
            return;
        }
        List<Field> fields = getConfidentialFields(object.getClass());
        if (CollectionUtils.isEmpty(fields)) {
            return;
        }
        entities.forEach(entity -> updateConfidentialProperties(entity, fields, codec));
    }

    /**
     * 更新敏感属性数据
     *
     * @param entity 数据实体
     * @param codec  敏感数据编解码回调方法
     * @param <T>    数据类型
     */
    public static <T> void updateConfidentialProperties(T entity, @NonNull Function<String, String> codec) {
        List<Field> fields = ObjectUtils.ifNull(entity, e -> getConfidentialFields(entity.getClass()));
        updateConfidentialProperties(entity, fields, codec);
    }

    /**
     * 更新敏感属性数据
     *
     * @param entity 数据实体
     * @param fields 敏感字段列表
     * @param codec  敏感数据编解码回调方法
     * @param <T>    数据类型
     */
    private static <T> void updateConfidentialProperties(T entity, List<Field> fields,
                                                         @NonNull Function<String, String> codec) {
        if (entity == null || CollectionUtils.isEmpty(fields)) {
            return;
        }
        boolean debug = log.isDebugEnabled();
        for (Field field : fields) {
            String value = (String) ObjectUtils.getValue(entity, field);
            if (debug) {
                log.debug("Before codec[{}]: {}", field.getName(), value);
            }
            value = codec.apply(value);
            if (debug) {
                log.debug("After codec[{}]: {}", field.getName(), value);
            }
            ObjectUtils.setValue(entity, field, value);
        }
    }

    /**
     * 初始化实体默认属性值
     *
     * @param entity    数据实体
     * @param operation 操作类型
     */
    public static void initializeEntityProperties(Object entity, @NonNull SqlCommandType operation) {
        // 更新基础信息
        initializeTreeProperties(entity, operation);
        initializeCreatableProperties(entity, operation);
        initializeUpdatableProperties(entity, operation);
        initializeDeletableProperties(entity, operation);
        initializeIsolatableProperties(entity, operation);

        // 敏感数据加密
        encodeConfidentialProperties(entity);
    }

    /**
     * 获取指定树以及所有子节点列表
     *
     * @param clazz 对象类型
     * @param id    树对象主键
     * @param <ID>  主键类型
     * @param <T>   对象类型
     * @return 树对象列表
     */
    public static <ID extends Serializable, T extends TreeEntity<ID, T>> List<T>
    getTreeFamilies(@NonNull Class<T> clazz, @NonNull ID id) {
        return getTreeFamilies(getService(clazz), id);
    }

    /**
     * 获取指定树以及所有子节点列表
     *
     * @param service 对象业务操作接口
     * @param id      树对象主键
     * @param <ID>    主键类型
     * @param <T>     对象类型
     * @return 树对象列表
     */
    public static <ID extends Serializable, T extends TreeEntity<ID, T>> List<T>
    getTreeFamilies(@NonNull IService<T> service, @NonNull ID id) {
        return service.lambdaQuery().likeRight(TreeEntity::getSequence, EntityUtils.buildTreeSequence(id)).list();
    }

    /**
     * 获取指定树以及所有子节点列表
     *
     * @param clazz 对象类型
     * @param ids   树对象主键集合
     * @param <ID>  主键类型
     * @param <T>   对象类型
     * @return 树对象列表
     */
    public static <ID extends Serializable, T extends TreeEntity<ID, T>> List<T>
    listTreeFamilies(@NonNull Class<T> clazz, @NonNull Collection<ID> ids) {
        return listTreeFamilies(getService(clazz), ids);
    }

    /**
     * 获取指定树以及所有子节点列表
     *
     * @param service 对象业务操作接口
     * @param ids     树对象主键集合
     * @param <ID>    主键类型
     * @param <T>     对象类型
     * @return 树对象列表
     */
    public static <ID extends Serializable, T extends TreeEntity<ID, T>> List<T>
    listTreeFamilies(@NonNull IService<T> service, @NonNull Collection<ID> ids) {
        if (CollectionUtils.isEmpty(ids)) {
            return Collections.emptyList();
        }
        LambdaQueryWrapper<T> query = Wrappers.lambdaQuery();
        ids.forEach(id -> query.or(c -> c.likeRight(TreeEntity::getSequence, EntityUtils.buildTreeSequence(id))));
        return service.list(query);
    }

    /**
     * 软删除对象实体
     *
     * @param clazz 对象类型
     * @param id    对象主键
     * @param <ID>  主键类型
     * @param <T>   对象类型
     * @return true/false
     */
    public static <ID extends Serializable, T extends Deletable<ID>> boolean softDelete(@NonNull Class<T> clazz,
                                                                                        ID id) {
        return softDelete(getService(clazz), id);
    }

    /**
     * 软删除对象实体
     *
     * @param service 对象业务操作接口
     * @param id      对象主键
     * @param <ID>    主键类型
     * @param <T>     对象类型
     * @return true/false
     */
    public static <ID extends Serializable, T extends Deletable<ID>> boolean softDelete(@NonNull IService<T> service,
                                                                                        ID id) {
        return softDelete(service, ObjectUtils.ifNull(id, service::getById));
    }

    /**
     * 软删除对象实体
     *
     * @param clazz  对象类型
     * @param entity 可删除对象实体
     * @param <ID>   主键类型
     * @param <T>    对象类型
     * @return true/false
     */
    public static <ID extends Serializable, T extends Deletable<ID>> boolean softDelete(@NonNull Class<T> clazz,
                                                                                        T entity) {
        return softDelete(getService(clazz), entity);
    }

    /**
     * 软删除对象实体
     *
     * @param clazz    对象类型
     * @param entities 可删除对象实体集合
     * @param <ID>     主键类型
     * @param <T>      对象类型
     * @return true/false
     */
    public static <ID extends Serializable, T extends Deletable<ID>> boolean softDelete(@NonNull Class<T> clazz,
                                                                                        Collection<T> entities) {
        return softDelete(getService(clazz), entities);
    }

    /**
     * 软删除对象实体
     *
     * @param service 对象业务操作接口
     * @param entity  可删除对象实体
     * @param <ID>    主键类型
     * @param <T>     对象类型
     * @return true/false
     */
    public static <ID extends Serializable, T extends Deletable<ID>> boolean softDelete(@NonNull IService<T> service,
                                                                                        T entity) {
        if (entity != null) {
            entity.setDeleted(true);
            return service.updateById(entity);
        }
        return false;
    }

    /**
     * 软删除对象实体
     *
     * @param service  对象业务操作接口
     * @param entities 可删除对象实体集合
     * @param <ID>     主键类型
     * @param <T>      对象类型
     * @return true/false
     */
    public static <ID extends Serializable, T extends Deletable<ID>> boolean softDelete(@NonNull IService<T> service,
                                                                                        Collection<T> entities) {
        if (!CollectionUtils.isEmpty(entities)) {
            entities.forEach(entity -> entity.setDeleted(true));
            return service.updateBatchById(entities);
        }
        return false;
    }

    /**
     * 获取非空实例
     *
     * @param clazz 对象类型
     * @param id    对象主键
     * @param <ID>  主键类型
     * @param <T>   对象类型
     * @return 对象实例
     */
    public static <ID extends Serializable, T extends Entity<ID>> T getRequireEntity(@NonNull Class<T> clazz,
                                                                                     @NonNull ID id) {
        return getRequireEntity(getService(clazz), id);
    }

    /**
     * 获取非空实例
     *
     * @param clazz   对象类型
     * @param id      对象主键
     * @param message 异常消息
     * @param <ID>    主键类型
     * @param <T>     对象类型
     * @return 对象实例
     */
    public static <ID extends Serializable, T extends Entity<ID>> T getRequireEntity(@NonNull Class<T> clazz,
                                                                                     @NonNull ID id, String message) {
        return getRequireEntity(getService(clazz), id, message);
    }

    /**
     * 获取非空实例
     *
     * @param service 对象业务操作接口
     * @param id      对象主键
     * @param <ID>    主键类型
     * @param <T>     对象类型
     * @return 对象实例
     */
    public static <ID extends Serializable, T extends Entity<ID>> T getRequireEntity(@NonNull IService<T> service,
                                                                                     @NonNull ID id) {
        return getRequireEntity(service, id, "Unknown entity");
    }

    /**
     * 获取非空实例
     *
     * @param service 对象业务操作接口
     * @param id      对象主键
     * @param message 异常消息
     * @param <ID>    主键类型
     * @param <T>     对象类型
     * @return 对象实例
     */
    public static <ID extends Serializable, T extends Entity<ID>> T getRequireEntity(@NonNull IService<T> service,
                                                                                     @NonNull ID id, String message) {
        T entity = service.getById(id);
        ErrorUtils.requireResource(entity, message);
        return entity;
    }

    @Bean
    public Transactor transactor() {
        return new Transactor() {
            @Override
            @Transactional(rollbackFor = Exception.class)
            public void execute(@NonNull Executor executor) {
                executor.execute();
            }

            @Override
            @Transactional(rollbackFor = Exception.class)
            public <T> T execute(@NonNull Supplier<T> executor) {
                return executor.get();
            }

            @Override
            @Transactional(rollbackFor = Exception.class)
            public void execute(@NonNull Handler executor) throws Exception {
                executor.handle();
            }

            @Override
            @Transactional(rollbackFor = Exception.class)
            public <T> T execute(@NonNull Provider<T> executor) throws Exception {
                return executor.get();
            }
        };
    }

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        // 初始化事务处理器
        TRANSACTOR = applicationContext.getBean(Transactor.class);

        // 初始化Session模版实例
        SESSION_TEMPLATE = applicationContext.getBean(SqlSessionTemplate.class);

        // 初始化数据编解码配置
        CODEC = applicationContext.getBean(DatabaseSecurityProperties.class).getCodec();
        if (!StringUtils.isEmpty(CODEC.getSecret())) {
            CODEC.setSecret(Base64.encodeBase64String(AESUtils.revise(CODEC.getSecret()).getBytes()));
        }
        if (!StringUtils.isEmpty(CODEC.getIv())) {
            CODEC.setIv(Base64.encodeBase64String(AESUtils.revise(CODEC.getIv()).getBytes()));
        }
    }

    /**
     * 事务处理器
     */
    public interface Transactor {
        /**
         * 事务执行
         *
         * @param executor 事务方法接口
         */
        void execute(Executor executor);

        /**
         * 事务执行
         *
         * @param executor 事务方法接口
         * @param <T>      对象类型
         * @return 执行结果
         */
        <T> T execute(Supplier<T> executor);

        /**
         * 事务执行
         *
         * @param executor 事务方法接口
         * @throws Exception 操作异常
         */
        void execute(Handler executor) throws Exception;

        /**
         * 事务执行
         *
         * @param executor 事务方法接口
         * @param <T>      对象类型
         * @return 执行结果
         * @throws Exception 操作异常
         */
        <T> T execute(Provider<T> executor) throws Exception;
    }
}