package com.chelvc.framework.database.interceptor;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.chelvc.framework.base.exception.ParameterInvalidException;
import com.chelvc.framework.base.model.Result;
import com.chelvc.framework.base.util.ErrorUtils;
import com.chelvc.framework.base.util.HttpUtils;
import lombok.extern.slf4j.Slf4j;
import org.mybatis.spring.MyBatisSystemException;
import org.springframework.core.Ordered;
import org.springframework.core.annotation.Order;
import org.springframework.http.HttpStatus;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.bind.annotation.ResponseStatus;
import org.springframework.web.bind.annotation.RestControllerAdvice;

/**
 * 数据库异常处理拦截器
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@RestControllerAdvice
@Order(Ordered.HIGHEST_PRECEDENCE)
public class DatabaseExceptionInterceptor {
    /**
     * 参数异常转换
     *
     * @param response  请求响应对象
     * @param exception 异常对象
     * @return 请求结果
     */
    @ResponseStatus(HttpStatus.PRECONDITION_FAILED)
    @ExceptionHandler(ParameterInvalidException.class)
    public Result<?> exception(HttpServletRequest request, HttpServletResponse response,
                               ParameterInvalidException exception) {
        log.warn(HttpUtils.getExceptionMessage(request, HttpStatus.PRECONDITION_FAILED, exception));
        Result<?> result = Result.build(HttpStatus.PRECONDITION_FAILED, exception.getErrors());
        response.setStatus(result.getCode());
        HttpUtils.initializeUnifiedResponseHeader(response);
        return result;
    }

    /**
     * 数据持久化异常转换
     *
     * @param response  请求响应对象
     * @param exception 异常对象
     * @return 请求结果
     */
    @ExceptionHandler(MyBatisSystemException.class)
    public Result<?> exception(HttpServletRequest request, HttpServletResponse response,
                               MyBatisSystemException exception) {
        ParameterInvalidException parameterInvalidException =
                ErrorUtils.lookup(exception, ParameterInvalidException.class);
        if (parameterInvalidException != null) {
            return this.exception(request, response, parameterInvalidException);
        }
        log.error(HttpUtils.getExceptionMessage(request, HttpStatus.INTERNAL_SERVER_ERROR, exception), exception);
        Result<?> result = Result.build(HttpStatus.INTERNAL_SERVER_ERROR);
        response.setStatus(result.getCode());
        HttpUtils.initializeUnifiedResponseHeader(response);
        return result;
    }
}
