package com.chelvc.framework.database.interceptor;

import java.util.Collections;
import java.util.Map;
import java.util.Set;

import com.baomidou.dynamic.datasource.spring.boot.autoconfigure.DynamicDataSourceProperties;
import com.baomidou.dynamic.datasource.toolkit.DynamicDataSourceContextHolder;
import com.chelvc.framework.base.util.StringUtils;
import com.google.common.collect.Maps;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.springframework.aop.framework.AopProxyUtils;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.config.ConfigurableListableBeanFactory;
import org.springframework.beans.factory.support.BeanDefinitionRegistry;
import org.springframework.beans.factory.support.BeanDefinitionRegistryPostProcessor;
import org.springframework.beans.factory.support.RootBeanDefinition;
import org.springframework.boot.autoconfigure.condition.ConditionalOnClass;
import org.springframework.core.annotation.Order;
import org.springframework.util.CollectionUtils;

/**
 * 动态数据源拦截器
 * <p>
 * 拦截使用了{@link org.springframework.stereotype.Controller}注解的类方法
 * 拦截使用了{@link org.springframework.web.bind.annotation.RestController}注解的类方法
 * 拦截使用了{@link org.springframework.stereotype.Service}注解的类方法
 * 拦截实现了{@link com.baomidou.mybatisplus.extension.service.IService}接口的类方法
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@ConditionalOnClass(DynamicDataSourceContextHolder.class)
public class DynamicDatasourceInterceptor implements BeanDefinitionRegistryPostProcessor {
    @Override
    public void postProcessBeanFactory(ConfigurableListableBeanFactory factory) throws BeansException {
    }

    @Override
    public void postProcessBeanDefinitionRegistry(BeanDefinitionRegistry registry) throws BeansException {
        boolean standard = true;
        try {
            // 尝试加载@DubboService注解
            this.getClass().getClassLoader().loadClass("org.apache.dubbo.config.annotation.DubboService");
        } catch (ClassNotFoundException e) {
            standard = false;
        }
        if (standard) {
            registry.registerBeanDefinition("standardDynamicDatasourceInterceptor",
                    new RootBeanDefinition(StandardDynamicDatasourceInterceptor.class));
        } else {
            registry.registerBeanDefinition("basicDynamicDatasourceInterceptor",
                    new RootBeanDefinition(BasicDynamicDatasourceInterceptor.class));
        }
    }

    /**
     * 动态数据源拦截器基础实现
     */
    @Aspect
    @Order(Integer.MIN_VALUE)
    public static class BasicDynamicDatasourceInterceptor extends AbstractDynamicDatasourceInterceptor {
        public BasicDynamicDatasourceInterceptor(DynamicDataSourceProperties properties) {
            super(properties);
        }

        /**
         * 数据源路由
         *
         * @param point 方法拦截点
         * @return 方法调用结果
         * @throws Throwable 方法调用异常
         */
        @Around("@within(org.springframework.stereotype.Service) " +
                "|| this(com.baomidou.mybatisplus.core.mapper.BaseMapper) " +
                "|| @within(org.apache.ibatis.annotations.Mapper)")
        public Object routing(ProceedingJoinPoint point) throws Throwable {
            return super.routing(point);
        }
    }

    /**
     * 动态数据源拦截器标准实现
     */
    @Aspect
    @Order(Integer.MIN_VALUE)
    public static class StandardDynamicDatasourceInterceptor extends AbstractDynamicDatasourceInterceptor {
        public StandardDynamicDatasourceInterceptor(DynamicDataSourceProperties properties) {
            super(properties);
        }

        /**
         * 数据源路由
         *
         * @param point 方法拦截点
         * @return 方法调用结果
         * @throws Throwable 方法调用异常
         */
        @Around("@within(org.springframework.stereotype.Service) " +
                "|| @within(org.apache.dubbo.config.annotation.DubboService) " +
                "|| this(com.baomidou.mybatisplus.core.mapper.BaseMapper) " +
                "|| @within(org.apache.ibatis.annotations.Mapper)")
        public Object routing(ProceedingJoinPoint point) throws Throwable {
            return super.routing(point);
        }
    }

    /**
     * 动态数据源拦截器抽象实现
     */
    public static abstract class AbstractDynamicDatasourceInterceptor {
        private final Map<String, String> datasourceContext;

        public AbstractDynamicDatasourceInterceptor(@NonNull DynamicDataSourceProperties properties) {
            // 初始化数据源标识集合
            Set<String> keys = properties.getDatasource().keySet();
            if (CollectionUtils.isEmpty(keys)) {
                this.datasourceContext = Collections.emptyMap();
            } else {
                this.datasourceContext = Maps.newHashMapWithExpectedSize(keys.size());
                keys.forEach(value -> {
                    for (String key : StringUtils.splitActives(value)) {
                        this.datasourceContext.put(key, value);
                    }
                });
            }
            if (log.isDebugEnabled()) {
                log.debug("Initialized datasource context: {}", this.datasourceContext);
            }
        }

        /**
         * 查找数据源标识
         *
         * @param point 方法拦截点
         * @return 数据源标识
         */
        private String lookupDatasourceKey(ProceedingJoinPoint point) {
            Object target = point.getTarget();

            // 根据目标类对象查找数据源标识
            String className = AopProxyUtils.ultimateTargetClass(target).getName();
            Set<Map.Entry<String, String>> routes = this.datasourceContext.entrySet();
            String key = routes.stream().filter(entry -> className.startsWith(entry.getKey())).findAny()
                    .map(Map.Entry::getValue).orElse(null);
            if (!StringUtils.isEmpty(key)) {
                return key;
            }

            // 根据目标接口对象查找数据源标识
            Class<?>[] interfaces = AopProxyUtils.proxiedUserInterfaces(target);
            if (interfaces.length > 0) {
                for (Class<?> clazz : interfaces) {
                    String interfaceName = clazz.getName();
                    String route = routes.stream().filter(entry -> interfaceName.startsWith(entry.getKey())).findAny()
                            .map(Map.Entry::getValue).orElse(null);
                    if (!StringUtils.isEmpty(route)) {
                        return route;
                    }
                }
            }
            return null;
        }

        /**
         * 数据源路由
         *
         * @param point 方法拦截点
         * @return 方法调用结果
         * @throws Throwable 方法调用异常
         */
        protected Object routing(ProceedingJoinPoint point) throws Throwable {
            String key = this.lookupDatasourceKey(point);
            if (log.isDebugEnabled()) {
                log.debug("Initializing datasource context for {}: {}", point.getSignature().toShortString(), key);
            }
            DynamicDataSourceContextHolder.push(key);
            try {
                return point.proceed(point.getArgs());
            } finally {
                DynamicDataSourceContextHolder.poll();
            }
        }
    }
}