package com.chelvc.framework.database.interceptor;

import java.io.Serializable;
import java.sql.Connection;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import com.chelvc.framework.base.function.Provider;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.chelvc.framework.database.annotation.Confidential;
import com.chelvc.framework.database.context.DatabaseContextHolder;
import com.chelvc.framework.database.entity.Creatable;
import com.chelvc.framework.database.entity.Deletable;
import com.chelvc.framework.database.entity.Entity;
import com.chelvc.framework.database.entity.Isolatable;
import com.chelvc.framework.database.entity.TreeEntity;
import com.chelvc.framework.database.entity.Updatable;
import com.chelvc.framework.database.util.EntityUtils;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.ibatis.cache.CacheKey;
import org.apache.ibatis.executor.Executor;
import org.apache.ibatis.executor.statement.StatementHandler;
import org.apache.ibatis.mapping.BoundSql;
import org.apache.ibatis.mapping.MappedStatement;
import org.apache.ibatis.mapping.ParameterMapping;
import org.apache.ibatis.mapping.SqlCommandType;
import org.apache.ibatis.plugin.Interceptor;
import org.apache.ibatis.plugin.Intercepts;
import org.apache.ibatis.plugin.Invocation;
import org.apache.ibatis.plugin.Plugin;
import org.apache.ibatis.plugin.Signature;
import org.apache.ibatis.reflection.MetaObject;
import org.apache.ibatis.reflection.SystemMetaObject;
import org.apache.ibatis.session.ResultHandler;
import org.apache.ibatis.session.RowBounds;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.annotation.Order;
import org.springframework.stereotype.Component;

/**
 * 基础属性自动更新维护拦截器
 *
 * @author Woody
 * @date 2023/4/5
 * @see Creatable
 * @see Updatable
 * @see Deletable
 * @see Isolatable
 * @see TreeEntity
 * @see Confidential
 */
@Slf4j
@Component
@Order(20000)
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
@Intercepts({
        @Signature(type = Executor.class, method = "update", args = {MappedStatement.class, Object.class}),
        @Signature(type = Executor.class, method = "query", args = {MappedStatement.class, Object.class,
                RowBounds.class, ResultHandler.class}),
        @Signature(type = Executor.class, method = "query", args = {MappedStatement.class, Object.class,
                RowBounds.class, ResultHandler.class, CacheKey.class, BoundSql.class}),
        @Signature(type = StatementHandler.class, method = "prepare", args = {Connection.class, Integer.class})
})
public class PropertyUpdateInterceptor implements Interceptor {
    /**
     * 刷新树形数据子节点序列号SQL语句
     */
    private static final String REFRESH_CHILDREN_SEQUENCE_SQL = "update `%s` t " +
            "left join (select id, level, sequence from `%s` where id = %s) c on 1=1 " +
            "left join (select level, sequence from `%s` where id = %s) p on 1=1 " +
            "set t.level = abs(cast(t.level as signed) - cast(c.level as signed) + ifnull(p.level, 0) + 1), " +
            "t.sequence = replace(t.sequence, c.sequence, concat(ifnull(p.sequence, ''), c.id, '" +
            StringUtils.HORIZONTAL + "')) where t.sequence like concat(c.sequence, '%%')";

    /**
     * 构建树形子节点刷新SQL语句
     *
     * @param table    数据表名称
     * @param id       被更新的数据主键
     * @param parentId 替换的目标上级节点主键
     * @return SQL语句
     */
    private String buildChildrenRefreshSql(String table, Serializable id, Serializable parentId) {
        id = ObjectUtils.ifNull(id, -1);
        parentId = ObjectUtils.ifNull(parentId, -1);
        return String.format(REFRESH_CHILDREN_SEQUENCE_SQL, table, table, id, table, parentId);
    }

    /**
     * 构建树形子节点数据创建SQL语句，用于覆盖默认序列号
     * 如果存在上级节点，则将"上级节点序列号-当前主键"作为完成序列号
     * 为了防止在并发情况下造成更新到不准确的数据，所以采用insert into ... select ... 的单条SQL执行方式
     *
     * @param tree     树形数据实体
     * @param boundSql SQL绑定对象
     * @return SQL语句
     */
    private String buildChildrenCreateSql(TreeEntity<?, ?> tree, BoundSql boundSql) {
        String sql = boundSql.getSql();
        String table = EntityUtils.getTableInfo(tree.getClass()).getTableName();
        StringBuilder builder = new StringBuilder(sql.substring(0, sql.indexOf("VALUES")));
        builder.append(" select ");
        List<String> placeholders = new LinkedList<>();
        Iterator<ParameterMapping> iterator = boundSql.getParameterMappings().iterator();
        while (iterator.hasNext()) {
            String property = iterator.next().getProperty();
            if (property.equals("level")) {
                placeholders.add("level + 1");
                iterator.remove();
            } else if (property.equals("sequence")) {
                placeholders.add(String.format("concat(sequence, '%s')", tree.getSequence()));
                iterator.remove();
            } else {
                placeholders.add("?");
            }
        }
        return builder.append(String.join(" ,", placeholders)).append(" from `").append(table)
                .append("` where id = ").append(tree.getParentId()).toString();
    }

    /**
     * 拦截{@link Executor}方法调用
     *
     * @param invocation 调用信息
     * @return 调用结果
     * @throws Throwable 调用异常
     */
    private Object interceptExecutorInvocation(Invocation invocation) throws Throwable {
        MappedStatement mappedStatement = (MappedStatement) invocation.getArgs()[0];
        SqlCommandType operation = mappedStatement.getSqlCommandType();
        if (operation == SqlCommandType.SELECT) {
            // 敏感数据解密
            Object result = invocation.proceed();
            if (result instanceof Collection) {
                DatabaseContextHolder.decodeConfidentialProperties((Collection<?>) result);
            } else {
                DatabaseContextHolder.decodeConfidentialProperties(result);
            }
            return result;
        } else if (operation != SqlCommandType.INSERT && operation != SqlCommandType.UPDATE) {
            return invocation.proceed();
        }

        Object entity = EntityUtils.getTargetObject(invocation.getArgs()[1]);
        if (!(entity instanceof Entity)) {
            return invocation.proceed();
        }

        // 初始化默认属性值
        DatabaseContextHolder.initializeEntityProperties(entity, operation);

        // 树形数据更新时，刷新子节点序列号
        if (entity instanceof TreeEntity && operation == SqlCommandType.UPDATE) {
            TreeEntity<?, ?> tree = (TreeEntity<?, ?>) entity;
            String table = EntityUtils.getTableInfo(entity.getClass()).getTableName();
            return DatabaseContextHolder.transactional((Provider<?>) () -> {
                Object result = invocation.proceed();
                String refreshSql = this.buildChildrenRefreshSql(table, tree.getId(), tree.getParentId());
                DatabaseContextHolder.executeNative(refreshSql);
                return result;
            });
        }
        return invocation.proceed();
    }

    /**
     * 拦截{@link StatementHandler}方法调用
     *
     * @param invocation 调用信息
     * @return 调用结果
     * @throws Throwable 调用异常
     */
    private Object interceptStatementHandlerInvocation(Invocation invocation) throws Throwable {
        StatementHandler statementHandler = (StatementHandler) invocation.getTarget();
        MetaObject metaObject = SystemMetaObject.forObject(statementHandler);
        MappedStatement mappedStatement = (MappedStatement) metaObject.getValue("delegate.mappedStatement");
        SqlCommandType operation = mappedStatement.getSqlCommandType();
        if (operation != SqlCommandType.INSERT) {
            return invocation.proceed();
        }

        // 获取操作实体，如果实体类型为Map，则需要通过属性"et"进一步获取原始对象
        BoundSql boundSql = (BoundSql) metaObject.getValue("delegate.boundSql");
        MetaObject parameter = SystemMetaObject.forObject(boundSql.getParameterObject());
        Object entity = parameter.getOriginalObject();
        if (entity instanceof Map && parameter.hasGetter("et")) {
            entity = parameter.metaObjectForProperty("et").getOriginalObject();
        }
        if (entity instanceof TreeEntity) {
            TreeEntity<?, ?> tree = (TreeEntity<?, ?>) entity;
            tree.setLevel(1);
            Serializable id = Objects.requireNonNull(tree.getId(), "Identity unspecified");
            tree.setSequence(EntityUtils.buildTreeSequence(id));
            if (tree.getParentId() != null) {
                String sql = this.buildChildrenCreateSql(tree, boundSql);
                metaObject.setValue("delegate.boundSql.sql", sql);
            }
        }
        return invocation.proceed();
    }

    @Override
    public Object intercept(Invocation invocation) throws Throwable {
        if (invocation.getTarget() instanceof Executor) {
            return this.interceptExecutorInvocation(invocation);
        } else if (invocation.getTarget() instanceof StatementHandler) {
            return this.interceptStatementHandlerInvocation(invocation);
        }
        return invocation.proceed();
    }

    @Override
    public Object plugin(Object target) {
        return target instanceof Executor || target instanceof StatementHandler ? Plugin.wrap(target, this) : target;
    }
}
