package com.chelvc.framework.database.util;

import java.util.Collections;
import java.util.List;
import java.util.function.Function;

import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.chelvc.framework.base.model.Pagination;
import com.chelvc.framework.base.model.Paging;
import com.chelvc.framework.base.util.ObjectUtils;
import lombok.NonNull;
import org.springframework.util.CollectionUtils;

/**
 * 分页工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
public final class PagingUtils {
    private PagingUtils() {
    }

    /**
     * 构建分页信息
     *
     * @param number 页码
     * @param size   页面大小
     * @param <T>    数据类型
     * @return 分页信息
     */
    public static <T> Page<T> page(long number, long size) {
        return page(number, size, true);
    }

    /**
     * 构建分页信息
     *
     * @param number   页码
     * @param size     页面大小
     * @param counting 是否获取总数
     * @param <T>      数据类型
     * @return 分页信息
     */
    public static <T> Page<T> page(long number, long size, boolean counting) {
        return new Page<>(number, size, counting);
    }

    /**
     * 将分页请求参数转换成MyBatis Plus分页参数
     *
     * @param paging 分页请求参数
     * @param <T>    目标对象类型泛型
     * @return MyBatis Plus分页参数
     */
    public static <T> Page<T> convert(Paging paging) {
        return paging == null ? null : page(paging.getNumber(), paging.getSize(), paging.isCounting());
    }

    /**
     * 将分页请求参数转换成MyBatis Plus分页参数
     *
     * @param paging   分页请求参数
     * @param counting 是否获取总数
     * @param <T>      目标对象类型泛型
     * @return MyBatis Plus分页参数
     */
    public static <T> Page<T> convert(Paging paging, boolean counting) {
        return paging == null ? null : page(paging.getNumber(), paging.getSize(), counting);
    }

    /**
     * 将MyBatis Plus分页参数转换成分页返回参数
     *
     * @param page MyBatis Plus分页参数
     * @param <T>  目标对象类型泛型
     * @return 分页返回参数
     */
    public static <T> Pagination<T> convert(Page<T> page) {
        return convert(page, (List<T>) ObjectUtils.ifNull(page, Page::getRecords));
    }

    /**
     * 将MyBatis Plus分页参数转换成分页返回参数
     *
     * @param page MyBatis Plus分页参数
     * @param type 目标类型
     * @param <T>  目标对象类型泛型
     * @return 分页返回参数
     */
    public static <T> Pagination<T> convert(Page<?> page, Class<T> type) {
        return convert(page, ObjectUtils.copying((List<?>) ObjectUtils.ifNull(page, Page::getRecords), type));
    }

    /**
     * 将MyBatis Plus分页参数转换成分页返回参数
     *
     * @param page    MyBatis Plus分页参数
     * @param adapter 对象适配函数
     * @param <T>     原始对象类型泛型
     * @param <R>     目标对象类型泛型
     * @return 分页返回参数
     */
    public static <T, R> Pagination<R> convert(Page<T> page, @NonNull Function<List<T>, List<R>> adapter) {
        List<T> records = ObjectUtils.ifNull(page, Page::getRecords);
        return convert(page, CollectionUtils.isEmpty(records) ? Collections.emptyList() : adapter.apply(records));
    }

    /**
     * 将MyBatis Plus分页参数转换成分页返回参数
     *
     * @param page    MyBatis Plus分页参数
     * @param records 分页数据列表
     * @param <T>     目标对象类型泛型
     * @return 分页返回参数
     */
    public static <T> Pagination<T> convert(Page<?> page, List<T> records) {
        return page == null ? null : Pagination.<T>builder().total(page.getTotal()).pages(page.getPages())
                .records(records).build();
    }

    /**
     * 分页记录转换
     *
     * @param page    原始分页信息
     * @param records 分页数据列表
     * @param <T>     目标对象类型泛型
     * @return 分页返回参数
     */
    public static <T> Pagination<T> convert(Pagination<?> page, List<T> records) {
        return page == null ? null : page.convert(records);
    }
}
