package com.chelvc.framework.dubbo.adapter;

import java.util.Collection;
import java.util.Map;

import com.google.common.collect.Maps;
import lombok.extern.slf4j.Slf4j;
import org.springframework.aop.framework.AopProxyUtils;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.core.ResolvableType;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;

/**
 * Dubbo对象适配器工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Component
public class DubboObjectAdapterHolder implements ApplicationContextAware {
    /**
     * 对象适配器映射表
     */
    private static Map<Class<?>, DubboObjectAdapterWrapper<?, ?>> CLASS_WRAPPER_MAPPING = null;

    /**
     * 查找对象适配器
     *
     * @param type 对象类型
     * @param <O>  原始对象类型
     * @param <T>  目标对象类型
     * @return 对象适配器实例
     */
    @SuppressWarnings("unchecked")
    public static <O, T> DubboObjectAdapterWrapper<O, T> lookupObjectAdapter(Class<?> type) {
        return CLASS_WRAPPER_MAPPING == null ? null : (DubboObjectAdapterWrapper<O, T>) CLASS_WRAPPER_MAPPING.get(type);
    }

    /**
     * 将对象适配器转换成包装对象实例
     *
     * @param adapter 对象适配器
     * @param <O>     原始对象类型
     * @param <T>     目标对象类型
     * @return 对象适配器包装对象实例
     */
    @SuppressWarnings("unchecked")
    private <O, T> DubboObjectAdapterWrapper<O, T> adapter2wrapper(DubboObjectAdapter<O, T> adapter) {
        Class<?> clazz = AopProxyUtils.ultimateTargetClass(adapter);
        ResolvableType type = ResolvableType.forClass(clazz);
        ResolvableType[] interfaces = type.getInterfaces();
        Class<O> original = (Class<O>) interfaces[0].getGeneric(0).resolve();
        Class<T> target = (Class<T>) interfaces[0].getGeneric(1).resolve();
        return DubboObjectAdapterWrapper.<O, T>builder().original(original).target(target).adapter(adapter).build();
    }

    @Override
    @SuppressWarnings({"rawtypes", "unchecked"})
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        Collection<DubboObjectAdapter> adapters =
                applicationContext.getBeansOfType(DubboObjectAdapter.class).values();
        if (!CollectionUtils.isEmpty(adapters)) {
            log.info("Loading dubbo object adapters: {}", adapters);
            CLASS_WRAPPER_MAPPING = Maps.newHashMapWithExpectedSize(adapters.size());
            adapters.forEach(adapter -> {
                DubboObjectAdapterWrapper<?, ?> wrapper = this.adapter2wrapper(adapter);
                CLASS_WRAPPER_MAPPING.put(wrapper.getOriginal(), wrapper);
            });
        }
    }
}
