package com.chelvc.framework.email;

import java.io.File;
import java.io.UnsupportedEncodingException;
import java.nio.charset.StandardCharsets;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMessage;

import com.chelvc.framework.base.util.EmailUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.chelvc.framework.email.config.EmailProperties;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

/**
 * 邮件操作接口默认实现
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Component
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class DefaultEmailHandler implements EmailHandler {
    /**
     * 空邮件附件数组
     */
    private static final MimeBodyPart[] EMPTY_ATTACHMENT_ARRAY = new MimeBodyPart[0];

    private final Session session;
    private final EmailProperties properties;

    /**
     * 将邮件体转换成邮件消息
     *
     * @param body        邮件体对象
     * @param attachments 邮件附件数组
     * @return 邮件消息对象
     */
    private Message body2message(@NonNull Body body, @NonNull MimeBodyPart... attachments) {
        Message message = new MimeMessage(this.session);
        try {
            message.addHeader("X-Mailer", "Microsoft Outlook Express 6.00.2900.2869");

            // 设置发件人
            if (StringUtils.isEmpty(this.properties.getNickname())) {
                message.setFrom(new InternetAddress(this.properties.getUsername()));
            } else {
                message.setFrom(new InternetAddress(this.properties.getUsername(), this.properties.getNickname(),
                        StandardCharsets.UTF_8.name()));
            }

            // 设置目标账户
            message.setRecipients(Message.RecipientType.TO, InternetAddress.parse(body.getTo()));

            // 设置抄送人
            if (!StringUtils.isEmpty(body.getCc())) {
                message.setRecipients(Message.RecipientType.CC, InternetAddress.parse(body.getCc()));
            }

            // 设置暗送人
            if (!StringUtils.isEmpty(body.getBcc())) {
                message.setRecipients(Message.RecipientType.BCC, InternetAddress.parse(body.getBcc()));
            }

            // 设置邮件主题
            message.setSubject(body.getSubject());
            message.setContent(EmailUtils.mixed(body.getContent(), attachments));
        } catch (MessagingException | UnsupportedEncodingException e) {
            throw new RuntimeException(e);
        }
        return message;
    }

    @Override
    public boolean send(@NonNull Body body) {
        return this.send(body, EMPTY_ATTACHMENT_ARRAY);
    }

    @Override
    public boolean send(@NonNull Body body, @NonNull File... attachments) {
        return this.send(body, EmailUtils.file2mixed(attachments));
    }

    @Override
    public boolean send(@NonNull Body body, @NonNull MimeBodyPart... attachments) {
        Message message = this.body2message(body, attachments);
        Transport transport = null;
        try {
            transport = this.session.getTransport();
            transport.connect();
            transport.sendMessage(message, message.getAllRecipients());
        } catch (Exception e) {
            log.warn("Email send failed: {}", e.getMessage());
            return false;
        } finally {
            if (transport != null) {
                try {
                    transport.close();
                } catch (Exception e) {
                    log.warn("Close email transport failed: {}", e.getMessage(), e);
                }
            }
        }
        return true;
    }
}
