package com.chelvc.framework.jpush.util;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;

import cn.jpush.api.push.model.Message;
import cn.jpush.api.push.model.Options;
import cn.jpush.api.push.model.PushPayload;
import cn.jpush.api.push.model.audience.Audience;
import cn.jpush.api.push.model.notification.IosAlert;
import cn.jpush.api.push.model.notification.IosNotification;
import cn.jpush.api.push.model.notification.Notification;
import com.chelvc.framework.base.model.Platform;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.gson.JsonObject;
import lombok.NonNull;
import org.springframework.util.CollectionUtils;

/**
 * JPush工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
public final class JPushUtils {
    private JPushUtils() {
    }

    /**
     * 将对象转换成别名
     *
     * @param objects 对象数组
     * @return 别名数组
     */
    public static Audience alias(Object... objects) {
        if (objects == null || objects.length == 0) {
            return null;
        }
        List<String> aliases = Lists.newArrayListWithCapacity(objects.length);
        for (Object object : objects) {
            if (object != null) {
                aliases.add(String.valueOf(object));
            }
        }
        return aliases.isEmpty() ? null : Audience.alias(aliases);
    }

    /**
     * 将对象转换成别名
     *
     * @param objects 对象集合
     * @return 别名数组
     */
    public static Audience alias(Collection<?> objects) {
        if (CollectionUtils.isEmpty(objects)) {
            return null;
        }
        List<String> aliases = objects.stream().filter(Objects::nonNull).map(String::valueOf)
                .collect(Collectors.toList());
        return aliases.isEmpty() ? null : Audience.alias(aliases);
    }

    /**
     * 平台信息适配
     *
     * @param platform 平台信息
     * @return 平台信息
     */
    public static cn.jpush.api.push.model.Platform adapter(Platform platform) {
        if (platform == Platform.IOS) {
            return cn.jpush.api.push.model.Platform.ios();
        } else if (platform == Platform.ANDROID) {
            return cn.jpush.api.push.model.Platform.android();
        }
        return null;
    }

    /**
     * 构建通知对象
     *
     * @param platform 平台信息
     * @param title    通知标题
     * @param content  通知内容
     * @param extras   扩展信息
     * @return 通知对象实例
     */
    public static Notification notification(@NonNull com.chelvc.framework.base.model.Platform platform, String title,
                                            String content, Map<String, String> extras) {
        if (platform == com.chelvc.framework.base.model.Platform.IOS) {
            return Notification.newBuilder().addPlatformNotification(IosNotification.newBuilder()
                    .addExtras(extras)
                    //设置ios标题
                    .setAlert(IosAlert.newBuilder().setTitleAndBody(title, null, content).build())
                    //如果要设置右侧大图，推送的时候携带 "mutable-content":true 说明是支持 iOS10 的 UNNotificationServiceExtension
                    .setMutableContent(true)
                    //如果不需要自定义声音，推送的时候，iOS 的 sound 值保持默认，传 default ，不要传空！
                    .setSound("default")
                    //设置角标自动+1
                    .incrBadge(1)
                    .build()).build();
        } else if (platform == com.chelvc.framework.base.model.Platform.ANDROID) {
            return Notification.android(content, title, extras);
        }
        throw new IllegalArgumentException("Not support platform: " + platform.name());
    }

    /**
     * 初始化安卓通知可选参数
     *
     * @param builder 消息推送载体构造对象
     */
    public static void initializeAndroidNotifyOptions(@NonNull PushPayload.Builder builder) {
        JsonObject huawei = new JsonObject();
        huawei.addProperty("importance", "NORMAL");
        huawei.addProperty("distribution", "secondary_push");

        JsonObject honor = new JsonObject();
        honor.addProperty("distribution", "secondary_push");

        JsonObject xiaomi = new JsonObject();
        xiaomi.addProperty("channel_id", "100458");
        xiaomi.addProperty("distribution", "secondary_push");

        JsonObject oppo = new JsonObject();
        oppo.addProperty("channel_id", "daily_message");
        oppo.addProperty("distribution", "secondary_push");

        JsonObject vivo = new JsonObject();
        vivo.addProperty("classification", 1);
        vivo.addProperty("distribution", "secondary_push");

        Options options = Options.newBuilder().setThirdPartyChannelV2(ImmutableMap.of(
                "huawei", huawei, "honor", honor, "xiaomi", xiaomi, "oppo", oppo, "vivo", vivo
        )).build();
        builder.setOptions(options);
    }

    /**
     * 构建消息推送载体对象
     *
     * @param platform 目标平台
     * @param message  透传消息
     * @param aliases  用户别名数组
     * @return 消息推送载体
     */
    public static PushPayload payload(@NonNull Platform platform, @NonNull Message message,
                                      @NonNull String... aliases) {
        return new PushPayload.Builder().setPlatform(adapter(platform)).setAudience(Audience.alias(aliases))
                .setMessage(message).build();
    }

    /**
     * 构建消息推送载体对象
     *
     * @param platform 目标平台
     * @param message  透传消息
     * @param aliases  用户别名集合
     * @return 消息推送载体
     */
    public static PushPayload payload(@NonNull Platform platform, @NonNull Message message,
                                      @NonNull Collection<String> aliases) {
        return new PushPayload.Builder().setPlatform(adapter(platform)).setAudience(Audience.alias(aliases))
                .setMessage(message).build();
    }

    /**
     * 构建消息推送载体对象
     *
     * @param platform     目标平台
     * @param notification 推送通知
     * @param aliases      用户别名数组
     * @return 消息推送载体
     */
    public static PushPayload payload(@NonNull Platform platform, @NonNull Notification notification,
                                      @NonNull String... aliases) {
        PushPayload.Builder builder = new PushPayload.Builder().setPlatform(adapter(platform))
                .setAudience(Audience.alias(aliases)).setNotification(notification);
        if (platform == Platform.ANDROID) {
            initializeAndroidNotifyOptions(builder);
        }
        return builder.build();
    }

    /**
     * 构建消息推送载体对象
     *
     * @param platform     目标平台
     * @param notification 推送通知
     * @param aliases      用户别名集合
     * @return 消息推送载体
     */
    public static PushPayload payload(@NonNull Platform platform, @NonNull Notification notification,
                                      @NonNull Collection<String> aliases) {
        PushPayload.Builder builder = new PushPayload.Builder().setPlatform(adapter(platform))
                .setAudience(Audience.alias(aliases)).setNotification(notification);
        if (platform == Platform.ANDROID) {
            initializeAndroidNotifyOptions(builder);
        }
        return builder.build();
    }

    /**
     * 构建消息推送载体对象
     *
     * @param target   目标用户
     * @param platform 目标平台
     * @param message  透传消息
     * @return 消息推送载体
     */
    public static PushPayload target(@NonNull String target, @NonNull Platform platform, @NonNull Message message) {
        return new PushPayload.Builder().setPlatform(adapter(platform)).setMessage(message).setTarget(target).build();
    }

    /**
     * 构建消息推送载体对象
     *
     * @param target       目标用户
     * @param platform     目标平台
     * @param notification 推送通知
     * @return 消息推送载体
     */
    public static PushPayload target(@NonNull String target, @NonNull Platform platform,
                                     @NonNull Notification notification) {
        PushPayload.Builder builder = new PushPayload.Builder().setTarget(target).setPlatform(adapter(platform))
                .setNotification(notification);
        if (platform == Platform.ANDROID) {
            initializeAndroidNotifyOptions(builder);
        }
        return builder.build();
    }
}
