package com.chelvc.framework.location.support;

import java.time.Duration;
import java.util.Date;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import com.chelvc.framework.base.util.DateUtils;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.chelvc.framework.location.MobileHandler;
import com.chelvc.framework.location.config.JuheLocationProperties;
import com.chelvc.framework.redis.util.RedisUtils;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Component;
import org.springframework.web.client.RestTemplate;

/**
 * 基于聚合平台手机号定位实现
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Component
@ConditionalOnBean(JuheLocationProperties.class)
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class JuheMobileHandler implements MobileHandler {
    /**
     * 手机号定位接口地址
     */
    private static final String MOBILE2ADDRESS_URL = "http://apis.juhe.cn/mobile/get?phone=%s&key=%s";

    private final RestTemplate restTemplate;
    private final RedisTemplate<String, Object> redisTemplate;
    private final JuheLocationProperties properties;

    /**
     * 执行定位请求
     *
     * @param mobile 手机号
     * @return 地址信息
     */
    private String execute(String mobile) {
        String url = String.format(MOBILE2ADDRESS_URL, mobile, this.properties.getKey());
        boolean debug = log.isDebugEnabled();
        if (debug) {
            log.debug("Juhe location request: {}", url);
        }
        JuheLocationResponse response;
        try {
            response = this.restTemplate.getForObject(url, JuheLocationResponse.class);
        } catch (Exception e) {
            log.warn("Juhe locate failed: {}", e.getMessage());
            return null;
        }
        if (debug) {
            log.debug("Juhe locate response: {}", response);
        }
        if (response == null || response.getStatus() == null || response.getStatus() != 0) {
            log.warn("Juhe locate failed: {}", response);
            return null;
        }
        String province = ObjectUtils.ifNull(response.getResult(), JuheLocationResponse.Location::getProvince);
        String city = ObjectUtils.ifNull(response.getResult(), JuheLocationResponse.Location::getCity);
        String address = Stream.of(province, city).filter(StringUtils::nonEmpty).collect(Collectors.joining());
        return StringUtils.ifEmpty(address, (String) null);
    }

    @Override
    public String getAddress(@NonNull String mobile) {
        String key = String.format("juhe:location:%s:%s", mobile, DateUtils.date2number(new Date()));
        String address = (String) this.redisTemplate.opsForValue().get(key);
        if (address == null) {
            address = RedisUtils.lockAround("juhe:location:lock:" + mobile, () -> {
                String temp = (String) this.redisTemplate.opsForValue().get(key);
                if (temp == null && (temp = this.execute(mobile)) != null) {
                    this.redisTemplate.opsForValue().set(key, temp, Duration.ofDays(1));
                }
                return temp;
            });
        }
        return address;
    }
}
