package com.chelvc.framework.location.support;

import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import com.chelvc.framework.base.model.Region;
import com.chelvc.framework.base.util.StringUtils;
import com.chelvc.framework.location.Address;
import com.chelvc.framework.location.Area;
import com.chelvc.framework.location.LocationHandler;
import com.chelvc.framework.location.config.TencentLocationProperties;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.stereotype.Component;
import org.springframework.web.client.RestTemplate;

/**
 * 腾讯地址定位处理实现
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Component
@ConditionalOnBean(TencentLocationProperties.class)
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class TencentLocationHandler implements LocationHandler {
    /**
     * IP定位接口地址
     */
    private static final String IP2ADDRESS_URL = "https://apis.map.qq.com/ws/location/v1/ip?ip=%s&key=%s";

    /**
     * 经纬度定位接口地址
     */
    private static final String LOCATION2ADDRESS_URL = "https://apis.map.qq.com/ws/geocoder/v1/?location=%s&key=%s";

    private final RestTemplate restTemplate;
    private final TencentLocationProperties properties;

    /**
     * 执行定位请求
     *
     * @param url       定位接口地址
     * @param parameter 定位参数
     * @return 地址信息
     */
    private Address execute(String url, String parameter) {
        String key = Objects.requireNonNull(this.properties.getKey(), "Tencent location key unspecified");
        boolean debug = log.isDebugEnabled();
        url = String.format(url, parameter, key);
        if (debug) {
            log.debug("Tencent location request: {}", url);
        }
        TencentLocationResponse response;
        try {
            response = this.restTemplate.getForObject(url, TencentLocationResponse.class);
        } catch (Exception e) {
            log.warn("Tencent locate failed: {}", e.getMessage());
            return null;
        }
        if (debug) {
            log.debug("Tencent locate response: {}", response);
        }
        if (response == null || response.getStatus() == null || response.getStatus() != 0) {
            log.warn("Tencent locate failed: {}", response);
            return null;
        }
        TencentLocationResponse.Location location = response.getResult().getLocation();
        if (!Region.isRegion(location.getAdcode())) {
            log.warn("Tencent locate adcode invalid: {}", location.getAdcode());
            return null;
        }

        // 构建地址信息
        Region region = Objects.requireNonNull(Region.of(location.getAdcode()));
        Address address = Address.builder().id((long) region.getCode()).build();
        address.setName(
                Stream.of(location.getProvince(), location.getCity(), location.getDistrict())
                        .filter(name -> !StringUtils.isEmpty(name)).collect(Collectors.joining("·"))
        );
        address.setProvince(
                Area.builder().id((long) region.getProvince().getCode()).name(location.getProvince()).build()
        );
        Region city = region.getCity();
        if (city != null) {
            address.setCity(Area.builder().id((long) city.getCode()).name(location.getCity()).build());
        }
        if (region.getType() == Region.Type.DISTRICT) {
            address.setDistrict(Area.builder().id((long) region.getCode()).name(location.getDistrict()).build());
        }
        return address;
    }

    @Override
    public Address ip2address(@NonNull String ip) {
        return this.execute(IP2ADDRESS_URL, ip);
    }

    @Override
    public Address location2address(@NonNull String location) {
        return this.execute(LOCATION2ADDRESS_URL, location);
    }
}
