package com.chelvc.framework.nacos.util;

import java.util.Map;
import java.util.Objects;
import java.util.StringTokenizer;
import java.util.stream.Collectors;

import com.alibaba.nacos.api.config.ConfigService;
import com.alibaba.nacos.api.exception.NacosException;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;
import lombok.NonNull;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.core.env.Environment;
import org.springframework.stereotype.Component;

/**
 * Nacos工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
@Component
public final class NacosUtils implements ApplicationContextAware {
    /**
     * Nacos配置ID
     */
    private static String ID;

    /**
     * Nacos配置组
     */
    private static String GROUP;

    /**
     * 配置中心业务处理对象实例
     */
    private static ConfigService CONFIG_SERVICE;

    private NacosUtils() {
    }

    /**
     * 获取配置中心业务处理对象实例
     *
     * @return 配置中心业务处理对象实例
     */
    public static ConfigService getConfigService() {
        return Objects.requireNonNull(CONFIG_SERVICE, "Nacos config service has not been initialized");
    }

    /**
     * 初始化配置信息
     *
     * @param key   配置键
     * @param value 配置值
     */
    public static void initializeConfiguration(@NonNull String key, @NonNull String value) {
        initializeConfiguration(ImmutableMap.of(key, value));
    }

    /**
     * 初始化配置信息
     *
     * @param configuration 配置信息键/值映射表
     */
    public static void initializeConfiguration(@NonNull Map<String, String> configuration) {
        ConfigService service = getConfigService();
        Map<String, String> mapping = Maps.newHashMap();

        // 加载当前配置信息
        String content;
        try {
            content = service.getConfig(ID, GROUP, 3000);
        } catch (NacosException e) {
            throw new RuntimeException(e);
        }
        if (!StringUtils.isEmpty(content)) {
            StringTokenizer tokenizer = new StringTokenizer(content, "\n");
            while (tokenizer.hasMoreElements()) {
                String[] splits = tokenizer.nextToken().split("=");
                String key = splits[0], value = splits.length > 1 ? splits[1] : null;
                mapping.put(key, value);
            }
        }

        // 更新并保存配置
        mapping.putAll(configuration);
        if (!mapping.isEmpty()) {
            content = mapping.entrySet().stream().map(
                    entry -> entry.getKey() + "=" + ObjectUtils.ifNull(entry.getValue(), StringUtils.EMPTY)
            ).collect(Collectors.joining("\n"));
            try {
                service.publishConfig(ID, GROUP, content);
            } catch (NacosException e) {
                throw new RuntimeException(e);
            }
        }
    }

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        CONFIG_SERVICE = applicationContext.getBean(ConfigService.class);
        Environment environment = applicationContext.getEnvironment();
        if (StringUtils.isEmpty(ID = environment.getProperty("nacos.config.id"))) {
            ID = environment.getProperty("spring.application.name");
        }
        GROUP = environment.getProperty("spring.cloud.nacos.config.group", "DEFAULT_GROUP");
    }
}
