package com.chelvc.framework.redis.config;

import java.time.Duration;
import java.util.Collection;

import com.chelvc.framework.base.context.JacksonContextHolder;
import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.annotation.PropertyAccessor;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.MapperFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.databind.jsontype.impl.LaissezFaireSubTypeValidator;
import io.lettuce.core.event.EventBus;
import io.lettuce.core.resource.ClientResources;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.digest.DigestUtils;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.autoconfigure.AutoConfigureAfter;
import org.springframework.boot.autoconfigure.data.redis.RedisAutoConfiguration;
import org.springframework.cache.CacheManager;
import org.springframework.cache.annotation.CachingConfigurerSupport;
import org.springframework.cache.annotation.EnableCaching;
import org.springframework.cache.interceptor.KeyGenerator;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Primary;
import org.springframework.core.task.TaskExecutor;
import org.springframework.data.redis.cache.RedisCacheConfiguration;
import org.springframework.data.redis.cache.RedisCacheManager;
import org.springframework.data.redis.cache.RedisCacheWriter;
import org.springframework.data.redis.connection.lettuce.LettuceConnectionFactory;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.listener.RedisMessageListenerContainer;
import org.springframework.data.redis.serializer.Jackson2JsonRedisSerializer;
import org.springframework.data.redis.serializer.RedisSerializationContext;
import org.springframework.data.redis.serializer.StringRedisSerializer;
import org.springframework.util.CollectionUtils;

/**
 * Redis缓存配置
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Configuration
@EnableCaching
@AutoConfigureAfter(RedisAutoConfiguration.class)
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class RedisConfigurer extends CachingConfigurerSupport implements ApplicationContextAware {
    private final TaskExecutor executor;
    private final ClientResources clientResources;
    private final LettuceConnectionFactory lettuceConnectionFactory;

    /**
     * Key前缀
     */
    @Value("${spring.profiles.active}:")
    private String prefix;

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        // Redis事件配置
        EventBus bus = this.clientResources.eventBus();
        Collection<RedisEventConfigurerAdapter> adapters =
                applicationContext.getBeansOfType(RedisEventConfigurerAdapter.class).values();
        if (!CollectionUtils.isEmpty(adapters)) {
            log.info("Loading redis event configurer adapters: {}", adapters);
            adapters.forEach(adapter -> adapter.configure(bus));
        }
    }

    @Bean
    public KeyGenerator keyGenerator() {
        return (target, method, params) -> {
            StringBuilder builder = new StringBuilder();
            builder.append(target.getClass().getName());
            builder.append(method.getName());
            builder.append("&");
            for (Object param : params) {
                if (param != null) {
                    builder.append(param.getClass().getName());
                    builder.append("&");
                    builder.append(JacksonContextHolder.serialize(param));
                    builder.append("&");
                }
            }
            return DigestUtils.sha256Hex(builder.toString());
        };
    }

    @Bean
    public CacheManager cacheManager() {
        RedisCacheWriter writer = RedisCacheWriter.nonLockingRedisCacheWriter(this.lettuceConnectionFactory);
        RedisSerializationContext.SerializationPair<?> serialization =
                RedisSerializationContext.SerializationPair.fromSerializer(this.jackson2JsonRedisSerializer());
        RedisCacheConfiguration redisCacheConfiguration = RedisCacheConfiguration.defaultCacheConfig()
                .disableCachingNullValues().entryTtl(Duration.ofMinutes(10)).serializeValuesWith(serialization);
        return RedisCacheManager.builder(writer).cacheDefaults(redisCacheConfiguration).build();
    }

    @Bean
    @Primary
    public RedisMessageListenerContainer redisMessageListenerContainer(RedisProperties properties) {
        RedisProperties.Listener config = properties.getListener();
        RedisMessageListenerContainer container = new RedisMessageListenerContainer();
        container.setTaskExecutor(this.executor);
        container.setConnectionFactory(this.lettuceConnectionFactory);
        container.setRecoveryInterval(config.getRecoveryInterval());
        container.setMaxSubscriptionRegistrationWaitingTime(config.getMaxSubscriptionRegistrationWaitingTime());
        return container;
    }

    @Bean
    public RedisTemplate<String, Object> redisTemplate() {
        RedisTemplate<String, Object> template = new RedisTemplate<>();
        template.setConnectionFactory(this.lettuceConnectionFactory);
        template.setKeySerializer(new PrefixRedisSerializer(this.prefix));
        template.setValueSerializer(this.jackson2JsonRedisSerializer());
        template.setHashKeySerializer(new StringRedisSerializer());
        template.setHashValueSerializer(this.jackson2JsonRedisSerializer());
        template.afterPropertiesSet();
        return template;
    }

    @Bean
    public Jackson2JsonRedisSerializer<Object> jackson2JsonRedisSerializer() {
        // objectMapper for redis serialization
        ObjectMapper mapper = new ObjectMapper();
        mapper.setVisibility(PropertyAccessor.ALL, JsonAutoDetect.Visibility.ANY);
        mapper.setSerializationInclusion(JsonInclude.Include.ALWAYS);
        mapper.activateDefaultTyping(LaissezFaireSubTypeValidator.instance, ObjectMapper.DefaultTyping.NON_FINAL);
        mapper.configure(MapperFeature.USE_ANNOTATIONS, true);
        mapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
        mapper.configure(SerializationFeature.FAIL_ON_EMPTY_BEANS, false);
        // jackson2JsonRedisSerializer
        Jackson2JsonRedisSerializer<Object> serializer = new Jackson2JsonRedisSerializer<>(Object.class);
        serializer.setObjectMapper(mapper);
        return serializer;
    }
}
