package com.chelvc.framework.redis.util;

import java.nio.charset.StandardCharsets;
import java.time.Duration;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.UUID;
import java.util.concurrent.ThreadLocalRandom;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import com.chelvc.framework.base.context.ApplicationContextHolder;
import com.chelvc.framework.base.function.Executor;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.chelvc.framework.base.util.ThreadUtils;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.data.redis.connection.RedisConnection;
import org.springframework.data.redis.connection.RedisConnectionFactory;
import org.springframework.data.redis.connection.RedisStandaloneConfiguration;
import org.springframework.data.redis.connection.RedisStringCommands;
import org.springframework.data.redis.connection.ReturnType;
import org.springframework.data.redis.connection.jedis.JedisConnectionFactory;
import org.springframework.data.redis.core.RedisConnectionUtils;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.core.script.DefaultRedisScript;
import org.springframework.data.redis.core.script.DigestUtils;
import org.springframework.data.redis.core.script.RedisScript;
import org.springframework.data.redis.core.types.Expiration;
import org.springframework.data.redis.serializer.Jackson2JsonRedisSerializer;
import org.springframework.data.redis.serializer.RedisSerializer;
import org.springframework.data.redis.serializer.StringRedisSerializer;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;

/**
 * Redis工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Component
public class RedisUtils implements ApplicationContextAware {
    /**
     * 默认锁超时时间（秒）
     */
    private static final long DEFAULT_LOCK_TIMEOUT = 60;

    /**
     * 随机串Rdids Key前缀
     */
    private static final String RANDOM_REDIS_PREFIX = "random:";

    /**
     * 序列号Rdids Key前缀
     */
    private static final String SEQUENCE_REDIS_PREFIX = "sequence:";

    /**
     * 唯一标识日期格式化处理器
     */
    private static final DateTimeFormatter IDENTITY_DATETIME_FORMATTER = DateTimeFormatter.ofPattern("yyyyMMddHHmmss");

    /**
     * Redis解锁脚本（原子操作）
     */
    private static final String UNLOCK_SCRIPT =
            "if redis.call(\"get\",KEYS[1]) == ARGV[1] then return redis.call(\"del\",KEYS[1]) else return 0 end";

    /**
     * Redis解锁脚本SHA（原子操作）
     */
    private static final String UNLOCK_SCRIPT_SHA = DigestUtils.sha1DigestAsHex(UNLOCK_SCRIPT);

    /**
     * Redis获取脚本（如果值不存在则初始化）
     */
    private static final String GET_WITH_INITIAL_SCRIPT =
            "local value = redis.call('GET', KEYS[1]) if value then return value " +
                    "else redis.call('SET', KEYS[1], ARGV[1]) return ARGV[1] end";

    /**
     * Redis获取脚本（如果值不存在则根据过期时间初始化）
     */
    private static final String GET_WITH_INITIAL_DURATION_SCRIPT =
            "local value = redis.call('GET', KEYS[1]) if value then return value " +
                    "else redis.call('SETEX', KEYS[1], ARGV[2], ARGV[1]) return ARGV[1] end";

    /**
     * Redis自增1脚本（如果值不存在则初始化）
     */
    private static final String INCR_WITH_INITIAL_SCRIPT =
            "if redis.call('EXISTS', KEYS[1]) == 1 then return redis.call('INCR', KEYS[1]) " +
                    "else redis.call('SET', KEYS[1], ARGV[1]) return redis.call('INCR', KEYS[1]) end";

    /**
     * Redis自增指定减量值脚本（如果值不存在则初始化）
     */
    private static final String INCRBY_WITH_INITIAL_SCRIPT =
            "if redis.call('EXISTS', KEYS[1]) == 1 then return redis.call('INCRBY', KEYS[1], ARGV[1]) " +
                    "else redis.call('SET', KEYS[1], ARGV[2]) return redis.call('INCRBY', KEYS[1], ARGV[1]) end";

    /**
     * Redis自增脚本（如果值不存在则根据过期时间初始化）
     */
    private static final String INCR_WITH_INITIAL_DURATION_SCRIPT =
            "if redis.call('EXISTS', KEYS[1]) == 1 then return redis.call('INCR', KEYS[1]) " +
                    "else redis.call('SETEX', KEYS[1], ARGV[2], ARGV[1]) return redis.call('INCR', KEYS[1]) end";

    /**
     * Redis自增指定减量值脚本（如果值不存在则根据过期时间初始化）
     */
    private static final String INCRBY_WITH_INITIAL_DURATION_SCRIPT =
            "if redis.call('EXISTS', KEYS[1]) == 1 then return redis.call('INCRBY', KEYS[1], ARGV[1]) " +
                    "else redis.call('SETEX', KEYS[1], ARGV[3], ARGV[2]) " +
                    "return redis.call('INCRBY', KEYS[1], ARGV[1]) end";

    /**
     * Redis自减1脚本（如果值不存在则初始化）
     */
    private static final String DECR_WITH_INITIAL_SCRIPT =
            "if redis.call('EXISTS', KEYS[1]) == 1 then return redis.call('DECR', KEYS[1]) " +
                    "else redis.call('SET', KEYS[1], ARGV[1]) return redis.call('DECR', KEYS[1]) end";

    /**
     * Redis自减指定减量值脚本（如果值不存在则初始化）
     */
    private static final String DECRBY_WITH_INITIAL_SCRIPT =
            "if redis.call('EXISTS', KEYS[1]) == 1 then return redis.call('DECRBY', KEYS[1], ARGV[1]) " +
                    "else redis.call('SET', KEYS[1], ARGV[2]) return redis.call('DECRBY', KEYS[1], ARGV[1]) end";

    /**
     * Redis自减脚本（如果值不存在则根据过期时间初始化）
     */
    private static final String DECR_WITH_INITIAL_DURATION_SCRIPT =
            "if redis.call('EXISTS', KEYS[1]) == 1 then return redis.call('DECR', KEYS[1]) " +
                    "else redis.call('SETEX', KEYS[1], ARGV[2], ARGV[1]) return redis.call('DECR', KEYS[1]) end";

    /**
     * Redis自减指定减量值脚本（如果值不存在则根据过期时间初始化）
     */
    private static final String DECRBY_WITH_INITIAL_DURATION_SCRIPT =
            "if redis.call('EXISTS', KEYS[1]) == 1 then return redis.call('DECRBY', KEYS[1], ARGV[1]) " +
                    "else redis.call('SETEX', KEYS[1], ARGV[3], ARGV[2]) " +
                    "return redis.call('DECRBY', KEYS[1], ARGV[1]) end";

    /**
     * Redis批量设值脚本（如果Key已存在则忽略）
     */
    private static final String SETNX_WITH_DURATION_SCRIPT =
            "local success = {} for i = 1, #KEYS do if redis.call('SET', KEYS[i], ARGV[i], 'EX', ARGV[#ARGV], 'NX') " +
                    "then table.insert(success, i - 1) end end return success";

    /**
     * Redis批量设值脚本（如果成功则更新过期时间）
     */
    private static final String MSETNX_WITH_DURATION_SCRIPT =
            "local args = {} for i = 1, #KEYS do table.insert(args, KEYS[i]) table.insert(args, ARGV[i]) end " +
                    "if redis.call('MSETNX', unpack(args)) == 1 then for i = 1, #KEYS do " +
                    "redis.call('EXPIRE', KEYS[i], ARGV[#ARGV]) end return 1 else return 0 end";

    /**
     * Redis批量设置字典脚本（如果成功则更新过期时间）
     */
    private static final String HMSET_WITH_DURATION_SCRIPT =
            "local args = {} table.insert(args, KEYS[1]) for i = 2, #KEYS do table.insert(args, KEYS[i]) " +
                    "table.insert(args, ARGV[i]) end local result = redis.call('HMSET', unpack(args)) " +
                    "if result[1] == false then return 0 else return redis.call('EXPIRE', KEYS[1], ARGV[1]) end";

    /**
     * Redis连接工厂
     */
    private static RedisConnectionFactory CONNECTION_FACTORY;

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        CONNECTION_FACTORY = applicationContext.getBean(RedisConnectionFactory.class);
    }

    /**
     * 获取Redis连接工厂对象
     *
     * @return Redis连接工厂对象
     */
    public static RedisConnectionFactory getConnectionFactory() {
        return Objects.requireNonNull(CONNECTION_FACTORY, "Redis connection factory has not been initialized");
    }

    /**
     * 执行Redis操作
     *
     * @param function 操作方法
     * @param <T>      返回数据类型泛型
     * @return 执行结果
     */
    public static <T> T execute(Function<RedisConnection, T> function) {
        return execute(function, e -> {
            log.warn("Redis execute failed: {}", e.getMessage());
            return null;
        });
    }

    /**
     * 执行Redis操作
     *
     * @param function 操作方法
     * @param failure  操作失败回调方法
     * @param <T>      返回数据类型泛型
     * @return 执行结果
     */
    public static <T> T execute(@NonNull Function<RedisConnection, T> function,
                                @NonNull Function<Exception, T> failure) {
        RedisConnection connection = null;
        RedisConnectionFactory factory = getConnectionFactory();
        try {
            connection = factory.getConnection();
            return function.apply(connection);
        } catch (Exception e) {
            return failure.apply(e);
        } finally {
            if (connection != null) {
                RedisConnectionUtils.releaseConnection(connection, factory, false);
            }
        }
    }

    /**
     * 加锁（阻塞），加锁成功则返回锁标识
     *
     * @param name 锁名称
     * @return 锁标识
     */
    public static String lock(String name) {
        return lock(name, Duration.ofSeconds(DEFAULT_LOCK_TIMEOUT));
    }

    /**
     * 加锁（阻塞），加锁成功则返回锁标识
     *
     * @param name     锁名称
     * @param duration 锁有效期
     * @return 锁标识
     */
    public static String lock(@NonNull String name, @NonNull Duration duration) {
        String token;
        long expiration = System.currentTimeMillis() + duration.toMillis();
        do {
            if (StringUtils.isEmpty(token = tryLock(name, duration))) {
                ThreadUtils.sleep(100);
            }
        } while (StringUtils.isEmpty(token) && System.currentTimeMillis() <= expiration && !Thread.interrupted());
        return token;
    }

    /**
     * 判断锁标识是否已加锁
     *
     * @param name 锁名称
     * @return true/false
     */
    public static boolean locked(@NonNull String name) {
        byte[] key = name.getBytes(StandardCharsets.UTF_8);
        return execute(connection -> {
            byte[] value = connection.get(key);
            return value != null && StringUtils.nonEmpty(new String(value, StandardCharsets.UTF_8));
        }, e -> false);
    }

    /**
     * 尝试加锁（无阻塞），如果加锁成功则返回锁标识，否则返回null
     *
     * @param name 锁名称
     * @return 锁标识
     */
    public static String tryLock(String name) {
        return tryLock(name, Duration.ofSeconds(DEFAULT_LOCK_TIMEOUT));
    }

    /**
     * 尝试加锁（无阻塞），如果加锁成功则返回锁标识，否则返回null
     *
     * @param name     锁名称
     * @param duration 锁有效期
     * @return 锁标识
     */
    public static String tryLock(@NonNull String name, @NonNull Duration duration) {
        String token = UUID.randomUUID().toString();
        byte[] key = name.getBytes(StandardCharsets.UTF_8);
        byte[] value = token.getBytes(StandardCharsets.UTF_8);
        return execute(connection -> {
            Boolean success = connection.set(key, value, Expiration.from(duration),
                    RedisStringCommands.SetOption.SET_IF_ABSENT);
            return Boolean.TRUE.equals(success) ? token : null;
        }, e -> {
            log.warn("Redis locking failed: {}", e.getMessage());
            return null;
        });
    }

    /**
     * 解锁
     *
     * @param name  锁名称
     * @param token 锁标识
     * @return true/false
     */
    public static boolean unlock(@NonNull String name, @NonNull String token) {
        byte[][] args = new byte[][]{name.getBytes(StandardCharsets.UTF_8), token.getBytes(StandardCharsets.UTF_8)};
        return execute(connection -> {
            Long result;
            try {
                result = connection.scriptingCommands().evalSha(UNLOCK_SCRIPT_SHA, ReturnType.INTEGER, 1, args);
            } catch (Exception e) {
                result = connection.scriptingCommands().eval(UNLOCK_SCRIPT.getBytes(), ReturnType.INTEGER, 1, args);
            }
            return result != null && result > 0;
        }, e -> {
            log.warn("Redis unlock failed: {}", e.getMessage());
            return false;
        });
    }

    /**
     * 加/解锁环绕执行方法
     *
     * @param name     锁名称
     * @param executor 回调方法
     */
    public static void lockAround(String name, @NonNull Executor executor) {
        lockAround(name, Duration.ofSeconds(DEFAULT_LOCK_TIMEOUT), executor);
    }

    /**
     * 加/解锁环绕执行方法
     *
     * @param name     锁名称
     * @param duration 有效时间（毫秒）
     * @param executor 回调方法
     */
    public static void lockAround(String name, Duration duration, @NonNull Executor executor) {
        String token = lock(name, duration);
        if (!StringUtils.isEmpty(token)) {
            try {
                executor.execute();
            } finally {
                unlock(name, token);
            }
        }
    }

    /**
     * 加/解锁环绕执行方法
     *
     * @param name     锁名称
     * @param executor 回调方法
     * @param <T>      返回数据类型泛型
     * @return 执行结果
     */
    public static <T> T lockAround(String name, Supplier<T> executor) {
        return lockAround(name, Duration.ofSeconds(DEFAULT_LOCK_TIMEOUT), executor);
    }

    /**
     * 加/解锁环绕执行方法
     *
     * @param name     锁名称
     * @param duration 有效时间
     * @param executor 回调方法
     * @param <T>      返回数据类型泛型
     * @return 执行结果
     */
    public static <T> T lockAround(String name, Duration duration, Supplier<T> executor) {
        return lockAround(name, duration, executor, () -> null);
    }

    /**
     * 加/解锁环绕执行方法
     *
     * @param name     锁名称
     * @param executor 回调方法
     * @param failure  枷锁失败回调方法
     * @param <T>      返回数据类型泛型
     * @return 执行结果
     */
    public static <T> T lockAround(String name, @NonNull Supplier<T> executor, @NonNull Supplier<T> failure) {
        return lockAround(name, Duration.ofSeconds(DEFAULT_LOCK_TIMEOUT), executor, failure);
    }

    /**
     * 加/解锁环绕执行方法
     *
     * @param name     锁名称
     * @param duration 有效时间
     * @param executor 回调方法
     * @param failure  枷锁失败回调方法
     * @param <T>      返回数据类型泛型
     * @return 执行结果
     */
    public static <T> T lockAround(String name, Duration duration, @NonNull Supplier<T> executor,
                                   @NonNull Supplier<T> failure) {
        String token = lock(name, duration);
        if (StringUtils.isEmpty(token)) {
            return failure.get();
        }
        try {
            return executor.get();
        } finally {
            unlock(name, token);
        }
    }

    /**
     * 加/解锁环绕执行方法
     *
     * @param name     锁名称
     * @param executor 回调方法
     */
    public static void tryLockAround(String name, @NonNull Executor executor) {
        tryLockAround(name, Duration.ofSeconds(DEFAULT_LOCK_TIMEOUT), executor);
    }

    /**
     * 加/解锁环绕执行方法
     *
     * @param name     锁名称
     * @param duration 有效时间
     * @param executor 回调方法
     */
    public static void tryLockAround(String name, Duration duration, @NonNull Executor executor) {
        String token = tryLock(name, duration);
        if (!StringUtils.isEmpty(token)) {
            try {
                executor.execute();
            } finally {
                unlock(name, token);
            }
        }
    }

    /**
     * 加/解锁环绕执行方法
     *
     * @param name     锁名称
     * @param supplier 回调方法
     * @param <T>      返回数据类型泛型
     * @return 执行结果
     */
    public static <T> T tryLockAround(String name, Supplier<T> supplier) {
        return tryLockAround(name, Duration.ofSeconds(DEFAULT_LOCK_TIMEOUT), supplier);
    }

    /**
     * 加/解锁环绕执行方法
     *
     * @param name     锁名称
     * @param duration 有效时间
     * @param supplier 回调方法
     * @param <T>      返回数据类型泛型
     * @return 执行结果
     */
    public static <T> T tryLockAround(String name, Duration duration, Supplier<T> supplier) {
        return tryLockAround(name, duration, supplier, () -> null);
    }

    /**
     * 加/解锁环绕执行方法
     *
     * @param name     锁名称
     * @param supplier 回调方法
     * @param failure  枷锁失败回调方法
     * @param <T>      返回数据类型泛型
     * @return 执行结果
     */
    public static <T> T tryLockAround(String name, @NonNull Supplier<T> supplier, @NonNull Supplier<T> failure) {
        return tryLockAround(name, Duration.ofSeconds(DEFAULT_LOCK_TIMEOUT), supplier, failure);
    }

    /**
     * 加/解锁环绕执行方法
     *
     * @param name     锁名称
     * @param duration 有效时间
     * @param supplier 回调方法
     * @param failure  枷锁失败回调方法
     * @param <T>      返回数据类型泛型
     * @return 执行结果
     */
    public static <T> T tryLockAround(String name, Duration duration, @NonNull Supplier<T> supplier,
                                      @NonNull Supplier<T> failure) {
        String token = tryLock(name, duration);
        if (StringUtils.isEmpty(token)) {
            return failure.get();
        }
        try {
            return supplier.get();
        } finally {
            unlock(name, token);
        }
    }

    /**
     * 生成唯一标识，默认1分钟有效期
     *
     * @return 唯一标识
     */
    public static long generateIdentity() {
        return generateIdentity(Duration.ofMinutes(1));
    }

    /**
     * 生成唯一标识
     * <p>
     * 唯一标识格式：yyyyMMddHHmmss + 五位随机数，例如：202108251933453756
     *
     * @param duration 唯一标识有效期
     * @return 唯一标识
     */
    public static long generateIdentity(@NonNull Duration duration) {
        ThreadLocalRandom random = ThreadLocalRandom.current();
        for (int i = 0; i < 10; i++) {
            String time = IDENTITY_DATETIME_FORMATTER.format(LocalDateTime.now());
            String value = StringUtils.rjust(String.valueOf(random.nextInt(100000)), 5, '0');
            long identity = Long.parseLong(time + value);
            byte[] key = (SEQUENCE_REDIS_PREFIX + identity).getBytes(StandardCharsets.UTF_8);
            Boolean success = execute(connection -> connection.set(key, ObjectUtils.EMPTY_BYTES,
                    Expiration.from(duration), RedisStringCommands.SetOption.SET_IF_ABSENT));
            if (Boolean.TRUE.equals(success)) {
                return identity;
            }
        }
        throw new RuntimeException("Generate identity attempts limit exceeded");
    }

    /**
     * 从队列中弹出数据
     *
     * @param template Redis操作模版实例
     * @param queue    队列名称
     * @param supplier 数据提供者
     * @param <T>      数据类型
     * @return 数据对象实例
     */
    @SuppressWarnings("unchecked")
    public static <T> T pop(@NonNull RedisTemplate<String, Object> template, @NonNull String queue,
                            @NonNull Supplier<List<T>> supplier) {
        // 尝试从队列中获取数据
        T object = (T) template.opsForList().rightPop(queue);
        if (object == null) {
            // 获取redis锁
            object = lockAround(queue + ":lock", () -> {
                T t = (T) template.opsForList().rightPop(queue);
                if (t == null) {
                    // 加载队列数据
                    List<T> objects = supplier.get();
                    if (CollectionUtils.isEmpty(objects)) {
                        return null;
                    }
                    template.opsForList().leftPushAll(queue, (Collection<Object>) objects);
                    t = (T) template.opsForList().rightPop(queue);
                }
                return t;
            });
        }
        return object;
    }

    /**
     * 批量获取Redis值
     *
     * @param keys 键名称列表
     * @return 键/值映射表
     */
    public static Map<String, String> get(@NonNull List<String> keys) {
        return get(keys, key -> key, value -> new String(value, StandardCharsets.UTF_8));
    }

    /**
     * 批量获取Redis值
     *
     * @param keys         键名称列表
     * @param keyAdapter   键适配器
     * @param valueAdapter 值适配器
     * @param <K>          键类型
     * @param <V>          值类型
     * @return 键/值映射表
     */
    public static <K, V> Map<K, V> get(List<String> keys, @NonNull Function<String, K> keyAdapter,
                                       @NonNull Function<byte[], V> valueAdapter) {
        if (CollectionUtils.isEmpty(keys)) {
            return Collections.emptyMap();
        }
        return execute(connection -> {
            List<byte[]> values = connection.mGet(
                    keys.stream().map(name -> name.getBytes(StandardCharsets.UTF_8)).toArray(byte[][]::new)
            );
            if (CollectionUtils.isEmpty(values)) {
                return Collections.emptyMap();
            }
            Map<K, V> mapping = Maps.newHashMapWithExpectedSize(keys.size());
            for (int i = 0; i < keys.size(); i++) {
                String key = keys.get(i);
                byte[] bytes = values.get(i);
                if (StringUtils.isEmpty(key) || bytes == null || bytes.length == 0) {
                    continue;
                }
                V value = valueAdapter.apply(bytes);
                if (StringUtils.nonEmpty(value)) {
                    mapping.put(keyAdapter.apply(key), value);
                }
            }
            return mapping;
        }, e -> Collections.emptyMap());
    }

    /**
     * 获取键值，如果键不存在则初始化
     *
     * @param template     Redis操作模版实例
     * @param key          键名称
     * @param initialValue 初始化值
     * @param <K>          键类型
     * @param <V>          值类型
     * @return 值
     */
    @SuppressWarnings("unchecked")
    public static <K, V> V get(@NonNull RedisTemplate<K, ?> template, @NonNull K key, @NonNull V initialValue) {
        RedisScript<V> script = new DefaultRedisScript<>(GET_WITH_INITIAL_SCRIPT, (Class<V>) initialValue.getClass());
        return template.execute(script, Collections.singletonList(key), initialValue);
    }

    /**
     * 获取键值，如果键不存在则初始化
     *
     * @param template     Redis操作模版实例
     * @param key          键名称
     * @param initialValue 初始化值
     * @param duration     有效时间
     * @param <K>          键类型
     * @param <V>          值类型
     * @return 值
     */
    @SuppressWarnings("unchecked")
    public static <K, V> V get(@NonNull RedisTemplate<K, ?> template, @NonNull K key, @NonNull V initialValue,
                               @NonNull Duration duration) {
        RedisScript<V> script =
                new DefaultRedisScript<>(GET_WITH_INITIAL_DURATION_SCRIPT, (Class<V>) initialValue.getClass());
        return template.execute(script, Collections.singletonList(key), initialValue, duration.getSeconds());
    }

    /**
     * 批量设置带过期时间的值，如果Key已存在则忽略
     *
     * @param template Redis操作模版实例
     * @param mapping  键/值映射表
     * @param duration 有效时间
     * @return 设置成功Key列表
     */
    @SuppressWarnings({"rawtypes", "unchecked"})
    public static List<String> setnx(@NonNull RedisTemplate<String, ?> template, @NonNull Map<String, ?> mapping,
                                     @NonNull Duration duration) {
        if (CollectionUtils.isEmpty(mapping)) {
            return Collections.emptyList();
        }
        int i = 0;
        List<String> keys = Lists.newArrayListWithCapacity(mapping.size());
        Object[] values = new Object[mapping.size() + 1];
        for (Map.Entry<String, ?> entry : mapping.entrySet()) {
            keys.add(entry.getKey());
            values[i++] = entry.getValue();
        }
        values[i] = duration.getSeconds();
        RedisScript<List> script = new DefaultRedisScript<>(SETNX_WITH_DURATION_SCRIPT, List.class);
        List<Number> success = template.execute(script, keys, values);
        if (CollectionUtils.isEmpty(success)) {
            return Collections.emptyList();
        }
        return success.stream().map(Number::intValue).map(keys::get).collect(Collectors.toList());
    }

    /**
     * 批量设置带过期时间的值，只有在全部设置成功的情况才更新过期时间
     *
     * @param template Redis操作模版实例
     * @param mapping  键/值映射表
     * @param duration 有效时间
     * @param <K>      键类型
     * @return true/false
     */
    public static <K> boolean msetnx(@NonNull RedisTemplate<K, ?> template, @NonNull Map<K, ?> mapping,
                                     @NonNull Duration duration) {
        if (CollectionUtils.isEmpty(mapping)) {
            return false;
        }
        int i = 0;
        List<K> keys = Lists.newArrayListWithCapacity(mapping.size());
        Object[] values = new Object[mapping.size() + 1];
        for (Map.Entry<K, ?> entry : mapping.entrySet()) {
            keys.add(entry.getKey());
            values[i++] = entry.getValue();
        }
        values[i] = duration.getSeconds();
        RedisScript<Boolean> script = new DefaultRedisScript<>(MSETNX_WITH_DURATION_SCRIPT, Boolean.class);
        return Boolean.TRUE.equals(template.execute(script, keys, values));
    }

    /**
     * 批量设置带过期时间的字典值
     *
     * @param template Redis操作模版实例
     * @param key      字典标识
     * @param mapping  键/值映射表
     * @param duration 有效时间
     * @param <K>      键类型
     * @return true/false
     */
    public static <K> boolean hmset(@NonNull RedisTemplate<K, ?> template, @NonNull K key,
                                    @NonNull Map<K, ?> mapping, @NonNull Duration duration) {
        if (CollectionUtils.isEmpty(mapping)) {
            return false;
        }
        int i = 0;
        List<K> keys = Lists.newArrayListWithCapacity(mapping.size() + 1);
        keys.add(key);
        Object[] values = new Object[mapping.size() + 1];
        values[i++] = duration.getSeconds();
        for (Map.Entry<K, ?> entry : mapping.entrySet()) {
            keys.add(entry.getKey());
            values[i++] = entry.getValue();
        }
        RedisScript<Boolean> script = new DefaultRedisScript<>(HMSET_WITH_DURATION_SCRIPT, Boolean.class);
        return Boolean.TRUE.equals(template.execute(script, keys, values));
    }

    /**
     * 数字自增1，如果键不存在则对值初始化
     *
     * @param template     Redis操作模版实例
     * @param key          键名称
     * @param initialValue 初始化值
     * @param <K>          键类型
     * @return 自增后的值
     */
    public static <K> Long increment(@NonNull RedisTemplate<K, ?> template, @NonNull K key, long initialValue) {
        RedisScript<Long> script = new DefaultRedisScript<>(INCR_WITH_INITIAL_SCRIPT, Long.class);
        return template.execute(script, Collections.singletonList(key), initialValue);
    }

    /**
     * 数字自增指定减量值，如果键不存在则对值初始化
     *
     * @param template     Redis操作模版实例
     * @param key          键名称
     * @param delta        减量值
     * @param initialValue 初始化值
     * @param <K>          键类型
     * @return 自增后的值
     */
    public static <K> Long increment(@NonNull RedisTemplate<K, ?> template, @NonNull K key, long delta,
                                     long initialValue) {
        RedisScript<Long> script = new DefaultRedisScript<>(INCRBY_WITH_INITIAL_SCRIPT, Long.class);
        return template.execute(script, Collections.singletonList(key), delta, initialValue);
    }

    /**
     * 数字自增1，如果键不存在则根据有效时间对值初始化
     *
     * @param template     Redis操作模版实例
     * @param key          键名称
     * @param initialValue 初始化值
     * @param duration     有效时间
     * @param <K>          键类型
     * @return 自增后的值
     */
    public static <K> Long increment(@NonNull RedisTemplate<K, ?> template, @NonNull K key, long initialValue,
                                     @NonNull Duration duration) {
        RedisScript<Long> script = new DefaultRedisScript<>(INCR_WITH_INITIAL_DURATION_SCRIPT, Long.class);
        return template.execute(script, Collections.singletonList(key), initialValue, duration.getSeconds());
    }

    /**
     * 数字自增指定减量值，如果键不存在则根据有效时间对值初始化
     *
     * @param template     Redis操作模版实例
     * @param key          键名称
     * @param delta        减量值
     * @param initialValue 初始化值
     * @param duration     有效时间
     * @param <K>          键类型
     * @return 自增后的值
     */
    public static <K> Long increment(@NonNull RedisTemplate<K, ?> template, @NonNull K key, long delta,
                                     long initialValue, @NonNull Duration duration) {
        RedisScript<Long> script = new DefaultRedisScript<>(INCRBY_WITH_INITIAL_DURATION_SCRIPT, Long.class);
        return template.execute(script, Collections.singletonList(key), delta, initialValue, duration.getSeconds());
    }

    /**
     * 数字自减1，如果键不存在则对值初始化
     *
     * @param template     Redis操作模版实例
     * @param key          键名称
     * @param initialValue 初始化值
     * @param <K>          键类型
     * @return 自减后的值
     */
    public static <K> Long decrement(@NonNull RedisTemplate<K, ?> template, @NonNull K key, long initialValue) {
        RedisScript<Long> script = new DefaultRedisScript<>(DECR_WITH_INITIAL_SCRIPT, Long.class);
        return template.execute(script, Collections.singletonList(key), initialValue);
    }

    /**
     * 数字自减指定减量值，如果键不存在则对值初始化
     *
     * @param template     Redis操作模版实例
     * @param key          键名称
     * @param delta        减量值
     * @param initialValue 初始化值
     * @param <K>          键类型
     * @return 自减后的值
     */
    public static <K> Long decrement(@NonNull RedisTemplate<K, ?> template, @NonNull K key, long delta,
                                     long initialValue) {
        RedisScript<Long> script = new DefaultRedisScript<>(DECRBY_WITH_INITIAL_SCRIPT, Long.class);
        return template.execute(script, Collections.singletonList(key), delta, initialValue);
    }

    /**
     * 数字自减1，如果键不存在则根据有效时间对值初始化
     *
     * @param template     Redis操作模版实例
     * @param key          键名称
     * @param initialValue 初始化值
     * @param duration     有效时间
     * @param <K>          键类型
     * @return 自减后的值
     */
    public static <K> Long decrement(@NonNull RedisTemplate<K, ?> template, @NonNull K key, long initialValue,
                                     @NonNull Duration duration) {
        RedisScript<Long> script = new DefaultRedisScript<>(DECR_WITH_INITIAL_DURATION_SCRIPT, Long.class);
        return template.execute(script, Collections.singletonList(key), initialValue, duration.getSeconds());
    }

    /**
     * 数字自减指定减量值，如果键不存在则根据有效时间对值初始化
     *
     * @param template     Redis操作模版实例
     * @param key          键名称
     * @param delta        减量值
     * @param initialValue 初始化值
     * @param duration     有效时间
     * @param <K>          键类型
     * @return 自减后的值
     */
    public static <K> Long decrement(@NonNull RedisTemplate<K, ?> template, @NonNull K key, long delta,
                                     long initialValue, @NonNull Duration duration) {
        RedisScript<Long> script = new DefaultRedisScript<>(DECRBY_WITH_INITIAL_DURATION_SCRIPT, Long.class);
        return template.execute(script, Collections.singletonList(key), delta, initialValue, duration.getSeconds());
    }

    /**
     * 生成随机字符串
     *
     * @return 随机字符串
     */
    public static String random() {
        return random(StringUtils::random, Duration.ofSeconds(60));
    }

    /**
     * 生成随机字符串
     *
     * @param length 随机字符串长度
     * @return 随机字符串
     */
    public static String random(int length) {
        return random(() -> StringUtils.random(length), Duration.ofSeconds(60));
    }

    /**
     * 生成随机字符串
     *
     * @param chars 随机串字符数组
     * @return 随机字符串
     */
    public static String random(@NonNull char[] chars) {
        return random(chars, 4);
    }

    /**
     * 生成随机字符串
     *
     * @param duration 随机串保存时间
     * @return 随机字符串
     */
    public static String random(@NonNull Duration duration) {
        return random(StringUtils::random, duration);
    }

    /**
     * 生成随机字符串
     *
     * @param chars  随机串字符数组
     * @param length 随机字符串长度
     * @return 随机字符串
     */
    public static String random(@NonNull char[] chars, int length) {
        return random(chars, length, Duration.ofSeconds(60));
    }

    /**
     * 生成随机字符串
     *
     * @param chars    随机串字符数组
     * @param length   随机字符串长度
     * @param duration 随机串保存时间
     * @return 随机字符串
     */
    public static String random(@NonNull char[] chars, int length, @NonNull Duration duration) {
        return random(() -> StringUtils.random(chars, length), duration);
    }

    /**
     * 生成随机字符串
     *
     * @param supplier 随机字符串生成器
     * @param duration 随机串保存时间
     * @return 随机字符串
     */
    public static String random(@NonNull Supplier<String> supplier, @NonNull Duration duration) {
        for (int i = 0; i < 10; i++) {
            String value = supplier.get();
            byte[] key = (RANDOM_REDIS_PREFIX + value).getBytes(StandardCharsets.UTF_8);
            Boolean success = execute(connection -> connection.set(key, ObjectUtils.EMPTY_BYTES,
                    Expiration.from(duration), RedisStringCommands.SetOption.SET_IF_ABSENT));
            if (Boolean.TRUE.equals(success)) {
                return value;
            }
        }
        throw new RuntimeException("Generate random attempts limit exceeded");
    }

    /**
     * 构建RedisTemplate实例
     *
     * @param configuration 配置信息
     * @return RedisTemplate实例
     */
    public static RedisTemplate<String, Object> template(@NonNull RedisStandaloneConfiguration configuration) {
        RedisSerializer<?> keySerializer = new StringRedisSerializer();
        RedisSerializer<?> valueSerializer = ApplicationContextHolder.getBean(Jackson2JsonRedisSerializer.class);
        RedisTemplate<String, Object> redisTemplate = new RedisTemplate<>();
        redisTemplate.setConnectionFactory(new JedisConnectionFactory(configuration));
        redisTemplate.setKeySerializer(keySerializer);
        redisTemplate.setValueSerializer(valueSerializer);
        redisTemplate.setHashKeySerializer(keySerializer);
        redisTemplate.setHashValueSerializer(valueSerializer);
        redisTemplate.afterPropertiesSet();
        return redisTemplate;
    }

    /**
     * 获取当前时间距过期时间的时长
     *
     * @param expiration 过期时间
     * @return 持续时长
     */
    public static Duration duration(@NonNull Date expiration) {
        return Duration.ofMillis(Math.max(expiration.getTime() - System.currentTimeMillis(), 1000));
    }
}
