package com.chelvc.framework.rocketmq.config;


import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.nio.charset.StandardCharsets;
import java.util.Objects;
import java.util.stream.Collectors;

import com.chelvc.framework.base.context.ApplicationContextHolder;
import com.chelvc.framework.base.context.JacksonContextHolder;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.rocketmq.context.RocketMQContextHolder;
import com.chelvc.framework.rocketmq.interceptor.SessionRocketMQListener;
import com.chelvc.framework.rocketmq.model.ConsumeStatus;
import com.chelvc.framework.rocketmq.model.MessageContext;
import lombok.extern.slf4j.Slf4j;
import org.apache.rocketmq.client.AccessChannel;
import org.apache.rocketmq.client.consumer.DefaultMQPushConsumer;
import org.apache.rocketmq.client.consumer.listener.ConsumeConcurrentlyStatus;
import org.apache.rocketmq.client.consumer.listener.ConsumeOrderlyStatus;
import org.apache.rocketmq.client.consumer.listener.MessageListenerConcurrently;
import org.apache.rocketmq.client.consumer.listener.MessageListenerOrderly;
import org.apache.rocketmq.common.message.MessageExt;
import org.apache.rocketmq.spring.annotation.ConsumeMode;
import org.apache.rocketmq.spring.annotation.MessageModel;
import org.apache.rocketmq.spring.annotation.RocketMQMessageListener;
import org.apache.rocketmq.spring.annotation.SelectorType;
import org.apache.rocketmq.spring.core.RocketMQListener;
import org.apache.rocketmq.spring.core.RocketMQReplyListener;
import org.apache.rocketmq.spring.support.DefaultRocketMQListenerContainer;
import org.springframework.aop.framework.AopProxyUtils;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.core.MethodParameter;
import org.springframework.core.env.Environment;
import org.springframework.messaging.Message;
import org.springframework.messaging.converter.MessageConverter;
import org.springframework.messaging.converter.SmartMessageConverter;
import org.springframework.messaging.support.MessageBuilder;

/**
 * 会话消息监听器容器
 *
 * @author Woody
 * @date 2023/5/18
 */
@Slf4j
public class SessionRocketMQListenerContainer extends DefaultRocketMQListenerContainer {
    private final DefaultRocketMQListenerContainer container;
    private Type type;
    private MethodParameter parameter;

    public SessionRocketMQListenerContainer(DefaultRocketMQListenerContainer container) {
        this.container = container;
    }

    /**
     * 容器初始化
     */
    @SuppressWarnings({"unchecked", "rawtypes"})
    private void initialize() {
        // 获取目标消息监听器对象类型
        Class<?> target;
        if (this.getRocketMQListener() != null) {
            target = AopProxyUtils.ultimateTargetClass(this.getRocketMQListener());
        } else {
            target = AopProxyUtils.ultimateTargetClass(this.getRocketMQReplyListener());
        }

        // 初始化消息消费者
        try {
            Method method = DefaultRocketMQListenerContainer.class.getDeclaredMethod("initRocketMQPushConsumer");
            method.setAccessible(true);
            method.invoke(this.container);
        } catch (NoSuchMethodException | IllegalAccessException | InvocationTargetException e) {
            throw new RuntimeException(e);
        }

        // 初始化消息类型
        try {
            Field field = DefaultRocketMQListenerContainer.class.getDeclaredField("messageType");
            field.setAccessible(true);
            field.set(this.container, this.type = RocketMQContextHolder.lookupConsumerMessageType(target));
        } catch (NoSuchFieldException | IllegalAccessException e) {
            throw new RuntimeException(e);
        }

        // 初始化消息处理方法
        Class<?> clazz;
        if (this.type instanceof ParameterizedType && this.getMessageConverter() instanceof SmartMessageConverter) {
            clazz = (Class<?>) ((ParameterizedType) this.type).getRawType();
        } else if (this.type instanceof Class) {
            clazz = (Class<?>) this.type;
        } else {
            throw new RuntimeException("parameterType:" + this.type + " of onMessage method is not supported");
        }
        try {
            Field field = DefaultRocketMQListenerContainer.class.getDeclaredField("methodParameter");
            field.setAccessible(true);
            field.set(this.container, this.parameter = new MethodParameter(target.getMethod("onMessage", clazz), 0));
        } catch (NoSuchFieldException | IllegalAccessException | NoSuchMethodException e) {
            throw new RuntimeException(e);
        }

        // 初始化会话消息监听器
        if (this.getRocketMQListener() instanceof SessionRocketMQListener) {
            // 获取MQ消费者组及主体
            Environment environment = ApplicationContextHolder.getEnvironment();
            RocketMQMessageListener annotation = target.getAnnotation(RocketMQMessageListener.class);
            String topic = ObjectUtils.ifNull(
                    ObjectUtils.ifNull(annotation, RocketMQMessageListener::topic), environment::resolvePlaceholders
            );
            String group = ObjectUtils.ifNull(
                    ObjectUtils.ifNull(annotation, RocketMQMessageListener::consumerGroup),
                    environment::resolvePlaceholders
            );

            // 根据消费模式初始化消息监听器
            DefaultMQPushConsumer consumer = this.getConsumer();
            SessionRocketMQListener listener = (SessionRocketMQListener) this.getRocketMQListener();
            ConsumeMode mode = ObjectUtils.ifNull(annotation, RocketMQMessageListener::consumeMode);
            if (mode == ConsumeMode.ORDERLY) {
                consumer.setMessageListener((MessageListenerOrderly) (messages, context) -> {
                    ConsumeStatus status;
                    try {
                        status = listener.onMessageContexts(
                                messages.stream().map(this::convert).collect(Collectors.toList())
                        );
                    } catch (Exception e) {
                        String ids = messages.stream().map(MessageExt::getMsgId).collect(Collectors.joining(","));
                        log.error("Orderly message consume failed [{}][{}]: {}", group, topic, ids, e);
                        return ConsumeOrderlyStatus.SUSPEND_CURRENT_QUEUE_A_MOMENT;
                    } finally {
                        consumer.setConsumeMessageBatchMaxSize(listener.getBatchConsumeSize());
                        context.setSuspendCurrentQueueTimeMillis(listener.getOrderlyRetryInterval());
                    }
                    return status == null ? ConsumeOrderlyStatus.SUCCESS : status.orderly();
                });
            } else {
                consumer.setMessageListener((MessageListenerConcurrently) (messages, context) -> {
                    ConsumeStatus status;
                    try {
                        status = listener.onMessageContexts(
                                messages.stream().map(this::convert).collect(Collectors.toList())
                        );
                    } catch (Exception e) {
                        String ids = messages.stream().map(MessageExt::getMsgId).collect(Collectors.joining(","));
                        log.error("Concurrently message consume failed [{}][{}]: {}", group, topic, ids, e);
                        return ConsumeConcurrentlyStatus.RECONSUME_LATER;
                    } finally {
                        consumer.setConsumeMessageBatchMaxSize(listener.getBatchConsumeSize());
                        context.setDelayLevelWhenNextConsume(listener.getConcurrentlyRetryInterval());
                    }
                    return status == null ? ConsumeConcurrentlyStatus.CONSUME_SUCCESS : status.concurrently();
                });
            }
        }
    }

    /**
     * 消息转换
     *
     * @param message 消息
     * @return 消息
     */
    private Object convert(MessageExt message) {
        if (Objects.equals(this.type, MessageExt.class)) {
            return message;
        }
        MessageConverter converter = this.getMessageConverter();
        Message<?> body = MessageBuilder.withPayload(new String(message.getBody(), StandardCharsets.UTF_8)).build();
        MessageContext<?> context =
                (MessageContext<?>) Objects.requireNonNull(converter.fromMessage(body, MessageContext.class));
        Object payload = context.getTarget();
        if (payload == null || (payload instanceof String && this.type == String.class)) {
            return context;
        }
        if (!(payload instanceof String)) {
            payload = JacksonContextHolder.serialize(payload);
        }
        Message<?> target = MessageBuilder.withPayload(payload).build();
        if (this.type instanceof Class) {
            payload = converter.fromMessage(target, (Class<?>) this.type);
        } else {
            Class<?> clazz = (Class<?>) ((ParameterizedType) this.type).getRawType();
            payload = ((SmartMessageConverter) converter).fromMessage(target, clazz, this.parameter);
        }
        return MessageContext.builder().target(payload).session(context.getSession()).build();
    }

    @Override
    public long getSuspendCurrentQueueTimeMillis() {
        return this.container.getSuspendCurrentQueueTimeMillis();
    }

    @Override
    public void setSuspendCurrentQueueTimeMillis(long suspendCurrentQueueTimeMillis) {
        this.container.setSuspendCurrentQueueTimeMillis(suspendCurrentQueueTimeMillis);
    }

    @Override
    public int getDelayLevelWhenNextConsume() {
        return this.container.getDelayLevelWhenNextConsume();
    }

    @Override
    public void setDelayLevelWhenNextConsume(int delayLevelWhenNextConsume) {
        this.container.setDelayLevelWhenNextConsume(delayLevelWhenNextConsume);
    }

    @Override
    public String getNameServer() {
        return this.container.getNameServer();
    }

    @Override
    public void setNameServer(String nameServer) {
        this.container.setNameServer(nameServer);
    }

    @Override
    public AccessChannel getAccessChannel() {
        return this.container.getAccessChannel();
    }

    @Override
    public void setAccessChannel(AccessChannel accessChannel) {
        this.container.setAccessChannel(accessChannel);
    }

    @Override
    public String getConsumerGroup() {
        return this.container.getConsumerGroup();
    }

    @Override
    public void setConsumerGroup(String consumerGroup) {
        this.container.setConsumerGroup(consumerGroup);
    }

    @Override
    public String getTopic() {
        return this.container.getTopic();
    }

    @Override
    public void setTopic(String topic) {
        this.container.setTopic(topic);
    }

    @Override
    public int getConsumeThreadMax() {
        return this.container.getConsumeThreadMax();
    }

    @Override
    public String getCharset() {
        return this.container.getCharset();
    }

    @Override
    public void setCharset(String charset) {
        this.container.setCharset(charset);
    }

    @Override
    public MessageConverter getMessageConverter() {
        return this.container.getMessageConverter();
    }

    @Override
    public DefaultRocketMQListenerContainer setMessageConverter(MessageConverter messageConverter) {
        this.container.setMessageConverter(messageConverter);
        return this;
    }

    @Override
    public RocketMQListener<?> getRocketMQListener() {
        return this.container.getRocketMQListener();
    }

    @Override
    public void setRocketMQListener(RocketMQListener rocketMQListener) {
        this.container.setRocketMQListener(rocketMQListener);
    }

    @Override
    public RocketMQReplyListener<?, ?> getRocketMQReplyListener() {
        return this.container.getRocketMQReplyListener();
    }

    @Override
    public void setRocketMQReplyListener(RocketMQReplyListener rocketMQReplyListener) {
        this.container.setRocketMQReplyListener(rocketMQReplyListener);
    }

    @Override
    public RocketMQMessageListener getRocketMQMessageListener() {
        return this.container.getRocketMQMessageListener();
    }

    @Override
    public void setRocketMQMessageListener(RocketMQMessageListener anno) {
        this.container.setRocketMQMessageListener(anno);
    }

    @Override
    public ConsumeMode getConsumeMode() {
        return this.container.getConsumeMode();
    }

    @Override
    public SelectorType getSelectorType() {
        return this.container.getSelectorType();
    }

    @Override
    public void setSelectorExpression(String selectorExpression) {
        this.container.setSelectorExpression(selectorExpression);
    }

    @Override
    public String getSelectorExpression() {
        return this.container.getSelectorExpression();
    }

    @Override
    public MessageModel getMessageModel() {
        return this.container.getMessageModel();
    }

    @Override
    public DefaultMQPushConsumer getConsumer() {
        return this.container.getConsumer();
    }

    @Override
    public void setConsumer(DefaultMQPushConsumer consumer) {
        this.container.setConsumer(consumer);
    }

    @Override
    public void destroy() {
        this.container.destroy();
    }

    @Override
    public boolean isAutoStartup() {
        return this.container.isAutoStartup();
    }

    @Override
    public void stop(Runnable callback) {
        this.container.stop(callback);
    }

    @Override
    public void start() {
        this.container.start();
    }

    @Override
    public void stop() {
        this.container.stop();
    }

    @Override
    public boolean isRunning() {
        return this.container.isRunning();
    }

    @Override
    public int getPhase() {
        return this.container.getPhase();
    }

    @Override
    public void afterPropertiesSet() throws Exception {
        this.initialize();
    }

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        this.container.setApplicationContext(applicationContext);
    }

    @Override
    public String toString() {
        return this.container.toString();
    }

    @Override
    public void setName(String name) {
        this.container.setName(name);
    }
}
