package com.chelvc.framework.rocketmq.context;

import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.time.Duration;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.atomic.AtomicLong;
import java.util.function.Consumer;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import com.chelvc.framework.base.context.ApplicationContextHolder;
import com.chelvc.framework.base.context.JacksonContextHolder;
import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.base.model.Session;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.chelvc.framework.rocketmq.annotation.MessageDelayer;
import com.chelvc.framework.rocketmq.interceptor.SessionRocketMQListener;
import com.chelvc.framework.rocketmq.model.Delay;
import com.chelvc.framework.rocketmq.model.DelayMessage;
import com.chelvc.framework.rocketmq.model.MessageContext;
import com.chelvc.framework.rocketmq.producer.RocketMQTransactionChecker;
import com.chelvc.framework.rocketmq.producer.RocketMQTransactionExecutor;
import com.chelvc.framework.rocketmq.producer.RocketMQTransactionHandler;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtMethod;
import javassist.LoaderClassPath;
import javassist.bytecode.AnnotationsAttribute;
import javassist.bytecode.ClassFile;
import javassist.bytecode.ConstPool;
import javassist.bytecode.SignatureAttribute;
import javassist.bytecode.annotation.Annotation;
import javassist.bytecode.annotation.StringMemberValue;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.rocketmq.client.producer.SendCallback;
import org.apache.rocketmq.client.producer.SendResult;
import org.apache.rocketmq.client.producer.SendStatus;
import org.apache.rocketmq.client.producer.TransactionSendResult;
import org.apache.rocketmq.spring.annotation.RocketMQMessageListener;
import org.apache.rocketmq.spring.annotation.RocketMQTransactionListener;
import org.apache.rocketmq.spring.autoconfigure.ListenerContainerConfiguration;
import org.apache.rocketmq.spring.core.RocketMQListener;
import org.apache.rocketmq.spring.core.RocketMQLocalTransactionListener;
import org.apache.rocketmq.spring.core.RocketMQLocalTransactionState;
import org.apache.rocketmq.spring.core.RocketMQReplyListener;
import org.apache.rocketmq.spring.core.RocketMQTemplate;
import org.apache.rocketmq.spring.support.RocketMQMessageConverter;
import org.springframework.aop.framework.AopProxyUtils;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.SmartInitializingSingleton;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.core.env.Environment;
import org.springframework.messaging.Message;
import org.springframework.messaging.converter.MessageConverter;
import org.springframework.messaging.support.GenericMessage;
import org.springframework.messaging.support.MessageBuilder;
import org.springframework.stereotype.Component;
import org.springframework.util.Assert;
import org.springframework.util.CollectionUtils;

/**
 * RocketMQ上下文工具类
 *
 * @author Woody
 * @date 2021/9/6
 */
@Slf4j
@Component
@RocketMQTransactionListener
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class RocketMQContextHolder implements RocketMQLocalTransactionListener, ApplicationContextAware,
        SmartInitializingSingleton {
    /**
     * RocketMQ主题消息头标识
     */
    private static final String ROCKETMQ_TYPE_HEADER = "type";

    /**
     * 对象池
     */
    private static final ClassPool CLASS_POOL = ClassPool.getDefault();

    /**
     * 自定义延时消息延时器计数器
     */
    private static final AtomicLong DELAYER_COUNTER = new AtomicLong(0);

    /**
     * 自定义延时消息主题名称集合
     */
    private static final Set<String> DELAYER_TOPIC_NAME = Sets.newConcurrentHashSet();

    /**
     * 异常上下文
     */
    private static final ThreadLocal<RuntimeException> EXCEPTION_CONTEXT = new ThreadLocal<>();

    /**
     * 自定义延时消息延时器包名
     */
    private static final String DELAYER_PACKAGE =
            RocketMQContextHolder.class.getPackage().getName().replace("context", "delayer");

    /**
     * 环境名称
     */
    private static String PROFILE;

    /**
     * RocketMQ模版实例
     */
    private static RocketMQTemplate ROCKETMQ_TEMPLATE;

    /**
     * 消息转换器
     */
    private static MessageConverter MESSAGE_CONVERTER;

    /**
     * 消息类型/本地事务检测器映射表
     */
    private static Map<String, RocketMQTransactionChecker<?>> MESSAGE_CHECKER_MAPPING = Collections.emptyMap();

    /**
     * 消息类型/本地事务执行器映射表
     */
    private static Map<String, RocketMQTransactionExecutor<?>> MESSAGE_EXECUTOR_MAPPING = Collections.emptyMap();

    static {
        CLASS_POOL.appendClassPath(new LoaderClassPath(Thread.currentThread().getContextClassLoader()));
    }

    @Override
    @SuppressWarnings("rawtypes")
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        // 初始化环境名称
        PROFILE = applicationContext.getEnvironment().getProperty("spring.profiles.active");

        // 初始化RocketMQ模版实例
        ROCKETMQ_TEMPLATE = applicationContext.getBean(RocketMQTemplate.class);

        // 初始化消息转换器
        MESSAGE_CONVERTER = applicationContext.getBean(RocketMQMessageConverter.class).getMessageConverter();

        // 初始化消息类型/本地事务检测实例映射表
        Collection<RocketMQTransactionHandler> handlers =
                applicationContext.getBeansOfType(RocketMQTransactionHandler.class).values();
        List<RocketMQTransactionChecker> checkers = handlers.stream()
                .filter(handler -> handler instanceof RocketMQTransactionChecker)
                .map(handler -> (RocketMQTransactionChecker) handler).collect(Collectors.toList());
        if (!CollectionUtils.isEmpty(checkers)) {
            log.info("Loading rocketmq transaction checkers: {}", checkers);
            MESSAGE_CHECKER_MAPPING = Maps.newHashMapWithExpectedSize(checkers.size());
            checkers.forEach(checker -> {
                Class<?> clazz = checker.getPayloadClass();
                Assert.notNull(clazz, "Rocketmq transaction checker payload class unspecified: " + checker);
                String type = clazz.getName();
                Assert.isTrue(!MESSAGE_CHECKER_MAPPING.containsKey(type),
                        "Rocketmq transaction checker payload class already exists: " + checker);
                MESSAGE_CHECKER_MAPPING.put(type, checker);
            });
        }

        // 初始化消息类型/本地事务执行实例映射表
        List<RocketMQTransactionExecutor> executors = handlers.stream()
                .filter(handler -> handler instanceof RocketMQTransactionExecutor)
                .map(handler -> (RocketMQTransactionExecutor) handler).collect(Collectors.toList());
        if (!CollectionUtils.isEmpty(executors)) {
            log.info("Loading rocketmq transaction executors: {}", executors);
            MESSAGE_EXECUTOR_MAPPING = Maps.newHashMapWithExpectedSize(executors.size());
            executors.forEach(executor -> {
                Class<?> clazz = executor.getPayloadClass();
                Assert.notNull(clazz, "Rocketmq transaction executor payload class unspecified: " + executor);
                String type = clazz.getName();
                Assert.isTrue(!MESSAGE_EXECUTOR_MAPPING.containsKey(type),
                        "Rocketmq transaction executor payload class already exists: " + executor);
                MESSAGE_EXECUTOR_MAPPING.put(type, executor);
            });
        }
    }

    @Override
    public void afterSingletonsInstantiated() {
        // 初始化延时消息延时器
        ApplicationContext context = ApplicationContextHolder.getApplicationContext();
        Map<String, Object> consumers = context.getBeansWithAnnotation(MessageDelayer.class);
        consumers.values().forEach(this::registerMessageDelayerContainer);
    }

    @Override
    @SuppressWarnings({"rawtypes", "unchecked"})
    public RocketMQLocalTransactionState checkLocalTransaction(@NonNull Message message) {
        RocketMQTransactionChecker checker = Objects.requireNonNull(
                MESSAGE_CHECKER_MAPPING.get(getMessageType(message)), "Rocketmq transaction checker not found"
        );
        return checker.check(message, message2payload(message, checker.getPayloadClass()));
    }

    @Override
    @SuppressWarnings({"rawtypes", "unchecked"})
    public RocketMQLocalTransactionState executeLocalTransaction(@NonNull Message message, Object arg) {
        try {
            if (arg instanceof Supplier) {
                return ((Supplier<RocketMQLocalTransactionState>) arg).get();
            }
            RocketMQTransactionExecutor executor = Objects.requireNonNull(
                    MESSAGE_EXECUTOR_MAPPING.get(getMessageType(message)), "Rocketmq transaction executor not found"
            );
            return executor.execute(message, message2payload(message, executor.getPayloadClass()), arg);
        } catch (RuntimeException e) {
            // 将异常存入上下文
            EXCEPTION_CONTEXT.set(e);
            throw e;
        }
    }

    /**
     * 查找消费者消息类型
     *
     * @param target 消费者对象类型
     * @return 消息对象类型
     */
    public static Type lookupConsumerMessageType(@NonNull Class<?> target) {
        Type parameterized = null;
        while (Objects.nonNull(target)) {
            Type[] interfaces = target.getGenericInterfaces();
            for (Type type : interfaces) {
                if (type instanceof ParameterizedType) {
                    Type raw = ((ParameterizedType) type).getRawType();
                    if (Objects.equals(raw, SessionRocketMQListener.class)
                            || Objects.equals(raw, RocketMQListener.class)
                            || Objects.equals(raw, RocketMQReplyListener.class)) {
                        parameterized = type;
                        break;
                    }
                }
            }
            target = target.getSuperclass();
        }
        if (Objects.isNull(parameterized)) {
            return Object.class;
        }

        Type[] arguments = ((ParameterizedType) parameterized).getActualTypeArguments();
        if (Objects.nonNull(arguments) && arguments.length > 0) {
            return arguments[0];
        }
        return Object.class;
    }

    /**
     * 注册消息延时器消费者容器
     *
     * @param consumer 消费者实例
     */
    private void registerMessageDelayerContainer(Object consumer) {
        // 获取延时消息主题
        Class<?> clazz = AopProxyUtils.ultimateTargetClass(consumer);
        MessageDelayer annotation = clazz.getAnnotation(MessageDelayer.class);
        String topic = annotation.topic();
        if (StringUtils.isEmpty(topic)) {
            RocketMQMessageListener listener = clazz.getAnnotation(RocketMQMessageListener.class);
            topic = ObjectUtils.ifNull(listener, RocketMQMessageListener::topic);
        }
        Assert.isTrue(StringUtils.nonEmpty(topic), "Message delayer topic unassigned: " + clazz.getName());
        Environment environment = ApplicationContextHolder.getApplicationContext().getEnvironment();
        topic = environment.resolvePlaceholders(topic);
        if (!DELAYER_TOPIC_NAME.add(topic)) {
            return;
        }

        // 获取延时消息类型
        Class<?> type = annotation.type();
        if (type == Object.class) {
            type = (Class<?>) lookupConsumerMessageType(clazz);
        }
        Assert.isTrue(type != Object.class, "Message delayer type unassigned: " + clazz.getName());

        // 注册消息延时器消费者容器
        try {
            Class<?> delayer = this.generateMessageDelayerClass(topic, type);
            ListenerContainerConfiguration configuration =
                    ApplicationContextHolder.getApplicationContext().getBean(ListenerContainerConfiguration.class);
            Method method = ListenerContainerConfiguration.class.getDeclaredMethod(
                    "registerContainer", String.class, Object.class
            );
            method.setAccessible(true);
            method.invoke(configuration, delayer.getName(), delayer.newInstance());
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 动态生成消息延时器类对象
     *
     * @param topic 延时器消息主题
     * @param type  消息类型
     * @return 消息延时器类对象
     * @throws Exception 处理异常
     */
    private Class<?> generateMessageDelayerClass(String topic, Class<?> type) throws Exception {
        // 构建Class及实现接口
        CtClass delayer = CLASS_POOL.makeClass(
                String.format("%s.CustomMessageDelayer_%d", DELAYER_PACKAGE, DELAYER_COUNTER.incrementAndGet())
        );
        delayer.setInterfaces(new CtClass[]{CLASS_POOL.get(RocketMQListener.class.getName())});
        delayer.setGenericSignature(new SignatureAttribute.ClassSignature(null, null,
                new SignatureAttribute.ClassType[]{new SignatureAttribute.ClassType(
                        RocketMQListener.class.getName(), new SignatureAttribute.TypeArgument[]{
                        new SignatureAttribute.TypeArgument(
                                new SignatureAttribute.ClassType(DelayMessage.class.getName())
                        )})}).encode()
        );

        // 添加无参构造方法
        CtConstructor constructor = new CtConstructor(new CtClass[]{}, delayer);
        constructor.setBody("{}");
        delayer.addConstructor(constructor);

        // 添加自定义延时消息消费方法实现
        delayer.addMethod(CtMethod.make(String.format(
                "public void onMessage(%s message){%s.processing(\"%s\", %s.class, message);}",
                DelayMessage.class.getName(),
                RocketMQContextHolder.class.getName(),
                topic,
                type.getName()
        ), delayer));

        // 添加泛型参数接口方法桥接
        delayer.addMethod(CtMethod.make(String.format(
                "public void onMessage(Object message){this.onMessage((%s) message);}",
                DelayMessage.class.getName()
        ), delayer));

        // 设置@RocketMQMessageListener注解
        String unique = getDelayerTopic(topic);
        ClassFile file = delayer.getClassFile();
        ConstPool constPool = file.getConstPool();
        AnnotationsAttribute attribute = new AnnotationsAttribute(constPool, AnnotationsAttribute.visibleTag);
        Annotation annotation = new Annotation(RocketMQMessageListener.class.getName(), constPool);
        annotation.addMemberValue("topic", new StringMemberValue(unique, constPool));
        annotation.addMemberValue("consumerGroup", new StringMemberValue(unique, constPool));
        attribute.addAnnotation(annotation);
        file.addAttribute(attribute);
        return delayer.toClass();
    }

    /**
     * 根据当前环境名称获取实际消息主题
     *
     * @param topic 消息主题
     * @return 消息主题
     */
    public static String getProfileTopic(@NonNull String topic) {
        return Objects.requireNonNull(PROFILE, "Profile has not been initialized") + "-" + topic;
    }

    /**
     * 获取自定义延时消息实际消息主题
     *
     * @param topic 消息主题
     * @return 消息主题
     */
    public static String getDelayerTopic(@NonNull String topic) {
        return topic + "-delay";
    }

    /**
     * 获取RocketMQ模版实例
     *
     * @return RocketMQ模版实例
     */
    public static RocketMQTemplate getRocketMQTemplate() {
        return Objects.requireNonNull(ROCKETMQ_TEMPLATE, "Rocketmq template has not been initialized");
    }

    /**
     * 获取消息转换器实例
     *
     * @return 消息转换器实例
     */
    public static MessageConverter getMessageConverter() {
        return Objects.requireNonNull(MESSAGE_CONVERTER, "Message converter has not been initialized");
    }

    /**
     * 将消息内容转换成消息对象
     *
     * @param payload 消息内容
     * @return 消息对象实例
     */
    public static Message<?> payload2message(@NonNull Object payload) {
        Object target = JacksonContextHolder.serialize(payload);
        Session session = SessionContextHolder.getSession(false);
        MessageContext<?> context = MessageContext.builder().target(target).session(session).build();
        return MessageBuilder.withPayload(context).setHeader(ROCKETMQ_TYPE_HEADER, payload.getClass().getName()).build();
    }

    /**
     * 将消息对象转换成消息体
     *
     * @param message 消息对象
     * @param type    消息体类型
     * @param <T>     消息体对象类型
     * @return 消息体实例
     */
    @SuppressWarnings("unchecked")
    public static <T> T message2payload(@NonNull Message<?> message, @NonNull Class<T> type) {
        MessageConverter converter = getMessageConverter();
        MessageContext<?> context =
                (MessageContext<?>) Objects.requireNonNull(converter.fromMessage(message, MessageContext.class));
        Object payload = context.getTarget();
        if (payload == null || (payload instanceof String && type == String.class)) {
            return (T) payload;
        }
        return (T) converter.fromMessage(new GenericMessage<>(payload), type);
    }

    /**
     * 获取消息类型
     *
     * @param message 消息对象实例
     * @return 消息类型
     */
    public static String getMessageType(@NonNull Message<?> message) {
        String type = message.getHeaders().get(ROCKETMQ_TYPE_HEADER, String.class);
        Assert.notNull(type, "Rocketmq message type not found: " + message);
        return type;
    }

    /**
     * 发送消息
     *
     * @param topic   消息主题
     * @param payload 消息内容
     */
    public static void send(@NonNull String topic, @NonNull Object payload) {
        send(topic, payload, true);
    }

    /**
     * 发送消息
     *
     * @param topic   消息主题
     * @param payload 消息内容
     * @param isolate 是否需要环境隔离
     */
    public static void send(@NonNull String topic, @NonNull Object payload, boolean isolate) {
        getRocketMQTemplate().send(isolate ? getProfileTopic(topic) : topic, payload2message(payload));
    }

    /**
     * 发送延时消息
     *
     * @param topic   消息主题
     * @param payload 消息内容
     * @param delay   延时类型
     */
    public static void send(@NonNull String topic, @NonNull Object payload, @NonNull Delay delay) {
        send(topic, payload, delay, true);
    }

    /**
     * 发送延时消息
     *
     * @param topic   消息主题
     * @param payload 消息内容
     * @param delay   延时类型
     * @param isolate 是否需要环境隔离
     */
    private static void send(@NonNull String topic, @NonNull Object payload, @NonNull Delay delay, boolean isolate) {
        RocketMQTemplate template = getRocketMQTemplate();
        int timeout = template.getProducer().getSendMsgTimeout();
        SendResult result = template.syncSend(
                isolate ? getProfileTopic(topic) : topic, payload2message(payload), timeout, delay.level()
        );
        SendStatus status = ObjectUtils.ifNull(result, SendResult::getSendStatus);
        if (status != SendStatus.SEND_OK) {
            throw new IllegalStateException("Send delay message failed: " + status);
        }
    }

    /**
     * 发送自定义延时消息
     *
     * @param topic    消息主题
     * @param payload  消息内容
     * @param delaying 延时时长
     */
    public static void send(@NonNull String topic, @NonNull Object payload, @NonNull Duration delaying) {
        Delay delay = Delay.close(delaying);
        if (delay == Delay.S1) {
            send(topic, payload);
        } else if (delay.time() == delaying.getSeconds()) {
            send(topic, payload, delay);
        } else {
            long expiration = System.currentTimeMillis() + Math.max(delaying.toMillis(), 0);
            send(getDelayerTopic(getProfileTopic(topic)), DelayMessage.of(payload, expiration), delay, false);
        }
    }

    /**
     * 发送自定义延时消息
     *
     * @param topic    消息主题
     * @param payload  消息内容
     * @param delaying 延时时间
     */
    public static void send(@NonNull String topic, @NonNull Object payload, @NonNull Date delaying) {
        long duration = (delaying.getTime() - System.currentTimeMillis()) / 1000;
        Delay delay = Delay.close(duration);
        if (delay == Delay.S1) {
            send(topic, payload);
        } else if (delay.time() == duration) {
            send(topic, payload, delay);
        } else {
            send(getDelayerTopic(getProfileTopic(topic)), DelayMessage.of(payload, delaying.getTime()), delay, false);
        }
    }

    /**
     * 延时消息处理，供消息延时器调用
     *
     * @param topic   消息主题
     * @param type    消息类型
     * @param message 延时消息实例
     */
    public static void processing(@NonNull String topic, @NonNull Class<?> type, @NonNull DelayMessage message) {
        long duration = (message.getExpiration() - System.currentTimeMillis()) / 1000;
        Delay delay = Delay.close(duration);
        if (delay == Delay.S1) {
            send(topic, JacksonContextHolder.deserialize(message.getPayload(), type), false);
        } else if (delay.time() == duration) {
            send(topic, JacksonContextHolder.deserialize(message.getPayload(), type), delay, false);
        } else {
            send(getDelayerTopic(topic), message, delay, false);
        }
    }

    /**
     * 批量发送消息
     *
     * @param topic    消息主题
     * @param payloads 消息内容集合
     */
    public static void send(@NonNull String topic, @NonNull Collection<?> payloads) {
        List<Message<?>> messages = payloads.stream().filter(Objects::nonNull)
                .map(RocketMQContextHolder::payload2message).collect(Collectors.toList());
        if (!messages.isEmpty()) {
            getRocketMQTemplate().syncSend(getProfileTopic(topic), messages);
        }
    }

    /**
     * 异步发送消息
     *
     * @param topic   消息主题
     * @param payload 消息内容
     */
    public static void asyncSend(@NonNull String topic, @NonNull Object payload) {
        asyncSend(topic, payload, true);
    }

    /**
     * 异步发送消息
     *
     * @param topic   消息主题
     * @param payload 消息内容
     * @param isolate 是否需要环境隔离
     */
    public static void asyncSend(@NonNull String topic, @NonNull Object payload, boolean isolate) {
        asyncSend(topic, payload, isolate, new SendCallback() {
            @Override
            public void onSuccess(SendResult result) {
            }

            @Override
            public void onException(Throwable e) {
                log.warn("RocketMQ message send failed: {}", e.getMessage(), e);
            }
        });
    }

    /**
     * 异步发送消息
     *
     * @param topic    消息主题
     * @param payload  消息内容
     * @param callback 发送结果回调接口
     */
    public static void asyncSend(@NonNull String topic, @NonNull Object payload, @NonNull SendCallback callback) {
        asyncSend(topic, payload, true, callback);
    }

    /**
     * 异步发送消息
     *
     * @param topic    消息主题
     * @param payload  消息内容
     * @param isolate  是否需要环境隔离
     * @param callback 发送结果回调接口
     */
    public static void asyncSend(@NonNull String topic, @NonNull Object payload, boolean isolate,
                                 @NonNull SendCallback callback) {
        getRocketMQTemplate().asyncSend(
                isolate ? getProfileTopic(topic) : topic, payload2message(payload), callback
        );
    }

    /**
     * 发送顺序消息
     *
     * @param topic   消息主题
     * @param payload 消息内容
     * @param key     唯一标识
     */
    public static void sendOrderly(@NonNull String topic, @NonNull Object payload, @NonNull Object key) {
        sendOrderly(topic, payload, key, true);
    }

    /**
     * 发送顺序消息
     *
     * @param topic   消息主题
     * @param payload 消息内容
     * @param key     唯一标识
     * @param isolate 是否需要环境隔离
     */
    public static void sendOrderly(@NonNull String topic, @NonNull Object payload, @NonNull Object key,
                                   boolean isolate) {
        getRocketMQTemplate().syncSendOrderly(
                isolate ? getProfileTopic(topic) : topic, payload2message(payload), String.valueOf(key)
        );
    }

    /**
     * 异步发送顺序消息
     *
     * @param topic   消息主题
     * @param payload 消息内容
     * @param key     唯一标识
     */
    public static void asyncSendOrderly(@NonNull String topic, @NonNull Object payload, @NonNull Object key) {
        asyncSendOrderly(topic, payload, key, true);
    }

    /**
     * 异步发送顺序消息
     *
     * @param topic   消息主题
     * @param payload 消息内容
     * @param key     唯一标识
     * @param isolate 是否需要环境隔离
     */
    public static void asyncSendOrderly(@NonNull String topic, @NonNull Object payload, @NonNull Object key,
                                        boolean isolate) {
        asyncSendOrderly(topic, payload, key, isolate, new SendCallback() {
            @Override
            public void onSuccess(SendResult result) {
            }

            @Override
            public void onException(Throwable e) {
                log.warn("RocketMQ message send failed: {}", e.getMessage(), e);
            }
        });
    }

    /**
     * 异步发送顺序消息
     *
     * @param topic    消息主题
     * @param payload  消息内容
     * @param key      唯一标识
     * @param callback 发送结果回调接口
     */
    public static void asyncSendOrderly(@NonNull String topic, @NonNull Object payload, @NonNull Object key,
                                        @NonNull SendCallback callback) {
        asyncSendOrderly(topic, payload, key, true, callback);
    }

    /**
     * 异步发送顺序消息
     *
     * @param topic    消息主题
     * @param payload  消息内容
     * @param key      唯一标识
     * @param isolate  是否需要环境隔离
     * @param callback 发送结果回调接口
     */
    public static void asyncSendOrderly(@NonNull String topic, @NonNull Object payload, @NonNull Object key,
                                        boolean isolate, @NonNull SendCallback callback) {
        getRocketMQTemplate().asyncSendOrderly(
                isolate ? getProfileTopic(topic) : topic, payload2message(payload), String.valueOf(key), callback
        );
    }

    /**
     * 发送单向消息
     *
     * @param topic   消息主题
     * @param payload 消息内容
     */
    public static void sendOneWay(@NonNull String topic, @NonNull Object payload) {
        sendOneWay(topic, payload, true);
    }

    /**
     * 发送单向消息
     *
     * @param topic   消息主题
     * @param payload 消息内容
     * @param isolate 是否需要环境隔离
     */
    public static void sendOneWay(@NonNull String topic, @NonNull Object payload, boolean isolate) {
        getRocketMQTemplate().sendOneWay(isolate ? getProfileTopic(topic) : topic, payload2message(payload));
    }

    /**
     * 发送单向顺序消息
     *
     * @param topic   消息主题
     * @param payload 消息内容
     * @param key     唯一标识
     */
    public static void sendOneWayOrderly(@NonNull String topic, @NonNull Object payload, @NonNull Object key) {
        sendOneWayOrderly(topic, payload, key, true);
    }

    /**
     * 发送单向顺序消息
     *
     * @param topic   消息主题
     * @param payload 消息内容
     * @param key     唯一标识
     * @param isolate 是否需要环境隔离
     */
    public static void sendOneWayOrderly(@NonNull String topic, @NonNull Object payload, @NonNull Object key,
                                         boolean isolate) {
        getRocketMQTemplate().sendOneWayOrderly(
                isolate ? getProfileTopic(topic) : topic, payload2message(payload), String.valueOf(key)
        );
    }

    /**
     * 发送事务消息
     *
     * @param topic    消息主题
     * @param payload  消息内容
     * @param callback 事务消息回调方法
     * @return 事务消息发送结果
     */
    public static TransactionSendResult sendTransactionalMessage(@NonNull String topic, @NonNull Object payload,
                                                                 @NonNull Supplier<RocketMQLocalTransactionState> callback) {
        return sendTransactionalMessage(topic, payload, callback, true);
    }

    /**
     * 发送事务消息
     *
     * @param topic    消息主题
     * @param payload  消息内容
     * @param callback 事务消息回调方法
     * @param isolate  是否需要环境隔离
     * @return 事务消息发送结果
     */
    public static TransactionSendResult sendTransactionalMessage(@NonNull String topic, @NonNull Object payload,
                                                                 @NonNull Supplier<RocketMQLocalTransactionState> callback,
                                                                 boolean isolate) {
        String type = payload.getClass().getName();
        Assert.isTrue(MESSAGE_CHECKER_MAPPING.containsKey(type),
                String.format("Rocketmq transaction checker with payload class '%s' not found", type));

        RocketMQTemplate template = getRocketMQTemplate();
        Message<?> message = payload2message(payload);
        try {
            // 发送事务消息
            TransactionSendResult result = template.sendMessageInTransaction(
                    isolate ? getProfileTopic(topic) : topic, message, callback
            );

            // 判断执行本地事务是否抛出了异常，如果存在异常则直接抛出
            RuntimeException exception = EXCEPTION_CONTEXT.get();
            if (exception != null) {
                throw exception;
            }
            return result;
        } finally {
            // 清空异常上下文
            EXCEPTION_CONTEXT.remove();
        }
    }

    /**
     * 消费消息上下文消息
     *
     * @param context  消息上下文对象
     * @param consumer 消息消费者
     * @param <T>      消息类型
     */
    public static <T> void consumeMessageContext(@NonNull MessageContext<T> context, @NonNull Consumer<T> consumer) {
        SessionContextHolder.setSession(context.getSession());
        try {
            consumer.accept(context.getTarget());
        } finally {
            SessionContextHolder.clearSessionContext();
        }
    }
}
