package com.chelvc.framework.security.config;

import java.io.Serializable;
import java.util.Objects;
import java.util.stream.Stream;

import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.base.model.Platform;
import com.chelvc.framework.base.model.Terminal;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.security.context.SecurityContextHolder;
import org.springframework.security.access.PermissionEvaluator;
import org.springframework.security.access.expression.SecurityExpressionRoot;
import org.springframework.security.access.expression.method.MethodSecurityExpressionOperations;
import org.springframework.security.access.hierarchicalroles.RoleHierarchy;
import org.springframework.security.authentication.AuthenticationTrustResolver;
import org.springframework.security.core.Authentication;
import org.springframework.security.oauth2.jwt.Jwt;

/**
 * 方法权限表达式处理包装类，自定义hasAuthority()、hasAnyAuthority()的处理逻辑
 *
 * @author Woody
 * @date 2023/4/5
 * @see MethodSecurityConfigurer
 */
public class MethodSecurityExpressionWrapper implements MethodSecurityExpressionOperations {
    private final SecurityExpressionRoot root;
    private Object filterObject;
    private Object returnObject;
    private Object target;

    public MethodSecurityExpressionWrapper(Authentication authentication) {
        this.root = new SecurityExpressionRoot(authentication) {
        };
    }

    @Override
    public Object getFilterObject() {
        return this.filterObject;
    }

    @Override
    public void setFilterObject(Object filterObject) {
        this.filterObject = filterObject;
    }

    @Override
    public Object getReturnObject() {
        return this.returnObject;
    }

    @Override
    public void setReturnObject(Object returnObject) {
        this.returnObject = returnObject;
    }

    @Override
    public Object getThis() {
        return this.target;
    }

    public void setThis(Object target) {
        this.target = target;
    }

    @Override
    public Authentication getAuthentication() {
        return this.root.getAuthentication();
    }

    /**
     * 判断主体是否是指定账户
     *
     * @param accounts 主体账户数组
     * @return true/false
     */
    public boolean isAccount(Serializable... accounts) {
        if (accounts == null || accounts.length == 0) {
            return false;
        }
        Jwt jwt = SecurityContextHolder.getJwt(false);
        if (jwt == null) {
            return false;
        }
        String id = ObjectUtils.ifNull(SecurityContextHolder.getId(jwt), String::valueOf);
        String user = SecurityContextHolder.getUser(jwt);
        String mobile = SecurityContextHolder.getMobile(jwt);
        String client = SecurityContextHolder.getClient(jwt);
        return Stream.of(accounts).filter(Objects::nonNull).map(String::valueOf).anyMatch(account ->
                Objects.equals(account, id) || Objects.equals(account, user)
                        || Objects.equals(account, mobile) || Objects.equals(account, client)
        );
    }

    /**
     * 判断是否是指定业务类型
     *
     * @param businesses 业务类型数组
     * @return true/false
     */
    public boolean isBusiness(String... businesses) {
        if (businesses == null || businesses.length == 0) {
            return false;
        }
        String current = SecurityContextHolder.getBusiness();
        return current != null && Stream.of(businesses).filter(Objects::nonNull)
                .anyMatch(business -> Objects.equals(business, current));
    }

    /**
     * 判断当前请求是否为指定客户端平台
     *
     * @param platforms 客户端平台数组
     * @return true/false
     */
    public boolean isPlatform(Platform... platforms) {
        if (platforms == null || platforms.length == 0) {
            return false;
        }
        Platform current = SessionContextHolder.getPlatform(false);
        return Stream.of(platforms).filter(Objects::nonNull).anyMatch(platform -> Objects.equals(platform, current));
    }

    /**
     * 判断当前请求是否为指定客户终端
     *
     * @param terminals 客户终端数组
     * @return true/false
     */
    public boolean isTerminal(Terminal... terminals) {
        if (terminals == null || terminals.length == 0) {
            return false;
        }
        Terminal current = SessionContextHolder.getTerminal(false);
        return Stream.of(terminals).filter(Objects::nonNull).anyMatch(terminal -> Objects.equals(terminal, current));
    }

    /**
     * 判断当前请求是否为指定版本号
     *
     * @param versions 客户端版本号数组
     * @return true/false
     */
    public boolean isVersion(String... versions) {
        if (versions == null || versions.length == 0) {
            return false;
        }
        String current = SessionContextHolder.getVersion(false);
        return Stream.of(versions).filter(Objects::nonNull).anyMatch(version -> Objects.equals(version, current));
    }

    @Override
    public boolean hasAuthority(String authority) {
        return this.root.hasAuthority(authority);
    }

    @Override
    public boolean hasAnyAuthority(String... authorities) {
        return this.root.hasAnyAuthority(authorities);
    }

    @Override
    public boolean hasRole(String role) {
        return this.root.hasRole(role);
    }

    @Override
    public boolean hasAnyRole(String... roles) {
        return this.root.hasAnyRole(roles);
    }

    @Override
    public boolean permitAll() {
        return this.root.permitAll();
    }

    @Override
    public boolean denyAll() {
        return this.root.denyAll();
    }

    @Override
    public boolean isAnonymous() {
        return this.root.isAnonymous();
    }

    @Override
    public boolean isAuthenticated() {
        return this.root.isAuthenticated();
    }

    @Override
    public boolean isRememberMe() {
        return this.root.isRememberMe();
    }

    @Override
    public boolean isFullyAuthenticated() {
        return this.root.isFullyAuthenticated();
    }

    @Override
    public boolean hasPermission(Object target, Object permission) {
        return this.root.hasPermission(target, permission);
    }

    @Override
    public boolean hasPermission(Object targetId, String targetType, Object permission) {
        return this.root.hasPermission(targetId, targetType, permission);
    }

    public void setPermissionEvaluator(PermissionEvaluator permissionEvaluator) {
        this.root.setPermissionEvaluator(permissionEvaluator);
    }

    public void setTrustResolver(AuthenticationTrustResolver trustResolver) {
        this.root.setTrustResolver(trustResolver);
    }

    public void setRoleHierarchy(RoleHierarchy roleHierarchy) {
        this.root.setRoleHierarchy(roleHierarchy);
    }

    public void setDefaultRolePrefix(String defaultRolePrefix) {
        this.root.setDefaultRolePrefix(defaultRolePrefix);
    }
}
