package com.chelvc.framework.security.config;

import java.security.KeyPair;
import java.security.interfaces.RSAPublicKey;
import java.util.Collection;
import java.util.List;
import javax.crypto.spec.SecretKeySpec;
import javax.servlet.Filter;

import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.base.model.Result;
import com.chelvc.framework.base.util.AESUtils;
import com.chelvc.framework.base.util.HttpUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.chelvc.framework.security.context.SecurityContextHolder;
import com.chelvc.framework.security.token.TokenActiveValidator;
import com.google.common.collect.Lists;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.boot.autoconfigure.security.oauth2.resource.OAuth2ResourceServerProperties;
import org.springframework.boot.web.servlet.FilterRegistrationBean;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.Bean;
import org.springframework.core.Ordered;
import org.springframework.core.io.ClassPathResource;
import org.springframework.http.HttpMethod;
import org.springframework.http.HttpStatus;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.authentication.AuthenticationProvider;
import org.springframework.security.config.annotation.authentication.builders.AuthenticationManagerBuilder;
import org.springframework.security.config.annotation.web.builders.HttpSecurity;
import org.springframework.security.config.annotation.web.builders.WebSecurity;
import org.springframework.security.config.annotation.web.configuration.EnableWebSecurity;
import org.springframework.security.config.annotation.web.configuration.WebSecurityConfigurerAdapter;
import org.springframework.security.config.annotation.web.configurers.ExpressionUrlAuthorizationConfigurer;
import org.springframework.security.config.http.SessionCreationPolicy;
import org.springframework.security.oauth2.core.DelegatingOAuth2TokenValidator;
import org.springframework.security.oauth2.core.OAuth2AuthenticationException;
import org.springframework.security.oauth2.core.OAuth2Error;
import org.springframework.security.oauth2.core.OAuth2TokenValidator;
import org.springframework.security.oauth2.core.OAuth2TokenValidatorResult;
import org.springframework.security.oauth2.jwt.Jwt;
import org.springframework.security.oauth2.jwt.JwtDecoder;
import org.springframework.security.oauth2.jwt.JwtTimestampValidator;
import org.springframework.security.oauth2.jwt.NimbusJwtDecoder;
import org.springframework.security.oauth2.provider.token.store.KeyStoreKeyFactory;
import org.springframework.security.oauth2.server.resource.BearerTokenError;
import org.springframework.security.oauth2.server.resource.authentication.JwtAuthenticationConverter;
import org.springframework.security.oauth2.server.resource.authentication.JwtGrantedAuthoritiesConverter;
import org.springframework.security.web.AuthenticationEntryPoint;
import org.springframework.security.web.access.AccessDeniedHandler;
import org.springframework.security.web.authentication.logout.LogoutHandler;
import org.springframework.security.web.authentication.logout.LogoutSuccessHandler;
import org.springframework.util.CollectionUtils;
import org.springframework.web.cors.CorsConfiguration;
import org.springframework.web.cors.UrlBasedCorsConfigurationSource;
import org.springframework.web.filter.CorsFilter;

/**
 * Web安全配置
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@EnableWebSecurity
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class WebSecurityConfigurer extends WebSecurityConfigurerAdapter {
    private final ApplicationContext applicationContext;
    private final SecurityProperties securityProperties;
    private final OAuth2ResourceServerProperties resourceServerProperties;

    @Autowired(required = false)
    private LogoutHandler logoutHandler;

    @Autowired(required = false)
    private LogoutSuccessHandler logoutSuccessHandler;

    @Bean
    public FilterRegistrationBean<Filter> crossDomainAccessRegistration() {
        FilterRegistrationBean<Filter> registration = new FilterRegistrationBean<>();
        CorsConfiguration corsConfiguration = new CorsConfiguration();
        corsConfiguration.addAllowedOrigin("*");
        corsConfiguration.addAllowedHeader("*");
        corsConfiguration.addAllowedMethod("*");
        corsConfiguration.setAllowCredentials(true);
        UrlBasedCorsConfigurationSource corsConfigurationSource = new UrlBasedCorsConfigurationSource();
        corsConfigurationSource.registerCorsConfiguration("/**", corsConfiguration);
        registration.setFilter(new CorsFilter(corsConfigurationSource));
        registration.setOrder(Ordered.HIGHEST_PRECEDENCE);
        return registration;
    }

    @Bean
    @ConditionalOnProperty(prefix = "platform.security.token.key-pair",
            value = {"alias", "resource", "key-store-password", "key-pair-password"})
    public KeyPair keyPair() {
        SecurityProperties.KeyPair keyPair = this.securityProperties.getToken().getKeyPair();
        KeyStoreKeyFactory keyStoreKeyFactory = new KeyStoreKeyFactory(new ClassPathResource(keyPair.getResource()),
                keyPair.getKeyStorePassword().toCharArray());
        return keyStoreKeyFactory.getKeyPair(keyPair.getAlias(), keyPair.getKeyPairPassword().toCharArray());
    }

    @Bean
    public NimbusJwtDecoder nimbusJwtDecoder() {
        if (!StringUtils.isEmpty(this.securityProperties.getToken().getSecret())) {
            String secret = this.securityProperties.getToken().getSecret();
            return NimbusJwtDecoder.withSecretKey(new SecretKeySpec(secret.getBytes(), AESUtils.ALGORITHM)).build();
        } else if (!StringUtils.isEmpty(this.resourceServerProperties.getJwt().getJwkSetUri())) {
            return NimbusJwtDecoder.withJwkSetUri(this.resourceServerProperties.getJwt().getJwkSetUri()).build();
        }
        return NimbusJwtDecoder.withPublicKey((RSAPublicKey) this.keyPair().getPublic()).build();
    }

    /**
     * 构建JWT解码器
     *
     * @return JWT解码器
     */
    private JwtDecoder jwtDecoder() {
        NimbusJwtDecoder decoder = this.nimbusJwtDecoder();

        // 添加自定义令牌验证器
        Collection<OAuth2TokenValidator<Jwt>> validators = Lists.newLinkedList();
        validators.add(new OAuth2TokenValidator<Jwt>() {
            private final JwtTimestampValidator timestampValidator = new JwtTimestampValidator();

            @Override
            public OAuth2TokenValidatorResult validate(Jwt jwt) {
                // 设置令牌/认证主体上下文
                SecurityContextHolder.setJwt(jwt);
                SessionContextHolder.initializeSessionPrincipal(
                        SecurityContextHolder.getId(jwt),
                        SecurityContextHolder.getBusiness(jwt),
                        SecurityContextHolder.getMobile(jwt)
                );
                return this.timestampValidator.validate(jwt);
            }
        });

        // 添加自定义令牌有效性验证器
        this.applicationContext.getBeansOfType(TokenActiveValidator.class).forEach((k, v) -> validators.add(v));
        log.info("Loading token validity validators: {}", validators);
        decoder.setJwtValidator(new DelegatingOAuth2TokenValidator<>(validators));
        return decoder;
    }

    @Bean
    public JwtAuthenticationConverter jwtAuthenticationConverter() {
        JwtAuthenticationConverter jwtAuthenticationConverter = new JwtAuthenticationConverter();
        JwtGrantedAuthoritiesConverter jwtGrantedAuthoritiesConverter = new JwtGrantedAuthoritiesConverter();
        jwtGrantedAuthoritiesConverter.setAuthorityPrefix(StringUtils.EMPTY);
        jwtGrantedAuthoritiesConverter.setAuthoritiesClaimName(SecurityContextHolder.JWT_CLAIM_AUTHORITIES);
        jwtAuthenticationConverter.setJwtGrantedAuthoritiesConverter(jwtGrantedAuthoritiesConverter);
        return jwtAuthenticationConverter;
    }

    @Bean
    @Override
    public AuthenticationManager authenticationManager() throws Exception {
        return super.authenticationManager();
    }

    @Override
    public void configure(WebSecurity web) {
        // 设置忽略的资源配置
        String[] ignores = StringUtils.splitActives(this.securityProperties.getResource().getIgnore());
        if (ignores != null && ignores.length > 0) {
            web.ignoring().antMatchers(ignores);
        }
    }

    @Bean
    public AccessDeniedHandler accessDeniedHandler() {
        return (request, response, exception) -> {
            log.warn(HttpUtils.getExceptionMessage(request, HttpStatus.FORBIDDEN, exception));
            HttpUtils.writeUnifiedResponse(response, Result.build(HttpStatus.FORBIDDEN));
        };
    }

    @Bean
    public AuthenticationEntryPoint authenticationEntryPoint() {
        return (request, response, exception) -> {
            HttpStatus status = HttpStatus.UNAUTHORIZED;
            if (exception instanceof OAuth2AuthenticationException) {
                OAuth2Error error = ((OAuth2AuthenticationException) exception).getError();
                if (error instanceof BearerTokenError) {
                    status = ((BearerTokenError) error).getHttpStatus();
                }
            }
            log.warn(HttpUtils.getExceptionMessage(request, status, exception));
            HttpUtils.writeUnifiedResponse(response, Result.build(status));
        };
    }

    @Override
    protected void configure(HttpSecurity http) throws Exception {
        // Http安全通用配置
        http.httpBasic().and().cors().and().csrf().disable()
                .authorizeRequests()
                .antMatchers(HttpMethod.OPTIONS)
                .permitAll()
                .and()
                .sessionManagement()
                .sessionCreationPolicy(SessionCreationPolicy.STATELESS)
                .and()
                .oauth2ResourceServer(server -> server.accessDeniedHandler(this.accessDeniedHandler())
                        .authenticationEntryPoint(this.authenticationEntryPoint()).jwt().decoder(this.jwtDecoder())
                        .jwtAuthenticationConverter(this.jwtAuthenticationConverter()));

        // 注销登录配置
        if (this.logoutHandler != null) {
            http.logout().addLogoutHandler(this.logoutHandler);
        }
        if (this.logoutSuccessHandler != null) {
            http.logout().logoutSuccessHandler(this.logoutSuccessHandler);
        }

        // 拒绝资源访问配置
        String[] denies = StringUtils.splitActives(this.securityProperties.getResource().getDeny());
        if (denies != null && denies.length > 0) {
            http.authorizeRequests().antMatchers(denies).denyAll();
        }

        // 允许资源访问配置
        String[] permits = StringUtils.splitActives(this.securityProperties.getResource().getPermit());
        if (permits != null && permits.length > 0) {
            http.authorizeRequests().antMatchers(permits).permitAll();
        }

        // 资源访问控制配置
        List<SecurityProperties.Access> accesses = this.securityProperties.getResource().getAccesses();
        if (!CollectionUtils.isEmpty(accesses)) {
            for (SecurityProperties.Access access : accesses) {
                String[] uris = StringUtils.splitActives(access.getUri());
                if (uris == null || uris.length == 0) {
                    continue;
                }
                ExpressionUrlAuthorizationConfigurer<?>.AuthorizedUrl authorizedUrl =
                        http.authorizeRequests().antMatchers(uris);
                if (!StringUtils.isEmpty(access.getIp())) {
                    authorizedUrl.hasIpAddress(access.getIp());
                }
                if (!StringUtils.isEmpty(access.getRole())) {
                    authorizedUrl.hasRole(access.getRole());
                }
                if (!StringUtils.isEmpty(access.getAuthority())) {
                    authorizedUrl.hasAuthority(access.getAuthority());
                }
                if (!StringUtils.isEmpty(access.getPermission())) {
                    authorizedUrl.access(access.getPermission());
                }
            }
        }

        //  认证资源访问配置
        String[] authenticates = StringUtils.splitActives(this.securityProperties.getResource().getAuthenticated());
        if (authenticates == null || authenticates.length == 0) {
            http.authorizeRequests().anyRequest().authenticated();
        } else {
            http.authorizeRequests().antMatchers(authenticates).authenticated();
        }
    }

    @Override
    protected void configure(AuthenticationManagerBuilder auth) {
        // 加载自定义认证方式
        Collection<AuthenticationProvider> providers =
                this.applicationContext.getBeansOfType(AuthenticationProvider.class).values();
        if (!CollectionUtils.isEmpty(providers)) {
            log.info("Loading authentication providers: {}", providers);
            providers.forEach(auth::authenticationProvider);
        }
    }
}
