package com.chelvc.framework.security.context;

import java.io.Serializable;
import java.time.Duration;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.TreeMap;
import java.util.regex.Pattern;
import javax.servlet.ServletRequestEvent;
import javax.servlet.ServletRequestListener;

import com.beust.jcommander.internal.Lists;
import com.chelvc.framework.base.context.ApplicationContextHolder;
import com.chelvc.framework.base.context.JacksonContextHolder;
import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.base.model.Session;
import com.chelvc.framework.base.model.Terminal;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.chelvc.framework.redis.util.RedisUtils;
import com.chelvc.framework.security.config.SecurityProperties;
import com.fasterxml.jackson.core.type.TypeReference;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.digest.HmacAlgorithms;
import org.apache.commons.codec.digest.HmacUtils;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.context.annotation.Configuration;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.security.oauth2.common.OAuth2AccessToken;
import org.springframework.security.oauth2.jwt.Jwt;
import org.springframework.util.CollectionUtils;

/**
 * 安全处理上下文工具类
 * <p>
 * 角色资源授权信息从Nacos配置中心获取并动态更新
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Configuration
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class SecurityContextHolder implements ApplicationContextAware, ServletRequestListener {
    /**
     * 主体身份标识
     */
    public static final String JWT_CLAIM_ID = "id";

    /**
     * JWT身份标识
     */
    public static final String JWT_CLAIM_JTI = "jti";

    /**
     * JWT手机号标识
     */
    public static final String JWT_CLAIM_MOBILE = "mobile";

    /**
     * 业务类型标识
     */
    public static final String JWT_CLAIM_BUSINESS = "business";

    /**
     * 用户账号标识
     */
    public static final String JWT_CLAIM_USER_NAME = "user_name";

    /**
     * 客户端账号标识
     */
    public static final String JWT_CLAIM_CLIENT_ID = "client_id";

    /**
     * 用户权限标识
     */
    public static final String JWT_CLAIM_AUTHORITIES = "authorities";

    /**
     * 权限属性名称
     */
    public static final String PERMISSION_PROPERTY = "platform.permission";

    /**
     * 方法资源拆分匹配模式
     */
    private static final Pattern METHOD_RESOURCE_SPLIT_PATTERN = StringUtils.getPattern(" +");

    /**
     * 权限类型引用
     */
    private static final TypeReference<Map<String, Set<String>>> PERMISSION_TYPE_REFERENCE =
            new TypeReference<Map<String, Set<String>>>() {
            };

    /**
     * JWT上下文
     */
    private static final ThreadLocal<Jwt> JWT_CONTEXT = new ThreadLocal<>();

    /**
     * 数字签名工具类实例
     */
    private static HmacUtils HMAC_UTILS = null;

    /**
     * 安全配置属性
     */
    private final SecurityProperties securityProperties;

    /**
     * 获取JWT信息
     *
     * @return JWT信息
     */
    public static Jwt getJwt() {
        return getJwt(true);
    }

    /**
     * 设置JWT信息
     *
     * @param jwt JWT信息
     */
    public static void setJwt(Jwt jwt) {
        JWT_CONTEXT.set(jwt);
    }

    /**
     * 获取JWT信息
     *
     * @param required JWT是否必须存在
     * @return JWT信息
     */
    public static Jwt getJwt(boolean required) {
        Jwt jwt = JWT_CONTEXT.get();
        return required ? Objects.requireNonNull(jwt, "Jwt has not been initialized") : jwt;
    }

    /**
     * 获取主体身份
     *
     * @return 主体身份标识
     */
    public static Long getId() {
        return getId(getJwt(false));
    }

    /**
     * 获取主体身份
     *
     * @param jwt JWT对象
     * @return 主体身份标识
     */
    public static Long getId(Jwt jwt) {
        return ObjectUtils.ifNull(jwt, t -> ObjectUtils.object2long(t.getClaim(JWT_CLAIM_ID)));
    }

    /**
     * 获取主体身份
     *
     * @param token OAuth2访问令牌对象
     * @return 主体身份标识
     */
    public static Long getId(OAuth2AccessToken token) {
        return ObjectUtils.ifNull(
                ObjectUtils.ifNull(token, OAuth2AccessToken::getAdditionalInformation),
                info -> ObjectUtils.object2long(info.get(JWT_CLAIM_ID))
        );
    }

    /**
     * 获取JWT身份标识
     *
     * @return JWT身份标识
     */
    public static String getJti() {
        return getJti(getJwt(false));
    }

    /**
     * 获取JWT身份标识
     *
     * @param jwt JWT对象
     * @return JWT身份标识
     */
    public static String getJti(Jwt jwt) {
        return jwt == null ? null : StringUtils.ifEmpty(jwt.getClaimAsString(JWT_CLAIM_JTI), (String) null);
    }

    /**
     * 获取用户标识
     *
     * @return 用户标识
     */
    public static String getUser() {
        return getUser(getJwt(false));
    }

    /**
     * 获取用户标识
     *
     * @param jwt JWT对象
     * @return 用户标识
     */
    public static String getUser(Jwt jwt) {
        return jwt == null ? null : StringUtils.ifEmpty(jwt.getClaimAsString(JWT_CLAIM_USER_NAME), (String) null);
    }

    /**
     * 获取手机号
     *
     * @return 手机号
     */
    public static String getMobile() {
        return getMobile(getJwt(false));
    }

    /**
     * 获取手机号
     *
     * @param jwt JWT对象
     * @return 手机号
     */
    public static String getMobile(Jwt jwt) {
        return jwt == null ? null : StringUtils.ifEmpty(jwt.getClaimAsString(JWT_CLAIM_MOBILE), (String) null);
    }

    /**
     * 获取手机号
     *
     * @param token OAuth2访问令牌对象
     * @return 手机号
     */
    public static String getMobile(OAuth2AccessToken token) {
        return ObjectUtils.ifNull(
                ObjectUtils.ifNull(token, OAuth2AccessToken::getAdditionalInformation),
                info -> StringUtils.ifEmpty((String) info.get(JWT_CLAIM_MOBILE), (String) null)
        );
    }

    /**
     * 获取业务类型
     *
     * @return 业务类型
     */
    public static String getBusiness() {
        return getBusiness(getJwt(false));
    }

    /**
     * 获取业务类型
     *
     * @param jwt JWT对象
     * @return 业务类型
     */
    public static String getBusiness(Jwt jwt) {
        return jwt == null ? null : StringUtils.ifEmpty(jwt.getClaimAsString(JWT_CLAIM_BUSINESS), (String) null);
    }

    /**
     * 获取业务类型
     *
     * @param id            会话ID
     * @param redisTemplate Redis操作模版实例
     * @return 业务类型
     */
    public static String getBusiness(@NonNull Serializable id, @NonNull RedisTemplate<String, Object> redisTemplate) {
        return (String) redisTemplate.opsForHash().get(buildAccessTokenKey(id), "business");
    }

    /**
     * 获取业务类型
     *
     * @param token OAuth2访问令牌对象
     * @return 业务类型
     */
    public static String getBusiness(OAuth2AccessToken token) {
        return ObjectUtils.ifNull(
                ObjectUtils.ifNull(token, OAuth2AccessToken::getAdditionalInformation),
                info -> StringUtils.ifEmpty((String) info.get(JWT_CLAIM_BUSINESS), (String) null)
        );
    }

    /**
     * 获取客户端标识
     *
     * @return 客户端标识
     */
    public static String getClient() {
        return getClient(getJwt(false));
    }

    /**
     * 获取客户端标识
     *
     * @param jwt JWT对象
     * @return 客户端标识
     */
    public static String getClient(Jwt jwt) {
        return jwt == null ? null : StringUtils.ifEmpty(jwt.getClaimAsString(JWT_CLAIM_CLIENT_ID), (String) null);
    }

    /**
     * 获取用户权限集合
     *
     * @return 权限标识集合
     */
    public static Set<String> getAuthorities() {
        return getAuthorities(getJwt(false));
    }

    /**
     * 获取用户权限集合
     *
     * @param jwt JWT对象
     * @return 权限标识集合
     */
    public static Set<String> getAuthorities(Jwt jwt) {
        List<String> authorities = jwt == null ? null : jwt.getClaimAsStringList(JWT_CLAIM_AUTHORITIES);
        return CollectionUtils.isEmpty(authorities) ? Collections.emptySet() : Sets.newHashSet(authorities);
    }

    /**
     * 构建令牌标识
     *
     * @param id 会话ID
     * @return 令牌标识
     */
    public static String buildAccessTokenKey(@NonNull Serializable id) {
        return buildAccessTokenKey(id, SessionContextHolder.getTerminal());
    }

    /**
     * 构建令牌标识
     *
     * @param id       会话ID
     * @param terminal 终端信息
     * @return 令牌标识
     */
    public static String buildAccessTokenKey(@NonNull Serializable id, @NonNull Terminal terminal) {
        return String.format("token:%s:%s", id, terminal);
    }

    /**
     * 存储访问令牌
     *
     * @param token         token OAuth2访问令牌对象
     * @param redisTemplate Redis操作模版实例
     */
    public static void storeAccessToken(@NonNull OAuth2AccessToken token,
                                        @NonNull RedisTemplate<String, Object> redisTemplate) {
        String key = buildAccessTokenKey(getId(token));
        if (log.isDebugEnabled()) {
            log.debug("Saving redis token: {}", key);
        }
        Map<String, Object> context = Maps.newHashMapWithExpectedSize(2);
        context.put("value", token.getValue());
        String business = getBusiness(token);
        if (StringUtils.nonEmpty(business)) {
            context.put("business", business);
        }
        long millis = Math.max(token.getExpiration().getTime() - System.currentTimeMillis(), 0);
        RedisUtils.hmset(redisTemplate, key, context, Duration.ofMillis(millis));
    }

    /**
     * 删除访问令牌
     *
     * @param token         token OAuth2访问令牌对象
     * @param redisTemplate Redis操作模版实例
     */
    public static void removeAccessToken(@NonNull OAuth2AccessToken token,
                                         @NonNull RedisTemplate<String, Object> redisTemplate) {
        String key = buildAccessTokenKey(getId(token));
        if (log.isDebugEnabled()) {
            log.debug("Saving redis token: {}", key);
        }
        redisTemplate.delete(key);
    }

    /**
     * 获取访问令牌
     *
     * @param jwt           JWT实例
     * @param redisTemplate Redis操作模版实例
     * @return 访问令牌
     */
    public static String getAccessToken(@NonNull Jwt jwt, @NonNull RedisTemplate<String, Object> redisTemplate) {
        return (String) redisTemplate.opsForHash().get(buildAccessTokenKey(getId(jwt)), "value");
    }

    /**
     * 获取上下文中角色/权限映射表
     *
     * @return 角色/权限映射表
     */
    public static Map<String, Set<String>> getContextPermissions() {
        String permission = ApplicationContextHolder.getSafeProperty(PERMISSION_PROPERTY);
        return StringUtils.isEmpty(permission) ? Collections.emptyMap() :
                JacksonContextHolder.deserialize(permission, PERMISSION_TYPE_REFERENCE);
    }

    /**
     * 判断当前用户是否拥有指定资源权限
     *
     * @param method 请求方式
     * @param uri    资源地址
     * @return true/false
     */
    public static boolean hasPermission(String method, String uri) {
        if (StringUtils.isEmpty(uri)) {
            return false;
        }

        // 基于用户角色及对应的资源判断是否拥有权限
        Set<String> roles = getAuthorities(getJwt(false));
        if (CollectionUtils.isEmpty(roles)) {
            return false;
        }
        Map<String, Set<String>> permissions = getContextPermissions();
        if (CollectionUtils.isEmpty(permissions)) {
            return false;
        }
        return roles.stream().anyMatch(role -> {
            Set<String> resources = permissions.get(role);
            return !CollectionUtils.isEmpty(resources) && resources.stream().anyMatch(resource -> {
                if (StringUtils.isEmpty(resource) || (resource = resource.trim()).isEmpty()) {
                    return false;
                }
                String[] sections = METHOD_RESOURCE_SPLIT_PATTERN.split(resource);
                if (sections.length == 1) {
                    return StringUtils.isPath(sections[0], uri);
                }
                return Objects.equals(method, sections[0]) && StringUtils.isPath(sections[1], uri);
            });
        });
    }

    /**
     * 获取数字签名工具类实例
     *
     * @return 数字签名工具类实例
     */
    public static HmacUtils getHmacUtils() {
        return Objects.requireNonNull(HMAC_UTILS, "Hmac utils has not been initialized");
    }

    /**
     * 数字签名
     *
     * @param parameters 签名参数
     * @return 签名信息
     */
    public static String sign(Map<?, ?> parameters) {
        Session session = SessionContextHolder.getSession();
        StringBuilder plaintext = new StringBuilder().append(session.getTenant()).append(session.getTerminal())
                .append(session.getId());
        if (parameters instanceof TreeMap) {
            parameters.forEach((key, value) -> plaintext.append(key).append(value));
        } else if (!CollectionUtils.isEmpty(parameters)) {
            List<?> keys = Lists.newArrayList(parameters.keySet());
            keys.sort(Comparator.comparing(Object::toString));
            keys.forEach(key -> plaintext.append(key).append(parameters.get(key)));
        }
        return getHmacUtils().hmacHex(plaintext.toString()) + System.currentTimeMillis();
    }

    /**
     * 验证数字签名
     *
     * @param signature  数字签名信息
     * @param parameters 签名参数
     * @return true/false
     */
    public static boolean validateSignature(String signature, Map<?, ?> parameters) {
        return !StringUtils.isEmpty(signature)
                && Objects.equals(signature.substring(0, 64), sign(parameters).substring(0, 64));
    }

    /**
     * 清空上下文信息
     */
    public static void clearSecurityContext() {
        JWT_CONTEXT.remove();
    }

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        HMAC_UTILS = new HmacUtils(HmacAlgorithms.HMAC_SHA_256, this.securityProperties.getSignature().getSecret());
    }

    @Override
    public void requestDestroyed(ServletRequestEvent event) {
        clearSecurityContext();
    }
}
