package com.chelvc.framework.security.interceptor;

import java.lang.reflect.Parameter;
import java.util.Collections;
import java.util.Map;
import java.util.Set;
import javax.annotation.PostConstruct;

import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.chelvc.framework.security.annotation.Sign;
import com.chelvc.framework.security.annotation.Signable;
import com.chelvc.framework.security.signature.Signer;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.aspectj.lang.JoinPoint;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.Before;
import org.aspectj.lang.reflect.MethodSignature;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.stereotype.Component;

/**
 * 方法签名验证拦截器
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Aspect
@Component
@ConditionalOnBean(Signer.class)
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class SignatureValidateInterceptor {
    private final Signer signer;

    @PostConstruct
    public void initialize() {
        log.info("Loading signer: {}", this.signer);
    }

    /**
     * 接口签名验证
     *
     * @param point 接口签名拦截点
     * @param sign  接口签名注解实例
     */
    @Before("@annotation(sign)")
    public void validate(JoinPoint point, Sign sign) {
        // 获取接口签名信息
        String signature = SessionContextHolder.getSession().getSignature();
        if (StringUtils.isEmpty(signature)) {
            throw new AccessDeniedException("Signature unspecified");
        }

        // 初始化签名参数
        Object[] args = point.getArgs();
        Parameter[] parameters = ((MethodSignature) point.getSignature()).getMethod().getParameters();
        Map<String, Object> signatures = args.length == 0 ? Collections.emptyMap() : Maps.newTreeMap(String::compareTo);
        for (int i = 0; i < parameters.length; i++) {
            Object arg = args[i];
            Signable signable = parameters[i].getAnnotation(Signable.class);
            if (arg == null || signable == null) {
                continue;
            }
            Class<?> clazz = arg.getClass();
            Set<String> excludes = signable.exclude().length == 0 ?
                    Collections.emptySet() : Sets.newHashSet(signable.exclude());
            ObjectUtils.iterateFields(clazz, (index, field) -> {
                String name = field.getName();
                if (!excludes.contains(name)) {
                    signatures.put(name, ObjectUtils.getValue(arg, field));
                }
            });
        }

        // 验证签名
        if (!this.signer.validate(signature, signatures)) {
            throw new AccessDeniedException("Invalid signature");
        }
    }
}
