package com.chelvc.framework.security.signature;

import java.nio.charset.StandardCharsets;
import java.util.Map;
import java.util.concurrent.TimeUnit;

import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.redis.util.RedisUtils;
import com.chelvc.framework.security.context.SecurityContextHolder;
import lombok.NonNull;
import org.springframework.data.redis.connection.RedisStringCommands;
import org.springframework.data.redis.core.types.Expiration;
import org.springframework.util.Assert;

/**
 * 基于Redis的数字签名实现
 *
 * @author Woody
 * @date 2023/4/5
 */
public class RedisSigner implements Signer {
    /**
     * 数字签名Rdids Key前缀
     */
    private static final String SIGNATURE_REDIS_PREFIX = "signature:";

    @Override
    public String sign(Map<?, ?> parameters, int expiration) {
        Assert.isTrue(expiration > 0, "expiration must greater than 0");
        String signature = SecurityContextHolder.sign(parameters);
        byte[] key = (SIGNATURE_REDIS_PREFIX + signature).getBytes(StandardCharsets.UTF_8);
        Boolean success = RedisUtils.execute(connection -> connection.set(key, ObjectUtils.EMPTY_BYTES,
                Expiration.from(expiration, TimeUnit.SECONDS), RedisStringCommands.SetOption.SET_IF_ABSENT));
        if (!Boolean.TRUE.equals(success)) {
            throw new RuntimeException("Sign failed");
        }
        return signature;
    }

    @Override
    public boolean validate(@NonNull String signature, @NonNull Map<?, ?> parameters) {
        if (SecurityContextHolder.validateSignature(signature, parameters)) {
            byte[] key = (SIGNATURE_REDIS_PREFIX + signature).getBytes(StandardCharsets.UTF_8);
            Long result = RedisUtils.execute(connection -> connection.del(key));
            return result != null && result > 0;
        }
        return false;
    }
}
