package com.chelvc.framework.security.token;

import java.util.Objects;
import java.util.concurrent.TimeUnit;
import javax.servlet.http.HttpServletRequest;

import com.chelvc.framework.security.context.SecurityContextHolder;
import com.chelvc.framework.security.interceptor.TokenReplayValidateInterceptor;
import lombok.NonNull;
import org.apache.commons.codec.digest.DigestUtils;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.core.ValueOperations;
import org.springframework.security.oauth2.jwt.Jwt;
import org.springframework.web.method.HandlerMethod;

/**
 * 基于Redis的令牌重放验证实现
 *
 * @author Woody
 * @date 2023/4/5
 * @see TokenReplayValidator
 * @see TokenReplayValidateInterceptor
 */
public class RedisTokenReplayValidator implements TokenReplayValidator {
    private final RedisTemplate<String, Object> redisTemplate;

    public RedisTokenReplayValidator(@NonNull RedisTemplate<String, Object> redisTemplate) {
        this.redisTemplate = redisTemplate;
    }

    @Override
    public boolean isReplayed(HttpServletRequest request, HandlerMethod method, Jwt jwt) {
        if (request == null || method == null || jwt == null) {
            return false;
        }
        // 根据服务名称、请求方法、请求令牌ID构建唯一标识
        String key = DigestUtils.sha256Hex(String.join(":",
                request.getServerName(),
                method.getBeanType().getName(),
                method.getMethod().getName(),
                SecurityContextHolder.getJti(jwt)
        ));
        Long id = SecurityContextHolder.getId(jwt);
        long timeout = Objects.requireNonNull(jwt.getExpiresAt()).toEpochMilli() - System.currentTimeMillis();
        ValueOperations<String, Object> operations = this.redisTemplate.opsForValue();
        return Boolean.FALSE.equals(operations.setIfAbsent(key, id, timeout, TimeUnit.MILLISECONDS));
    }
}
