package com.chelvc.framework.sms.support;

import java.util.Map;
import java.util.Objects;

import com.aliyun.dysmsapi20170525.Client;
import com.aliyun.dysmsapi20170525.models.SendSmsRequest;
import com.aliyun.dysmsapi20170525.models.SendSmsResponse;
import com.aliyun.dysmsapi20170525.models.SendSmsResponseBody;
import com.chelvc.framework.base.context.JacksonContextHolder;
import com.chelvc.framework.base.exception.ResourceUnavailableException;
import com.chelvc.framework.sms.TemplateSmsHandler;
import com.chelvc.framework.sms.config.AliyunSmsProperties;
import com.google.common.collect.ImmutableMap;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.stereotype.Component;

/**
 * 基于阿里云短信操作实现
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Component
@ConditionalOnBean(AliyunSmsProperties.class)
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class AliyunSmsHandler implements TemplateSmsHandler {
    private final Client client;
    private final AliyunSmsProperties properties;

    @Override
    public void send(String mobile, String template, String... parameters) {
        throw new UnsupportedOperationException("Not support method");
    }

    @Override
    public void send(@NonNull String mobile, @NonNull String template, @NonNull Map<String, ?> parameters) {
        String param = JacksonContextHolder.serialize(parameters);
        boolean debug = log.isDebugEnabled();
        if (debug) {
            log.debug("Aliyun sms request: {}, {}", mobile, param);
        }
        SendSmsRequest request = new SendSmsRequest().setPhoneNumbers(mobile)
                .setSignName(this.properties.getSignature()).setTemplateCode(template).setTemplateParam(param);
        SendSmsResponse response;
        try {
            response = this.client.sendSms(request);
        } catch (Exception e) {
            log.warn("Aliyun sms send failed: {}", e.getMessage());
            return;
        }
        if (debug) {
            log.debug("Aliyun sms response: {}", JacksonContextHolder.serialize(response));
        }
        SendSmsResponseBody body = response.getBody();
        if (!Objects.equals(body.getCode(), "OK")) {
            log.warn("Aliyun sms send failed: [{}] {}", body.getCode(), body.getMessage());
            if (Objects.equals(body.getCode(), "isv.BUSINESS_LIMIT_CONTROL")) {
                throw new ResourceUnavailableException(SMS_SEND_FREQUENTLY, body.getMessage());
            }
            throw new ResourceUnavailableException(body.getMessage());
        }
    }

    @Override
    public void sendCaptcha(@NonNull String mobile, @NonNull String template, @NonNull String code) {
        this.send(mobile, template, ImmutableMap.of("code", code));
    }
}
