package com.chelvc.framework.sms.support;

import java.time.Duration;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;

import com.chelvc.framework.base.context.ApplicationContextHolder;
import com.chelvc.framework.base.exception.ResourceUnavailableException;
import com.chelvc.framework.base.util.StringUtils;
import com.chelvc.framework.redis.util.RedisUtils;
import com.chelvc.framework.sms.Captcha;
import com.chelvc.framework.sms.CaptchaSmsHandler;
import com.chelvc.framework.sms.SmsSession;
import com.chelvc.framework.sms.TemplateSmsHandler;
import com.chelvc.framework.sms.config.CaptchaSmsProperties;
import com.fasterxml.jackson.core.type.TypeReference;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;

/**
 * 短信验证码操作默认实现
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Component
@ConditionalOnBean(CaptchaSmsProperties.class)
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class DefaultCaptchaSmsHandler implements CaptchaSmsHandler {
    /**
     * 验证码配置引用
     */
    private static final TypeReference<List<Captcha>> CAPTCHA_REFERENCE = new TypeReference<List<Captcha>>() {
    };

    private final CaptchaSmsProperties properties;
    private final RedisTemplate<String, Object> redisTemplate;

    /**
     * 构建令牌Redis标识
     *
     * @param mobile 手机号
     * @return Redis标识
     */
    private String key(String mobile) {
        return "sms:token:" + mobile;
    }

    /**
     * 获取模版短信处理器
     *
     * @return 模版短信处理器
     */
    private TemplateSmsHandler getTemplateSmsHandler() {
        return ApplicationContextHolder.getBean(ApplicationContextHolder.getSafeProperty(
                "platform.sms.captcha.handler", AliyunSmsHandler.class.getName()
        ));
    }

    /**
     * 查找手机号测试验证码
     *
     * @param mobile 手机号
     * @return 验证码
     */
    private String lookupTestingCaptcha(String mobile) {
        List<Captcha> testes = ApplicationContextHolder.getSafeProperty("platform.sms.captcha.test", CAPTCHA_REFERENCE);
        if (CollectionUtils.isEmpty(testes)) {
            return null;
        }
        return testes.stream().filter(captcha -> Objects.equals(captcha.getMobile(), mobile)).map(Captcha::getCode)
                .filter(StringUtils::nonEmpty).findFirst().orElse(null);
    }

    @Override
    public SmsSession send(@NonNull String mobile) {
        // 如果目标手机号属于测试手机号，则直接返回配置的验证码信息，否则获取真实的验证码
        String test = this.lookupTestingCaptcha(mobile);
        boolean testing = StringUtils.nonEmpty(test);
        String code = testing ? test : String.valueOf(ThreadLocalRandom.current().nextInt(100000, 1000000));
        Duration duration = Duration.ofSeconds(this.properties.getInterval());
        return RedisUtils.tryLockAround("sms:interval:" + mobile, duration, () -> {
            if (!testing) {
                this.getTemplateSmsHandler().sendCaptcha(mobile, this.properties.getTemplate(), code);
            }
            String key = this.key(mobile), token = StringUtils.uuid();
            Captcha captcha = Captcha.builder().token(token).code(code).mobile(mobile).build();
            this.redisTemplate.opsForValue().set(key, captcha, this.properties.getExpiration(), TimeUnit.SECONDS);
            return SmsSession.builder().token(token).expiration(this.properties.getExpiration()).build();
        }, () -> {
            throw new ResourceUnavailableException(TemplateSmsHandler.SMS_SEND_FREQUENTLY, "短信发送过于频繁，请稍后再试");
        });
    }

    @Override
    public Captcha getCaptcha(@NonNull String mobile) {
        return (Captcha) this.redisTemplate.opsForValue().get(this.key(mobile));
    }

    @Override
    public void using(@NonNull String mobile, @NonNull Consumer<Captcha> consumer) {
        Captcha captcha = this.getCaptcha(mobile);
        consumer.accept(captcha);
        if (captcha != null) {
            this.redisTemplate.delete(this.key(mobile));
        }
    }

    @Override
    public boolean check(@NonNull String token, @NonNull String mobile, @NonNull String code) {
        Captcha captcha = this.getCaptcha(mobile);
        return captcha != null && Objects.equals(token, captcha.getToken())
                && Objects.equals(mobile, captcha.getMobile()) && Objects.equals(code, captcha.getCode());
    }

    @Override
    public boolean validate(@NonNull String token, @NonNull String mobile, @NonNull String code) {
        boolean available = this.check(token, mobile, code);
        if (available) {
            this.redisTemplate.delete(this.key(mobile));
        }
        return available;
    }
}
