package com.chelvc.framework.sms.support;

import java.util.Arrays;
import java.util.Map;

import com.chelvc.framework.base.context.JacksonContextHolder;
import com.chelvc.framework.base.exception.ResourceUnavailableException;
import com.chelvc.framework.sms.TemplateSmsHandler;
import com.chelvc.framework.sms.config.TencentSmsProperties;
import com.github.qcloudsms.SmsSingleSender;
import com.github.qcloudsms.SmsSingleSenderResult;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.stereotype.Component;

/**
 * 基于腾讯云的模版短信操作实现
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Component
@ConditionalOnBean(TencentSmsProperties.class)
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class TencentSmsHandler implements TemplateSmsHandler {
    private final SmsSingleSender sender;
    private final TencentSmsProperties properties;

    @Override
    public void send(@NonNull String mobile, @NonNull String template, @NonNull String... parameters) {
        boolean debug = log.isDebugEnabled();
        if (debug) {
            log.debug("Tencent sms request: {}, {}", mobile, Arrays.toString(parameters));
        }
        SmsSingleSenderResult result;
        try {
            result = this.sender.sendWithParam(this.properties.getRegion(), mobile, Integer.parseInt(template),
                    parameters, this.properties.getSignature(), null, null);
        } catch (Exception e) {
            log.warn("Tencent sms send failed: {}", e.getMessage());
            return;
        }
        if (debug) {
            log.debug("Tencent sms response: {}", JacksonContextHolder.serialize(result));
        }
        if (result.result != 0) {
            log.warn("Tencent sms send failed: [{}] {}", result.result, result.errMsg);
            if (result.result == 1023) {
                throw new ResourceUnavailableException(SMS_SEND_FREQUENTLY, result.errMsg);
            }
            throw new ResourceUnavailableException(result.errMsg);
        }
    }

    @Override
    public void send(String mobile, String template, Map<String, ?> parameters) {
        throw new UnsupportedOperationException("Not support method");
    }

    @Override
    public void sendCaptcha(@NonNull String mobile, @NonNull String template, @NonNull String code) {
        this.send(mobile, template, code);
    }
}
