package com.chelvc.framework.wechat;

import java.time.Duration;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.UUID;
import java.util.concurrent.TimeUnit;
import java.util.function.Supplier;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;

import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.base.util.DecimalUtils;
import com.chelvc.framework.base.util.HostUtils;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.chelvc.framework.redis.util.RedisUtils;
import com.chelvc.framework.wechat.config.WechatProperties;
import com.chelvc.framework.wechat.util.WechatUtils;
import com.github.wxpay.sdk.WXPayUtil;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.connection.RedisStandaloneConfiguration;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;
import org.springframework.web.client.RestTemplate;

/**
 * 微信处理默认实现
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Component
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class DefaultWechatHandler implements WechatHandler {
    /**
     * Scheme过期时间（天）
     */
    private static final int SCHEME_EXPIRE_INTERVAL = 30;

    /**
     * 访问令牌标识前缀
     */
    private static final String ACCESS_TOKEN_PREFIX = "wechat:token:";

    /**
     * 获取用户信息接口地址
     */
    private static final String USER_URL =
            "https://api.weixin.qq.com/cgi-bin/user/info?access_token=%s&openid=%s&lang=zh_CN";

    /**
     * 微信下单接口地址
     */
    private static final String UNIFIEDORDER_URL = "https://api.mch.weixin.qq.com/pay/unifiedorder";

    /**
     * 微信获取手机号接口地址
     */
    private static final String CODE2MOBILE_URL =
            "https://api.weixin.qq.com/wxa/business/getuserphonenumber?access_token=%s";

    /**
     * 微信登录接口地址
     */
    private static final String JSCODE2SESSION_URL =
            "https://api.weixin.qq.com/sns/jscode2session?appid=%s&secret=%s&js_code=%s&grant_type=authorization_code";

    /**
     * 获取访问令牌接口地址
     */
    private static final String ACCESS_TOKEN_URL =
            "https://api.weixin.qq.com/cgi-bin/token?grant_type=client_credential&appid=%s&secret=%s";

    /**
     * 获取小程序schemeCode接口地址
     */
    private static final String GENERATE_SCHEME_URL = "https://api.weixin.qq.com/wxa/generatescheme?access_token=%s";

    private final RestTemplate restTemplate;
    private final RedisTemplate<String, Object> redisTemplate;
    private final WechatProperties properties;
    private final Map<String, String> tokens = Maps.newHashMap();
    private RedisTemplate<String, Object> cache;

    /**
     * 获取缓存RedisTemplate实例
     *
     * @return RedisTemplate实例
     */
    private RedisTemplate<String, Object> getCache() {
        // 获取缓存配置
        WechatProperties.Cache config = this.properties.getCache();
        String host = ObjectUtils.ifNull(config, WechatProperties.Cache::getHost);
        Integer port = ObjectUtils.ifNull(config, WechatProperties.Cache::getPort);
        if (StringUtils.isEmpty(host) || Objects.isNull(port)) {
            return null;
        }

        // 初始化RedisTemplate实例
        if (this.cache == null) {
            synchronized (this) {
                if (this.cache == null) {
                    String password = ObjectUtils.ifNull(config, WechatProperties.Cache::getPassword);
                    Integer database = ObjectUtils.ifNull(config, WechatProperties.Cache::getDatabase, () -> 0);
                    RedisStandaloneConfiguration configuration = new RedisStandaloneConfiguration();
                    configuration.setPort(port);
                    configuration.setHostName(host);
                    configuration.setPassword(password);
                    configuration.setDatabase(database);
                    this.cache = RedisUtils.template(configuration);
                }
            }
        }
        return this.cache;
    }

    /**
     * 获取支付配置
     *
     * @param type 支付类型
     * @return 支付配置
     */
    private WechatProperties.Payment getPayment(String type) {
        List<WechatProperties.Payment> payments = this.properties.getPayments();
        WechatProperties.Payment payment = CollectionUtils.isEmpty(payments) ? null :
                payments.stream().filter(config -> Objects.equals(config.getType(), type)).findAny().orElse(null);
        return Objects.requireNonNull(payment, "Not support payment type: " + type);
    }

    /**
     * 生成支付参数
     *
     * @param payment 支付配置
     * @param request 支付请求参数
     * @return 支付参数键/值对
     */
    private Map<String, String> generatePaymentParameters(WechatProperties.Payment payment, WechatPayRequest request) {
        String mchid = Objects.requireNonNull(payment.getMchid(), "Wechat mchid unspecified");
        String callback = Objects.requireNonNull(payment.getCallback(), "Wechat callback unspecified");
        Map<String, String> parameters = Maps.newHashMap();
        parameters.put("appid", payment.getAppid());
        parameters.put("mch_id", mchid);
        parameters.put("body", request.getComment());
        parameters.put("out_trade_no", request.getOrder());
        if (!StringUtils.isEmpty(request.getContext())) {
            parameters.put("attach", request.getContext());
        }
        parameters.put("fee_type", "CNY");
        parameters.put("total_fee", String.valueOf(request.getAmount().multiply(DecimalUtils.HUNDRED).intValue()));
        parameters.put("spbill_create_ip", HostUtils.LOCAL_ADDRESS);
        parameters.put("notify_url", callback);
        parameters.put("trade_type", payment.getMode().name());
        parameters.put("nonce_str", WXPayUtil.generateNonceStr());
        return parameters;
    }

    /**
     * 参数签名
     *
     * @param payment    支付配置
     * @param parameters 签名参数
     * @return 签名信息
     */
    private String sign(WechatProperties.Payment payment, Map<String, String> parameters) {
        String mchkey = Objects.requireNonNull(payment.getMchkey(), "Wechat mchkey unspecified");
        return WechatUtils.sign(mchkey, parameters);
    }

    @Override
    public String getAppid() {
        return this.properties.getAppid();
    }

    @Override
    public String getAccessToken() {
        return this.getAccessToken(this.properties.getAppid(), this.properties.getSecret());
    }

    @Override
    public String refreshAccessToken() {
        return this.refreshAccessToken(this.properties.getAppid(), this.properties.getSecret());
    }

    @Override
    public String getAccessToken(@NonNull String appid, @NonNull String secret) {
        String key = ACCESS_TOKEN_PREFIX + appid;
        String token = this.tokens.get(key);
        if (StringUtils.isEmpty(token)) {
            synchronized (key.intern()) {
                if (StringUtils.isEmpty(token = this.tokens.get(key))) {
                    return this.refreshAccessToken(appid, secret);
                }
            }
        }
        return token;
    }

    @Override
    public String refreshAccessToken(@NonNull String appid, @NonNull String secret) {
        String key = ACCESS_TOKEN_PREFIX + appid;
        RedisTemplate<String, Object> cache = this.getCache();
        synchronized (key.intern()) {
            String token = cache == null ? null : (String) cache.opsForValue().get(key);
            if (StringUtils.isEmpty(token)
                    && (StringUtils.isEmpty(token = (String) this.redisTemplate.opsForValue().get(key))
                    || Objects.equals(token, this.tokens.get(key)))) {
                String url = String.format(ACCESS_TOKEN_URL, appid, secret);
                if (log.isDebugEnabled()) {
                    log.debug("Wechat request: {}", url);
                }
                token = RedisUtils.lockAround(key + ":lock", () -> {
                    WechatAccessToken accessToken =
                            WechatUtils.execute(() -> this.restTemplate.getForObject(url, WechatAccessToken.class));
                    long duration = accessToken.getDuration() - 60;
                    this.redisTemplate.opsForValue().set(key, accessToken.getToken(), duration, TimeUnit.SECONDS);
                    return accessToken.getToken();
                }, () -> {
                    throw new IllegalStateException("Get wechat access token timeout");
                });
            }
            this.tokens.put(key, token);
            return token;
        }
    }

    @Override
    public String getScheme() {
        return this.getScheme(null, null, null);
    }

    @Override
    public String getScheme(String path, String query, String environment) {
        // 尝试从Redis中获取Scheme
        path = StringUtils.ifEmpty(path, StringUtils.EMPTY);
        query = StringUtils.ifEmpty(query, StringUtils.EMPTY);
        environment = StringUtils.ifEmpty(environment, StringUtils.EMPTY);
        HttpSession session = ObjectUtils.ifNull(SessionContextHolder.getRequest(), HttpServletRequest::getSession);
        String id = ObjectUtils.ifNull(
                ObjectUtils.ifNull(session, HttpSession::getId), () -> UUID.randomUUID().toString()
        );
        String key = String.format("wechat:scheme:%s:%s:%s:%s", id, environment, path, query);
        String scheme = (String) this.redisTemplate.opsForValue().get(key);
        if (StringUtils.nonEmpty(scheme)) {
            return scheme;
        }

        // 重新获取Scheme
        Map<String, Object> body = ImmutableMap.of(
                "jump_wxa", ImmutableMap.of(
                        "path", path,
                        "query", query,
                        "env_version", environment
                ),
                "expire_type", 1,
                "expire_interval", SCHEME_EXPIRE_INTERVAL
        );
        String url = String.format(GENERATE_SCHEME_URL, this.getAccessToken());
        if (log.isDebugEnabled()) {
            log.debug("Wechat request: {}, body: {}", url, body);
        }
        HttpHeaders headers = new HttpHeaders();
        headers.setContentType(MediaType.APPLICATION_JSON);
        HttpEntity<?> entity = new HttpEntity<>(body, headers);
        scheme = WechatUtils.execute(
                () -> this.restTemplate.exchange(url, HttpMethod.POST, entity, WechatScheme.class).getBody(),
                (status, message) -> {
                    // 如果访问令牌过期或无效则更新访问令牌后重试
                    if (status == 40001 || status == 42001) {
                        return WechatUtils.execute(() -> this.restTemplate.exchange(
                                String.format(GENERATE_SCHEME_URL, this.refreshAccessToken()),
                                HttpMethod.POST, entity, WechatScheme.class
                        ).getBody());
                    }
                    throw new RuntimeException(String.format("[%d] %s", status, message));
                }
        ).getOpenlink();
        this.redisTemplate.opsForValue().set(key, scheme, Duration.ofDays(SCHEME_EXPIRE_INTERVAL - 1));
        return scheme;
    }

    @Override
    public WechatUser getUser(@NonNull String accessToken, @NonNull String openid) {
        return this.getUser(accessToken, openid, this::refreshAccessToken);
    }

    @Override
    public WechatUser getUser(@NonNull String accessToken, @NonNull String openid,
                              @NonNull Supplier<String> tokenRefresher) {
        String url = String.format(USER_URL, accessToken, openid);
        if (log.isDebugEnabled()) {
            log.debug("Wechat request: {}", url);
        }
        return WechatUtils.execute(
                () -> this.restTemplate.getForObject(url, WechatUser.class),
                (status, message) -> {
                    // 如果访问令牌过期或无效则更新访问令牌后重试
                    if (status == 40001 || status == 42001) {
                        return WechatUtils.execute(() -> this.restTemplate.getForObject(
                                String.format(USER_URL, tokenRefresher.get(), openid), WechatUser.class
                        ));
                    }
                    throw new RuntimeException(String.format("[%d] %s", status, message));
                });
    }

    @Override
    public String code2mobile(@NonNull String code) {
        String url = String.format(CODE2MOBILE_URL, this.getAccessToken());
        if (log.isDebugEnabled()) {
            log.debug("Wechat request: {}", url);
        }
        WechatMobile.Mobile mobile = WechatUtils.execute(
                () -> this.restTemplate.postForObject(url, ImmutableMap.of("code", code), WechatMobile.class),
                (status, message) -> {
                    // 如果访问令牌过期或无效则更新访问令牌后重试
                    if (status == 40001 || status == 42001) {
                        return WechatUtils.execute(() -> this.restTemplate.postForObject(
                                String.format(CODE2MOBILE_URL, this.refreshAccessToken()),
                                ImmutableMap.of("code", code), WechatMobile.class
                        ));
                    }
                    throw new RuntimeException(String.format("[%d] %s", status, message));
                }
        ).getMobile();
        return Objects.requireNonNull(mobile).getNumber();
    }

    @Override
    public WechatSession code2session(@NonNull String code) {
        String url = String.format(JSCODE2SESSION_URL, this.properties.getAppid(), this.properties.getSecret(), code);
        if (log.isDebugEnabled()) {
            log.debug("Wechat request: {}", url);
        }
        return WechatUtils.execute(() -> this.restTemplate.getForObject(url, WechatSession.class));
    }

    @Override
    public String sign(@NonNull String type, @NonNull Map<String, String> parameters) {
        return this.sign(this.getPayment(type), parameters);
    }

    @Override
    public WechatUnifiedOrder unifiedorder(@NonNull String type, @NonNull WechatPayRequest request) {
        // 获取支付配置
        WechatProperties.Payment payment = this.getPayment(type);

        // 构建请求参数
        Map<String, String> parameters = this.generatePaymentParameters(payment, request);

        // 如果是小程序支付则需要指定openid
        if (payment.getMode() == PayMode.JSAPI) {
            parameters.put("openid", Objects.requireNonNull(request.getOpenid(), "openid unspecified"));
        }

        boolean debug = log.isDebugEnabled();
        if (debug) {
            log.debug("Wechat parameters: {}", parameters);
        }

        // 请求参数签名
        parameters.put("sign", this.sign(payment, parameters));

        // 将请求参数转换成xml请求体
        String body = WechatUtils.map2xml(parameters);
        if (debug) {
            log.debug("Wechat request: {}, {}", UNIFIEDORDER_URL, body);
        }

        // 发起下单请求
        HttpHeaders headers = new HttpHeaders();
        headers.setContentType(MediaType.TEXT_XML);
        HttpEntity<?> entity = new HttpEntity<>(body, headers);
        String response = this.restTemplate.exchange(UNIFIEDORDER_URL, HttpMethod.POST, entity, String.class).getBody();
        if (debug) {
            log.debug("Wechat response: {}", response);
        }

        // 验证相应结果
        Map<String, String> result = ObjectUtils.ifNull(response, WechatUtils::xml2map);
        if (StringUtils.isEmpty(response) || !Objects.equals(result.get("return_code"), "SUCCESS")
                || !Objects.equals(result.get("result_code"), "SUCCESS")
                || !Objects.equals(result.remove("sign"), this.sign(payment, result))) {
            log.error("Wechat unifiedorder failed: {}", response);
            throw new RuntimeException("Wechat unifiedorder failed");
        }

        // 构建微信统一下单信息
        return WechatUnifiedOrder.builder().mode(payment.getMode()).appid(result.get("appid"))
                .mchid(result.get("mch_id")).nonce(result.get("nonce_str")).prepayid(result.get("prepay_id"))
                .qrcode(result.get("code_url")).build();
    }
}
